import tzlocal
import warnings
import mpmath
import pandas as pd
from datetime import timedelta


class Time(object):
    """Time class encapsulates provides a set of methods for conversion between various time formats.
    In addition it provides a generator for days between given periods"""

    __units = {'ns': 1, 'us': 1e3, 'ms': 1e6, 's': 1e9}

    def __init__(self, input, unit=None, tz=None):

        if isinstance(input, (int, float)) and unit in Time.__units.keys():
            input = Time.__convert_real_numbers_to_unix_time(input, Time.__units[unit])
            unit = 'ns'

        if tz:
            self.__pd_timestamp = pd.Timestamp(input, tz=tz, unit=unit)
        else:
            self.__pd_timestamp = pd.Timestamp(input, tz=tzlocal.get_localzone(), unit=unit)

        # region properties

    @property
    def unix_timestamp(self):
        return self.__pd_timestamp.value

    @property
    def pandas_timestamp(self):
        return self.__pd_timestamp

    @property
    def string_timestamp(self):
        return self.__pd_timestamp.__str__()

    @property
    def string_timestamp_without_time_zone_indication(self):
        return self.__pd_timestamp.__str__()[:-6]

    @property
    def datetime_timestamp(self):
        return self.__pd_timestamp.to_pydatetime()

    # endregion

    # region conversion methods

    @staticmethod
    def to_unix_timestamp(input, unit='ns', tz=None):
        """
            **Converts input time variable into a unix timestamp integer type**

            Method allowing users to convert time from four types
            (integer, string, datetime, pandas.Timestamp) into unix timestamp integer.
            The method relies on pandas.timestamp for convertion.

            :param input: input date and time
            :type input: int, string, datetime, pandas.datetime
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  string_date = "2018-05-01 00:00:00"
                  Time.convertToUnixTimestamp(string_date, unit = "ns")
                  >>> 1525125600000000000

                  from lhcsmapi.Time import Time
                  from datetime import datetime
                  date_time = datetime(2018, 5, 1, 0, 0, 0, 0)
                  Time.convertToUnixTimestamp(date_time, unit = "ns")
                  >>> 1525125600000000000

                  from lhcsmapi.Time import Time
                  timestamp = 1525125600000000000
                  Time.convertToUnixTimestamp(timestamp, unit = "ns")
                  >>> 1525125600000000000

                  from lhcsmapi.Time import Time
                  import pandas as pd
                  pd_timestamp = pd.Timestamp(2018, 5, 1, 0)
                  Time.convertToUnixTimestamp(pd_timestamp, unit = "ns")
                  >>> 1525125600000000000

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  string_date = "2018-05-01 00:00:00"
                  Time.convertToUnixTimestamp(string_date, unit = "n")
                  >>> ValueError: cannot cast unit n
        """
        time_converter = Time.__validate_conversion(input, 'unix', unit, tz)
        return time_converter.unix_timestamp

    @staticmethod
    def to_pandas_timestamp(input, unit='ns', tz=None):
        """
            **Converts input time variable into a pandas.Timestamp type**

            Method allowing users to convert time from four types
            (integer, string, datetime, pandas.Timestamp) into pandas.Timestamp.
            The method relies on pandas.timestamp for convertion.

            :param input: input date and time
            :type input: int, string, datetime, pandas.datetime
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  string_date = "2019-07-10 10:23:27.013"
                  Time.convertToPandasTimestamp(string_date, unit = "ns")
                  >>> Timestamp('2019-07-10 10:23:27.013000+0200', tz='Europe/Zurich')

                  from lhcsmapi.Time import Time
                  from datetime import datetime
                  date_time = datetime(2019, 7, 10, 10, 23, 27, 13)
                  Time.convertToPandasTimestamp(date_time, unit = "ns")
                  >>> Timestamp('2019-07-10 10:23:27.013000+0200', tz='Europe/Zurich')

                  from lhcsmapi.Time import Time
                  timestamp = 1562747007013000000
                  Time.convertToPandasTimestamp(timestamp, unit = "ns")
                  >>> Timestamp('2019-07-10 10:23:27.013000+0200', tz='Europe/Zurich')

                  from lhcsmapi.Time import Time
                  import pandas as pd
                  pd_timestamp = pd.Timestamp(2019, 7, 10, 10, 23, 27, 13000)
                  Time.convertToPandasTimestamp(pd_timestamp, unit = "ns")
                  >>> Timestamp('2019-07-10 10:23:27.013000+0200', tz='Europe/Zurich')

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  string_date = "2019-07-10 10:23:27.013"
                  Time.convertToPandasTimestamp(string_date, unit = "n")
                  >>> ValueError: cannot cast unit n
        """
        time_converter = Time.__validate_conversion(input, 'pandas', unit, tz)
        return time_converter.pandas_timestamp

    @staticmethod
    def to_string(input, unit='ns', tz=None):
        """
            **Converts input time variable into a string type**

            Method allowing users to convert time from four types
            (integer, string, datetime, pandas.Timestamp) into string.
            The method relies on pandas.timestamp for convertion.

            :param input: input date and time
            :type input: int, string, datetime, pandas.datetime
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  string_date = "2017-07-02 01:23:27"
                  Time.convertToString(string_date, unit = "ns")
                  >>> '2017-07-02 01:23:27'

                  from lhcsmapi.Time import Time
                  from datetime import datetime
                  date_time = datetime(2017, 7, 2, 1, 23, 27, 0)
                  Time.convertToString(date_time, unit = "ns")
                  >>> '2017-07-02 01:23:27'

                  from lhcsmapi.Time import Time
                  timestamp = 1498951407000000000
                  Time.convertToString(timestamp, unit = "ns")
                  >>> '2017-07-02 01:23:27'

                  from lhcsmapi.Time import Time
                  import pandas as pd
                  pd_timestamp = pd.Timestamp(2017, 7, 2, 1, 23, 27)
                  Time.convertToString(pd_timestamp, unit = "ns")
                  >>> '2017-07-02 01:23:27'

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  string_date = "2017-07-02 01:23:27"
                  Time.convertToString(string_date, unit = "n")
                  >>> ValueError: cannot cast unit n
        """
        time_converter = Time.__validate_conversion(input, 'string formatted', unit, tz)
        return time_converter.string_timestamp

    @staticmethod
    def to_string_short(input, unit='ns', tz=None, n_dec_digits=3):
        time_converter = Time.__validate_conversion(input, 'string formatted', unit, tz)
        time_without_tz = time_converter.string_timestamp.split('+')[0]
        # Cut trailing zeros up to the number of digits
        if '.' in time_without_tz:
            time_without_tz = time_without_tz.rstrip('0')
            n_dec_digits_act = len(time_without_tz.split('.')[-1])
            if n_dec_digits_act < n_dec_digits:
                time_without_tz += '0'*(n_dec_digits-n_dec_digits_act)
            return time_without_tz
        else:
            return time_without_tz

    @staticmethod
    def to_datetime(input, unit='ns', tz=None):
        """
            **Converts input time variable into a datetime type**

            Method allowing users to convert time from four types
            (integer, string, datetime, pandas.Timestamp) into datetime.
            The method relies on pandas.timestamp for convertion.

            :param input: input date and time
            :type input: int, string, datetime, pandas.datetime
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  string_date = "2019-01-02 23:59:11"
                  Time.convertToDatetime(string_date, unit = "ns")
                  >>> datetime.datetime(2019, 1, 2, 23, 59, 11, tzinfo=<DstTzInfo 'Europe/Zurich' CET+1:00:00 STD>)

                  from lhcsmapi.Time import Time
                  from datetime import datetime
                  date_time = datetime(2019, 1, 2, 23, 59, 11, 0)
                  Time.convertToDatetime(string_date, unit = "ns")
                  >>> datetime.datetime(2019, 1, 2, 23, 59, 11, tzinfo=<DstTzInfo 'Europe/Zurich' CET+1:00:00 STD>)

                  from lhcsmapi.Time import Time
                  timestamp = 1546469951000000000
                  Time.convertToDatetime(timestamp, unit = "ns")
                  >>> datetime.datetime(2019, 1, 2, 23, 59, 11, tzinfo=<DstTzInfo 'Europe/Zurich' CET+1:00:00 STD>)

                  from lhcsmapi.Time import Time
                  import pandas as pd
                  pd_timestamp = pd.Timestamp(2019, 1, 2, 23, 59, 11, 0)
                  Time.convertToDatetime(pd_timestamp, unit = "ns")
                  >>> datetime.datetime(2019, 1, 2, 23, 59, 11, tzinfo=<DstTzInfo 'Europe/Zurich' CET+1:00:00 STD>)

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  string_date = "2019-01-02 23:59:11"
                  Time.convertToDatetime(string_date, unit = "n")
                  >>> ValueError: cannot cast unit n
        """
        time_converter = Time.__validate_conversion(input, 'datetime', unit, tz)
        return time_converter.datetime_timestamp

    @staticmethod
    def to_unix_timestamp_in_sec(input, unit='ns', tz=None, warn=False):
        """
            **Converts input time variable into a float type with us precision**

            Method returns input converted to unix time in seconds with us precision

            :param input: input unix timestamp in nanoseconds
            :type input: int unix timestamp
            :param unit: unit of the input: s, ms, us, ns (optional, ns by default)
            :type unit: string
            :param durationData: time zone (optional, local by default)
            :type durationData:
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::



            - Expected Fail Response::


        """

        unix_time = Time.to_unix_timestamp(input, unit, tz)
        return Time.__convert_unix_time_in_nanoseconds_to_seconds(unix_time, warn)

    # endregion
    @staticmethod
    def get_query_period_in_unix_time(start_time_date=None, end_time_date=None, duration_date=None, tz=None):
        """
            **Calculates start and end time from start time and either end time or duration**

            Method calculates t_start and t_end or duration values based on the arguments and then returns a tuple
            with these converted to unix time. Three operation modes are possible:

            - if start_time_date is not None:

            (t_start, t_end) = (start_time_date, end_time_date)

            - elif len(duration_date) is 1:

            (t_start, t_end) = (start_time_date, startTimeData+end_time_date[0])

            - elif len(duration_date) is 2:

            (t_start, t_end) = (start_time_date-duration_date[0], start_time_date+duration_date[1])

            :param start_time_date: input date and time
            :type start_time_date: int unix timestamp
            :param end_time_date: unit of the input (None if durationData defined)
            :type end_time_date: int unix timestamp
            :param duration_date: time zone (None if endTimeData defined)
            :type duration_date: list of (time, unit) tuples
            :param tz: time zone (optional, local by default)
            :type tz: tzlocal variable
            :return: time expressed as a unix time stamp

            - Example::

                  from lhcsmapi.Time import Time
                  start_timestamp = 1426220459491000000
                  end_timestamp =   1426220479491000000
                  Time.getQueryPeriodInUnixTime(start_timestamp, end_timestamp, None, tz='Europe/Zurich')
                  >>> (1426220459491000000, 1426220479491000000)

                  from lhcsmapi.Time import Time
                  start_timestamp = 1426220459491000000
                  duration =  [(20, 's')]
                  Time.getQueryPeriodInUnixTime(start_timestamp, None, duration, tz='Europe/Zurich')
                  >>> (1426220459491000000, 1426220479491000000)

                  from lhcsmapi.Time import Time
                  start_timestamp = 1426220469491000000
                  duration =  [(10, 's'), (10, 's')]
                  Time.getQueryPeriodInUnixTime(start_timestamp, None, duration, tz='Europe/Zurich')
                  >>> (1426220459491000000, 1426220479491000000)

            - Expected Fail Response::

                  from lhcsmapi.Time import Time
                  start_timestamp = 1426220469491000000
                  Time.getQueryPeriodInUnixTime(start_timestamp, None, None, tz='Europe/Zurich')
                  >>> ValueError: end_time_date and duration_date can not be both None at the same time!
        """

        # convert t_start to unixTime:
        start_time, unit = Time.__validate_time_tuple(start_time_date)
        if start_time:
            start_time = Time.to_unix_timestamp(start_time, unit, tz)

        # extact t_end if provided if so convert to unixTime:
        end_time, unit = Time.__validate_time_tuple(end_time_date)
        if end_time:
            end_time = Time.to_unix_timestamp(end_time, unit, tz)

        elif duration_date:

            if isinstance(duration_date, tuple):
                duration_date = [duration_date]

            # duration is list with single entry or tuple , calculate t_end from duration:
            if isinstance(duration_date, list) and len(duration_date) == 1:

                duration, unit = Time.__validate_time_tuple(duration_date[0])
                end_time = start_time + Time.to_unix_timestamp(duration, unit, tz)

            # duration is list with two entries (tuple) , calculate t_start and t_end from duration:
            elif isinstance(duration_date, list) and len(duration_date) == 2:

                dt1, unit1 = Time.__validate_time_tuple(duration_date[0])
                dt2, unit2 = Time.__validate_time_tuple(duration_date[1])

                end_time = start_time + Time.to_unix_timestamp(dt2, unit2, tz)
                start_time -= Time.to_unix_timestamp(dt1, unit1, tz)

        return (start_time, end_time)

    @staticmethod
    def daterange(start_date, end_date):
        """
            **Creates a generator of consecutive days between two dates**

            Method allowing users to create a generator of days from a given
            start_date, with an increment of a day, to end_date - 1 day

            :param start_date: starting date for the generator
            :type input: datetime.date(YYYY, MM, DD)
            :param end_date: end date
            :type input: datetime.date(YYYY, MM, DD)
            :return: generator providing a day from start_date to end_date - 1 day

            - Example::

                  from lhcsmapi.Time import Time
                  from datetime import date
                  start_date = date(2014, 1, 1)
                  end_date = date(2014, 1, 3)
                  for single_date in Time.daterange(start_date, end_date):
                      print(single_date)
                  >>> 2014-01-01
                  >>> 2014-01-02

            - Expected Fail Response::

                  from datetime import date
                  start_date = date(2014, 1, 5)
                  end_date = date(2014, 1, 3)
                  for single_date in Time.daterange(start_date, end_date):
                      print(single_date)
                  >>>
        """
        for n in range(int((end_date - start_date).days)):
            yield start_date + timedelta(n)

    @staticmethod
    def modulo_day_in_unix_timestamp(start_time, end_time):
        start_time_unix = Time.to_unix_timestamp(start_time)
        end_time_unix = Time.to_unix_timestamp(end_time)
        if start_time_unix > end_time_unix:
            raise AttributeError('t_start {} is greater than the t_end {}'.format(start_time, end_time))

        diff_time_unix = end_time_unix - start_time_unix
        day_in_nanos = int(24*60*60*1e9)

        quotient = diff_time_unix//day_in_nanos
        remainder = diff_time_unix%day_in_nanos
        return (quotient, remainder)


    # region private helper methods
    @staticmethod
    def __validate_time_tuple(time_data):
        """Method returns a tuple that can be safely unpacked regardless of whether timedata is a  tuple: (time, unit)
        or single: time object"""

        time = None
        unit = None

        if isinstance(time_data, tuple):
            if len(time_data) > 0:
                time = time_data[0]

            if len(time_data) > 1:
                unit = time_data[1]

        else:
            time = time_data

        return time, unit

    @staticmethod
    def __validate_conversion(input, conversion_type, unit='ns', tz=None):
        try:
            return Time(input, unit, tz)
        except TypeError:
            raise TypeError('Cannot convert input {} to {} timestamp'.format(type(input), conversion_type))

    @staticmethod
    def __convert_real_numbers_to_unix_time(input, ratio):
        """Method converts input of int or float to unix time based on ratio"""
        with mpmath.mp.workdps(20):
            input_full = mpmath.mpf(str(input))

            ratio = mpmath.mpf(ratio)
            unix_time = input_full * ratio
            return int(unix_time)

    @staticmethod
    def __convert_unix_time_in_nanoseconds_to_seconds(unix_time, warn=False):
        """Method converts unix time in nanoseconds to unix time in seconds with us precision"""

        ns_remainder = unix_time % 1000

        if warn and ns_remainder > 0:
            warnings.warn('discarding nonzero {} nanoseconds in conversion'.format(ns_remainder), stacklevel=2)

        ns = unix_time - ns_remainder
        unix_time_seconds = mpmath.mpf(ns) / mpmath.mpf(1e9)

        return float(unix_time_seconds)
    #endregion
