import pandas as pd
import scipy as sp
import numpy as np
from copy import deepcopy
import matplotlib.pyplot as plt
import warnings

from lhcsmapi.Time import Time
from lhcsmapi.analysis.diode.DiodeLeadResistanceAnalysis import DiodeLeadResistanceAnalysis
from lhcsmapi.analysis.diode.DiodeLeadResistancePlot import DiodeLeadResistancePlot
from lhcsmapi.dbsignal.ExpSignalAnalysis import ExpSignalAnalysis
from lhcsmapi.dbsignal.SignalProcessing import SignalProcessing
from lhcsmapi.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities
from lhcsmapi.pyedsl.AssertionBuilder import AssertionBuilder
from lhcsmapi.pyedsl.FeatureBuilder import FeatureBuilder
from lhcsmapi.pyedsl.PlotBuilder import create_title
from lhcsmapi.reference.Reference import Reference


def get_expert_comment(system, is_automatic=False, automatic_analysis_rating=''):
    # if in the automatic mode (is_automatic=True), always return False to avoid breaking the analysis loop
    if is_automatic:
        if automatic_analysis_rating == '':
            system_comment = 'Automatic analysis, no expert feedback.'
        else:
            system_comment = automatic_analysis_rating
    else:
        system_check = input("Is {} OK? [NOT_OK]:".format(system))
        system_check = system_check.upper() or 'NOT_OK'
        if system_check != 'OK':
            system_check = 'NOT_OK'
            system_comment = input("Please insert some comments: ")
            system_comment = system_comment or system_check
        else:
            system_comment = 'OK'

    print("-----")
    print("{} Analysis comment is: {}".format(system, system_comment))
    print("-----")
    return system_comment


def check_show_next(index, max_index, is_automatic=False):
    # if in the automatic mode (is_automatic=True), always return False to avoid breaking the analysis loop
    if is_automatic:
        return False
    # if the list has been exhausted
    if index == max_index:
        return True
    else:
        show_next = input("Show next? [Y or N]:")
        show_next = show_next.upper() or 'Y'
        if show_next == 'N':
            return True
        else:
            return False


def get_expert_decision(user_prompt, answers):
    answers_upper = [answer.upper() for answer in answers]
    expert_decision = input("{} {}:".format(user_prompt, answers))
    while expert_decision.upper() not in answers_upper:
        expert_decision = input("{} {}:".format(user_prompt, answers))

    # find the index of matching the input
    index_decision = answers_upper.index(expert_decision.upper())
    return answers[index_decision]


def convert_plot_features_map_to_text(features_map):
    max_key_len = max(map(len, features_map))
    return '\n\n'.join(['{}:{} {} {}'.format(key, (max_key_len - len(key)) * ' ', value[0], value[1]) for key, value in
                        features_map.items()])


def calculate_char_time_profile(i_meas_df, timestamp_first_quench=0):
    tau_i_meas_df = deepcopy(i_meas_df)
    signal = tau_i_meas_df.columns[0]
    der_name = 'd{}'.format(signal)
    SignalProcessing.calculate_time_derivative_of_signals(tau_i_meas_df, tau_i_meas_df.index, [signal], [der_name])

    # Calculate time constant as a ratio of a signal by its derivative
    char_time_name = '{0}/d{0}'.format(signal)
    tau_i_meas_df[char_time_name] = -tau_i_meas_df[signal] / sp.signal.medfilt(tau_i_meas_df[der_name], 3)
    tau_i_meas_df.loc[tau_i_meas_df.index < timestamp_first_quench, [char_time_name]] = np.nan
    tau_i_meas_df.drop(columns=[der_name], inplace=True)

    return tau_i_meas_df[char_time_name]


# DIODE
def calculate_resistance_nxcals(circuit_name, i_circuit_sync, u_diode_sync):
    U_DIODE = u_diode_sync.columns[0]

    # Resample circuit current
    t_end = 350 if 'RB' in circuit_name else 150
    i_circuit_sync_resampled = SignalAnalysis.resample(i_circuit_sync, ts=0.1, t0=-50, t_end=t_end)
    if len(i_circuit_sync_resampled) > len(u_diode_sync):
        i_circuit_sync_resampled = i_circuit_sync_resampled.head(len(u_diode_sync))

    # Join current and diode voltage into a single dataframe
    u_diode_sync['I_MEAS'] = i_circuit_sync_resampled.values

    # Recover resistive and diode minimum voltage
    DiodeLeadResistanceAnalysis.calculate_minimum_diode_and_current_leads_voltage(u_diode_sync, U_DIODE)

    # Calculate resistance
    u_diode_sync['RES'] = u_diode_sync.apply(lambda row: 1e+6 * row['U_RES'] / row['I_MEAS'], axis=1, raw=False)

    # Set to 0 resistance values for too low current and with too high diode voltage derivative
    DiodeLeadResistanceAnalysis.set_to_zero_resistance_outside_of_range(u_diode_sync, 'I_MEAS')

    # Filter resistance
    DiodeLeadResistanceAnalysis.filter_resistance(u_diode_sync)

    return u_diode_sync


# nQPS
def plot_i_meas_nqps_voltage(i_meas_df, nqps_signal_dfs, timestamp_fgc_act, circuit_name, *,
                             signal, xlim=(-5, 150), ylim=None):
    title = '{} {} I_MEAS(t) and {}(t)'.format(Time.to_string_short(timestamp_fgc_act), circuit_name,
                                               signal)
    ax1 = i_meas_df.plot(title=title, color='C0', figsize=(15, 7.5), grid=True)
    ax1.title.set_size(20)
    ax1.legend(['I_MEAS'], loc='lower left')
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.set_ylabel("I_MEAS, [A]", fontsize=15, color='C0')
    ax1.tick_params(labelsize=15)

    ax2 = ax1.twinx()
    ax2.set_ylabel("{}, [V]".format(signal), fontsize=15)
    ax2.tick_params(labelsize=15)
    for nqps_signal_df in nqps_signal_dfs:
        if not nqps_signal_df.empty:
            nqps_signal_df.plot(ax=ax2, legend=False)

    ax2.set_xlim(xlim)
    if ylim is not None:
        ax2.set_ylim(ylim)

    plt.show()


def calculate_char_time_value(series, t_start_decay, t_end_decay):
    # ToDo: possible duplicate of tau_energy - to be fixed with the DSL
    index_start_decay = series.index.get_loc(t_start_decay, method='nearest')

    index_end_decay = series.index.get_loc(t_end_decay, method='nearest')
    df_decay = series.iloc[index_start_decay:index_end_decay] - series.min()
    df_decay_squared = df_decay.pow(2)
    df_decay_tau = ExpSignalAnalysis.calculate_characteristic_time_in_seconds_with_integral(
        df_decay_squared, exponent=2)
    return df_decay_tau[0]


def plot_earth_current(i_a_df, i_earth_df, i_earth_ref_df, *, title='', hwc_test=''):
    ax1 = i_a_df.plot(title=title, figsize=(15, 7.5), color='C0', grid=True, legend=True)
    ax1.title.set_size(20)
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.set_ylabel("I_A [A]", color='C0', fontsize=15)
    ax1.legend(loc='lower left')
    ax1.tick_params(labelsize=15)

    ax2 = ax1.twinx()
    i_earth_df.plot(ax=ax2, xlim=(-1, 5), color='C2', legend=True)
    i_earth_ref_df.plot(ax=ax2, xlim=(-1, 5), color='C3', style='--')
    ax2.set_ylabel("IEARTH [A]", color='C2', fontsize=15)
    ax2.legend(['IEARTH', 'IEARTH ({})'.format(hwc_test)], loc='upper right')
    ax2.tick_params(labelsize=15)
    ax1.set_xlim(-0.1, 0.3)
    # correct first axis
    ax1_ylim_min = int(i_a_df.reindex([0.3], method='nearest', tolerance=.1).values[0])
    ax1_ylim_max = int(i_a_df.reindex([-0.1], method='nearest', tolerance=.1).values[0]) * 1.001
    ax1.set_ylim(ax1_ylim_min, ax1_ylim_max)

    plt.show()


def plot_earth_current_pcnt(i_meas_df, i_earth_pcnt_df, i_earth_pcnt_ref_df, *, title='', hwc_test=''):
    ax1 = i_meas_df.plot(title=title, figsize=(15, 7.5), color='C0', grid=True, legend=True)
    ax1.title.set_size(20)
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.legend(loc='lower left')
    ax1.set_ylabel("I_MEAS [A]", color='C0', fontsize=15)
    ax1.tick_params(labelsize=15)

    ax2 = ax1.twinx()
    i_earth_pcnt_df.plot(ax=ax2, color='C2', legend=True)
    i_earth_pcnt_ref_df.plot(ax=ax2, color='C3', yerr=3)
    ax2.legend(['I_EARTH_PCNT', 'I_EARTH_PCNT ({})'.format(hwc_test)], loc='upper right')
    ax2.set_ylabel("I_EARTH_PCNT [%]", color='C2', fontsize=15)
    ax2.tick_params(labelsize=15)
    ax2.set_xlim(0.3, 120)
    plt.show()


# EE
def plot_ee_fgc(u_dump_res_dfs, i_meas_df, *, title='', plot_text=''):
    ax1 = i_meas_df.plot(title=title, figsize=(15, 7.5), color='C0', legend=True, grid=True)
    ax1.title.set_size(20)
    ax1.set_ylabel("I_MEAS, [A]", color='C0', fontsize=15)
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.legend(loc='lower left')
    ax1.tick_params(labelsize=15)

    ax2 = ax1.twinx()
    colors = ['C2', 'C3']
    for u_dump_res_df, color in zip(u_dump_res_dfs, colors):
        u_dump_res_df.plot(ax=ax2, color=color, legend=True)
    ax2.set_ylabel("U_DUMP_RES, [V]", color='C2', fontsize=15)

    ax2.annotate(plot_text, xytext=(max(u_dump_res_dfs[0].index.max() * 1.3, i_meas_df.index.max()) * 1.1, 0),
                 xy=(0.6, 0), fontsize=15, family='monospace')
    ax2.legend(loc='upper right')
    ax2.tick_params(labelsize=15)
    ax2.set_xlim((-50, ax2.get_xlim()[1]))
    plt.show()


def plot_i_a_u_dump_res(i_a_df: pd.DataFrame, i_ref_df: pd.DataFrame, u_dump_res_dfs: list, timestamp_fgc: int,
                        timestamp_ees: list, timestamp_pic: int, *, title='', plot_text=''):
    i_a_max = i_a_df.max().values[0]
    ylim = (i_a_max - 100, i_a_max + 10)

    ax1 = i_a_df.plot(title=title, xlim=(-0.2, 1), ylim=ylim, color='C0', legend=True, figsize=(15, 7.5),
                      grid=True)
    i_ref_df.plot(ax=ax1, color='C1', marker='o', style='-', xlim=(-0.2, 1), legend=True, grid=True)
    ax1.title.set_size(20)
    ax1.set_ylabel("I_A, [A]", color='C0', fontsize=15)
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.legend(loc='lower left')
    ax1.tick_params(labelsize=15)

    ax2 = ax1.twinx()
    colors = ['C2', 'C3']
    for u_dump_res_df, color in zip(u_dump_res_dfs, colors):
        u_dump_res_df.plot(ax=ax2, color=color, legend=True)
    ax2.set_ylabel("U_DUMP_RES, [V]", color='C2', fontsize=15)

    ax2.annotate(plot_text, xytext=(1.1, 0), xy=(0.6, 0), fontsize=15, family='monospace')

    ax2.legend(loc='upper right')
    for index, timestamp_ee in enumerate(timestamp_ees):
        ax2.annotate('', xy=((timestamp_ee - timestamp_fgc) * 1e-9, (0.2 - index * 0.1) * ax2.get_ylim()[1]),
                     xytext=((timestamp_pic - timestamp_fgc) * 1e-9, (0.2 - index * 0.1) * ax2.get_ylim()[1]),
                     arrowprops=dict(arrowstyle='<->'))

    ax2.tick_params(labelsize=15)

    ax2.set_xlim(-0.2, 1)
    # correct first axis
    ax1_ylim_min = int(i_a_df.reindex([1], method='nearest', tolerance=.1).values[0])
    ax1_ylim_max = int(i_a_df.reindex([-0.1], method='nearest', tolerance=.1).values[0]) * 1.001
    ax1.set_ylim(ax1_ylim_min, ax1_ylim_max)

    plt.axvline(x=(timestamp_pic - timestamp_fgc) * 1e-9, color='green', linestyle='--', linewidth=1.5);
    plt.show()


# Plot FGC should go to the PC analysis
def plot_fgc_current(circuit_type, i_meas_df, i_meas_ref_df, tau_i_meas_df, tau_i_meas_ref_df, title='',
                     plot_text=''):
    hwc_test = 'PNO.b2' if circuit_type == 'RB' else 'PNO.b3'
    ylim = (50, 110) if circuit_type == 'RB' else (0, 35)

    ax1 = i_meas_df.plot(color='C0', legend=True, grid=True)
    i_meas_ref_df.plot(ax=ax1, color='C1', style='--', grid=True)
    ax1.set_ylabel("I_MEAS [A]", color='C0', fontsize=15)
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.legend(['I_MEAS', 'I_MEAS ({})'.format(hwc_test)], loc='lower left')
    ax1.tick_params(labelsize=16)

    ax2 = ax1.twinx()

    tau_i_meas_df.plot(ax=ax2, title=title, figsize=(15, 10), color='C3', legend=True)
    tau_i_meas_ref_df.plot(ax=ax2, color='C2', legend=True, style='--')
    ax2.title.set_size(20)
    ax2.set_ylabel("-I_MEAS/dI_MEAS/dt [s]", color='C3', fontsize=15)
    ax2.legend(['-I_MEAS/dI_MEAS', '-I_MEAS/dI_MEAS ({})'.format(hwc_test)], loc='upper right')

    ax2.set_ylim(ylim)
    ax2.set_xlim(-50, ax2.get_xlim()[1])

    ax2.annotate(plot_text, xytext=(ax2.get_xlim()[1] * 1.2, ax2.get_ylim()[0]), xy=(20, ax2.get_ylim()[0]),
                 fontsize=15, family='monospace')
    ax2.tick_params(labelsize=16)
    plt.show()


# DIODE
def plot_current_voltage_nxcals(magnet, u_diode, ax=None, *, title='NXCALS'):
    if ax is None:
        ax1 = u_diode.filter(regex='I_MEAS').plot(title=title, figsize=(15, 7.5), legend=True, lw=2, color='C0')
    else:
        ax1 = u_diode.filter(regex='I_MEAS').plot(ax=ax, title=title, legend=True, lw=2, color='C0')

    ax1.title.set_size(20)
    ax1.set_xlabel("time, [s]", fontsize=20)
    ax1.tick_params(axis='both', which='major', labelsize=15)
    ax1.legend(['I_MEAS'], loc='lower left')
    ax1.set_ylabel("I_MEAS [A]", color='C0', fontsize=20)
    ax1.grid()

    ax2 = ax1.twinx()
    u_diode.filter(regex=magnet).plot(ax=ax2, legend=True, lw=2, color='C2')
    ax2.set_ylabel("U_DIODE [V]", color='C2', fontsize=20)
    ax2.tick_params(axis='both', which='major', labelsize=15)

    ax2.set_ylim([0, ax2.get_ylim()[1]])

    if ax is None:
        plt.show()


def plot_current_resistance_nxcals(magnet, u_diode, ax=None, *, title='NXCALS'):
    if ax is None:
        ax1 = u_diode.filter(regex='I_MEAS').plot(title=title, figsize=(15, 7.5), legend=None, lw=2, color='C0')
    else:
        ax1 = u_diode.filter(regex='I_MEAS').plot(ax=ax, title=title, legend=None, lw=2, color='C0')

    ax1.title.set_size(30)
    ax1.set_xlabel("time, [s]", fontsize=20)
    ax1.tick_params(axis='both', which='major', labelsize=15)
    ax1.legend(['I_MEAS'], loc='lower left')
    ax1.set_ylabel("I_MEAS [A]", color='C0', fontsize=20)
    ax1.grid()

    ax2 = ax1.twinx()
    u_diode['RES_filtered'].plot(ax=ax2, legend=True, lw=2, color='C2')
    ax2.set_ylabel("R_DIODE_LEADS(t) (Calculated) [uOhm]", color='C2', fontsize=20)
    ax2.tick_params(axis='both', which='major', labelsize=15)
    ax2.legend(['R_DIODE_LEADS(t)'], loc='upper right')

    x_text = 200 if "MB" in magnet else 100
    res_max_cals, t_res_max_cals, i_res_max_cals = DiodeLeadResistanceAnalysis.extract_res_max_time_current_nxcals(
        u_diode, 'I_MEAS')
    plot_text = ('r_max = {:.1f} uOhm\ni_r_max = {} A\nt_r_max = {} s'.
                 format(res_max_cals, int(i_res_max_cals), int(t_res_max_cals)))
    ax2.annotate(plot_text, xytext=(x_text, res_max_cals / 2), xy=(t_res_max_cals, res_max_cals),
                 arrowprops=dict(facecolor='white', shrink=0.05), fontsize=15)
    if ax is None:
        plt.show()


def plot_current_voltage_resistance_pm_nxcals(source_qds, circuit_name, timestamp_qds, u_diode_pm, u_diode_nxcals):
    fig, ax = plt.subplots(2, 2, figsize=(30, 15))
    title = "Circuit: {}, Magnet: {}, Time Stamp: {}, I_MEAS(t), U_DIODE(t), " \
            "and R_DIODE_LEADS(t)".format(circuit_name, source_qds, Time.to_string_short(timestamp_qds))
    fig.suptitle(title, fontsize=30)
    # # PM
    DiodeLeadResistancePlot.plot_current_voltage_pm(source_qds, u_diode_pm, timestamp_qds, ax=ax[0, 0])
    DiodeLeadResistancePlot.plot_current_resistance_pm(circuit_name, source_qds, u_diode_pm, timestamp_qds,
                                                       ax=ax[1, 0])
    # # NXCALS
    plot_current_voltage_nxcals(source_qds, u_diode_nxcals, ax=ax[0, 1])
    plot_current_resistance_nxcals(source_qds, u_diode_nxcals, ax=ax[1, 1])
    plt.show()


def plot_current_voltage_resistance_nxcals(source_qds, u_diode_nxcals):
    plot_current_voltage_nxcals(source_qds, u_diode_nxcals)
    plot_current_resistance_nxcals(source_qds, u_diode_nxcals)


def calculate_ee_resistance(u_dump_res_df, i_meas_df):
    index_i_meas_u_dump_res_max = i_meas_df.index.get_loc(u_dump_res_df.idxmax().values[0],
                                                          method='nearest')
    i_meas_u_dump_res_max = i_meas_df.iloc[index_i_meas_u_dump_res_max].values[0]
    return u_dump_res_df.max().values[0] / i_meas_u_dump_res_max


def append(data_structure, col, values):
    for index, tau in enumerate(values):
        data_structure['{}_{}'.format(col, index + 1)] = tau


def get_current_plateau_start_end(i_meas_df, i_meas_threshold=500):
    I_MEAS = i_meas_df.columns[0]
    i_meas_df['acqStamp'] = i_meas_df.index
    i_meas_df = SignalUtilities.synchronize_df(i_meas_df)
    i_meas_df = SignalUtilities.convert_indices_to_sec(i_meas_df)

    t0 = i_meas_df['acqStamp'].values[0]

    pd.options.mode.chained_assignment = None
    # take the current only above 500 A
    i_meas_df = i_meas_df[i_meas_df[I_MEAS] > i_meas_threshold]

    # calculate time derivative
    i_meas_df.loc[:, 'time'] = i_meas_df.index
    i_meas_df['dI_MEAS'] = i_meas_df[I_MEAS].diff()
    i_meas_df['dtime'] = i_meas_df['time'].diff()
    i_meas_df = i_meas_df.dropna()
    i_meas_df['dI_MEAS_dt'] = i_meas_df.apply(lambda col: col['dI_MEAS'] / col['dtime'], axis=1)
    # round to the nearest integer
    i_meas_df['round_dI_MEAS_dt'] = i_meas_df.apply(lambda row: np.around(row['dI_MEAS_dt']), axis=1)
    # get sign of the round value
    i_meas_df['sign_dI_MEAS_dt'] = i_meas_df.apply(lambda row: np.sign(row['round_dI_MEAS_dt']), axis=1)
    # find change of the sign of the first derivative
    i_meas_df['change_sign_dI_MEAS_dt'] = i_meas_df['sign_dI_MEAS_dt'].diff()
    di_meas_dt_change = i_meas_df[i_meas_df['change_sign_dI_MEAS_dt'] != 0]
    di_meas_dt_change.index = di_meas_dt_change['acqStamp']
    di_meas_dt_change.drop(columns=['acqStamp'], inplace=True)

    # Get time_start, time_end of plateaus
    pattern_ramp_up = [-1, 1]
    exact_ramp_up_df, _ = SignalAnalysis.find_rows_with_exact_pattern_in_series(
        di_meas_dt_change['change_sign_dI_MEAS_dt'], pattern_ramp_up)
    exact_ramp_up_df.rename(columns={-1: "plateau_start", 1: 'plateau_end'}, inplace=True)

    pattern_ramp_down = [-1, -1]
    exact_ramp_down_df, _ = SignalAnalysis.find_rows_with_exact_pattern_in_series(
        di_meas_dt_change['change_sign_dI_MEAS_dt'], pattern_ramp_down)
    exact_ramp_down_df.columns = ['plateau_start', 'plateau_end']

    plateau_timing = pd.concat([exact_ramp_up_df, exact_ramp_down_df])

    # ToDo: sort plateau_timing by plateau_start
    plateau_timing['plateau_start_sync'] = plateau_timing['plateau_start'].apply(lambda x: (x - t0) / 1e9)
    plateau_timing['plateau_end_sync'] = plateau_timing['plateau_end'].apply(lambda x: (x - t0) / 1e9)

    pd.options.mode.chained_assignment = 'warn'

    return plateau_timing


class CircuitAnalysis(object):

    def __init__(self, circuit_type, results_table, is_automatic=True, colormap='default'):
        self.circuit_type = circuit_type
        self.results_table = results_table
        self.is_automatic = is_automatic
        self.colormap = colormap

    # Timestamp table
    def create_timestamp_table(self, timestamp_dct, circuit_name=''):

        timestamp_dct_filtered = {}
        for key, value in timestamp_dct.items():
            if isinstance(value, int) or isinstance(value, float) or isinstance(value, np.int64):
                if not np.isnan(value):
                    timestamp_dct_filtered[key] = value
            elif isinstance(value, pd.DataFrame):
                if not value.empty:
                    timestamp_dct_filtered[key] = value.loc[0, 'timestamp']
            else:
                raise ValueError('Type {} not supported'.format(type(value)))
        if ('nQPS' in timestamp_dct_filtered) and ('iQPS' in timestamp_dct_filtered):
            if timestamp_dct_filtered['nQPS'] < timestamp_dct_filtered['iQPS']:
                warnings.warn('nQPS PM timestamp ({}) was created before iQPS one ({})'
                              .format(timestamp_dct_filtered['nQPS'], timestamp_dct_filtered['iQPS']))

        timestamp_df = pd.DataFrame({'Source': list(timestamp_dct_filtered.keys()),
                                     'Timestamp': list(timestamp_dct_filtered.values())})

        timestamp_df['Date and time'] = timestamp_df.apply(lambda col: Time.to_string_short(col['Timestamp']), axis=1)
        timestamp_df.sort_values(by='Timestamp', inplace=True)
        timestamp_df.reset_index(inplace=True, drop=True)

        # append the timestamps to the results table
        for k, v in timestamp_dct_filtered.items():
            if not (('iqps' in k.lower()) or ('nqps' in k.lower())):
                self.results_table['timestamp_%s' % k.lower()] = v

        return timestamp_df

    # PIC
    def analyze_pic(self, timestamp_pic):
        diff_timestamp_pic = abs(timestamp_pic[0] - timestamp_pic[1])
        if diff_timestamp_pic > 1e6:
            warnings.warn('PIC timestamps {} differ by more than 1 ms ({} ns)'
                          .format(timestamp_pic, diff_timestamp_pic))
        else:
            print('PIC timestamps ({}) and ({}) are less then 1 ms away.'.format(timestamp_pic[0], timestamp_pic[1]))

        # Update results table
        if self.results_table is not None:
            self.results_table['%s_PIC' % self.circuit_type] = get_expert_comment('PIC', is_automatic=self.is_automatic)

    # PC
    def analyze_i_meas_pc(self, circuit_name, timestamp_fgc, timestamp_fgc_ref, timestamp_pic,
                          i_meas_df, i_meas_ref_df, duration_decay=(2, 100)):
        # Check Input
        if any([i_meas_df.empty, i_meas_ref_df.empty]):
            raise ValueError('One of the required signals missing, analysis aborted!')

        # Calculate Features
        # # characteristic time constant
        tau_i_meas, tau_i_meas_ref = FeatureBuilder().with_signal(i_meas_df, i_meas_ref_df) \
            .subtract_min_value() \
            .with_duration(t_start=duration_decay[0], t_end=duration_decay[1]) \
            .calculate_features('tau_energy') \
            .values[0]

        # Compare features to the reference
        pc_act_df = pd.DataFrame({'tau_i_meas': tau_i_meas}, index=['act'])

        is_tau_i_meas_correct = Reference.compare_features_to_reference(pc_act_df, self.circuit_type, 'PC')

        # Plot
        title = '{} {}: I_MEAS(t) and I_MEAS(t)/dI_MEAS(t) (+PNO.b3)\n'.format(
            Time.to_string_short(timestamp_fgc), circuit_name)

        tau_i_meas_df = calculate_char_time_profile(i_meas_df)
        tau_i_meas_ref_df = calculate_char_time_profile(i_meas_ref_df)

        plot_features = {'Circuit name': [circuit_name, ''],
                         'PIC Abort': [Time.to_string_short(timestamp_pic), ''],
                         'PC Off': [Time.to_string_short(timestamp_fgc), ''],
                         'PNO.b3': [Time.to_string_short(timestamp_fgc_ref), ''],
                         'max(I_MEAS)': [round(i_meas_df.max().values[0]), 'A'],
                         'max(I_MEAS, PNO.b3)': [round(i_meas_ref_df.max().values[0]), 'A'],
                         'I_MEAS: tau': [round(tau_i_meas), 's'],
                         'I_MEAS(PNO.b3):tau': [round(tau_i_meas_ref), 's']}
        plot_text = convert_plot_features_map_to_text(plot_features)

        # PlotBuilder
        # ylim = (50, 110) if self.circuit_type == 'RB' else (0, 35)
        # PlotBuilder().with_signal([i_meas_df, i_meas_ref_df], title=title, figsize=(15, 10), grid=True) \
        #     .with_ylabel(ylabel='I_MEAS, [A]') \
        #     .with_signal([tau_i_meas_df, tau_i_meas_ref_df]) \
        #     .with_ylabel(ylabel='-I_MEAS/dI_MEAS/dt [s]').with_ylim(ylim) \
        #     .plot()

        plot_fgc_current(self.circuit_type, i_meas_df, i_meas_ref_df, tau_i_meas_df, tau_i_meas_ref_df,
                         title, plot_text)

        # Update results table
        if self.results_table is not None:
            status = 'Within reference' if is_tau_i_meas_correct else 'Outside of reference'
            analysis = '%s_PC_I_MEAS_tau' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    def analyze_i_earth_pc(self, circuit_type, circuit_name, timestamp_fgc_rqd, i_a_rqd_df, i_earth_rqd_df,
                           i_earth_rqd_ref_df):
        # TODO check each signal separately to provide a meaningful warning
        if any([i_a_rqd_df.empty, i_earth_rqd_df.empty, i_earth_rqd_ref_df.empty]):
            warnings.warn('One of the required signals missing, analysis aborted!')
        else:
            hwc_test = 'PNO.b2' if circuit_type == 'RB' else 'PNO.b3'
            title = '{} {}: I_A and I_EARTH (+{})'.format(Time.to_string_short(timestamp_fgc_rqd), circuit_name,
                                                          hwc_test)
            plot_earth_current(i_a_rqd_df, i_earth_rqd_df, i_earth_rqd_ref_df, title=title, hwc_test=hwc_test)

        # Update results table
        if self.results_table is not None:
            analysis = '%s_PC_I_EARTH' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic)

    def analyze_i_earth_pcnt_pc(self, circuit_type, circuit_name, timestamp_fgc_rqd, i_meas_rqd_df, i_earth_pcnt_rqd_df,
                                i_earth_pcnt_rqd_ref_df):
        if any([i_meas_rqd_df.empty, i_earth_pcnt_rqd_df.empty, i_earth_pcnt_rqd_ref_df.empty]):
            warnings.warn('One of the required signals missing, analysis aborted!')
        else:
            hwc_test = 'PNO.b2' if circuit_type == 'RB' else 'PNO.b3'
            title = '{} {}: I_MEAS and I_EARTH (+{})'.format(Time.to_string_short(timestamp_fgc_rqd), circuit_name,
                                                             hwc_test)
            plot_earth_current_pcnt(i_meas_rqd_df, i_earth_pcnt_rqd_df, i_earth_pcnt_rqd_ref_df, title=title,
                                    hwc_test=hwc_test)

        # Update results table
        if self.results_table is not None:
            analysis = '%s_PC_I_EARTH_PCNT' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic)

    # EE
    def analyze_char_time_u_dump_res_ee(self, circuit_name, timestamp_fgc, u_dump_res_dfs, i_meas_df,
                                        duration_decay=(1, 120)):
        # Vectorize Input
        u_dump_res_dfs = u_dump_res_dfs if isinstance(u_dump_res_dfs, list) else [u_dump_res_dfs]

        # Check Input
        if u_dump_res_dfs[0].empty:
            raise ValueError('U_DUMP_RES empty, analysis aborted!')

        # Calculate Features
        # # characteristic time constant
        tau_u_dump_res = FeatureBuilder().with_signal(u_dump_res_dfs) \
            .subtract_min_value() \
            .with_duration(t_start=duration_decay[0], t_end=duration_decay[1]) \
            .calculate_features('tau_energy') \
            .values[0]

        # # resistance
        r_dump_res = [calculate_ee_resistance(u_dump_res_df, i_meas_df) for u_dump_res_df in u_dump_res_dfs]

        # Compare features to the reference
        ee_act_df = pd.DataFrame(index=['act'])
        append(ee_act_df, 'tau_u_dump_res', tau_u_dump_res)
        append(ee_act_df, 'R', r_dump_res)

        is_res_tau_u_dump_res_correct = Reference.compare_features_to_reference(ee_act_df, self.circuit_type, 'EE')

        # Plot
        title = '{} {} EE: I_MEAS(t) and U_DUMP_RES(t)'.format(Time.to_string_short(timestamp_fgc), circuit_name)

        plot_features = {'Circuit name': [circuit_name, '']}
        tau_u_dump_res = [(round(tau), 's') for tau in tau_u_dump_res]
        append(plot_features, 'tau: EE', tau_u_dump_res)

        r_dump_res = [(round(res * 1000, 3), 'mOhm') for res in r_dump_res]
        append(plot_features, 'res: EE', r_dump_res)

        plot_text = convert_plot_features_map_to_text(plot_features)
        plot_ee_fgc(u_dump_res_dfs, i_meas_df, title=title, plot_text=plot_text)

        # Update results table
        if self.results_table is not None:
            status = 'Within reference' if is_res_tau_u_dump_res_correct else 'Outside of reference'
            analysis = '%s_EE_U_DUMP_RES_tau_res' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    def analyze_delay_time_u_dump_res_ee(self, circuit_name, timestamp_fgc, timestamp_pic, timestamp_ees,
                                         i_a_df, i_ref_df, u_dump_res_dfs):
        # Vectorize Input
        u_dump_res_dfs = u_dump_res_dfs if isinstance(u_dump_res_dfs, list) else [u_dump_res_dfs]
        timestamp_ees = timestamp_ees if isinstance(timestamp_ees, list) else [timestamp_ees]

        # Check Input
        if any([i_a_df.empty, i_ref_df.empty, u_dump_res_dfs[0].empty]):
            raise ValueError('One of the required signals missing, analysis aborted!')

        # Calculate Features
        t_delay_ees = [(timestamp_ee - timestamp_pic) * 1e-9 for timestamp_ee in timestamp_ees]

        ee_act_df = pd.DataFrame(index=['act'])
        append(ee_act_df, 't_delay_ee', t_delay_ees)

        # Compare features to the reference
        is_t_delay_ee_correct = Reference.compare_features_to_reference(ee_act_df, self.circuit_type, 'EE',
                                                                        timestamp=timestamp_fgc)

        # Plot
        plot_features = {'Circuit name': [circuit_name, '']}
        # # add calculated features
        t_delay_ees_ms = [(round(t_delay_ee * 1e3), 'ms') for t_delay_ee in t_delay_ees]
        append(plot_features, 't_delay_ee', t_delay_ees_ms)

        plot_text = convert_plot_features_map_to_text(plot_features)

        title = '{} {} EE: I_A(t) and U_DUMP_RES(t)'.format(Time.to_string_short(timestamp_fgc), circuit_name)
        plot_i_a_u_dump_res(i_a_df, i_ref_df, u_dump_res_dfs, timestamp_fgc, timestamp_ees,
                            timestamp_pic, title=title, plot_text=plot_text)

        # Update results table
        if self.results_table is not None:
            status = 'Within reference' if is_t_delay_ee_correct else 'Outside of reference'
            analysis = '%s_EE_U_DUMP_RES_delay' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    def analyze_ee_temp(self, circuit_name, timestamp_ee, t_res_act_dfs, t_res_ref_dfs, *, abs_margin=25, scaling=1):
        title = create_title(circuit_name, timestamp_ee, 'T_RES_BODY')
        is_signal_correct = AssertionBuilder().with_signal(t_res_act_dfs) \
            .compare_to_reference(signal_ref_dfs=t_res_ref_dfs, abs_margin=abs_margin, scaling=scaling) \
            .show_plot(ylabel='T_RES_BODY, [degC]', title=title) \
            .get_assertion_result()

        if self.results_table is not None:
            status = 'Within reference' if is_signal_correct else 'Outside of reference'
            analysis = '%s_EE_T_RES_BODY' % circuit_name.split('.')[0]
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    # NQPS Voltage: U_DIODE, U_EARTH
    def analyze_u_diode_nqps(self, circuit_name, timestamp_fgc, i_meas_df, u_dfs, signal, system):
        if i_meas_df.empty or (len(u_dfs) == 0):
            warnings.warn('Required signals missing, analysis aborted!')
        else:
            plot_i_meas_nqps_voltage(i_meas_df, u_dfs, timestamp_fgc, circuit_name,
                                     signal=signal)
            plot_i_meas_nqps_voltage(i_meas_df, u_dfs, timestamp_fgc, circuit_name,
                                     signal=signal, xlim=(-2, 3))
        if self.results_table is not None:
            system_signal = '{}_{}'.format(system, signal)
            self.results_table[system_signal] = get_expert_comment(system_signal, self.is_automatic)

    # DIODE
    def analyze_diode_leads(self, circuit_name, i_meas_u_diode_u_ref_pm_dfs, i_meas_u_diode_nxcals_dfs):
        index_max = self.results_table.index[-1]
        for index, row in self.results_table.iterrows():
            current_col = 'I_MEAS_quench' if self.circuit_type == 'RB' else 'I_MEAS_RQD_quench'
            source_qds, timestamp_qds, u_diode_db, i_quench = row[['magnet', 'timestamp_iqps', 'nqps_db', current_col]]
            print('{}/{}: Analysing diode lead resistance of {} on {}, {}'
                  .format(index, index_max, source_qds, Time.to_string(timestamp_qds), timestamp_qds))

            if i_quench < 1000:
                print('The analysis skipped as the quench current is below 1 kA ({} A)'.format(i_quench))
                continue

            # NXCALS
            i_meas_diode_df, u_diode_df = deepcopy(i_meas_u_diode_nxcals_dfs[index])

            if len(u_diode_df) < 1000:
                print('The analysis skipped as the diode measurement has only {} data points'.format(len(u_diode_df)))
                continue

            try:
                res_diode_nxcals_df = calculate_resistance_nxcals(circuit_name, i_meas_diode_df, u_diode_df)
            except ValueError as err:
                print('The analysis skipped as as we encountered an exception: {}'.format(err))
                continue

            # PM
            if u_diode_db == 'PM':
                i_a_df, u_diode_a_df, u_diode_b_df, u_ref_df = deepcopy(i_meas_u_diode_u_ref_pm_dfs[index])
                if not i_a_df.empty and not u_diode_a_df.empty and not u_diode_b_df.empty and not u_ref_df.empty:
                    try:
                        res_diode_pm_df = \
                            DiodeLeadResistanceAnalysis.calculate_resistance_pm(i_a_df, u_diode_a_df, u_diode_b_df,
                                                                                u_ref_df, timestamp_qds, time_corr=0.0)
                        res_max_pm = res_diode_pm_df['RES'].max()
                        t_res_max_pm = res_diode_pm_df['time'].max()
                        plot_current_voltage_resistance_pm_nxcals(source_qds, circuit_name, timestamp_qds,
                                                                  res_diode_pm_df,
                                                                  res_diode_nxcals_df)
                    except:
                        warnings.warn('Computation of diode resistance failed, '
                                      'please check the input signals and analysis assumptions.')


                else:
                    res_max_pm = 0
                    t_res_max_pm = 0
            else:
                res_max_pm = 0
                t_res_max_pm = 0

                plot_current_voltage_resistance_nxcals(source_qds, res_diode_nxcals_df)

            # Maximum resistance value
            res_max_cals, t_res_max_cals, _ = DiodeLeadResistanceAnalysis \
                .extract_res_max_time_current_nxcals(res_diode_nxcals_df, 'I_MEAS')

            res_max = max(res_max_pm, res_max_cals)
            t_res_max = max(t_res_max_pm, t_res_max_cals)
            col_r = 'R_DIODE_LEADS_%s' % circuit_name.split('.')[0]
            col_t = 't_DIODE_LEADS_%s' % circuit_name.split('.')[0]

            # Update results table
            self.results_table.loc[index, col_r] = round(res_max, 1)
            self.results_table.loc[index, col_t] = round(t_res_max, 1)

            # Show next
            check_show_next(index, index_max, self.is_automatic)

    # DFB
    def analyze_leads_voltage(self, u_dfs, circuit_name, timestamp, *, signal, value_min, value_max):
        if all(u_df.empty for u_df in u_dfs):
            return None

        is_signal_correct = AssertionBuilder().with_signal(u_dfs) \
            .has_min_max_value(value_min=value_min, value_max=value_max) \
            .show_plot(y_label='{}, [V]'.format(signal),
                       title='%s, %s: %s(t)' % (Time.to_string_short(timestamp), circuit_name, signal)) \
            .get_assertion_result()

        if self.results_table is not None:
            status = 'Within reference' if is_signal_correct else 'Outside of reference'
            analysis = '%s_LEADS_%s' % (circuit_name, signal)
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    # VF
    def analyze_voltage_feelers(self, circuit_name, timestamp_fgc, i_meas_df, u_dfs, signal, system, ylim=None):
        # Plot
        if i_meas_df.empty or (len(u_dfs) == 0):
            warnings.warn('Required signals missing, analysis aborted!')
        else:
            plot_i_meas_nqps_voltage(i_meas_df, u_dfs, timestamp_fgc, circuit_name,
                                     signal=signal, ylim=ylim)
            plot_i_meas_nqps_voltage(i_meas_df, u_dfs, timestamp_fgc, circuit_name,
                                     signal=signal, xlim=(-2, 3), ylim=ylim)

        # Analyze
        not_communicating_cards = []
        disabled_cards = []
        status = 'OK'
        for u_df in u_dfs:
            if (float(u_df.min()) < -2000) and (float(u_df.max()) < -2000):
                not_communicating_cards.append(u_df.columns[0])
                status = 'NOT_OK'
            if (float(u_df.min()) > -1) and (float(u_df.max()) < 1):
                disabled_cards.append(u_df.columns[0])
                status = 'NOT_OK'
        if not_communicating_cards:
            status += '_NOT_COMMUNICATING_CARDS'
        if disabled_cards:
            status += '_DISABLED_CARDS'

        # Print a report
        print('Report on voltage feeler cards:')
        print('- there are {} cards.'.format(len(u_dfs)))
        print('- there are {} functioning cards.'
              .format(len(u_dfs) - len(not_communicating_cards) - len(disabled_cards)))
        print('- there are {} not communicating cards: {}.'
              .format(len(not_communicating_cards), not_communicating_cards))
        print('- there are {} disabled cards: {}.'.format(len(disabled_cards), disabled_cards))

        # Store analysis result
        if self.results_table is not None:
            system_signal = '{}_{}'.format(system, signal)
            self.results_table[system_signal] = get_expert_comment(system_signal, self.is_automatic, status)
