from typing import Union, List, Tuple

import pandas as pd
import numpy as np
from tqdm.notebook import tqdm
import warnings

from lhcsmapi.analysis.decorators import check_nan_timestamp, check_nan_timestamp_signals
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.Time import Time


class CircuitQuery(object):
    """Class containing methods for query of events and signals. Each query is implemented with pyeDSL.

    """

    def __init__(self, circuit_type, circuit_name):
        self.circuit_type = circuit_type
        self.circuit_name = circuit_name

    # PIC
    @check_nan_timestamp(return_type=np.nan)
    def find_timestamp_pic(self, timestamp_fgc: int, *, spark, duration=[(1, 's'), (60, 's')], verbose=True) -> float:
        """ Method searching PIC timestamps around an FGC PM event timestamp.
        If the input timestamp is NaN or the PIC timestamp was not found, then a NaN is returned.

        :param timestamp_fgc: FGC PM event timestamp
        :type timestamp_fgc: int
        :param spark: connector to spark and NXCALS queries
        :type spark: spark connector
        :param duration: search duration before and after the FGC PM timestamp
        :type duration: List[Tuple(int, str)]
        :param verbose: flag to display information about the queried signal
        :type verbose: bool
        :return: float value of PIC timestamp in ns precision
        """
        st_abort_pic_df = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PIC', signal='ST_ABORT_PIC') \
            .signal_query(verbose=verbose).dfs[0]

        return st_abort_pic_df.index[0] if len(st_abort_pic_df) > 0 else np.nan

    # PC
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_pc(self, t_start: Union[str, int, float], t_end: Union[str, int, float], *,
                                 verbose=True) -> pd.DataFrame:
        """ Method searching a PM FGC events between given start and end times.
        If either the start or end time is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param t_end: end time of a search
        :type t_end: Union[str, int, float]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: pd.DataFrame with source and timestamp of the PM FGC events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC') \
            .event_query(verbose=verbose).df

    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_pc_pm(self, timestamp_fgc: int, timestamp_sync: int, *, signal_names: List[str],
                    verbose=True) -> List[pd.DataFrame]:
        """ Method querying PM for power converter signals
        If any timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_fgc: FGC PM event timestamp (ns precision)
        :type timestamp_fgc: int
        :param timestamp_sync: timestamp to which the queried signals are synchronized (ns precision)
        :type timestamp_sync: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal=signal_names) \
            .signal_query(verbose=verbose) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_pc_pm_events(self, timestamp_fgc: int, *, signal_names: List[str], verbose=True) -> None:
        """ Method querying PM for FGC event signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        Note that the output signal may have duplicated indices and is neither synchronized nor converted to s.

        :param timestamp_fgc: FGC PM event timestamp (ns precision)
        :type timestamp_fgc: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal=signal_names) \
            .signal_query(allow_duplicate_index=True, verbose=verbose).dfs

    def get_timestamp_ref(self, col='fgcPm') -> int:
        """ Method returning a reference timestamp for either FGC or EE

        :param col: name of a column with information about reference timestamp (either fgcPm or eePm)
        :return: unix timestamp in ns
        """
        timestamp_ref = Reference.get_power_converter_reference_fpa(self.circuit_type, self.circuit_name, col)
        return Time.to_unix_timestamp(timestamp_ref)

    # EE
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_ee(self, t_start: Union[str, int, float], duration=[(10, 's'), (1000, 's')],
                                 verbose=True) -> pd.DataFrame:
        """ Method searching PM energy extraction events
        If either the start or end time is NaN, then an empty dataframe is returned.

        :param t_start: start time of a query
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param verbose: flag to display information about the queried signal
        :type verbose: bool
        return: pd.DataFrame with source and timestamp of the PM EE events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='EE') \
            .event_query(verbose=verbose).df

    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_ee_u_dump_res_pm(self, timestamp_ee: int, timestamp_sync: int, system='EE', *,
                               signal_names: List[str], verbose=True) -> List[pd.DataFrame]:
        """ Method querying PM for energy extraction voltage signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_ee: EE PM event timestamp (ns precision)
        :type timestamp_ee: int
        :param timestamp_sync: timestamp to which the queried signals are synchronized (ns precision)
        :type timestamp_sync: int
        :param system: name of EE system to query
        :type system: str
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param verbose: flag to display information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_ee) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query(verbose=verbose) \
            .remove_initial_offset(timestamp_ee) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_ee_t_res_pm(self, timestamp_ee: int, timestamp_sync: int, system='EE', *, signal_names: List[str],
                          verbose=True) -> List[pd.DataFrame]:
        """ Method querying PM for energy extraction temperature signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_ee: EE PM event timestamp (ns precision)
        :type timestamp_ee: int
        :param timestamp_sync: timestamp to which the queried signals are synchronized (ns precision)
        :type timestamp_sync: int
        :param system: name of EE system to query
        :type system: str
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param verbose: flag to display information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_ee) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query(verbose=verbose) \
            .remove_values_for_time_less_than(timestamp_ee) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_ee_cals(self, timestamp_fgc: Union[int, str, float], system='EE', *, signal_names: List[str], ldb,
                      duration=[(3 * 3600, 's')], t_thr=1000, verbose=True) -> List[pd.DataFrame]:
        """ Method querying CALS for energy extraction signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :param system: name of EE system
        :type system: str
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param ldb: CALS pytimber connector
        :type ldb: pytimber
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param t_thr: threshold of time for the query after which signals are taken for the output
        :type t_thr: int
        :param verbose: flag to display information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        return QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query(verbose=verbose) \
            .synchronize_time(timestamp_fgc / 1e9) \
            .remove_values_for_time_less_than(t_thr).dfs

    # iQPS
    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_iqps_nxcals(self, timestamp_fgc: int, *, signal_names: List[str], spark, duration=[(300, 's'), (10, 's')],
                          verbose=True) -> List[pd.DataFrame]:
        """ Method querying NXCALS for iQPS signals

        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param spark: connector to spark and NXCALS queries
        :type spark: spark connector
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_fgc, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names) \
            .signal_query(verbose=verbose) \
            .synchronize_time(timestamp_fgc) \
            .convert_index_to_sec() \
            .filter_median().dfs

    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_source_timestamp_qds(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')],
                                  verbose=True) -> pd.DataFrame:
        """ Method searching PM QDS events
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: pd.DataFrame with source and timestamp of the PM QDS events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', source='*') \
            .event_query(verbose=verbose) \
            .filter_source(self.circuit_name, 'QDS') \
            .drop_duplicate_source() \
            .sort_values(by='timestamp').df

    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_qds_pm(self, timestamp_qds: int, timestamp_qds_sync: int, *, signal_names: List[str],
                     verbose=True) -> List[pd.DataFrame]:
        """ Method querying PM for QDS signals
        If the timestamp is NaN, then a list of empty dataframes of the size of the signal names is returned.
        The signals are synchronized in time and converted from ns to s.

        :param timestamp_qds: PM QDS timestamp in ns precision
        :type timestamp_qds: int
        :param timestamp_qds: PM QDS timestamp in ns precision (for signal synchronization)
        :type timestamp_qds: int
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names) \
            .signal_query(verbose=verbose) \
            .synchronize_time(timestamp_qds_sync) \
            .convert_index_to_sec() \
            .filter_median().dfs

    def query_iqps_analog_pm(self, source_timestamp_qds_df: pd.DataFrame,
                             signal_names: List[str]) -> List[List[pd.DataFrame]]:
        """ Method querying PM for iQPS analog signals
        Signals are synchronized to the iQPS PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: List of list of iQPS signals. The outer list has as many elements as iQPS events and the lenght
        of the inner is equal to the number of iQPS signals.
        """
        iqps_analog_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying iQPS PM'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            iqps_analog_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names, source=source_qds,
                               wildcard={'CELL': source_qds}) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec() \
                .filter_median().dfs
            iqps_analog_dfs.append(iqps_analog_df)

        return iqps_analog_dfs

    def query_iqps_digital_pm(self, source_timestamp_qds_df: pd.DataFrame,
                              signal_names: List[str]) -> List[List[pd.DataFrame]]:
        """ Method querying PM for iQPS digital signals
        Signals are synchronized to the iQPS PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :return: List of list of iQPS signals. The outer list has as many elements as iQPS events and the lenght
        of the inner is equal to the number of iQPS signals.
        """
        iqps_digital_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying IQPS PM'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            iqps_digital_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signal_names, source=source_qds) \
                .signal_query() \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec().dfs
            iqps_digital_dfs.append(iqps_digital_df)

        return iqps_digital_dfs

    # nQPS
    def find_source_timestamp_nqps(self, t_start: Union[int, str, float], duration=[(10, 's'), (400, 's')],
                                   verbose=True) -> pd.DataFrame:
        """ Method searching PM nQPS events
        If the start is NaN, then an empty dataframe is returned.

        :param t_start: start time of a search
        :type t_start: Union[str, int, float]
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: pd.DataFrame with source and timestamp of the PM nQPS events
        """
        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, source='*') \
            .event_query(verbose=verbose) \
            .filter_source(self.circuit_name, 'QDS') \
            .drop_duplicate_source() \
            .sort_values(by='timestamp').df

    def query_nqps_voltage_pm(self, source_timestamp_qds_df: pd.DataFrame) -> List[List[pd.DataFrame]]:
        """ Method querying PM for nQPS signals
        Signals are synchronized to the nQPS PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :return: List of list of nQPS signals. The outer list has as many elements as nQPS events and the lenght
        of the inner is equal to the number of nQPS signals.
        """
        u_nqps_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM for nQPS analog signals'):
            timestamp_qds = row['timestamp']
            source_qds = row['source']
            qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
            system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
            signal = 'U_DIODE_{}'.format(self.circuit_name.split('.')[0])
            wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

            source_timestamp_nqps_df = self.find_source_timestamp_nqps_pm(source_qds, timestamp_qds)
            if not source_timestamp_nqps_df.empty:
                timestamp_nqps = source_timestamp_nqps_df.loc[0, 'timestamp']

                u_diode_rqd_df = QueryBuilder().with_pm() \
                    .with_timestamp(timestamp_nqps) \
                    .with_circuit_type(self.circuit_type) \
                    .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal, source=qps_crate,
                                   wildcard={'MAGNET': wildcard}) \
                    .signal_query() \
                    .synchronize_time(timestamp_nqps) \
                    .convert_index_to_sec().dfs[0]

                u_ref_rqd_df = QueryBuilder().with_pm() \
                    .with_timestamp(timestamp_nqps) \
                    .with_circuit_type(self.circuit_type) \
                    .with_metadata(circuit_name=self.circuit_name, system=system, signal='U_REF_N1', source=qps_crate,
                                   wildcard={'QPS_CRATE': qps_crate}) \
                    .signal_query() \
                    .synchronize_time(timestamp_nqps) \
                    .convert_index_to_sec().dfs[0]
                u_nqps_dfs.append([u_diode_rqd_df, u_ref_rqd_df])
            else:
                warnings.warn('nQPS buffer is empty for %s !' % source_qds)
                u_nqps_dfs.append([pd.DataFrame(), pd.DataFrame()])

        return u_nqps_dfs

    # QH
    def query_qh_pm(self, source_timestamp_qds_df: pd.DataFrame, *, signal_names: List[str],
                    is_ref=False) -> List[List[pd.DataFrame]]:
        """ Method querying PM for QH signals (either voltage or current or both)
        Signals are synchronized to the QH PM event and index is converted to seconds.

        :param source_timestamp_qds_df: dataframe with PM source and timestamp for QDS events
        :type source_timestamp_qds_df: pd.DataFrame
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param is_ref: boolean flag to determine whether a query of reference or actual discharge is necessary
        :type is_ref: bool
        :return: List of list of QH signals. The outer list has as many elements as QH events and the lenght
        of the inner is equal to the number of QH signals.
        """
        hds_dfs = []
        for index, row in tqdm(source_timestamp_qds_df.iterrows(), total=source_timestamp_qds_df.shape[0],
                               desc='Querying PM'):
            source_qds = row['source']
            if is_ref:
                timestamp_qh = Reference.get_quench_heater_reference_discharge(self.circuit_type, source_qds)
            else:
                timestamp_qh = row['timestamp'] + 1000000

            hds_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qh) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QH', signal=signal_names, source=source_qds,
                               wildcard={'CELL': source_qds}) \
                .signal_query() \
                .synchronize_time(timestamp_qh) \
                .convert_index_to_sec().dfs

            hds_dfs.append(hds_df)

        return hds_dfs

    # DIODE LEADS
    def find_source_timestamp_nqps_pm(self, source_qds: str, timestamp_qds: int, duration=[(600, 's'), (600, 's')],
                                      verbose=True) -> pd.DataFrame:
        """ Method searching PM nQPS events
        If the timestamp is NaN, then an empty dataframe is returned.

        :param source_qds: PM QDS source according to metadata
        :type source_qds: str
        :param timestamp_qds: PM QDS timestamp in ns precision
        :type timestamp_qds: int
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: pd.DataFrame with source and timestamp of the PM nQPS events
        """
        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])

        return QueryBuilder().with_pm() \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, source=qps_crate) \
            .event_query(verbose=verbose).df

    def _query_i_u_diode_pm(self, circuit_name: str, source_qds: str, timestamp_fgc: int, timestamp_qds_a: int,
                            timestamp_qds_b: int) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """ Method querying PM for current and voltage of a diode
        Signals are **not** synchronized to the input timestamp and index is **not** converted to seconds.

        :param circuit_name: name of a circuit to query
        :type circuit_name: str
        :param source_qds: PM QDS source according to metadata
        :type source_qds: str
        :param timestamp_fgc: FGC PM event timestamp (ns precision)
        :type timestamp_fgc: int
        :param timestamp_qds_a: QDS Board A PM event timestamp (ns precision)
        :type timestamp_qds_a: int
        :param timestamp_qds_b: QDS Board B PM event timestamp (ns precision)
        :type timestamp_qds_b: int
        :return: tuple of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        i_a_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_fgc) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_A') \
            .signal_query().dfs[0]

        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_a_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_a) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                           wildcard={'MAGNET': wildcard}) \
            .signal_query().dfs[0]

        u_diode_b_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_b) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                           wildcard={'MAGNET': wildcard}) \
            .signal_query().dfs[0]

        u_ref_rqd_df: pd.DataFrame = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_qds_a) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal='U_REF_N1', source=qps_crate,
                           wildcard={'QPS_CRATE': qps_crate}) \
            .signal_query().dfs[0]

        return i_a_rqd_df, u_diode_a_rqd_df, u_diode_b_rqd_df, u_ref_rqd_df

    def _query_i_u_diode_nxcals(self, circuit_name: str, source_qds: str, timestamp_qds: int, *, spark,
                                duration=[(50, 's'), (150, 's')]) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """ Method querying NXCALS for current and voltage of a diode
        Signals are synchronized to the input timestamp and index is converted to seconds.

        :param circuit_name: name of a circuit to query
        :type circuit_name: str
        :param source_qds: PM QDS source according to metadata
        :type source_qds: str
        :param timestamp_qds: QDS Board A PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_qds: int
        :param spark: connector to spark and NXCALS queries
        :type spark: spark connector
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :return: tuple of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        i_meas_diode_df: pd.DataFrame = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_MEAS') \
            .signal_query() \
            .synchronize_time(timestamp_qds) \
            .convert_index_to_sec().dfs[0]

        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_rqd_df: pd.DataFrame = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=circuit_name, system=system, signal=signal, wildcard={'MAGNET': wildcard}) \
            .signal_query() \
            .synchronize_time(timestamp_qds) \
            .convert_index_to_sec().dfs[0]

        return i_meas_diode_df, u_diode_rqd_df

    def query_current_voltage_diode_leads_pm(self, timestamp_fgc: int, results_table: pd.DataFrame) \
            -> List[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
        """ Method querying PM for current and voltage of a diode

        :param timestamp_fgc: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_fgc: int
        :param results_table: analysis results table
        :type results_table: pd.DataFrame
        :return: List of tuples of current and voltage pd.DataFrames
        """
        i_meas_u_diode_u_ref_pm_dfs = []
        for index, row in tqdm(results_table.iterrows(), total=results_table.shape[0], desc='Querying U, I Diode PM'):
            source_qds, timestamp_qds, u_diode_db = row[['magnet', 'timestamp_iqps', 'nqps_db']]

            if u_diode_db == 'PM':
                source_timestamp_nqps_df = self.find_source_timestamp_nqps_pm(source_qds, timestamp_qds)

                #ToDo: cover with a decorator
                if not source_timestamp_nqps_df.empty:
                    timestamp_nqps_board_a = source_timestamp_nqps_df.loc[0, 'timestamp']
                    timestamp_nqps_board_b = source_timestamp_nqps_df.loc[1, 'timestamp']
                    i_meas_u_diode_u_ref_pm_df = self._query_i_u_diode_pm(self.circuit_name, source_qds, timestamp_fgc,
                                                                          timestamp_nqps_board_a,
                                                                          timestamp_nqps_board_b)
                else:
                    i_meas_u_diode_u_ref_pm_df = (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame)
            else:
                i_meas_u_diode_u_ref_pm_df = (pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame)

            i_meas_u_diode_u_ref_pm_dfs.append(i_meas_u_diode_u_ref_pm_df)

        return i_meas_u_diode_u_ref_pm_dfs

    def query_current_voltage_diode_leads_nxcals(self, results_table: pd.DataFrame, *,
                                                 spark) -> List[Tuple[pd.DataFrame, pd.DataFrame]]:
        """ Method querying NXCALS for current and voltage of a diode

        :param results_table: analysis results table
        :type results_table: pd.DataFrame
        :param spark: connector to spark and NXCALS queries
        :type spark: spark connector
        :return: List of tuples of current and voltage pd.DataFrames
        """
        i_meas_u_diode_nxcals_dfs = []
        for index, row in tqdm(results_table.iterrows(), total=results_table.shape[0],
                               desc='Querying U, I Diode leads NXCALS'):
            source_qds, timestamp_qds, u_diode_db = row[['magnet', 'timestamp_iqps', 'nqps_db']]

            timestamp_qds = Time.to_unix_timestamp(timestamp_qds)
            i_meas_u_diode_nxcals_df = self._query_i_u_diode_nxcals(self.circuit_name, source_qds,
                                                                    timestamp_qds, spark=spark)

            i_meas_u_diode_nxcals_dfs.append(i_meas_u_diode_nxcals_df)

        return i_meas_u_diode_nxcals_dfs

    # U_DIODE, U_EARTH
    def query_voltage_cals(self, system: str, signal: str, timestamp_sync: int, *, ldb, duration=[(50, 's'), (150, 's')],
                           verbose=True) -> List[pd.DataFrame]:
        """ Method querying CALS for voltage signals (either U_DIODE or U_EARTH)
        Signals are synchronized to the input timestamp

        :param system: name of a leads system to query according to metadata
        :type system: str
        :param signal: name of leads signal to query according to metadata
        :param timestamp_sync: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_sync: int
        :param ldb: CALS pytimber connector
        :type ldb: pytimber
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        return QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=timestamp_sync, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal, wildcard={'MAGNET': '*'}) \
            .signal_query(verbose=verbose) \
            .synchronize_time(timestamp_sync / 1e9).dfs

    # LEADS
    @check_nan_timestamp(return_type=pd.DataFrame())
    def find_timestamp_leads(self, t_start: int, system='LEADS', duration=[(10, 's'), (400, 's')],
                             verbose=True) -> pd.DataFrame:
        """ Method searching PM leads events
        If the timestamp is NaN, then an empty dataframe is returned.

        :param t_start: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type t_start: int
        :param system: name of a leads system to query according to metadata
        :type system: str
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: pd.DataFrame with source and timestamp of the PM LEADS events
        """
        return QueryBuilder().with_pm() \
            .with_duration(t_start=t_start, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system) \
            .event_query(verbose=verbose).df

    @check_nan_timestamp_signals(return_type=pd.DataFrame())
    def query_leads(self, timestamp_sync: int, source_timestamp_leads_df: pd.DataFrame, *, system='LEADS',
                    signal_names: List[str], spark, duration=[(30, 's'), (90, 's')],
                    verbose=True) -> List[pd.DataFrame]:
        """ Method querying leads signals from either PM or NXCALS
        If any timestamp is NaN, then a list of empty dataframes is returned.
        Signals are synchronized to input timestamp and index is converted to seconds.


        :param timestamp_sync: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_sync: int
        :param source_timestamp_leads_df:
        :param system: name of a leads system to query according to metadata
        :type system: str
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param spark: connector to spark and NXCALS queries
        :type spark: spark connector
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        if source_timestamp_leads_df.empty:
            return self._query_leads_nxcals(duration, signal_names, spark, system, timestamp_sync, verbose)
        else:
            return self._query_leads_pm(signal_names, source_timestamp_leads_df, system, timestamp_sync, verbose)

    def _query_leads_nxcals(self, duration: List[Tuple[int, str]], signal_names: List[str], spark, system: str,
                            timestamp_sync: int, verbose=True) -> List[pd.DataFrame]:
        """ Method querying NXCALS for leads signals
        Signals are synchronized to input timestamp and index is converted to seconds.

        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param spark: connector to spark and NXCALS queries
        :type spark: spark connector
        :param system: name of a leads system to query according to metadata
        :type system: str
        :param timestamp_sync: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_sync: int
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        return QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=timestamp_sync, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal_names) \
            .signal_query(verbose=verbose) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

    def _query_leads_pm(self, signal_names: List[str], source_timestamp_leads_df: pd.DataFrame, system: str,
                        timestamp_sync: int, verbose: bool) -> List[pd.DataFrame]:
        """ Method querying PM for leads signals.
        Signals are synchronized to input timestamp and index is converted to seconds.

        :param signal_names: list of signal names to query according to metadata description
        :type signal_names: List[str]
        :param source_timestamp_leads_df:
        :param system: name of a leads system to query according to metadata
        :type system: str
        :param timestamp_sync: FGC PM event timestamp (ns precision) to which the query is synchronised
        :type timestamp_sync: int
        :param verbose: flag to print information about the queried signal
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        source_leads = source_timestamp_leads_df.loc[0, 'source']
        timestamp_leads = source_timestamp_leads_df.loc[0, 'timestamp']
        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_leads) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system=system,
                           signal=signal_names, source=source_leads) \
            .signal_query(verbose=verbose) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs
