from typing import Tuple, Dict

from IPython.display import display, HTML
import scipy.signal as spsig
from copy import deepcopy
import numpy as np
import pandas as pd
import warnings

from lhcsmapi.analysis.decorators import check_dataframe_empty
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.PlotBuilder import PlotBuilder, create_title
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis, get_expert_decision
from lhcsmapi.reference.Reference import Reference


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return '%s\n' % (message)


warnings.formatwarning = warning_on_one_line


def find_time_of_i_a_i_ref_deviation(i_ref_df: pd.DataFrame, i_a_df: pd.DataFrame) -> float:
    """ Function calculating the time at which the reference and actual power converter currents start to diverge.
    The algorithm firstly takes an absolute value of both signals and re-interpolates them to the same sampling time.
    Then, the maximum difference for time range [-1, t=i_max] s is obtained.
    The time of deviation is equal to the first index for which the difference exceeds the maximum one before an event.

    :param i_ref_df: reference power converter current
    :type i_ref_df: pd.DataFrame
    :param i_a_df: actual power converter current (non-filtered, high sampling rate)
    :type i_a_df: pd.DataFrame

    :return: time at which I_A and I_REF signals start to diverge
    """

    # take absolute value of each signal
    i_ref_abs_df = abs(i_ref_df)
    i_a_abs_df = abs(i_a_df)
    # re-interpolate both signals
    idx = i_ref_abs_df.index.union(i_a_abs_df.index)
    signal_comparison_df = i_ref_abs_df.reindex(idx).interpolate('index')
    signal_comparison_df = signal_comparison_df.join(i_a_abs_df).dropna()

    # find maximum difference between the time of maximum I_REF value and 1 second before
    t_i_ref_max = i_ref_abs_df[(i_ref_abs_df.index > -1) & (i_ref_abs_df.index < 1)].idxmax().values[0]
    diff_df = abs(signal_comparison_df['STATUS.I_REF'] - signal_comparison_df['IAB.I_A'])
    max_diff = diff_df[(diff_df.index > (t_i_ref_max - 1)) & (diff_df.index < t_i_ref_max)].max()

    # find time of start of deviation from maximum difference between maximum value of I_REF and 0 (FGC timestamp)
    mask = (diff_df.index > t_i_ref_max) & (diff_df.index < 0) & (diff_df.values > max_diff)

    return diff_df[mask].idxmin()


def flip_u_qs0_signal(u_qs0_ref: pd.DataFrame) -> pd.DataFrame:
    """ Function flipping the U_QS0 signal such that the first signal increase has positive slope.

    :param u_qs0_ref: U_QS0 QPS signal
    :type u_qs0_ref: pd.DataFrame

    :return: flipped U_QS0 signal
    """

    # # if there is only a negative peak, then flip the signal
    # # if there are two peaks and the first is negative, then flip the signal
    # # added the case when U_QS0 is saturated only to negative!
    # # otherwise, do nothing
    if all(u_qs0_ref.max() < 0.2) and all(u_qs0_ref.min() < -0.2):
        return -deepcopy(u_qs0_ref)
    elif (u_qs0_ref.max().values[0] > 0.2) & (u_qs0_ref.min().values[0] < -0.2) & (
            u_qs0_ref.idxmin().values[0] < u_qs0_ref.idxmax().values[0]):
        return -deepcopy(u_qs0_ref)
    else:
        return deepcopy(u_qs0_ref)


# ToDo: Move to SignalDerivative class
def calculate_time_derivative_of_signal(df: pd.DataFrame, kernel_size=51) -> pd.DataFrame:
    """ Function calculating time derivatives for signal contained in the passed DataFrame object.
    The returned signal name contains prefix d w.r.t. the input signal name, e.g. I_MEAS -> dI_MEAS.

    :param df: input DataFrame
    :param kernel_size: kernel size for gradient calculation
    :type kernel_size: int

    :return: pd.DataFrame with signal derivative
    """
    if isinstance(df, pd.DataFrame) and (df.shape[1] == 1) and not df.empty:
        signal_name = df.columns[0]
        der_name = 'd{}'.format(signal_name)
        return pd.DataFrame(spsig.medfilt(np.gradient(df[signal_name]) / np.gradient(df.index), kernel_size), df.index,
                            columns=[der_name])
    else:
        raise ValueError('Wrong input type, please check the input assumptions')


def calculate_last_const_di_dt(i_meas_df: pd.DataFrame, const_di_dt_duration=100) -> Tuple[float, float]:
    """ Function calculating the value and duration of the last di/dt prior to an event in the input signal.

    :param i_meas_df: pd.DataFrame with power converter current
    :type i_meas_df: pd.DataFrame
    :param const_di_dt_duration: minimum duration of constant di/dt in samples
    :type const_di_dt_duration: int

    :return: a tuple of last di/dt value and duration
    """
    # Calculate the first derivative
    d_i_meas_df = calculate_time_derivative_of_signal(i_meas_df[i_meas_df.index < 0], kernel_size=51)
    # To remove noise round the derivative to a single digit
    d_i_meas_df['round_dSTATUS.I_MEAS'] = d_i_meas_df.apply(lambda row: round(row['dSTATUS.I_MEAS'], 1), axis=1)
    # Calculate the 2nd derivative
    dd_i_meas_df = calculate_time_derivative_of_signal(pd.DataFrame(d_i_meas_df['round_dSTATUS.I_MEAS']), kernel_size=1)
    # Count duration of 2nd derivative equal to 0
    dd_i_meas_df['count'] = (dd_i_meas_df['dround_dSTATUS.I_MEAS'].diff(1) != 0).astype('int').cumsum()
    # Only take 2nd derivatives longer then a given number of points
    thresh = dd_i_meas_df['count'].value_counts() > const_di_dt_duration
    max_counts = thresh[thresh == True].index
    # Take indices of constant 2nd values
    index_max_counts = [dd_i_meas_df[dd_i_meas_df['count'] == max_count].index[0] for max_count in max_counts]

    # Sort according to the index_max_counts
    index_max_counts_sorted = sorted(index_max_counts, reverse=True)

    # Return the last constant didt
    last_di_dt = d_i_meas_df[dd_i_meas_df.index == max(index_max_counts_sorted)]['round_dSTATUS.I_MEAS'].values[0]
    index_max_count = index_max_counts_sorted.index(max(index_max_counts_sorted))
    if (index_max_count == 0) or (index_max_count == len(index_max_counts_sorted) - 1):
        last_di_dt_duration = -max(index_max_counts)
    else:
        last_di_dt_duration = index_max_counts_sorted[index_max_count] - index_max_counts_sorted[index_max_count + 1]

    return last_di_dt, last_di_dt_duration


# ToDo: to FeatureBuilder
def max_abs_min(df: pd.DataFrame) -> float:
    """ Function calculating the maximum absolute value of the signal.
    Let min_df and max_df be the global minimum and maximum value of the input dataframe.
    Then, the returned value is min_df if |min_df| > max_df, otherwise the reutrned value is max_df.
    As a result, should the absolute value of the minimum current be greater than the maximum value, then the
    minimum value with the negative sign is returned.

    :param df: input signal
    :type df: pd.DataFrame

    :return: maximum absolute value of the input signal
    """
    return df.max().values[0] if df.max().values[0] > abs(df.min().values[0]) else df.min().values[0]


def timestamp_check(timestamp_dct: Dict[str, int], qds_a_name: str, qds_b_name: str, fgc_ee_name: str) -> None:
    """ Function checks timestamps of QPS, FGC, and EE for 600A circuits:
    For all circuit families: with/without EE, RCBXH/V, RCD/O the
    - for RCBXH/V and RCD/O circuits a check between QPS A and B boards and QPS board A and FGC is performed
      - the difference between timestamps of QDS board A and B is at maximum equal to 1 ms
      - the difference between timestamps of QDS board A and FGC is at maximum equal to 20 ms

    - in addition for circuits with EE a check between QDS board A and EE timestamps is performed
      - the difference between timestamps of QDS board A and EE is at maximum equal to 20 ms

    :param timestamp_dct: a dictionary from timestamp name to its value (it is assumed that does not contain NaN)
    :type timestamp_dct: Dict[str, int]
    :param qds_a_name: name of QDS board A timestamp
    :type qds_a_name: str
    :param qds_b_name: name of QDS board B timestamp
    :type qds_b_name: str
    :param fgc_ee_name: name of either FGC or EE timestamp
    :type fgc_ee_name: str
    """
    if (qds_a_name in timestamp_dct) and (qds_b_name in timestamp_dct):
        if timestamp_dct[qds_b_name] - timestamp_dct[qds_a_name] != int(1e6):
            warnings.warn('QDS PM events (board A and B) are not synchronized to 1 ms delay!')
    else:
        warnings.warn('At least one QDS PM event missing!', stacklevel=2)

    fgc_or_ee = 'FGC' if 'FGC' in fgc_ee_name else 'EE'
    if (qds_a_name in timestamp_dct) and (fgc_ee_name in timestamp_dct):
        if abs(timestamp_dct[qds_a_name] - timestamp_dct[fgc_ee_name]) > int(20e6):
            warnings.warn('QDS_A and %s PM events are not synchronized to +/- 20 ms delay!' % fgc_or_ee)
    else:
        warnings.warn('Either QDS_A or %s PM event is missing. '
                      'Can not check synchronization of QDS and FGC timestamps!' % fgc_or_ee)


class _600ACircuitAnalysis(CircuitAnalysis):
    """Class for analysis of 600A circuits supporting three distinct cases:
    - circuits with/without EE
    - RCBXH/V circuits
    - RCD/O circuits
    """

    # Timestamp table
    def create_timestamp_table(self, timestamp_dct: Dict[str, int], circuit_name='') -> None:
        """ Method creating timestamp table as well as checking the timing between considered timestamps.
        Depending on the type of 600A circuit, an appropriate check is performed.

        :param timestamp_dct: a dictionary from timestamp name to its value (it is assumed that does not contain NaN)
        :type timestamp_dct: Dict[str, int]
        :param circuit_name: name of a circuit
        :type circuit_name: str
        """
        timestamp_table_df = super().create_timestamp_table(timestamp_dct)

        display(HTML(timestamp_table_df.to_html()))

        if timestamp_table_df.loc[0, 'Source'] == 'FGC':
            warnings.warn('FGC is the first timestamp, potentially a PC trip not a quench!', stacklevel=0)

        if 'RCD' in circuit_name:
            timestamp_check(timestamp_dct, qds_a_name='QDS_A_RCD', qds_b_name='QDS_B_RCD', fgc_ee_name='FGC_RCD')
            timestamp_check(timestamp_dct, qds_a_name='QDS_A_RCO', qds_b_name='QDS_B_RCO', fgc_ee_name='FGC_RCO')
        elif 'RCBX' in circuit_name:
            timestamp_check(timestamp_dct, qds_a_name='QDS_A_RCBXH', qds_b_name='QDS_B_RCBXH', fgc_ee_name='FGC_RCBXH')
            timestamp_check(timestamp_dct, qds_a_name='QDS_A_RCBXV', qds_b_name='QDS_B_RCBXV', fgc_ee_name='FGC_RCBXV')
        else:
            timestamp_check(timestamp_dct, qds_a_name='QDS_A', qds_b_name='QDS_B', fgc_ee_name='FGC')
            if 'EE' in SignalMetadata.get_system_types_per_circuit_name(self.circuit_type, circuit_name):
                timestamp_check(timestamp_dct, qds_a_name='QDS_A', qds_b_name='QDS_B', fgc_ee_name='EE')

    # PC
    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, PC trigger analysis skipped!')
    def analyze_i_meas_pc_miits(self, i_meas_df: pd.DataFrame, i_a_df: pd.DataFrame, i_ref_df: pd.DataFrame,
                                col_name='I_MEAS MIITs') -> None:
        """ Method calculating quench load (MIITs) of the power converter current.
        If any input signal is missing, then the analysis is aborted.
        The calculated MIITs is set in the results table.

        :param i_meas_df: measured and filtered power converter current
        :type i_meas_df: pd.DataFrame
        :param i_a_df: actual and un-filtered power converter current
        :type i_a_df: pd.DataFrame
        :param i_ref_df:  reference power converter current
        :type i_ref_df: pd.DataFrame
        :param col_name: name of a column to set the output results table
        :type col_name: str
        """
        t_deviation = find_time_of_i_a_i_ref_deviation(i_ref_df, i_a_df)
        i_meas_pos_df = i_meas_df[i_meas_df.index > t_deviation]
        miits_i_meas = np.trapz(i_meas_pos_df.values.reshape(len(i_meas_pos_df.index)) ** 2,
                                i_meas_pos_df.index) * 1e-6
        print('The I_MEAS(t) MIITs is %s MA^2s.' % round(miits_i_meas, 3))
        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = round(miits_i_meas, 3)

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, PC trigger analysis skipped!')
    def analyze_i_meas_pc_trigger(self, timestamp_fgc: int, events_action_df: pd.DataFrame,
                                  events_symbol_df: pd.DataFrame) -> None:
        """ Method finding and displaying the source of a PC trigger.
        If any input signal is missing, then the analysis is aborted.

        :param timestamp_fgc: FGC event timestamp
        :type timestamp_fgc: int
        :param events_action_df: PC events action
        :type events_action_df: pd.DataFrame
        :param events_symbol_df: PC events symbol
        :type events_symbol_df: pd.DataFrame
        """
        event_pc_df = pd.concat([events_action_df, events_symbol_df], axis=1)
        fault_event_pc_df = event_pc_df[event_pc_df['EVENTS.ACTION'] == 'FAULT']
        if not fault_event_pc_df.empty:
            array = np.asarray(fault_event_pc_df.index)
            idx = (np.abs(array - timestamp_fgc)).argmin()

            print('The reason for PC PM event is:')
            display(HTML(fault_event_pc_df[fault_event_pc_df.index == fault_event_pc_df.index[idx]].to_html()))

    @check_dataframe_empty(mode='all', warning='I_MEAS signal is empty, dI_MEAS/dt analys skipped!')
    def analyze_i_meas_pc_didt(self, i_meas_df: pd.DataFrame, col_name='Ramp Rate') -> None:
        """ Method analyzing the power converter current derivative.
        If the input signal is missing, then the analysis is skipped.
        In case the current derivative prior to an event was zero (less than 0.1 A/s), then
        - the value of the current plateau and its duration are reported in the results table
        otherwise only the value of the derivative (ramp-rate) is reported in the results table

        :param i_meas_df: measured power converter current
        :type i_meas_df: pd.DataFrame
        :param col_name: name of a column to set in results table
        :type col_name: str
        """
        # Ramp Rate and Plateau Duration
        pd.options.mode.chained_assignment = None
        idx = max([i_meas_df.idxmin().values[0], i_meas_df.idxmax().values[0]])
        i_meas_df_der = i_meas_df[i_meas_df.index < idx]

        last_di_dt, last_di_dt_duration = calculate_last_const_di_dt(i_meas_df_der)
        # Store analysis result
        if self.results_table is not None:
            if abs(last_di_dt) > 0.1:
                self.results_table[col_name] = last_di_dt
            else:
                self.results_table[col_name] = 0
                self.results_table['Plateau Duration'] = max(last_di_dt_duration,
                                                             self.results_table['Plateau Duration'].values[0])
        pd.options.mode.chained_assignment = 'warn'

    @check_dataframe_empty(mode='any', warning='At least one DataFrame is empty, I_MEAS quench analysis skipped!')
    def analyze_i_meas_pc_quench_current(self, i_meas_df: pd.DataFrame, i_ref_df: pd.DataFrame, i_a_df: pd.DataFrame,
                                         col_name='Quench Current') -> None:
        """ Method calculating the quench current, i.e., the current at the moment when I_A and I_REF start to
        deviate from one another.
        If any input signal is missing, then the analysis is skipped.
        The calculated quench current is set in the results table.

        :param i_meas_df: measured power converter current
        :type i_meas_df: pd.DataFrame
        :param i_ref_df: reference power converter current
        :type i_ref_df: pd.DataFrame
        :param i_a_df: actual power converter current
        :type i_a_df: pd.DataFrame
        :param col_name: name of a column to set in results table
        :type col_name: str
        """
        t_deviation = find_time_of_i_a_i_ref_deviation(i_ref_df, i_a_df)
        i_meas_pos_df = i_meas_df[i_meas_df.index > t_deviation]
        print('Quench Current is %s A.' % round(i_meas_pos_df['STATUS.I_MEAS'].iloc[0], 1))
        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = round(i_meas_pos_df['STATUS.I_MEAS'].iloc[0], 1)

    @check_dataframe_empty(mode='any', warning='I_MEAS, I_REF, I_A signals are empty, analysis skipped!')
    def plot_i_meas_pc(self, circuit_name: str, timestamp_fgc: int, i_meas_df: pd.DataFrame, i_ref_df: pd.DataFrame,
                       i_a_df: pd.DataFrame) -> None:
        """ Method plotting power converter currents in two views:
        - global view of the full signals
        - zoom view around the moment of the deviation between the I_A and I_REF currents
        If any input signal is missing, then the analysis is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param i_meas_df: measured power converter current
        :type i_meas_df: pd.DataFrame
        :param i_ref_df: reference power converter current
        :type i_ref_df: pd.DataFrame
        :param i_a_df: actual power converter current
        :type i_a_df: pd.DataFrame
        """
        title = create_title(circuit_name, timestamp_fgc, ['I_MEAS', 'I_REF', 'I_A'])

        # Global view
        PlotBuilder().with_signal([i_a_df, i_meas_df, i_ref_df], title=title, figsize=(20, 7), grid=True) \
                .with_ylabel(ylabel='I, [A]') \
            .plot()

        # Zoomed view
        # # in x
        t_deviation = find_time_of_i_a_i_ref_deviation(i_ref_df, i_a_df)
        # # in y
        idx_i_a = i_a_df.index.get_loc(t_deviation, 'nearest')
        mid = i_a_df[i_a_df.index == i_a_df.index[idx_i_a]].values[0][0]

        PlotBuilder().with_signal([i_meas_df, i_a_df, i_ref_df], title=title, figsize=(15, 15), grid=True) \
                .with_xlim([t_deviation - 0.01, 0.01]) \
                .with_ylabel(ylabel='I, [A]').with_ylim([mid - 10, mid + 10]) \
                .with_axvline(x=t_deviation, ls='--') \
            .plot()

    @check_dataframe_empty(mode='any', warning='I_EARTH signal is empty, analysis skipped!')
    def analyze_i_earth_pc(self, circuit_name: str, timestamp_fgc: int, i_earth_df: pd.DataFrame,
                           col_name='Earth Current') -> None:
        """ Method plotting the earth current and calculating its absolute maximum value.
        If the input signal is missing, then the analysis is skipped.
        The calculated absolute maximum value is set in the results table.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param i_earth_df: current of the earth circuit in the power converter
        :type i_earth_df: pd.DataFrame
        :param col_name: name of a column to set in results table
        :type col_name: str
        """
        i_earth_max = max_abs_min(i_earth_df)
        print('The maximum earth current is %s A.' % round(i_earth_max, 3))
        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = round(i_earth_max, 3)

        # Plot
        title = create_title(circuit_name, timestamp_fgc, 'I_EARTH')
        PlotBuilder().with_signal(i_earth_df, title=title, grid=True) \
                .with_ylabel(ylabel='IEARTH [A]') \
            .plot()

    # EE
    @check_dataframe_empty(mode='any', warning='I_MEAS, U_DUMP_RES signals are empty, analysis skipped!')
    def analyze_u_dump_res_ee(self, circuit_name: str, timestamp_fgc: int, i_meas_df: pd.DataFrame,
                              u_dump_res_df: pd.DataFrame) -> None:
        """ Method plotting and calculating the maximum energy extraction voltage.
        If any input signal is missing, then the analysis is skipped.
        The calculated maximum value is set in the results table.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_fgc: timestamp of the FGC PM event (for the title)
        :type timestamp_fgc: int
        :param i_meas_df: measured power converter current
        :type i_meas_df: pd.DataFrame
        :param u_dump_res_df: energy extraction voltage
        :type u_dump_res_df: pd.DataFrame
        """
        # Analyze
        u_dump_res_max = max_abs_min(u_dump_res_df)
        print('The maximum EE voltage is %s V.' % round(u_dump_res_max, 3))

        # Store analysis result
        if self.results_table is not None:
            self.results_table['EE max voltage'] = round(u_dump_res_max, 3)


        # Plot
        title = create_title(circuit_name, timestamp_fgc, ['I_MEAS', 'U_DUMP_RES'])
        PlotBuilder().with_signal(i_meas_df, title=title, figsize=(20, 7), grid=True) \
                .with_ylabel(ylabel='I_MEAS, [A]') \
                .with_xlim([-1, 2]) \
            .with_signal(u_dump_res_df) \
                .with_ylabel(ylabel='U_DUMP_RES, [V]') \
            .plot()

    # QPS
    @check_dataframe_empty(mode='all', warning='U_RES, U_DIFF, I_DCCT, I_DIDT signals are empty, analysis skipped!')
    def analyze_qds(self, circuit_name: str, timestamp_qps: int, i_dcct_df: pd.DataFrame, i_didt_df: pd.DataFrame,
                    u_diff_df: pd.DataFrame, u_res_df: pd.DataFrame, col_names=('Reason', 'Quench Origin')) -> None:
        """Method plotting the QDS signals.
        If any input signal is missing, then the analysis is skipped.
        The calculated maximum value is set in the results table.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_sync: timestamp of the FGC PM event (for the title)
        :type timestamp_sync: int
        :param i_dcct_df: current of the DCCT sensor
        :type i_dcct_df: pd.DataFrame
        :param i_didt_df: signal of the current derivative sensor
        :type i_didt_df: pd.DataFrame
        :param u_diff_df: differential voltage
        :type u_diff_df: pd.DataFrame
        :param u_res_df: resistive voltage
        :type u_res_df: pd.DataFrame
        :param col_names: names of columns to set in results table (first element of the tuple is the reason for quench,
        the second element is the Quench Origin)
        :type col_names: Tuple[str, str]
        """
        title = create_title(circuit_name, timestamp_qps, ['U_RES', 'U_DIFF', 'I_DCCT', 'I_DIDT'])
        PlotBuilder().with_signal(u_diff_df, title=title, figsize=(20, 7), color=['C0'], grid=True) \
                .with_ylabel(ylabel='U_DIFF, [V]').with_ylim([-10, 10]) \
                .with_legend(loc='upper left') \
            .with_signal(i_dcct_df, color=['C2']) \
                .with_ylabel(ylabel='I_DCCT, [A]') \
                .with_legend(loc='upper right') \
            .with_signal(i_didt_df, color=['C3']) \
                .with_ylabel(ylabel='I_DIDT, [A/s]') \
                .with_legend(loc='lower right') \
                .with_spines(axes='right', position=1.1) \
            .with_signal(u_res_df, color=['C1']) \
                .with_ylabel(ylabel='U_RES, [V]').with_ylim([-0.25, 0.25]) \
                .with_xlim([-0.1, 0.1]) \
                .with_legend(loc='lower left') \
                .with_spines(axes='right', position=-0.11) \
                .with_tick_params(axis="y", direction="in", pad=-50) \
                .with_ylabel_coords(position=(-0.175, 0.5)) \
            .plot()

        # Get expert input on reason
        self.results_table[col_names[0]] = get_expert_decision('Reason for FPA: ', ['quench', 'coupling', 'other'])

        # Get expert input on quench origin
        self.results_table[col_names[1]] = get_expert_decision('Origin of a quench', ['magnet', 'busbar', 'other'])

    @check_dataframe_empty(mode='all', warning='U_RES signal is empty, analysis skipped!')
    def plot_u_res(self, circuit_name: str, timestamp_qps: int, u_res_nxcals_df: pd.DataFrame) -> None:
        """ Method plotting the resistive voltage of Board A and B prior to a quench event
        If the input signal is missing, then the analysis is skipped.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_qps: timestamp of the QPS PM event (for the title)
        :type timestamp_qps: int
        :param u_res_nxcals_df: resistive voltage signal
        :type u_res_nxcals_df: pd.DataFrame
        """
        title = create_title(circuit_name, timestamp_qps, 'U_RES')

        PlotBuilder().with_signal(u_res_nxcals_df, title=title, grid=True) \
            .with_ylabel(ylabel='U_RES, [V]').with_ylim([-0.25, 0.25]) \
            .plot()

    @check_dataframe_empty(mode='all', warning='U_RES signal is empty, analysis skipped!')
    def calculate_u_res_slope(self, circuit_name: str, timestamp_qps: int, u_res_df: pd.DataFrame,
                              col_name='dUres/dt') -> None:
        """ Method calculating and plotting the slope of the resistive voltage increase
        If any input signal is missing, then the analysis is skipped.
        The calculated voltage derivative is set in the results table.

        :param circuit_name: name of the analyzed circuit (for the title)
        :type circuit_name: str
        :param timestamp_qps: timestamp of the QPS PM event (for the title)
        :type timestamp_qps: int
        :param u_res_df: resistive voltage signal
        :type u_res_df: pd.DataFrame
        :param col_name: name of a column to set the output results table
        :type col_name: str
        """
        # Skip first ten points to avoid initial noise
        u_res_slice_df = u_res_df[u_res_df.index > u_res_df.index[10]]
        # Take absolute value
        u_res_flip_df = flip_u_qs0_signal(u_res_slice_df)

        # Find first index of argmax (0.2 V)
        t_u_max = u_res_flip_df[u_res_flip_df >= 0.2].dropna().index[0]
        v_u_max = u_res_df[u_res_df.index == t_u_max].dropna().values[0]

        # Find index of voltage equal to 0.025 V
        temp = u_res_flip_df[(u_res_flip_df.index < t_u_max)]
        t_u_min = temp[(temp.values < 0.025)].index[-1]
        v_u_min = u_res_df[u_res_df.index == t_u_min].dropna().values[0][0]

        # Calculate dU_RES/dt
        d_u_res_dt = (v_u_max - v_u_min) / (t_u_max - t_u_min)
        print('The slope of the U_RES signal is %s V/s.' % round(d_u_res_dt[0], 3))
        # Store analysis result
        if self.results_table is not None:
            self.results_table[col_name] = d_u_res_dt[0]

        # Plot
        title = create_title(circuit_name, timestamp_qps, 'U_RES')
        slope_df = pd.DataFrame(data=[v_u_min, v_u_max[0]], index=[t_u_min, t_u_max], columns=['dU_RES/dt'])

        PlotBuilder().with_signal([u_res_slice_df, slope_df], title=title, grid=True) \
            .with_ylabel(ylabel='U_RES, [V]') \
            .plot()

    # MP3 Results Tables
    def create_mp3_results_table(self) -> pd.DataFrame:
        """ Method creating an MP3-compatible quench database row for 600A circuits with/without EE

        :return: pd.DataFrame with the results of analysis calculation
        """
        columns_mp3 = Reference.read_quench_database_columns(circuit_type='600A', table_type='MP3')
        mp3_table = pd.DataFrame(columns=columns_mp3, index=[0])

        # union of MP3 and results table columns
        columns_lhcsm = Reference.read_quench_database_columns(circuit_type='600A', table_type='LHCSM')
        same_columns = list(set(columns_mp3).intersection(set(columns_lhcsm)))
        mp3_table[same_columns] = self.results_table[same_columns]

        mp3_table['DCCT MIITs'] = self.results_table['I_MEAS MIITs']
        mp3_table['Person who analyzed'] = self.results_table['performed_by']

        return mp3_table

    def create_mp3_results_table_rcdo(self) -> pd.DataFrame:
        """ Method creating an MP3-compatible quench database row for RCD/RCO 600A circuits

        :return: pd.DataFrame with the results of analysis calculation
        """
        columns_mp3 = Reference.read_quench_database_columns(circuit_type='600A_RCDO', table_type='MP3')
        mp3_table = pd.DataFrame(columns=columns_mp3, index=[0])

        # union of MP3 and results table columns
        columns_lhcsm = Reference.read_quench_database_columns(circuit_type='600A_RCDO', table_type='LHCSM')
        same_columns = list(set(columns_mp3).intersection(set(columns_lhcsm)))
        mp3_table[same_columns] = self.results_table[same_columns]

        mp3_table['Reason'] = 'RCO: ' + self.results_table['Reason RCO'] + \
                              ' RCD: ' + self.results_table['Reason RCD']
        mp3_table['Quench Origin'] = 'RCO: ' + self.results_table['Quench Origin RCO'] + \
                                     ' RCD: ' + self.results_table['Quench Origin RCD']
        mp3_table['Person who analyzed'] = self.results_table['performed_by']

        return mp3_table

    def create_mp3_results_table_rcbxhv(self) -> pd.DataFrame:
        """ Method creating an MP3-compatible quench database row for RCBXH/RCBXV 600A circuits

        :return: pd.DataFrame with the results of analysis calculation
        """
        columns_mp3 = Reference.read_quench_database_columns(circuit_type='600A_RCBXHV', table_type='MP3')
        mp3_table = pd.DataFrame(columns=columns_mp3, index=[0])

        # union of MP3 and results table columns
        columns_lhcsm = Reference.read_quench_database_columns(circuit_type='600A_RCBXHV', table_type='LHCSM')
        same_columns = list(set(columns_mp3).intersection(set(columns_lhcsm)))

        mp3_table[same_columns] = self.results_table[same_columns]

        i_q_h = 0 if np.isnan(self.results_table['I_Q_H'].values[0]) else self.results_table['I_Q_H'].values[0]
        i_q_v = 0 if np.isnan(self.results_table['I_Q_V'].values[0]) else self.results_table['I_Q_V'].values[0]
        mp3_table['radius'] = (i_q_h ** 2 + i_q_v ** 2) ** 0.5
        mp3_table['Reason'] = 'H: ' + self.results_table['Reason H'] + \
                              ' V: ' + self.results_table['Reason V']
        mp3_table['Quench Origin'] = 'H: ' + self.results_table['Quench Origin H'] + \
                                     ' V: ' + self.results_table['Quench Origin V']
        mp3_table['Person who analyzed'] = self.results_table['performed_by']

        return mp3_table
