from typing import List

import pandas as pd
import numpy as np
import warnings
import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery
from lhcsmapi.reference.Reference import Reference


class R600ACircuitQuery(CircuitQuery):
    """Class providing extending the CircuitQuery class with methods for creation of templates of analysis tables
    in 600A circuits (with/without EE, RCBXH/V, RCD/O).

    """

    def create_report_analysis_template(self, timestamp_fgc: int, author='') -> pd.DataFrame:
        """ Method creating a report analysis template for 600A circuits with/without EE

        :param timestamp_fgc: timestamp of an FGC event
        :type timestamp_fgc: int
        :param author: NICE account name of the analysis author
        :type author: str
        :return: pre-filled pd.DataFrame with a template to be filled further with data from an analysis of
         a powering event in 600A circuits with/without EE
        """
        columns = Reference.read_quench_database_columns(circuit_type='600A', table_type='LHCSM')
        results_table = pd.DataFrame(columns=columns, index=[0])
        results_table['Circuit Name'] = self.circuit_name
        results_table['Family'] = self.circuit_name.split('.')[0]
        results_table['Date'] = Time.to_string_short(timestamp_fgc).split(' ')[0]
        results_table['Time'] = Time.to_string_short(timestamp_fgc).split(' ')[1]
        results_table['performed_by'] = author

        results_table['lhcsmapi_version'] = lhcsmapi.__version__

        return results_table

    def create_report_analysis_template_rcd(self, timestamp_fgc_rcd: int, timestamp_fgc_rco: int,
                                            circuit_names: List[str], author='') -> pd.DataFrame:
        """ Method creating a report analysis template for RCD/RCO 600A circuits

        :param timestamp_fgc_rcd: timestamp of an FGC event in RCD circuit
        :type timestamp_fgc_rcd: int
        :param timestamp_fgc_rco: timestamp of an FGC event in RCO circuit
        :type timestamp_fgc_rco: int
        :param circuit_names: list of RCD/RCO circuit names [RCD.xxx, RCO.xxx]
        :type: List[str]
        :param author: NICE account name of the analysis author
        :type author: str
        :return: pre-filled pd.DataFrame with a template to be filled further with data from analysis of
         a powering event in RCD/RCO 600A circuits
        """
        columns = Reference.read_quench_database_columns(circuit_type='600A_RCDO', table_type='LHCSM')
        results_table = pd.DataFrame(columns=columns, index=[0])

        results_table['Family'] = 'RCD'
        timestamp_fgc = min(timestamp_fgc_rcd, timestamp_fgc_rco)
        if not np.isnan(timestamp_fgc):
            results_table['Date'] = Time.to_string_short(timestamp_fgc).split(' ')[0]
            results_table['Time'] = Time.to_string_short(timestamp_fgc).split(' ')[1]

        if not np.isnan(timestamp_fgc_rcd) and not np.isnan(timestamp_fgc_rco):
            results_table['Circuit Name'] = circuit_names[0].replace('RCD', 'RCD/O')
        elif not np.isnan(timestamp_fgc_rcd):
            results_table['Circuit Name'] = circuit_names[0]
        elif not np.isnan(timestamp_fgc_rco):
            results_table['Circuit Name'] = circuit_names[1]
        else:
            warnings.warn('Both RCD and RCO FGC timestamps are missing.')

        results_table['performed_by'] = author
        results_table['lhcsmapi_version'] = lhcsmapi.__version__

        return results_table

    def create_report_analysis_template_rcbx(self, timestamp_fgc_rcbxh: int, timestamp_fgc_rcbxv: int,
                                             circuit_names: List[str], author='') -> pd.DataFrame:
        """ Method creating a report analysis template for RCBXH/RCBXV 600A circuits

        :param timestamp_fgc_rcbxh: timestamp of an FGC event in RCBXH circuit
        :type timestamp_fgc_rcbxh: int
        :param timestamp_fgc_rcbxv: timestamp of an FGC event in RCBXV circuit
        :type timestamp_fgc_rcbxv: int
        :param circuit_names: list of RCBXH/RCBXV circuit names [RCBXH.xxx, RCBXV.xxx]
        :type: List[str]
        :param author: NICE account name of the analysis author
        :type author: str
        :return: pre-filled pd.DataFrame with a template to be filled further with data from analysis
        of a powering event in RCBXH/RCBXV 600A circuits
        """
        columns = Reference.read_quench_database_columns(circuit_type='600A_RCBXHV', table_type='LHCSM')
        results_table = pd.DataFrame(columns=columns, index=[0])

        results_table['Family'] = 'MCBX'
        timestamp_fgc = min(timestamp_fgc_rcbxh, timestamp_fgc_rcbxv)
        if not np.isnan(timestamp_fgc):
            results_table['Date'] = Time.to_string_short(timestamp_fgc).split(' ')[0]
            results_table['Time'] = Time.to_string_short(timestamp_fgc).split(' ')[1]

        if not np.isnan(timestamp_fgc_rcbxh) and not np.isnan(timestamp_fgc_rcbxv):
            results_table['Circuit Name'] = circuit_names[0].replace('RCBXH', 'RCBXV')
        elif not np.isnan(timestamp_fgc_rcbxh):
            results_table['Circuit Name'] = circuit_names[0]
        elif not np.isnan(timestamp_fgc_rcbxv):
            results_table['Circuit Name'] = circuit_names[1]
        else:
            warnings.warn('Both RCBXH and RCBXV FGC timestamps are missing.')

        results_table['performed_by'] = author
        results_table['lhcsmapi_version'] = lhcsmapi.__version__

        return results_table
