import math
from copy import deepcopy
import pandas as pd
import matplotlib.pyplot as plt
from IPython.display import display, HTML

from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis, check_show_next, get_expert_decision, get_expert_comment
from lhcsmapi.dbsignal.ExpSignalAnalysis import ExpSignalAnalysis
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.AssertionBuilder import AssertionBuilder
from lhcsmapi.pyedsl.FeatureBuilder import FeatureBuilder
from lhcsmapi.reference.Reference import Reference
from lhcsmapi.metadata.MappingMetadata import MappingMetadata


def get_df_col(dfs, col):
    for df in dfs:
        if col in df.columns[0]:
            return df, df.columns[0]
    return None, None


def calculate_resistance(u_i_hds_dfs):
    no_signal = int(len(u_i_hds_dfs) / 2)
    r_hds_dfs = []
    for index in range(1, no_signal + 1):
        u_hds_df, col_u_hds_df = get_df_col(u_i_hds_dfs, 'U_HDS_{}'.format(index))
        i_hds_df, col_i_hds_df = get_df_col(u_i_hds_dfs, 'I_HDS_{}'.format(index))

        r_value = u_hds_df.values / i_hds_df.values
        column = col_u_hds_df.replace('U', 'R')
        time = u_hds_df.index
        r_hds_dfs.append(pd.DataFrame(data=r_value, index=time, columns=[column]))
    return r_hds_dfs


def plot_qh_discharge(circuit_type, timestamp_qh, source_qh, u_i_hds_dfs, r_hds_dfs, u_i_hds_ref_dfs, r_hds_ref_dfs):
    no_subplots = 3 if circuit_type == "RB" else 1
    no_signal = int(len(u_i_hds_dfs) / 2)

    fig, ax = plt.subplots(1, no_subplots, figsize=(20, 7))
    for index in range(1, no_signal + 1):
        u_hds_df = get_df_col(u_i_hds_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'U_HDS_{}'.format(index))[0]
        u_hds_df.plot(ax=ax[0])
        u_hds_ref_df.plot(ax=ax[0], style='--')
    ax[0].grid(True)
    ax[0].tick_params(labelsize=15)
    ax[0].set_xlabel("time, [s]", fontsize=15)
    ax[0].set_ylabel("U_HDS, [V]", fontsize=15)

    if circuit_type == "RB":
        title = "Magnet: {}, Time Stamp: {}, U_HDS(t), I_HDS(t), and R_HDS(t)". \
            format(source_qh, Time.to_string_short(timestamp_qh))

        for index in range(1, no_signal + 1):
            i_hds_df = get_df_col(u_i_hds_dfs, 'I_HDS_{}'.format(index))[0]
            i_hds_ref_df = get_df_col(u_i_hds_ref_dfs, 'I_HDS_{}'.format(index))[0]
            i_hds_df.plot(ax=ax[1], title=title)
            i_hds_ref_df.plot(ax=ax[1], style='--')

        ax[1].title.set_size(20)
        ax[1].grid(True)
        ax[1].tick_params(labelsize=15)
        ax[1].set_xlabel("time, [s]", fontsize=15)
        ax[1].set_ylabel("I_HDS, [A]", fontsize=15)

        for r_hds_df, r_hds_ref_df in zip(r_hds_dfs, r_hds_ref_dfs):
            r_hds_df.plot(ax=ax[2])
            r_hds_ref_df.plot(ax=ax[2], style='--')
        ax[2].grid(True)
        ax[2].tick_params(labelsize=15)
        ax[2].set_xlabel("time, [s]", fontsize=15)
        ax[2].set_ylabel("R_HDS (Calculated), [Ohm]", fontsize=15)

    plt.show()


def get_max_diff_val_and_idx(u_nqps_diff_dfs, t_st_magnet_ok, source_qds, quenched_magnets):
    # take only certain period of time
    # # a fix in case after t_st_magnet_ok QH resulted in voltage increase beyond the threshold
    u_nqps_diff_slice_dfs = []
    for u_nqps_diff_df in u_nqps_diff_dfs:
        mask = (u_nqps_diff_df.index > t_st_magnet_ok - 0.2) & (u_nqps_diff_df.index < t_st_magnet_ok)
        u_nqps_diff_slice_dfs.append(u_nqps_diff_df[mask])

    if all([u_nqps_diff_slice_df.empty for u_nqps_diff_slice_df in u_nqps_diff_slice_dfs]):
        return -math.inf, 0

    u_nqps_diff_no_quench_dfs = []
    for u_nqps_diff_slice_df in u_nqps_diff_slice_dfs:
        col = u_nqps_diff_slice_df.columns[0]
        if 'U_REF' in col or not all([quenched_magnet in col for quenched_magnet in quenched_magnets]):
            u_nqps_diff_no_quench_dfs.append(u_nqps_diff_slice_df)

    # ToDo with FeatureBuilder
    max_idx = [u_nqps_diff_no_quench_df.max(axis=1).idxmax() for u_nqps_diff_no_quench_df in
               u_nqps_diff_no_quench_dfs]
    max_val = [u_nqps_diff_no_quench_df.max(axis=0).max() for u_nqps_diff_no_quench_df in u_nqps_diff_no_quench_dfs]
    idx_max_val = max_val.index(max(max_val))
    return max_idx[idx_max_val], max_val[idx_max_val]


def analyze_qds_trigger(u_qs0_df, st_magnet_ok_df, st_nqd0_df):

    mask = (st_magnet_ok_df.index > st_magnet_ok_df.index[10]) & \
           (st_magnet_ok_df.values == False).reshape(len(st_magnet_ok_df), )
    st_qds_sub_df = st_magnet_ok_df[mask]

    if len(st_qds_sub_df) > 0:
        t_st_magnet_ok = st_qds_sub_df.index[0]

        mask = (st_nqd0_df.index > st_nqd0_df.index[10]) & \
               (st_nqd0_df.index < t_st_magnet_ok) & \
               (st_nqd0_df.values == True).reshape(len(st_nqd0_df), )

        t_st_nqd0 = st_nqd0_df[mask].index[-1]
        t_st_nqd0 = st_nqd0_df.index[st_nqd0_df.index.get_loc(t_st_nqd0) + 1]

        u_st_nqd0 = u_qs0_df[u_qs0_df.index == t_st_nqd0].values[0][0]

        if min(abs(u_qs0_df).values)[0] <= 0.01:
            u_qs0_before_quench_df = u_qs0_df[u_qs0_df.index < t_st_nqd0]

            mask = (abs(u_qs0_before_quench_df - u_qs0_before_quench_df.values[0]) <= 0.01).values.reshape(
                len(u_qs0_before_quench_df), )
            t_start_quench = u_qs0_before_quench_df[mask].index[-1]

            t_start_quench = u_qs0_df.index[u_qs0_df.index.get_loc(t_start_quench) - 1]
        else:
            t_start_quench = 0

        u_start_quench = \
            u_qs0_df.loc[u_qs0_df.index[u_qs0_df.index.get_loc(t_start_quench, method='nearest')]].values[0]

        du_dt = (u_st_nqd0 - u_start_quench) / (t_st_nqd0 - t_start_quench)

        t_delay_qh_trigger = -t_st_magnet_ok
        analysis_comment = ''
    else:
        t_st_magnet_ok = u_qs0_df.index[-1] - 0.075
        t_st_nqd0 = u_qs0_df.index[-1] - 0.075
        t_start_quench = u_qs0_df.index[0] + 0.025
        u_start_quench = math.nan
        du_dt = 0
        t_delay_qh_trigger = math.nan
        analysis_comment = 'Seems that the heaters were triggered by nQPS'

    COLUMN_NAMES = ["t_st_magnet_ok [s]", "t_st_nqd0 [s]", "u_st_nqd0 [V]", "t_start_quench [s]", "du_dt [V/s]",
                    "t_delay_qh_trigger [s]", "analysis_comment"]
    value = [t_st_magnet_ok, t_st_nqd0, u_start_quench, t_start_quench, du_dt, t_delay_qh_trigger, analysis_comment]
    return pd.DataFrame(columns=COLUMN_NAMES, data=[value])


def find_dfs_with_quenched_magnet(u_nqps_shift_diff_df, source_qds):
    u_nqps_shift_diff_filt_dfs = []
    for u_nqps_shift_diff_df_el in u_nqps_shift_diff_df:
        if source_qds in u_nqps_shift_diff_df_el.columns.values[0]:
            u_nqps_shift_diff_filt_dfs.append(u_nqps_shift_diff_df_el)
    return u_nqps_shift_diff_filt_dfs


def plot_u_diode_cals(u_nqps_dfs, timestamp_nqps, circuit_name, i_meas_df=None, ax=None, xlim=None):
    if ax is None:
        title = '{} {} U_DIODE_RB(t) - CALS'.format(Time.to_string_short(timestamp_nqps), circuit_name)
        fig, ax1 = plt.subplots(figsize=(15, 7))
    else:
        title = 'nQPS'
        ax1 = ax

    for u_nqps_df in u_nqps_dfs:
        u_nqps_df.plot(ax=ax1, title=title, grid=True)

    ax1.title.set_size(20)
    ax1.legend(loc='lower left')
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.set_ylabel("U_DIODE_RB, [V]", fontsize=15)
    ax1.tick_params(labelsize=15)

    if xlim is not None:
        ax1.set_xlim(xlim)

    if i_meas_df is not None:
        ax2 = ax1.twinx()
        i_meas_df.plot(ax=ax2, color='C4')
        ax2.legend(loc='upper right')
        ax2.tick_params(labelsize=15)
        ax2.set_ylabel("I_MEAS, [A]", color='C4', fontsize=15)

    ax1.grid(True)

    if ax is None:
        plt.show()


def plot_u_diode_diff_pm(u_nqps_diff_dfs, timestamp_nqps, circuit_name, t_st_magnet_ok, ax=None):
    if ax is None:
        title = '{} {} U_DIODE_RB(t) - CALS'.format(Time.to_string_short(timestamp_nqps), circuit_name)
        fig, ax1 = plt.subplots(figsize=(15, 7))
    else:
        title = 'nQPS'
        ax1 = ax

    for u_nqps_diff_df in u_nqps_diff_dfs:
        u_nqps_diff_df.plot(ax=ax1, title=title, grid=True)

    ax1.axvspan(xmin=-0.2, xmax=t_st_magnet_ok, ymin=0, ymax=0.5, facecolor='xkcd:yellowgreen')
    ax1.set_ylabel("U (Calculated diode differences), [V]", fontsize=15)
    ax1.legend(loc='lower left')
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.set_xlim(t_st_magnet_ok - 0.2, t_st_magnet_ok + 0.1)
    ax1.set_ylim(0, 2)
    ax1.tick_params(labelsize=15)


def plot_u_diode_pm_nqds(u_nqps_dfs, timestamp_nqps, circuit_name, ax=None):
    if ax is None:
        title = '{} {} U_DIODE_RB(t) and ST_MAGNET_OK_SQRB(t) - PM'.format(
            Time.to_string_short(timestamp_nqps), circuit_name)
        fig, ax1 = plt.subplots(figsize=(15, 7))
    else:
        title = 'nQPS'
        ax1 = ax

    for u_nqps_df in u_nqps_dfs:
        u_nqps_df.plot(ax=ax1, title=title, grid=True)

    ax1.title.set_size(20)
    ax1.legend(loc='lower left')
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.set_ylabel("U_DIODE_RB, [V]", fontsize=15)
    ax1.tick_params(labelsize=15)

    if ax is None:
        plt.show()


def _get_magnet_name(signal_name):
    magnet_name = signal_name.split('.')[1].split(':')[0]
    if 'U_REF' in signal_name:
        return magnet_name + ':U_REF'
    else:
        return magnet_name


def calculate_col_diff(u_diode_ref_dfs):
    diff_dfs = []
    for index, df_first in enumerate(u_diode_ref_dfs):
        magnet_first = _get_magnet_name(df_first.columns[0])
        for df_second in u_diode_ref_dfs[index + 1:]:
            magnet_second = _get_magnet_name(df_second.columns[0])
            diff_name = '{}-{}'.format(magnet_first, magnet_second)
            value = abs(df_first.values - df_second.values)
            time = df_first.index
            diff_df = pd.DataFrame(index=time, data=value, columns=[diff_name])
            diff_dfs.append(diff_df)

    return diff_dfs


def synchronize_nqps_to_iqps(u_diode_ref_dfs, timestamp_nqps, timestamp_qds):
    u_nqps_shift_dfs = deepcopy(u_diode_ref_dfs)
    for u_nqps_shift_df in u_nqps_shift_dfs:
        if not math.isnan(timestamp_nqps):
            u_nqps_shift_df.index = u_nqps_shift_df.index + (timestamp_nqps - timestamp_qds) * 1e-9
        else:
            u_nqps_shift_df.index = u_nqps_shift_df.index + 10

    return u_nqps_shift_dfs


def plot_u_qds(u_qs0_df, u_1_df, u_2_df, timestamp_qds, timestamp_pic, *, title, xmin, xmax, xlim, ax=None):
    if ax is None:
        ax1 = u_qs0_df.plot(title=title, figsize=(15, 7.5), color='C0', grid=True, xlim=xlim, legend=True)
    else:
        ax1 = u_qs0_df.plot(ax=ax, title='iQPS', color='C0', grid=True, xlim=xlim, legend=True)

    ax1.title.set_size(20)
    ax1.axvline(x=timestamp_pic - timestamp_qds * 1e-9, color='green', linestyle='--', linewidth=1.5)
    ax1.legend(['U_QS0'], loc='lower left')
    ax1.set_ylabel("U_QS0, [V]", color='C0', fontsize=15)
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.tick_params(labelsize=15)
    ax1.axhspan(ymin=-0.1, ymax=+0.1, facecolor='xkcd:yellowgreen')
    ax1.axvspan(xmin=xmin, xmax=xmax, facecolor='xkcd:goldenrod')

    ax2 = ax1.twinx()
    u_1_df.plot(ax=ax2, color='C1', xlim=xlim)
    u_2_df.plot(ax=ax2, color='C2', xlim=xlim)
    ax2.legend(['U_1', 'U_2'], loc='upper right')
    ax2.set_ylabel("U_1, U_2, [V]", fontsize=15)
    ax2.set_xlabel("time, [s]", fontsize=15)
    ax2.tick_params(labelsize=15)

    if ax is None:
        plt.show()


def plot_st_qds(u_qs0_df, st_magnet_ok_df, st_nqd0_df, *, title, xmin, xmax, xlim, ax=None):
    if ax is None:
        ax1 = u_qs0_df.plot(title=title, figsize=(15, 7.5), style='o-', grid=True, xlim=xlim, legend=True)
    else:
        ax1 = u_qs0_df.plot(ax=ax, style='o-', grid=True, xlim=xlim, legend=True)

    ax1.title.set_size(20)
    ax1.legend(['U_QS0'], loc='lower left')
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.set_ylabel("U_QS0, [V]", color='C0', fontsize=15)
    ax1.tick_params(labelsize=15)
    ax1.axhspan(ymin=-0.1, ymax=+0.1, facecolor='xkcd:yellowgreen')  # beige
    ax1.axvspan(xmin=xmin, xmax=xmax, facecolor='xkcd:goldenrod')

    ax2 = ax1.twinx()
    st_magnet_ok_df.astype(int).plot(ax=ax2, yticks=[0, 1], ylim=[-0.1, 1.1], drawstyle="steps-post",
                                     style={'ST_MAGNET_OK': 'ro-'}, xlim=xlim, grid=True)
    st_nqd0_df.astype(int).plot(ax=ax2, yticks=[0, 1], ylim=[-0.1, 1.1], style={'ST_NQD0': 'bx-'},
                                drawstyle="steps-post", xlim=xlim, grid=True)

    ax2.set_ylabel("ST_NDQ0, ST_MAGNET_OK, [-]", fontsize=15)
    ax2.legend(['ST_NDQ0', 'ST_MAGNET_OK'], loc='upper right')
    ax2.tick_params(labelsize=15)

    if ax is None:
        plt.show()


# QH
def analyze_single_qh(timestamp_qh, source_qh, u_hds_dfs, i_hds_dfs, u_hds_ref_dfs, i_hds_ref_dfs, circuit_type='RB'):

    def get_decay_start_index(hds_dfs):
        # find df with I_HDS_1
        i_hds_df, col_i_hds_df = get_df_col(hds_dfs, 'I_HDS_1')
        if i_hds_df is not None:
            decay_start_index = ExpSignalAnalysis.get_window_with_given_mean_and_std(i_hds_df[col_i_hds_df])
            if decay_start_index == 0:
                u_hds_df, col_u_hds_df = get_df_col(hds_dfs, 'U_HDS_1')
                if u_hds_df is not None:
                    return ExpSignalAnalysis.get_window_with_deviation_from_initial_value(u_hds_df[col_u_hds_df])
                else:
                    return 0
            else:
                return decay_start_index
        else:
            return 0

    def get_decay_only(hds_dfs):
        decay_start_index = get_decay_start_index(hds_dfs)
        if decay_start_index == 0:
            return hds_dfs
        else:
            return [hds_df.iloc[(decay_start_index + 1):len(hds_df)] for hds_df in hds_dfs]

    # Get decay only
    u_i_hds_dfs = get_decay_only(u_hds_dfs + i_hds_dfs)
    u_i_hds_ref_dfs = get_decay_only(u_hds_ref_dfs + i_hds_ref_dfs)

    # Filter
    u_i_hds_dfs = [u_i_hds_df.rolling(window=3, min_periods=1).median() for u_i_hds_df in u_i_hds_dfs]
    u_i_hds_ref_dfs = [u_i_hds_ref_df.rolling(window=3, min_periods=1).median()
                       for u_i_hds_ref_df in u_i_hds_ref_dfs]

    # Calculate resistance
    r_hds_dfs = calculate_resistance(u_i_hds_dfs)
    r_hds_ref_dfs = calculate_resistance(u_i_hds_ref_dfs)

    timestamp_ref = Reference.get_quench_heater_reference_discharge('RB', source_qh)

    # Calculate features
    # # compare to reference values
    first_last_u_df = FeatureBuilder().with_signal(u_i_hds_dfs) \
        .calculate_features(features=['first', 'last20mean'], index=timestamp_qh)

    is_first_last_u_correct = Reference.compare_features_to_reference(first_last_u_df, circuit_type, 'QH',
                                                                      wildcard={'CELL': source_qh})

    first_r_df = FeatureBuilder().with_signal(r_hds_dfs) \
        .calculate_features(features='first20mean', index=timestamp_qh)
    first_r_ref_df = FeatureBuilder().with_signal(r_hds_ref_dfs) \
        .calculate_features(features='first20mean', index=timestamp_qh)
    is_first_r_correct = Reference.compare_difference_of_features_to_reference(first_r_df, first_r_ref_df,
                                                                               circuit_type, 'QH',
                                                                               wildcard={'CELL': source_qh})

    # # compare to reference discharge
    tau_df = FeatureBuilder().with_signal(u_i_hds_dfs) \
        .calculate_features(features=['tau_charge'], index=timestamp_qh)
    tau_ref_df = FeatureBuilder().with_signal(u_i_hds_ref_dfs) \
        .calculate_features(features=['tau_charge'], index=timestamp_ref)

    is_tau_correct = Reference.compare_difference_of_features_to_reference(tau_df, tau_ref_df, circuit_type,
                                                                           'QH', wildcard={'CELL': source_qh})

    # Display features
    display(HTML(pd.concat([first_last_u_df, first_r_df], axis=1).to_html()))
    display(HTML(pd.concat([tau_df, tau_ref_df], axis=0).to_html()))

    # # Plot signals
    plot_qh_discharge('RB', timestamp_qh, source_qh, u_i_hds_dfs, r_hds_dfs, u_i_hds_ref_dfs, r_hds_ref_dfs)

    return is_first_last_u_correct and is_first_r_correct and is_tau_correct


class RbCircuitAnalysis(CircuitAnalysis):

    # EE
    def analyze_ee_temp_and_cooling(self, t_res_body_long_dfs, st_res_overtemp_long_dfs, *, t_greater, value_max):
        is_signal_correct = AssertionBuilder().with_signal(t_res_body_long_dfs) \
            .for_time_greater_than(t_greater=t_greater) \
            .has_min_max_value(value_max=value_max) \
            .show_plot(y_label='T_RES_BODY, [degC]',
                       data_sec=st_res_overtemp_long_dfs, drawstyle_sec='steps-post',
                       ylabel_sec='ST_RES_OVERTEMP, [-]') \
            .get_assertion_result()

        if self.results_table is not None:
            analysis = '%s_EE_T_RES_BODY_2_HOURS_AFTER_FPA' % self.circuit_type
            status = 'Within reference' if is_signal_correct else 'Outside of reference'
            self.results_table[analysis] = get_expert_comment(analysis, self.is_automatic, status)

    # QH
    def analyze_qh(self, u_hds_dfs, i_hds_dfs, u_hds_ref_dfs, i_hds_ref_dfs):
        index_max = self.results_table.index[-1]
        for index, row in self.results_table.iterrows():
            source_qh = row['magnet']
            timestamp_qh = row['timestamp_iqps'] + 1000000
            print('{}/{}: Analysing quench heater signals of {} on {}, {}' \
                  .format(index, index_max, source_qh, Time.to_string(timestamp_qh), timestamp_qh))

            is_qh_correct = analyze_single_qh(timestamp_qh, source_qh, u_hds_dfs[index], i_hds_dfs[index],
                                              u_hds_ref_dfs[index],i_hds_ref_dfs[index])

            # Update results table
            if self.results_table is not None:
                status = 'Within reference' if is_qh_correct else 'Outside of reference'
                self.results_table.loc[index, 'QH'] = status

            # Show next
            check_show_next(index, index_max, self.is_automatic)

    def analyze_qds(self, circuit_name, timestamp_fgc, timestamp_pic, u_qds_dfs, u_nqps_dfs, i_meas_df):
        index_max = self.results_table.index[-1]
        for index, row in self.results_table.iterrows():
            source_qds, source_nqps, timestamp_qds, timestamp_nqps, i_quench, nqps_db = \
                row[['magnet', 'nqps_crate', 'timestamp_iqps', 'timestamp_nqps', 'I_MEAS_quench', 'nqps_db']]

            print('{}/{}: Analysing quench detection signals of {} on {}, {}' \
                  .format(index, self.results_table.index[-1], source_qds, Time.to_string(timestamp_qds), timestamp_qds))

            u_qs0_df, u_1_df, u_2_df, st_nqd0_df, st_magnet_ok_df = u_qds_dfs[index]

            # Analysis
            report_qds_trigger = analyze_qds_trigger(u_qs0_df, st_magnet_ok_df, st_nqd0_df)
            t_start_quench = report_qds_trigger.at[0, 't_start_quench [s]']
            t_st_nqd0 = report_qds_trigger.at[0, 't_st_nqd0 [s]']
            t_st_magnet_ok = report_qds_trigger.at[0, 't_st_magnet_ok [s]']

            # # iQPS
            nqps_crate = MappingMetadata.get_crate_name_from_magnet_name(circuit_type='RB', magnet=source_qds)
            dt_from_first_quench = (timestamp_qds - self.results_table.at[0, 'timestamp_iqps'])*1e9
            title = "Magnet: {}, iQPS Time Stamp: {}, nQPS Crate: {}, Current = {} A, Time from the 1st quench {} s". \
                format(source_qds, Time.to_string_short(timestamp_qds), nqps_crate, i_quench, dt_from_first_quench)
            fig, ax = plt.subplots(2, 2, figsize=(30, 15))
            fig.suptitle(title, fontsize=20)
            xlim = (t_start_quench - 0.025, t_st_magnet_ok + 0.075)
            plot_u_qds(u_qs0_df, u_1_df, u_2_df, timestamp_qds, timestamp_pic, title=title, xmin=t_start_quench,
                       xmax=t_st_nqd0, xlim=xlim, ax=ax[0, 0])
            plot_st_qds(u_qs0_df, st_magnet_ok_df, st_nqd0_df, xmin=t_start_quench, xmax=t_st_nqd0, title=title,
                        xlim=xlim, ax=ax[1, 0])

            if nqps_db == 'PM':
                u_nqps_shift_df = synchronize_nqps_to_iqps(u_nqps_dfs[index], timestamp_nqps, timestamp_qds)
                u_nqps_shift_diff_df = calculate_col_diff(u_nqps_shift_df)
                quenched_magnets = self.results_table.loc[:index - 1, 'magnet'].values
                u_nqps_shift_diff_filt_dfs = find_dfs_with_quenched_magnet(u_nqps_shift_diff_df, source_qds)
                t_nqps_diff_max, u_nqps_diff_max = get_max_diff_val_and_idx(u_nqps_shift_diff_filt_dfs, t_st_magnet_ok,
                                                                            source_qds, quenched_magnets)

                plot_u_diode_pm_nqds(u_nqps_shift_df, timestamp_nqps, circuit_name, ax=ax[0, 1])
                # Plot only the differences with the quenched magnet
                if u_nqps_shift_diff_filt_dfs:
                    plot_u_diode_diff_pm(u_nqps_shift_diff_filt_dfs, timestamp_nqps, circuit_name, t_st_magnet_ok,
                                         ax=ax[1, 1])
                else:
                    ax[1, 1].axis('off')

                if (u_nqps_diff_max > 1) & (t_nqps_diff_max < t_st_magnet_ok):
                    is_iqps_or_nqps = 'nQPS'
                else:
                    is_iqps_or_nqps = 'iQPS'
            else:
                u_nqps_shift_df = u_nqps_dfs[index]

                i_meas_shift_df = deepcopy(i_meas_df)
                i_meas_shift_df.index = i_meas_shift_df.index + (timestamp_fgc - timestamp_qds) * 1e-9
                plot_u_diode_cals(u_nqps_shift_df, timestamp_nqps, circuit_name, i_meas_df=i_meas_shift_df, ax=ax[0, 1])
                plot_u_diode_cals(u_nqps_shift_df, timestamp_nqps, circuit_name, i_meas_df=None, ax=ax[1, 1],
                                  xlim=[-2, 2])

                if 'nQPS' in self.results_table.at[index, 'comments']:
                    is_iqps_or_nqps = 'nQPS'
                else:
                    is_iqps_or_nqps = '?'

                t_nqps_diff_max = math.nan
                u_nqps_diff_max = math.nan

            ax[0, 1].xaxis.set_tick_params(which='both', labelbottom=True)
            plt.show()

            report_qds_trigger['t_nqps_diff_max [s]'] = t_nqps_diff_max
            report_qds_trigger['u_nqps_diff_max [V]'] = u_nqps_diff_max
            display(HTML(report_qds_trigger.to_html()))

            du_dt = report_qds_trigger.at[0, 'du_dt [V/s]']
            self.results_table.at[index, 'dU_QS0/dt'] = du_dt
            self.results_table.at[index, 'U_QS0_polarity'] = 'Pos' if du_dt > 0 else 'Neg'
            self.results_table.at[index, 'comments'] = report_qds_trigger.at[0, 'analysis_comment']
            if self.is_automatic:
                self.results_table.at[index, 'iqps_or_nqps_trigger'] = is_iqps_or_nqps
            else:
                self.results_table.at[index, 'iqps_or_nqps_trigger'] = \
                    get_expert_decision('What is the source of trigger', ['iQPS', 'nQPS'])

            # Show next
            check_show_next(index, index_max, self.is_automatic)

    def create_mp3_results_table(self, circuit_name, timestamp_fgc):

        columns_mp3 = Reference.read_quench_database_columns(circuit_type='RB', table_type='MP3')
        mp3_table_df = pd.DataFrame(columns=columns_mp3, index=[0])

        def create_pm_file_name(timestamp, name):
            return Time.to_string_short(timestamp).replace(' ', '_').replace(':', '').replace('-', '') + '_' + name

        # Circuit
        mp3_table_df['Circuit'] = self.results_table['circuit']
        # Date
        mp3_table_df['Date'] = Time.to_string_short(timestamp_fgc).split(' ')[0]
        # Time
        mp3_table_df['Time'] = Time.to_string_short(timestamp_fgc).split(' ')[1].split('.')[0]
        # FGC PM file
        pc_name = SignalMetadata.get_circuit_name_metadata(self.circuit_type, circuit_name)['PC']['%PC%']
        mp3_table_df['FGC PM file'] = create_pm_file_name(timestamp_fgc, pc_name)
        # Position
        mp3_table_df['Position'] = self.results_table['magnet']
        # QPS PM file
        mp3_table_df['QPS PM file'] = self.results_table['timestamp_iqps']
        mp3_table_df['QPS PM file'] = mp3_table_df.apply(
            lambda row: create_pm_file_name(row['QPS PM file'], row['Position']), axis=1)
        # flagged in APEX
        mp3_table_df['flagged in APEX'] = 'no'
        # iQPS or nQPS
        mp3_table_df['iQPS or nQPS'] = self.results_table['iqps_or_nqps_trigger']
        # I_q
        mp3_table_df['I_q'] = self.results_table['I_MEAS_quench']
        # Nr in Q event
        mp3_table_df['Nr in Q event'] = mp3_table_df.apply(lambda row: row.name + 1, axis=1)
        # Magnet ID
        mp3_table_df['Magnet ID'] = self.results_table['magnet_id']
        # QS0 polarity
        mp3_table_df['QS0 polarity'] = self.results_table['U_QS0_polarity']
        # dV(QS0)/dt_aver
        mp3_table_df['dV(QS0)/dt_aver'] = self.results_table['dU_QS0/dt']
        # R_leads_max
        mp3_table_df['R_leads_max'] = self.results_table['R_DIODE_LEADS_RB']
        # Comments
        mp3_table_df['Comments'] = self.results_table['comments']
        # I_MEAS
        mp3_table_df['I_MEAS'] = self.results_table['RB_PC_I_MEAS_tau']
        # nQPS
        mp3_table_df['nQPS'] = self.results_table['DIODE_RB_U_DIODE_RB']
        # Heater
        mp3_table_df['Heater'] = self.results_table['QH']
        # EE
        mp3_table_df['EE'] = self.results_table['RB_EE_U_DUMP_RES_tau_res']
        # V feeler
        mp3_table_df['V feeler'] = self.results_table['VF_U_EARTH_RB']
        # Analysis performed by
        mp3_table_df['Analysis performed by'] = self.results_table['performed_by']

        return mp3_table_df
