import warnings
import requests

import numpy as np
import pandas as pd
from tqdm.notebook import tqdm

import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder


def _create_report_analysis_template(circuit_type, circuit_name, metadata_qds, event_source_timestamp_qds_df,
                                     i_meas_df, timestamp_fgc):
    # ToDo: From a csv file
    COLUMN_NAMES = ["circuit", "magnet", "magnet_id", "timestamp_pic", "timestamp_sync", "timestamp_ee_odd",
                    "timestamp_ee_even", "timestamp_leads_odd", "timestamp_leads_even", "timestamp_iqps",
                    "datetime_iqps", "I_MEAS_quench", "iqps_board_type", "nqps_crate", "iqps_or_nqps_trigger",
                    "quench_type", "U_QS0_polarity", "dU_QS0/dt", "R_DIODE_LEADS_RB", "nqps_db", "QH", "comments",
                    "performed_by", "lhcsmapi_version"]
    report_qds = pd.DataFrame(columns=COLUMN_NAMES)

    for index, row in tqdm(event_source_timestamp_qds_df.iterrows(), total=event_source_timestamp_qds_df.shape[0],
                           desc='Querying PM'):
        source_qds = row['source']
        timestamp_qds = row['timestamp']
        t_quench = (timestamp_qds - timestamp_fgc) * 1e-9
        if t_quench > i_meas_df.index[-1]:
            i_quench = i_meas_df.values[-1]
            warnings.warn('Quench time beyond PM buffer. Need to query CALS in order to get the quench current!')
        else:
            i_quench = int(i_meas_df.reindex([t_quench], method='nearest', tolerance=.1).values[0])

        # ToDo: It would be nice to add this query to the library
        pm_request = "http://pm-api-pro/v2/pmdata?system=QPS&className=" + metadata_qds[
            'className'] + "&source=" + source_qds + "&TimestampInNanos=" + str(
            timestamp_qds) + "&pageSize=1&pageIndex=0"
        resp = requests.get(pm_request)
        board_type = resp.json()["content"][0]["namesAndValues"][1]["value"]

        n_qps = MappingMetadata.get_crate_name_from_magnet_name(circuit_type, source_qds)
        magnet_id = MappingMetadata.get_magnet_id_from_magnet_name(circuit_type, 'MB.{}'.format(source_qds))
        u_diode_db = 'PM' if t_quench < 2 else 'NXCALS'
        values = [circuit_name, source_qds, magnet_id, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, timestamp_qds,
                  Time.to_string_short(timestamp_qds), i_quench, board_type, n_qps, '?', '?', '?', 0, 0, u_diode_db,
                  '', '', '', lhcsmapi.__version__]
        report_qds.loc[len(report_qds.index) + 1] = values

    return report_qds.drop_duplicates('magnet')


class RbCircuitQuery(CircuitQuery):

    # PIC
    def find_timestamp_pic(self, timestamp_fgc, *, spark, duration=[(1, 's'), (60, 's')], verbose=True):
        st_abort_pic_dfs = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=int(timestamp_fgc), duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='PIC', signal='ST_ABORT_PIC') \
            .signal_query(verbose=verbose).dfs

        return [st_abort_pic_df.index[0] for st_abort_pic_df in st_abort_pic_dfs]

    # EE
    def query_ee_u_dump_res_pm(self, source_ee, timestamp_ee, timestamp_sync, system, *, is_ref=False, verbose=True):
        # ToDo EE_EVEN, EE_ODD - merge with CircuitQuery
        metadata_ee = SignalMetadata.get_circuit_signal_database_metadata(self.circuit_type, self.circuit_name, 'EE',
                                                                          'PM')

        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_ee) \
            .with_query_parameters(signal='{}:U_DUMP_RES'.format(source_ee), system=metadata_ee['system'],
                                   className=metadata_ee['className'], source=source_ee) \
            .signal_query(verbose=verbose) \
            .remove_initial_offset(timestamp_ee) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

    def query_ee_t_res_pm(self, source_ee, timestamp_ee, timestamp_sync, system, *, signal_names, verbose=True):
        # ToDo EE_EVEN, EE_ODD - merge with CircuitQuery
        metadata_ee = SignalMetadata.get_circuit_signal_database_metadata(self.circuit_type, self.circuit_name, 'EE',
                                                                          'PM')

        return QueryBuilder().with_pm() \
            .with_timestamp(timestamp_ee) \
            .with_query_parameters(signal=signal_names, system=metadata_ee['system'],
                                   className=metadata_ee['className'], source=source_ee) \
            .signal_query(verbose=verbose) \
            .remove_values_for_time_less_than(timestamp_ee) \
            .synchronize_time(timestamp_sync) \
            .convert_index_to_sec().dfs

    def create_report_analysis(self, source_timestamp_nqps_df, timestamp_fgc, i_meas_df, author, verbose=True):
        source_timestamp_qds = QueryBuilder().with_pm() \
            .with_duration(t_start=timestamp_fgc, duration=[(600, 's'), (600, 's')]) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS', source='*') \
            .event_query(verbose=verbose) \
            .filter_source(self.circuit_name, 'QH') \
            .df

        # Create report analysis template
        metadata_qds = SignalMetadata.get_circuit_signal_database_metadata(self.circuit_type, self.circuit_name, 'QDS',
                                                                           'PM')
        report_analysis = _create_report_analysis_template(self.circuit_type, self.circuit_name,
                                                           metadata_qds,
                                                           source_timestamp_qds, i_meas_df,
                                                           timestamp_fgc)
        report_analysis['doneBy'] = author
        # nQPS - append timestamps
        source_timestamp_nqps_renamed_df = source_timestamp_nqps_df.rename(
            columns={'source': 'nqps_crate', 'timestamp': 'timestamp_nqps'})
        # ToDo If an nQPS PM event is missing, then NaN is set enforcing the entire row to be set to into float
        report_analysis = report_analysis.merge(source_timestamp_nqps_renamed_df[['timestamp_nqps', 'nqps_crate']],
                                                on='nqps_crate', how='left').sort_values(by='timestamp_iqps') \
            .reset_index(drop=True)

        return report_analysis

    def query_voltage_logic_iqps(self, report_analysis, *, signal_names, verbose=True):
        u_qds_dfs = []
        for index, row in tqdm(report_analysis.iterrows(), total=report_analysis.shape[0],
                               desc='Querying iQPS logic PM'):
            source_qds, timestamp_qds = row[['magnet', 'timestamp_iqps']]

            u_qds_df = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_qds) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', source=source_qds,
                               wildcard={'CELL': source_qds},
                               signal=signal_names) \
                .signal_query(verbose=verbose) \
                .synchronize_time(timestamp_qds) \
                .convert_index_to_sec().dfs

            u_qds_dfs.append(u_qds_df)

        return u_qds_dfs

    def query_voltage_nqps(self, results_table, *, ldb):
        u_nqps_dfs = []
        for index, row in tqdm(results_table.iterrows(), total=results_table.shape[0],
                               desc='Querying CALS U_DIODE_RB, U_REF_N1'):
            source_nqps, timestamp_nqps, timestamp_qds, nqps_db = row[['nqps_crate', 'timestamp_nqps',
                                                                        'timestamp_iqps', 'nqps_db']]

            magnet_names = MappingMetadata.get_magnet_names_from_crate_name(self.circuit_type, source_nqps)
            if nqps_db == 'PM':
                u_diode_dfs, u_ref_n1_dfs = self._query_nqps_pm(magnet_names, source_nqps, timestamp_nqps)
            else:
                u_diode_dfs, u_ref_n1_dfs = self._query_nqps_cals(source_nqps, magnet_names, timestamp_qds, ldb=ldb)

            u_nqps_dfs.append(u_diode_dfs + u_ref_n1_dfs)

        return u_nqps_dfs

    def _query_nqps_cals(self, source_nqps, magnet_names, timestamp_qds, *, ldb, duration=[(10, 's'), (350, 's')]):

        u_diode_dfs = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=timestamp_qds, duration=duration) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='DIODE_RB', source=source_nqps,
                           wildcard={'MAGNET': magnet_names}, signal='U_DIODE_RB') \
            .signal_query() \
            .synchronize_time(timestamp_qds / 1e9) \
            .dfs
        try:
            u_ref_n1_dfs = QueryBuilder().with_cals(ldb) \
                .with_duration(t_start=timestamp_qds, duration=duration) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='DIODE_RB', source=source_nqps,
                               wildcard={'QPS_CRATE': source_nqps}, signal='U_REF_N1') \
                .signal_query() \
                .synchronize_time(timestamp_qds / 1e9) \
                .dfs
        except:
            u_ref_n1_dfs = [pd.DataFrame()]
            print('U_REF_N1 signal not present for crate {}'.format(source_nqps))

        return u_diode_dfs, u_ref_n1_dfs

    def _query_nqps_pm(self, magnet_names, source_nqps, timestamp_nqps):

        u_diode_dfs = QueryBuilder().with_pm() \
            .with_timestamp(timestamp_nqps) \
            .with_circuit_type(self.circuit_type) \
            .with_metadata(circuit_name=self.circuit_name, system='DIODE_RB', source=source_nqps,
                           wildcard={'MAGNET': magnet_names}, signal='U_DIODE_RB') \
            .signal_query() \
            .synchronize_time(timestamp_nqps) \
            .convert_index_to_sec().dfs
        try:
            u_ref_n1_dfs = QueryBuilder().with_pm() \
                .with_timestamp(timestamp_nqps) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system='DIODE_RB', source=source_nqps,
                               wildcard={'QPS_CRATE': source_nqps}, signal='U_REF_N1') \
                .signal_query() \
                .synchronize_time(timestamp_nqps) \
                .convert_index_to_sec().dfs
        except:
            u_ref_n1_dfs = [pd.DataFrame()]
            print('U_REF_N1 signal not present for crate {}'.format(source_nqps))

        return u_diode_dfs, u_ref_n1_dfs
