from typing import List

import matplotlib.pyplot as plt
from IPython.display import display, HTML
import pandas as pd

from lhcsmapi.Time import Time
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis
from lhcsmapi.analysis.magnet.iqds.IqdsRqAnalysis import find_circuit_name_with_first_quenched_magnet, choose_u_qs0, \
    find_t_start_t_end_voltage_slope, calculate_voltage_slope
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.FeatureBuilder import FeatureBuilder
from lhcsmapi.pyedsl.PlotBuilder import PlotBuilder
from lhcsmapi.reference.Reference import Reference


class RqCircuitAnalysis(CircuitAnalysis):
    """Class for analysis of main quadrupole circuits
    """

    def analyze_qds(self, circuit_names: List[str], iqps_analog_dfs: List[pd.DataFrame],
                    iqps_digital_dfs: List[pd.DataFrame], u_nqps_rqd_dfs: List[pd.DataFrame],
                    u_nqps_rqf_dfs: List[pd.DataFrame]) -> None:
        """ Method analyzing QDS signals in RQ circuit. It is plotting all signals and calculating:
        - quench origin (RQD/RQF, INT/EXT)
        - time derivative of the U_QS0 signal

        :param circuit_names: names of RQ circuits to analyze ['RQD.XXX', 'RQF.XXX']
        :type circuit_names: List[str]
        :param iqps_analog_dfs: list of tuples of analog iQPS signals. There are as many outer lists as quenched
        magnets and as many elements in the inner tuple as there are analog iQPS signals to analyze
        :type iqps_analog_dfs: List[Tuple[pd.DataFrame]]
        :param iqps_analog_dfs: list of tuples of digital iQPS signals. There are as many outer lists as quenched
        magnets and as many elements in the inner tuple as there are digital iQPS signals to analyze
        :type iqps_analog_dfs: List[Tuple[pd.DataFrame]]
        :param u_nqps_rqd_dfs: list of tuples of nQPS RQD signals. There are as many outer lists as quenched
        magnets and as many elements in the inner tuple as there are nQPS RQD signals to analyze
        :type u_nqps_rqd_dfs: List[Tuple[pd.DataFrame]]
        :param u_nqps_rqf_dfs: list of tuples of nQPS RQD signals. There are as many outer lists as quenched
        magnets and as many elements in the inner tuple as there are nQPS RQD signals to analyze
        :type u_nqps_rqf_dfs: List[Tuple[pd.DataFrame]]
        """
        for index, row in self.results_table.iterrows():
            source_qds, timestamp_qds = row[['magnet', 'timestamp_iqps']]
            magnet_name = 'MQ.{}'.format(source_qds)

            u_qs0_ext_rq_df, u_qs0_int_rq_df, u_1_ext_rq_df, u_2_ext_rq_df, u_1_int_rq_df, u_2_int_rq_df = \
                iqps_analog_dfs[
                    index]
            st_magnet_ok_ext_rq_df, st_magnet_ok_int_rq_df, st_nqd0_ext_rq_df, st_nqd0_int_rq_df = iqps_digital_dfs[
                index]

            u_diode_rqd_df, u_ref_rqd_df = u_nqps_rqd_dfs[index]
            u_diode_rqf_df, u_ref_rqf_df = u_nqps_rqf_dfs[index]

            # Analysis
            circuit_name = find_circuit_name_with_first_quenched_magnet(circuit_names, u_diode_rqd_df, u_diode_rqf_df)
            aperture = MappingMetadata.get_rq_aperture(circuit_name, magnet_name) if len(circuit_name) > 0 else ''
            u_qs0_aperture_df = choose_u_qs0(aperture, u_qs0_int_rq_df, u_qs0_ext_rq_df) if len(aperture) > 0 \
                                                                                        else pd.DataFrame()
            t_slope_start, t_slope_end = find_t_start_t_end_voltage_slope(u_qs0_aperture_df)
            du_dt = calculate_voltage_slope(u_qs0_aperture_df, t_slope_start, t_slope_end)
            if len(aperture) > 0:
                print('The first aperture to quench is {} of magnet {}'.format(aperture, magnet_name))
                self.results_table.loc[index, 'origin'] = '{}/{}'.format(circuit_name.split('.')[0], aperture)
                print('The voltage increase is {:.3f} V/s'.format(du_dt))
            else:
                print('No aperture quenched in magnet {}'.format(magnet_name))
                self.results_table.loc[index, 'origin'] = 'None'

            self.results_table.loc[index, 'dU_QS0/dt'] = round(du_dt, 2)

            fig, ax = plt.subplots(2, 2, figsize=(30, 15))
            title = "Magnet: {}, Aperture: {}/{}, Time Stamp: {}".format(source_qds, circuit_name.split('.')[0],
                                                                         aperture, Time.to_string_short(timestamp_qds))
            fig.suptitle(title, fontsize=30)

            # iQPS analog
            if (not u_qs0_ext_rq_df.empty) and (not u_qs0_int_rq_df.empty):
                u_qs0_ext_rq_df.plot(ax=ax[0][0])
                u_qs0_int_rq_df.plot(ax=ax[0][0], grid=True)
                ax[0][0].tick_params(labelsize=15)
                ax[0][0].set_xlim((-0.1, 0.1))
                ax[0][0].set_ylim((-0.3, 0.3))
                ax[0][0].axhspan(ymin=-0.1, ymax=+0.1, facecolor='xkcd:yellowgreen')
                ax[0][0].axvspan(xmin=t_slope_start, xmax=t_slope_end, facecolor='xkcd:goldenrod')

            # nQPS analog
            try:
                u_diode_rqd_df.plot(ax=ax[0][1])
                u_ref_rqd_df.plot(ax=ax[0][1])
                u_diode_rqf_df.plot(ax=ax[0][1])
                u_ref_rqf_df.plot(ax=ax[0][1], grid=True)
            except Exception as e:
                pass
            ax[0][1].tick_params(labelsize=15)

            # iQPS digital
            try: # ToDo: Replace with a check if signal is not empty
                st_magnet_ok_ext_rq_df.astype(int).plot(ax=ax[1][0])
                st_magnet_ok_int_rq_df.astype(int).plot(ax=ax[1][0], grid=True)
            except Exception as e:
                pass
            ax[1][0].tick_params(labelsize=15)
            ax[1][0].set_xlim((-0.1, 0.1))
            st_nqd0_ext_rq_df.astype(int).plot(ax=ax[1][0])
            st_nqd0_int_rq_df.astype(int).plot(ax=ax[1][0], grid=True)
            ax[1][0].tick_params(labelsize=15)

            # nQPS analog - zoom
            try:
                u_diode_rqd_df.plot(ax=ax[1][1])
                u_ref_rqd_df.plot(ax=ax[1][1])
                u_diode_rqf_df.plot(ax=ax[1][1])
                u_ref_rqf_df.plot(ax=ax[1][1], grid=True)
            except Exception as e:
                pass
            ax[1][1].tick_params(labelsize=15)
            ax[1][1].set_xlim((-0.1, 0.1))
            plt.show()

    def analyze_qh(self, circuit_type: str, u_hds_rq_dfs: List[List[pd.DataFrame]],
                   u_hds_rq_ref_dfs: List[List[pd.DataFrame]]) -> None:
        """ Method analyzing QH voltage signals in RQ circuit. It is plotting all signals and calculating:
            - initial voltage
            - final voltage
            - time constant based on the charge approach

        :param circuit_type: circuit type to analyze ('RQ')
        :type circuit_type: str
        :param u_hds_rq_dfs: list of lists of QH voltage signals. There are as many outer lists as quenched
        magnets and as many elements in the inner list as there are QH voltage signals to analyze
        :type u_hds_rq_dfs: List[List[pd.DataFrame]]
        :param u_hds_rq_ref_dfs: list of lists of reference QH voltage signals. There are as many outer lists as
        quenched magnets and as many elements in the inner list as there are reference QH voltage signals to analyze
        :type u_hds_rq_ref_dfs: List[List[pd.DataFrame]]
        """

        for index, row in self.results_table.iterrows():
            source_qds, timestamp_qds = row[['magnet', 'timestamp_iqps']]
            timestamp_qds = Time.to_unix_timestamp(timestamp_qds)
            timestamp_ref = Reference.get_quench_heater_reference_discharge(circuit_type, source_qds)

            u_hds_rq_df = u_hds_rq_dfs[index]
            u_hds_rq_ref_df = u_hds_rq_ref_dfs[index]

            # Calculate features
            # # to reference values
            first_last_u_df = FeatureBuilder().with_signal(u_hds_rq_df) \
                .calculate_features(features=['first', 'last20mean'], index=timestamp_qds)

            is_first_last_u_ok = Reference.compare_features_to_reference(first_last_u_df, self.circuit_type,
                                                                         'QH', wildcard={'CELL': source_qds})

            # # to a reference discharge
            tau_df = FeatureBuilder().with_signal(u_hds_rq_df) \
                .calculate_features(features=['tau_charge'], index=timestamp_qds)

            tau_ref_df = FeatureBuilder().with_signal(u_hds_rq_ref_df) \
                .calculate_features(features=['tau_charge'], index=timestamp_ref)

            is_tau_ok = Reference.compare_difference_of_features_to_reference(tau_df, tau_ref_df, self.circuit_type,
                                                                              'QH', wildcard={'CELL': source_qds})

            # Update results table
            # Use OK rather than True always
            self.results_table.loc[index, 'QH'] = 'OK' if is_first_last_u_ok and is_tau_ok else 'NOT_OK'

            # Display features
            features_qh_concat = pd.concat([tau_df, tau_ref_df], axis=0)
            display(HTML(features_qh_concat.to_html()))
            display(HTML(first_last_u_df.to_html()))

            # Plot
            title = "Magnet: {}, Time Stamp: {}, Reference {}, U_HDS(t)" \
                .format(source_qds, Time.to_string_short(timestamp_qds + 1000000), Time.to_string_short(timestamp_ref))

            PlotBuilder().with_signal(u_hds_rq_df, title=title, figsize=(15, 7.5), grid=True) \
                .with_ylabel(ylabel='U_HDS, [V]') \
                .with_signal(u_hds_rq_ref_df, is_twinx=False, grid=True, style='--') \
                .with_ylabel(ylabel='U_HDS, [V]', color='C0') \
                .plot()

            PlotBuilder().with_signal(u_hds_rq_df, title=title, figsize=(15, 7.5), grid=True) \
                .with_ylabel(ylabel='U_HDS, [V]') \
                .with_signal(u_hds_rq_ref_df, is_twinx=False, grid=True, style='--') \
                .with_ylabel(ylabel='U_HDS, [V]', color='C0') \
                .plot(yscale='log')

    def create_mp3_results_table(self, circuit_names: List[str], timestamp_fgc_rqd: int) -> pd.DataFrame:
        """ Method creating an MP3-compatible quench database row for the main quadrupole circuit

        :param circuit_names: names of RQ circuits ['RQD.XXX', 'RQF.XXX']
        :type circuit_names: List[str]
        :param timestamp_fgc_rqd: unix timestamp of an FGC event in RQD circuit (ns precision)

        :return: pd.DataFrame with the results of analysis calculation
        """

        def create_pm_file_name(timestamp, name):
            return Time.to_string_short(timestamp).replace(' ', '_').replace(':', '').replace('-', '') + '_' + name

        columns_mp3 = Reference.read_quench_database_columns(circuit_type='RQ', table_type='MP3')
        mp3_table_df = pd.DataFrame(columns=columns_mp3, index=[0])

        # Circuit
        mp3_table_df['Circuit'] = self.results_table['circuit']
        # Date
        mp3_table_df['Date'] = Time.to_string_short(timestamp_fgc_rqd).split(' ')[0]
        # Time
        mp3_table_df['Time'] = Time.to_string_short(timestamp_fgc_rqd).split(' ')[1].split('.')[0]
        # FGC PM file
        pc_name = SignalMetadata.get_circuit_name_metadata(self.circuit_type, circuit_names[0])['PC']['%PC%']
        mp3_table_df['FGC PM file'] = create_pm_file_name(timestamp_fgc_rqd, pc_name).replace('RQD', 'RQx')
        # Position
        mp3_table_df['Position'] = self.results_table['magnet']
        # iQPS PM file
        mp3_table_df['iQPS PM file'] = self.results_table['timestamp_iqps']
        mp3_table_df['iQPS PM file'] = mp3_table_df.apply(
            lambda row: create_pm_file_name(row['iQPS PM file'], row['Position']), axis=1)
        # I_q_D
        mp3_table_df['I_q_D'] = self.results_table['I_MEAS_RQD_quench']
        # I_q_F
        mp3_table_df['I_q_F'] = self.results_table['I_MEAS_RQF_quench']
        # Origin
        mp3_table_df['Origin'] = self.results_table['origin']
        # R_lead_max_RQD
        mp3_table_df['R_lead_max_RQD'] = self.results_table['R_DIODE_LEADS_RQD']
        # R_lead_max_RQF
        mp3_table_df['R_lead_max_RQF'] = self.results_table['R_DIODE_LEADS_RQF']
        # t_lead_max_RQD
        if 't_DIODE_LEADS_RQD' in self.results_table.columns:
            mp3_table_df['t_lead_max_RQD'] = self.results_table['t_DIODE_LEADS_RQD']
        # t_lead_max_RQF
        if 't_DIODE_LEADS_RQF' in self.results_table.columns:
            mp3_table_df['t_lead_max_RQF'] = self.results_table['t_DIODE_LEADS_RQF']

        return mp3_table_df
