import warnings
from typing import List

import numpy as np
import pandas as pd

import lhcsmapi
from lhcsmapi.Time import Time
from lhcsmapi.analysis.CircuitQuery import CircuitQuery
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference


class RqCircuitQuery(CircuitQuery):
    """Class  extending the CircuitQuery class with methods for creation of templates of analysis tables and query
    of voltage across magnets in main quadrupole circuits.

    """

    def create_report_analysis_template(self, source_timestamp_qds_df: pd.DataFrame,
                                        source_timestamp_nqps_df: pd.DataFrame, timestamp_fgc_rqd: int,
                                        i_meas_rqd_df: pd.DataFrame, i_meas_rqf_df:pd.DataFrame,
                                        author: str) -> pd.DataFrame:
        """ Method creating a report analysis template for main quadrupole circuits

        :param source_timestamp_qds_df: table with QDS PM source and timestamp
        :type source_timestamp_qds_df: pd.DataFrame
        :param source_timestamp_nqps_df: table with QDS PM source and timestamp
        :type source_timestamp_nqps_df: pd.DataFrame
        :param timestamp_fgc_rqd: unix timestamp of an FGC event in RQD circuit in nanoseconds
        :type timestamp_fgc_rqd: int
        :param i_meas_rqd_df: measured power converter current in RQD circuit
        :type i_meas_rqd_df: pd.DataFrame
        :param i_meas_rqf_df: measured power converter current in RQF circuit
        :type i_meas_rqf_df: pd.DataFrame
        :param author: NICE account name of the analysis author
        :type author: str
        :return: pre-filled pd.DataFrame with a template to be filled further with data from analysis
        of a powering event in main quadrupole (RQ) circuits
        """
        columns = Reference.read_quench_database_columns(circuit_type='RQ', table_type='LHCSM')
        report_table = pd.DataFrame(columns=columns)

        for index, row in source_timestamp_qds_df.iterrows():
            timestamp_qds = row['timestamp']
            qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, row['source'])
            if source_timestamp_nqps_df.empty:
                timestamp_nqps = np.nan
            else:
                mask = source_timestamp_nqps_df['source'] == qps_crate
                if any(mask.values):
                    timestamp_nqps = source_timestamp_nqps_df[mask]['timestamp'].values[0]
                else:
                    timestamp_nqps = np.nan
            iqps_or_nqps_trigger = 'nQPS' if timestamp_nqps < timestamp_qds else 'iQPS'

            t_quench = (min(timestamp_qds, timestamp_nqps) - timestamp_fgc_rqd) * 1e-9
            if t_quench > i_meas_rqd_df.index[-1]:
                i_quench_rqd = i_meas_rqd_df.values[-1]
                i_quench_rqf = i_meas_rqf_df.values[-1]
                warnings.warn('Quench time beyond PM buffer. Need to query NXCALS in order to get the quench current!')
            else:
                i_quench_rqd = int(i_meas_rqd_df.reindex([t_quench], method='nearest', tolerance=.1).values[0])
                i_quench_rqf = int(i_meas_rqf_df.reindex([t_quench], method='nearest', tolerance=.1).values[0])

            nqps_db = 'PM' if t_quench < 2 else 'NXCALS'
            values = [self.circuit_name.replace('D', ''), np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan,
                      np.nan, Time.to_string_short(timestamp_fgc_rqd), Time.to_string_short(timestamp_qds),
                      timestamp_nqps, iqps_or_nqps_trigger, i_quench_rqd, i_quench_rqf, row['source'], qps_crate, '',
                      np.nan, '', np.nan, np.nan, nqps_db, '', author, lhcsmapi.__version__]
            report_table.loc[len(report_table.index)] = values

        return report_table

    def query_voltage_cals(self, system: str, signal: str, timestamp_fgc: int, *, ldb, duration=[(50, 's'), (150, 's')],
                           verbose=True) -> List[pd.DataFrame]:
        """ Method querying CALS for voltage signals (either voltage feelers or diodes).
        If voltage across diodes is queried, then in addition, there is a U_REF signal queried for those diodes
        without U_DIODE signal.

        :param system: name of a system to query according to metadata
        :type system: str
        :param signal: name of a signal to query
        :type signal: str
        :param timestamp_fgc: unix timestamp of an FGC event in a circuit in nanoseconds to which the query is synced
        :type timestamp_fgc: int
        :param ldb: CALS pytimber connector
        :type ldb: pytimber
        :param duration: search duration before and after the start time
        :type duration: List[Tuple(int, str)]
        :type verbose: bool
        :return: list of pd.DataFrames (if a signal was not queried, an empty pd.DataFrame is returned and warning
        containing the raw query is returned)
        """
        u_cals_dfs = super().query_voltage_cals(system, signal, timestamp_fgc, ldb=ldb, duration=duration,
                                                verbose=verbose)

        if 'DIODE' in system:
            rq_diode = MappingMetadata.read_magnet_to_cell_qps_crate(circuit_type='RQ')
            crates = list(rq_diode[rq_diode['Crate U_DIODE_RQx'].isna() & (rq_diode['Circuit'] == self.circuit_name)]
                          ['Crate U_REF_N1'].values)

            u_ref_dfs = QueryBuilder().with_cals(ldb) \
                .with_duration(t_start=int(timestamp_fgc), duration=duration) \
                .with_circuit_type(self.circuit_type) \
                .with_metadata(circuit_name=self.circuit_name, system=system, signal='U_REF_N1',
                               wildcard={'QPS_CRATE': crates}) \
                .signal_query() \
                .synchronize_time(timestamp_fgc / 1e9).dfs

            return u_cals_dfs + u_ref_dfs
        else:
            return u_cals_dfs
