from copy import deepcopy
import pandas as pd
import numpy as np

from pyspark.sql.types import IntegerType
import pyspark.sql.functions as F

from lhcsmapi.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.pyedsl.ResistanceBuilder import ResistanceBuilder


def generate_busbar_status(beam_mode_with_pattern_series, pattern):
    # find patterns long-enough
    beam_mode_spfr_ramp_pattern_df, _ = SignalAnalysis.find_rows_with_exact_pattern_in_series(
        beam_mode_with_pattern_series, pattern=list(pattern.values()))
    beam_mode_spfr_ramp_pattern_df['is_sb_long_enough'] = (beam_mode_spfr_ramp_pattern_df[14.0] -
                                                           beam_mode_spfr_ramp_pattern_df[8.0]) > 1800 * 1e9
    beam_mode_spfr_ramp_pattern_long_enough_df = beam_mode_spfr_ramp_pattern_df.query('is_sb_long_enough == True')

    t_setups = beam_mode_spfr_ramp_pattern_long_enough_df[2].values
    t_preramps = beam_mode_spfr_ramp_pattern_long_enough_df[6].values
    t_flattops = beam_mode_spfr_ramp_pattern_long_enough_df[8].values
    t_start_injs = np.array([max(t_s.item(), t_p.item() - int(1800 * 1e9)) for t_s, t_p in zip(t_setups, t_preramps)])
    t_end_injs = np.array([t_preramp.item() for t_preramp in t_preramps])
    t_start_sbs = np.array([t_flattop.item() for t_flattop in t_flattops])
    t_end_sbs = np.array([t_flattop.item() + int(1800 * 1e9) for t_flattop in t_flattops])
    busbar_df = pd.DataFrame({'t_start_injs': t_start_injs,
                              't_end_injs': t_end_injs,
                              't_start_sbs': t_start_sbs,
                              't_end_sbs': t_end_sbs})
    busbar_df.sort_values(by='t_start_injs', inplace=True)

    return busbar_df


def get_complete_busbar_df(t_start, t_end, pattern, spark):
    # {'SETUP': 2.0, 'PRERAMP': 6.0, 'FLATTOP': 8.0, 'RAMPDOWN': 14.0}

    # query beam mode
    beam_mode_with_pattern_df = QueryBuilder().with_nxcals(spark) \
        .with_duration(t_start=t_start, t_end=t_end) \
        .with_query_parameters(nxcals_system='CMW', signal='HX:BMODE') \
        .signal_query() \
        .filter_values(list(pattern.keys())) \
        .map_values(pattern) \
        .dfs[0]

    # generate busbar status
    return generate_busbar_status(beam_mode_with_pattern_df["HX:BMODE"], pattern)


def find_plateau_start_and_end(i_meas_raw_df, i_meas_threshold=500):
    i_meas_df = deepcopy(i_meas_raw_df)
    I_MEAS = i_meas_df.columns[0]

    i_meas_df['acqStamp'] = i_meas_df.index
    i_meas_df = SignalUtilities.synchronize_df(i_meas_df)
    i_meas_df = SignalUtilities.convert_indices_to_sec(i_meas_df)

    # take the current only above 500 A
    i_meas_df = i_meas_df[i_meas_df[I_MEAS] > i_meas_threshold]

    # calculate time derivative
    i_meas_df.loc[:, 'time'] = i_meas_df.index
    i_meas_df['dI_MEAS'] = i_meas_df[I_MEAS].diff()
    i_meas_df['dtime'] = i_meas_df['time'].diff()
    i_meas_df = i_meas_df.dropna()
    i_meas_df['dI_MEAS_dt'] = i_meas_df.apply(lambda col: col['dI_MEAS'] / col['dtime'], axis=1)

    # round to the nearest integer
    i_meas_df['round_dI_MEAS_dt'] = i_meas_df.apply(lambda row: np.around(row['dI_MEAS_dt']), axis=1)

    # get sign of the round value
    i_meas_df['sign_dI_MEAS_dt'] = i_meas_df.apply(lambda row: np.sign(row['round_dI_MEAS_dt']), axis=1)

    # Find change of derivative
    plateau_start_pattern_df, plateau_start_pattern_idx = SignalAnalysis.find_rows_with_exact_pattern_in_series(
        i_meas_df['sign_dI_MEAS_dt'], [1, 0])

    plateau_end_pattern_df, plateau_end_pattern_idx = SignalAnalysis.find_rows_with_exact_pattern_in_series(
        i_meas_df['sign_dI_MEAS_dt'], [0, 1])

    # plateau start, plateau_end
    plateau_start = list(i_meas_df.loc[i_meas_df.index[plateau_start_pattern_idx[1::2]], 'acqStamp'].values)
    plateau_end = list(i_meas_df.loc[i_meas_df.index[plateau_end_pattern_idx[1::2]], 'acqStamp'].values)

    if plateau_end[0] < plateau_start[0]:
        plateau_start = [i_meas_raw_df.index[0]] + plateau_start

    if len(plateau_start) > len(plateau_end):
        plateau_end.append(i_meas_raw_df.index[-1])

    if len(plateau_start) < len(plateau_end):
        plateau_start = i_meas_raw_df.index[0].values + plateau_start

    return plateau_start, plateau_end


def find_start_end_of_the_longest_ramp_up(plateau_start, plateau_end):
    # calculate differences between plateau_start[i+1] and plateau_end[i] for i=0...len(plateau_start)-2
    ramp_up_duration = []
    for i in range(len(plateau_start)-1):
        ramp_up_duration.append(plateau_start[i+1] - plateau_end[i])

    # find the index of the maximum ramp up
    index_max_duration = ramp_up_duration.index(max(ramp_up_duration))

    return plateau_end[index_max_duration], plateau_start[index_max_duration+1]


def find_start_end_of_the_longest_ramp_up(plateau_start, plateau_end):
    # calculate differences between plateau_start[i+1] and plateau_end[i] for i=0...len(plateau_start)-2
    ramp_up_duration = []
    for i in range(len(plateau_start)-1):
        ramp_up_duration.append(plateau_start[i+1] - plateau_end[i])

    # find the index of the maximum ramp up
    index_max_duration = ramp_up_duration.index(max(ramp_up_duration))

    return plateau_end[index_max_duration], plateau_start[index_max_duration+1]


def get_busbar_resistances(circuit_type, signal_type, t_start_inj, t_end_inj, t_start_sb, t_end_sb, spark):
    # based on plateau, define udf
    def plateau_translate(timestamp):
        if (timestamp >= t_end_inj) & (timestamp < t_start_sb):
            return 0
        elif (timestamp >= t_start_inj) & (timestamp <= t_end_inj):
            return 1
        elif (timestamp >= t_start_sb) & (timestamp <= t_end_sb):
            return 2
        else:
            return -1

    plateau_translate_udf = F.udf(plateau_translate, IntegerType())

    # query current
    i_meas_feature_df = QueryBuilder() \
        .with_nxcals(spark) \
        .with_duration(t_start=int(t_start_inj), t_end=int(t_end_sb)) \
        .with_circuit_type(circuit_type) \
        .with_metadata(circuit_name='*', system='PC', signal='I_MEAS') \
        .feature_query(['mean', 'std'], function=plateau_translate_udf) \
        .sort_values(by='class') \
        .df

    # query voltage
    u_res_feature_df = QueryBuilder() \
        .with_nxcals(spark) \
        .with_duration(t_start=int(t_start_inj), t_end=int(t_end_sb)) \
        .with_circuit_type(circuit_type) \
        .with_metadata(circuit_name='*', system='BUSBAR', signal=signal_type, wildcard={'BUSBAR': '*'}) \
        .feature_query(['mean', 'std'], function=plateau_translate_udf) \
        .sort_busbar_location(circuit_type, circuit_name='*') \
        .correct_voltage_sign() \
        .df

    # calculate resistance
    return ResistanceBuilder().with_busbar_voltage(u_res_feature_df) \
        .with_busbar_current(i_meas_feature_df) \
        .calculate_mean_resistance(circuit_type) \
        .convert_to_row(index=t_start_inj) \
        .fill_missing_columns() \
        .sort_columns_busbar_location() \
        .df
