import numpy as np
import pandas as pd

from lhcsmapi.analysis.diode.DiodeLeadResistanceQuery import DiodeLeadResistanceQuery
from lhcsmapi.Time import Time
from lhcsmapi.dbsignal.SignalAnalysis import SignalAnalysis


class DiodeLeadResistanceAnalysis(object):

    @staticmethod
    def set_to_zero_resistance_outside_of_range(u_diode, I_MEAS):
        calc_diff_u_diode = lambda x: x[-1] - x[0]
        u_diode['dU_DIODE'] = u_diode['U_DIODE_min'].rolling(11).apply(calc_diff_u_diode, raw=True)

        u_diode.loc[u_diode.index < u_diode.index[550], 'dU_DIODE'] = 0

        if u_diode['dU_DIODE'].max() > 0.05:
            index_cut = u_diode.query('dU_DIODE > 0.05').index[0]
        else:
            index_cut = max(u_diode.index)

        u_diode.loc[u_diode[I_MEAS] <= 100, 'RES'] = 0
        u_diode.loc[u_diode.index > index_cut, 'RES'] = 0

    @staticmethod
    def calculate_minimum_diode_and_current_leads_voltage(u_diode, U_DIODE):
        # "5" - toggling period of nQPS U_DIODE signals
        # "50s x 10 Hz + 20" delay to the diode full opening
        calc_u_diode_min = lambda x: min((x[0] + x[-1]) / 2, x[5])
        u_diode['U_DIODE_min'] = u_diode.loc[u_diode.index >= u_diode.index[535], U_DIODE].rolling(11).apply(
            calc_u_diode_min, raw=True).shift(-5)

        calc_u_res = lambda x: abs(x[5] - (x[0] + x[-1]) / 2)
        u_diode['U_RES'] = u_diode.loc[u_diode.index >= u_diode.index[535], U_DIODE].rolling(11).apply(calc_u_res, raw=True).shift(-5)

        u_diode.loc[u_diode.index < u_diode.index[540], ['U_RES', 'U_DIODE_min']] = 0

    @staticmethod
    def filter_resistance(u_diode):
        u_diode['RES'] = u_diode['RES'].rolling(41).median().values
        u_diode['RES_filtered'] = u_diode['RES'].rolling(window=10).mean()

    @staticmethod
    def extract_res_max_time_current_nxcals(u_diode, I_MEAS):
        res_max = u_diode['RES_filtered'].max()
        t_res_max = u_diode.idxmax(axis=0, skipna=True)['RES_filtered']
        i_res_max = u_diode.loc[u_diode.index == t_res_max, I_MEAS].values[0]
        return res_max, t_res_max, i_res_max

    @staticmethod
    def calculate_resistance_cals(magnet, circuit_name, i_circuit_sync, u_diode_sync):
        # Get metadata - signal SIGNAL_NAMES
        magnet_type, I_MEAS, U_DIODE = DiodeLeadResistanceQuery.get_magnet_type_i_meas_u_diode(magnet, circuit_name)

        # Resample circuit current
        t_end = 350 if magnet_type == "MB" else 150
        i_circuit_sync_resampled = SignalAnalysis.resample(i_circuit_sync, ts=0.1, t0=-50, t_end=t_end)

        # Join current and diode voltage into a single dataframe
        u_diode_sync[I_MEAS] = i_circuit_sync_resampled[I_MEAS].values

        # Recover resistive and diode minimum voltage
        DiodeLeadResistanceAnalysis.calculate_minimum_diode_and_current_leads_voltage(u_diode_sync, U_DIODE)

        # Calculate resistance
        u_diode_sync['RES'] = u_diode_sync.apply(lambda row: 1e+6 * row['U_RES'] / row[I_MEAS], axis=1, raw=False)

        # Set to 0 resistance values for too low current and with too high diode voltage derivative
        DiodeLeadResistanceAnalysis.set_to_zero_resistance_outside_of_range(u_diode_sync, I_MEAS)

        # Filter resistance
        DiodeLeadResistanceAnalysis.filter_resistance(u_diode_sync)

        return u_diode_sync

    @staticmethod
    def preprocess_pm_qps_signal(time_corr, timestamp_unix, pm_qps_signal, pm_signal_name):
        pm_qps_signal.drop(pm_qps_signal.index[:2], inplace=True)
        pm_qps_signal.drop(pm_qps_signal.index[-2:], inplace=True)
        pm_qps_signal.drop(pm_qps_signal[pm_qps_signal[pm_signal_name] == 0.0].index, inplace=True)
        pm_qps_signal = pm_qps_signal.set_index(((pm_qps_signal.index.values) - timestamp_unix) / 1e+9 + time_corr)
        return pm_qps_signal

    @staticmethod
    def preprocess_pm_pc_signal(iab_i_a, iab_i_a_name, index_end_diode, index_start_diode, timestamp_unix):
        iab_i_a.drop(iab_i_a[iab_i_a[iab_i_a_name] == 0.0].index, inplace=True)
        iab_i_a = iab_i_a.set_index(((iab_i_a.index.values) - timestamp_unix) / 1e+9)

        index_end = iab_i_a.index.get_loc(index_end_diode, method='nearest')
        index_start = iab_i_a.index.get_loc(index_start_diode, method='nearest')
        iab_i_a.drop(iab_i_a.index[index_end:], inplace=True)
        iab_i_a.drop(iab_i_a.index[:index_start], inplace=True)
        return iab_i_a

    @staticmethod
    def resample_pm_signal(df):
        df_resampled = df.reindex(
            df.index.union(np.linspace(df.index[0], df.index[-1], 2001))).interpolate('index')
        df_resampled.drop(df.index, axis=0, inplace=True)
        return df_resampled

    @staticmethod
    def calculate_resistence(t_max, df):
        df['U_RES'] = df['U_DIODE_A'] - df['U_DIODE_B']
        df.loc[df.index < t_max + 0.05, 'U_RES'] = 0
        df['RES'] = df['U_RES'] / df['IAB.I_A']
        df['RES'] = df['RES'].rolling(window=3, min_periods=1).median()
        df['RES'] = df['RES'] * 1e6
        df['time'] = t_max

    @staticmethod
    def extract_res_max_time_current_pm(df):
        res_max = int(df['RES'].max())
        t_res_max = df.idxmax(axis=0, skipna=True)['RES']
        i_res_max = df.reindex([t_res_max], method='nearest', tolerance=.0001)['IAB.I_A'].values[0]
        return res_max, t_res_max, i_res_max

    @staticmethod
    def calculate_resistance_pm(iab_i_a, u_diode_a, u_diode_b, u_ref, timestamp_iqps_qh, time_corr=0.0):

        timestamp_unix = Time.to_unix_timestamp(timestamp_iqps_qh)

        # U_DIODE
        u_diode_a = DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_diode_a,
                                                                         u_diode_a.columns[0])

        u_diode_b = DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix - 1e+6, u_diode_b,
                                                                         u_diode_b.columns[0])

        # U_REF
        u_ref = DiodeLeadResistanceAnalysis.preprocess_pm_qps_signal(time_corr, timestamp_unix, u_ref, u_ref.columns[0])

        # I_A
        index_start_diode = u_diode_a.index[0]
        index_end_diode = u_diode_a.index[-1]
        iab_i_a = DiodeLeadResistanceAnalysis.preprocess_pm_pc_signal(iab_i_a, iab_i_a.columns[0], index_end_diode,
                                                                      index_start_diode, timestamp_unix)

        # Resample signals
        u_diode_a_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(u_diode_a)
        u_diode_b_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(u_diode_b)
        u_ref_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(u_ref)
        iab_i_a_resampled = DiodeLeadResistanceAnalysis.resample_pm_signal(iab_i_a)

        # Concatenate signals
        res_df = pd.DataFrame()
        res_df['U_DIODE_A'] = u_diode_a_resampled[u_diode_a.columns[0]]
        res_df['U_DIODE_B'] = u_diode_b_resampled.values
        res_df['U_REF'] = u_ref_resampled.values
        res_df['IAB.I_A'] = iab_i_a_resampled.values

        # Calculate resistance
        t_max = u_diode_a.idxmax(axis=0, skipna=True)[0]
        DiodeLeadResistanceAnalysis.calculate_resistence(t_max, res_df)

        return res_df
