import numpy as np
import pandas as pd
from copy import deepcopy
from functools import reduce
from scipy import signal as s

from lhcsmapi.dbsignal.ExpSignalAnalysis import ExpSignalAnalysis
from lhcsmapi.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.dbsignal.SignalProcessing import SignalProcessing
from lhcsmapi.metadata.SignalMetadata import SignalMetadata


class QuenchHeaterAnalysis(object):

    @staticmethod
    def calculate_actual_reference_features(circuit_type, circuit_name, qh_sync_decay_df, qh_ref_sync_decay_df, magnet):
        features_qh = QuenchHeaterAnalysis.calculate_qh_discharge_features(circuit_type, circuit_name, qh_sync_decay_df,
                                                                           'Actual', magnet)
        features_qh_ref = QuenchHeaterAnalysis.calculate_qh_discharge_features(circuit_type, circuit_name,
                                                                               qh_ref_sync_decay_df, 'Reference', magnet)
        return pd.concat([features_qh, features_qh_ref], axis=0)

    @staticmethod
    def calculate_qh_discharge_features(circuit_type, circuit_name, qh_sync_decay_df, timestamp_qh, magnet):
        # Metadata
        U_HDS = SignalMetadata.get_signal_name(circuit_type, circuit_name, 'QH', 'PM', 'U_HDS')
        U_HDS = [U_HDS_el.replace('%CELL%', magnet) for U_HDS_el in U_HDS]
        I_HDS = SignalMetadata.get_signal_name(circuit_type, circuit_name, 'QH', 'PM', 'I_HDS')
        I_HDS = [I_HDS_el.replace('%CELL%', magnet) for I_HDS_el in I_HDS]

        char_time_qh = QuenchHeaterAnalysis.calculate_characteristic_decay_time(qh_sync_decay_df, timestamp_qh,
                                                                                U_HDS + I_HDS,
                                                                                methods=['tau_charge'])
        char_time_qh = char_time_qh.filter(regex='[UI]_HDS_\d{1}_tau_charge')
        char_time_qh.columns = [col.replace('tau_charge', 'tau [s]') for col in char_time_qh.columns]
        if circuit_type == 'RB':
            R_HDS = [u.replace('U', 'R') for u in U_HDS]
            u_i_initial_final_qh = QuenchHeaterAnalysis.calculate_initial_final_value(qh_sync_decay_df, timestamp_qh,
                                                                                      U_HDS + I_HDS)
            r_initial_qh = QuenchHeaterAnalysis.calculate_initial_resistance(u_i_initial_final_qh, timestamp_qh, U_HDS,
                                                                             I_HDS, R_HDS)
            r_initial_qh.columns = [col.replace('initial', 'initial [ohm]') for col in r_initial_qh.columns]
            features_qh = char_time_qh.join(r_initial_qh)
        else:
            features_qh = char_time_qh
        return features_qh

    @staticmethod
    def rate_qh_discharge(circuit_type, features_qh_concat):
        features_abs_diff_qh = features_qh_concat.diff().tail(1).abs()
        # reduce operation
        is_u_hds_tau_correct = reduce(np.logical_and,
                                      [features_abs_diff_qh['U_HDS_{}_tau [s]'.format(i)] < 0.003 for i in range(1, 4)])
        if circuit_type == 'RB':
            is_i_hds_tau_correct = reduce(np.logical_and,
                                          [features_abs_diff_qh['I_HDS_{}_tau [s]'.format(i)] < 0.003 for i in
                                           range(1, 4)])
            is_r_hds_initial_correct = reduce(np.logical_and,
                                              [features_abs_diff_qh['R_HDS_{}_initial [ohm]'.format(i)] < 0.5 for i in
                                               range(1, 4)])
            return is_u_hds_tau_correct.values[0] and is_i_hds_tau_correct.values[0] and \
                   is_r_hds_initial_correct.values[0]
        else:
            return is_u_hds_tau_correct.values[0]

    @staticmethod
    def calculate_initial_final_value(df, timestamp, SIGNAL_NAMES):
        row_feature_df = pd.DataFrame(index=[timestamp])
        for SIGNAL_NAME in SIGNAL_NAMES:
            row_feature_df[SIGNAL_NAME + '_initial'] = np.mean(s.medfilt(df[SIGNAL_NAME].values[0:19], 3))
            row_feature_df[SIGNAL_NAME + '_final'] = np.mean(s.medfilt(df[SIGNAL_NAME].values[-20:-1], 3))

        return row_feature_df

    @staticmethod
    def calculate_initial_resistance(row_feature_in_df, timestamp, U_NAMES, I_NAMES, R_NAMES):
        row_feature_df = pd.DataFrame(index=[timestamp])
        for u, i, r in zip(U_NAMES, I_NAMES, R_NAMES):
            row_feature_df[r + '_initial'] = row_feature_in_df[u + '_initial'] / row_feature_in_df[i + '_initial']
        return row_feature_df

    # ToDo: for the sake of flexibility, everything should be calculated as a method, including cTime_mean and std
    @staticmethod
    def calculate_characteristic_decay_time(df, timestamp, SIGNAL_NAMES,
                                            methods=["tau_charge", "tau_energy", "tau_lin_reg", "tau_exp_fit"]):
        row_feature_df = pd.DataFrame(index=[timestamp])
        for SIGNAL_NAME in SIGNAL_NAMES:
            c_time_exp_decay = ExpSignalAnalysis.get_characteristic_time_of_exponential_decay(
                df[SIGNAL_NAME].to_frame(), methods)
            row_feature_tau_df = SignalProcessing.convert_feature_dataframe_into_row(c_time_exp_decay, timestamp)
            row_feature_df = pd.concat([row_feature_df, row_feature_tau_df], axis=1, sort=False)
            tau_df = ExpSignalAnalysis.calculate_change_in_characteristic_of_exponential_decay(df[SIGNAL_NAME],
                                                                                               window=1000)
            row_feature_df[SIGNAL_NAME + '_cTime_mean'] = tau_df.mean()
            row_feature_df[SIGNAL_NAME + '_cTime_std'] = tau_df.std()

        return row_feature_df

    @staticmethod
    def calculate_similarity_matrix(df, timestamp, SIGNAL_NAMES):
        row_feature_df = pd.DataFrame(index=[timestamp])
        for SIGNAL_NAME in SIGNAL_NAMES:
            similarity_matrix = SignalAnalysis.calculate_similarity_matrix(df[SIGNAL_NAME]).fillna(-1)
            row_feature_sm_current_df = SignalProcessing.convert_feature_dataframe_into_row(similarity_matrix, timestamp)
            row_feature_df = pd.concat([row_feature_df, row_feature_sm_current_df], axis=1, sort=False)

        return row_feature_df

    @staticmethod
    def compare_curve_to_curve_in_group(decay_df, decay_ref_df, row_feature_df, SIGNAL_NAMES, envelope_amplitude,
                                        is_global_envelope_characteristic_time=False):
        # Filter Data
        decay_filtered_df = decay_df.rolling(window=50, min_periods=1).median()
        decay_filtered_ref = decay_ref_df.rolling(window=50, min_periods=1).median()

        # Normalize Data
        decay_filtered_norm_df = SignalAnalysis.normalize_df_of_signal_group(decay_filtered_df, [SIGNAL_NAMES])
        decay_filtered_norm_ref_df = SignalAnalysis.normalize_df_of_signal_group(decay_filtered_ref, [SIGNAL_NAMES])

        row_feature_c2c_df = pd.DataFrame(index=[row_feature_df.index])
        for SIGNAL_NAME in SIGNAL_NAMES:
            # Double the normal Time Constant
            if not is_global_envelope_characteristic_time:
                is_global_envelope_characteristic_time = 2 * row_feature_df[SIGNAL_NAME + '_tau_energy'].values[0]

            normalization_coefficient = 1 / (
                        decay_filtered_df[SIGNAL_NAME].max() - decay_filtered_df[SIGNAL_NAME].min())
            row_feature_c2c_df[
                SIGNAL_NAME + '_c2c_doubleTCEnvelope_globallynormalized'] = SignalAnalysis.compare_curve_to_curve(
                decay_filtered_norm_df[SIGNAL_NAME], decay_filtered_norm_ref_df[SIGNAL_NAME],
                envelope_amplitude * normalization_coefficient, is_global_envelope_characteristic_time)

        return row_feature_c2c_df

    @staticmethod
    def extract_features(df, event_time, U_HDS, I_HDS, R_HDS=None):

        feature_df = pd.DataFrame(index=[event_time])

        # Characteristic time of a pseudo-exponential decay
        c_time_df = QuenchHeaterAnalysis.calculate_characteristic_decay_time(df, event_time, I_HDS + U_HDS)
        # Initial final value
        init_final_val_df = QuenchHeaterAnalysis.calculate_initial_final_value(df, event_time, I_HDS + U_HDS)
        # Initial resistance
        if R_HDS is not None:
            initial_res_df = QuenchHeaterAnalysis.calculate_initial_resistance(init_final_val_df, event_time,
                                                                               U_HDS, I_HDS, R_HDS)
        else:
            initial_res_df = pd.DataFrame(index=[event_time])
        # Similarity matrix
        # # Normalize Data
        normalized_df = SignalAnalysis.normalize_df_of_signal_group(
            deepcopy(df), [U_HDS, I_HDS, R_HDS])
        similarity_matrix_df = QuenchHeaterAnalysis.calculate_similarity_matrix(normalized_df,
                                                                                timestamp=event_time,
                                                                                SIGNAL_NAMES=[U_HDS, I_HDS, R_HDS])
        return pd.concat([feature_df, c_time_df, init_final_val_df, initial_res_df, similarity_matrix_df], axis=1,
                         sort=False)

    @staticmethod
    def compare_curve_to_curve(df, df_ref, features_ref, I_HDS, R_HDS, U_HDS):
        row_feature_c2c_df = pd.DataFrame(index=[features_ref.index])
        c2c_voltage = QuenchHeaterAnalysis.compare_curve_to_curve_in_group(deepcopy(df[U_HDS]), deepcopy(df_ref[U_HDS]),
                                                                           features_ref, U_HDS, 20)

        c2c_current = QuenchHeaterAnalysis.compare_curve_to_curve_in_group(
            deepcopy(df[I_HDS]), deepcopy(df_ref[I_HDS]),
            features_ref, I_HDS, 5)

        c2c_resistance = QuenchHeaterAnalysis.compare_curve_to_curve_in_group(
            deepcopy(df[R_HDS]), deepcopy(df_ref[R_HDS]),
            features_ref, R_HDS, 0.5, is_global_envelope_characteristic_time=0.08)

        return pd.concat([row_feature_c2c_df, c2c_voltage, c2c_current, c2c_resistance], axis=1)

