import pandas as pd
from copy import deepcopy

from lhcsmapi.dbsignal.cals.CalsDbSignal import CalsDbSignal
from lhcsmapi.dbsignal.influx.InfluxDbSignal import InfluxDbSignal
from lhcsmapi.dbsignal.nxcals.NxcalsDbSignal import NxcalsDbSignal
from lhcsmapi.dbsignal.post_mortem.PmDbSignal import PmDbSignal
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities


class Signal(object):
    """Signal class encapsulates behavior for working with signal databases. It provides read and write methods for accessing
    Post Mortem, CALS, NXCALS, and InfluxDB APIs.
    """

    __time_units = {'PM': 'ns', 'CALS': 's', 'INFLUX': None, 'NXCALS': 'ns'}

    def __init__(self, df=None):

        self.__df_state = None
        self.__df = pd.DataFrame()
        if df is not None:
            self.df = df

    # region df property
    @property
    def df(self):
        return self.__df

    @df.setter
    def df(self, df):
        """Method validates that the passed DataFrame
        """
        if not isinstance(df, pd.DataFrame):
            raise TypeError(
                "You can only assign a pandas DataFrame to signal.df")

        self.__df = df

    # endregion
    def read(self, api, append=False, **kwargs):
        """Method allowing users to read signal from a given database api

            :param api: input signal database signal ("test_cals", "pm", "influx")
            :type api: string
            :param append: flag indicating whether the signal should be appended to the object's dataframe (True) or not (False). False emulates static behavior
            :type append: boolean
            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary
            :return: pandas.DataFrame

            - Example CALS::

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  I_MEAS = "RPTE.UA47.RB.A45:I_MEAS"
                  t_start = 1426220459491000000
                  t_end =   1426220479491000000
                  i_meas_df = Signal().read('cals', signal=I_MEAS, t_start=t_start, t_end=t_end, ldb=ldb)
                  print(i_meas_df.head())
                  >>>               RPTE.UA47.RB.A45:I_MEAS
                  >>> 1.426220e+09                  10071.5
                  >>> 1.426220e+09                  10076.5
                  >>> 1.426220e+09                  10081.5
                  >>> 1.426220e+09                  10086.5
                  >>> 1.426220e+09                  10091.5

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  I_MEAS = "RPTE.UA47.RB.A45:I_MEAS"
                  t_start = 1426220459491000000
                  duration = [(20, 's')]
                  i_meas_df = Signal().read('cals', signal=I_MEAS, t_start=t_start, duration=duration, ldb=ldb)
                  print(i_meas_df.head())
                  >>>               RPTE.UA47.RB.A45:I_MEAS
                  >>> 1.426220e+09                  10071.5
                  >>> 1.426220e+09                  10076.5
                  >>> 1.426220e+09                  10081.5
                  >>> 1.426220e+09                  10086.5
                  >>> 1.426220e+09                  10091.5

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  I_MEAS = "RPTE.UA47.RB.A45:I_MEAS"
                  t_start = 1426220469491000000
                  duration = [(10, 's'), (10, 's')]
                  i_meas_df = Signal().read('cals', signal=I_MEAS, t_start=t_start, duration=duration, ldb=ldb)
                  print(i_meas_df.head())
                  >>>               RPTE.UA47.RB.A45:I_MEAS
                  >>> 1.426220e+09                  10071.5
                  >>> 1.426220e+09                  10076.5
                  >>> 1.426220e+09                  10081.5
                  >>> 1.426220e+09                  10086.5
                  >>> 1.426220e+09                  10091.5

            - Expected Fail Response CALS::

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  I_MEAS = "RPTE.UA47.RB.A45:I_MEAS"
                  t_start = 1426220469491000000
                  duration = [(10, 's'), (10, 's')]
                  i_meas_df = Signal().read('cals', signal=I_MEAS, t_start=t_start, duration=duration)
                  print(i_meas_df.head())
                  >>> Exception: You have to provide a instance of pytimber.LoggingDB via the spark keyword

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  ldb = pytimber.LoggingDB()
                  I_MEAS = "RPTE.UA47.RB.A45:I_MEA"
                  t_start = 1426220469491000000
                  duration = [(10, 's'), (10, 's')]
                  i_meas_df = Signal().read('cals', signal=I_MEAS, t_start=t_start, duration=duration, ldb=ldb)
                  print(i_meas_df.head())
                  >>> WARNING:pytimber.pytimber:No variables found.
                  >>> Empty DataFrame
                  >>> Columns: []
                  >>> Index: []
                  >>> UserWarning: Querying CALS with the parameters signal: RPTE.UA47.RB.A45:I_MEA, start time: 1426220459.4910002 and end time: 1426220479.4910002 returned no data.
                  >>> An empty DataFrame was returned.
                  >>> ___________________________________________________________________________________
                  >>> Reproduce the error using the raw pytimber API:

                  >>> import pytimber
                  >>> ldb = pytimber.LoggingDB()
                  >>> ldb.get(RPTE.UA47.RB.A45:I_MEA, 1426220459.4910002, 1426220479.49100)

            - Example PM::

                  from lhcsmapi.dbsignal.Signal import Signal
                  system = "QPS"
                  className = "DQAMCNMB_PMHSU"
                  magnet = "B20L5"
                  signal = "I_HDS_4"
                  timestamp = 1426220469491000000
                  i_hds_4_df = Signal().read('pm', system=system, className=className, source=magnet,
                                             signal=signal, eventTime=timestamp)
                  print(i_hds_4_df.head())
                  >>>                       I_HDS_4
                  >>> 1426220469491020832  0.089648
                  >>> 1426220469491026040  0.091555
                  >>> 1426220469491031248  0.091555
                  >>> 1426220469491036456  0.093463
                  >>> 1426220469491041664  0.091555

            - Expected Fail Response PM::

                      from lhcsmapi.dbsignal.Signal import Signal
                      system = "QPS"
                      className = "DQAMCNMB_PMHSU"
                      magnet = "B20L5"
                      signal = "I_HDS_4"
                      timestamp = 1426220469491000000
                      i_hds_4_df = Signal().read('pm', system=system, className=className, source=magnet,
                                                    signal=signal, eventTime=timestamp)
                      print(i_hds_4_df.head())
                      >>> Querying Post Mortem failed using the following query:
                      >>> http://pm-api-pro/v2/pmdata/signal?system=QPS&className=DQAMCNMB_PMHS&source=B20L5&timestampInNanos=1426220469491000000&signal=I_HDS_4
        """

        api = api.upper()

        df = Signal._database_access_handler(api, 'read', **kwargs)
        if append:
            if self.__df_state is not None and self.__df_state != api:
                # if the the recent read request is different from the previous conversions are needed
                # convert both DataFrames
                self.df = SignalUtilities.convert_indices_to_timestamps(self.df, Signal.__time_units[self.__df_state])
                df = SignalUtilities.convert_indices_to_timestamps(df, Signal.__time_units[api])
                self.__df_state = 'INFLUX'
            else:
                self.__df_state = api
            df = pd.concat([df, self.df], axis=1)
        else:
            self.__df_state = api

        self.df = df

        return self.df

    @staticmethod
    def write(api, **kwargs):
        """
            **Writes df into a database**

            Method writes dataframe associated with an object to a selected database.

            :param api: input signal database signal ("influx")
            :type api: string
            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary

        """
        return Signal._database_access_handler(api, 'write', **kwargs)

    @staticmethod
    def _database_access_handler(api, access_method_identifier, **kwargs):

        db_access = Signal.execute_dbsignal_factory(api, **kwargs)
        try:
            access_method = getattr(db_access, access_method_identifier, None)
            return access_method(**kwargs)
        except AttributeError:
            raise AttributeError('{} Signal has no {} method'.format(api, access_method_identifier))

    @staticmethod
    def execute_dbsignal_factory(signal_type, **kwargs):
        """**Returns an object of one of the derived DbSignal classes according to the passed signalType**

            Method executes a factory producing signal database APIs.

            :param signal_type: input signal database signal ("test_cals", "pm", "influx")
            :type signal_type: string
            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary
            :return: datafrane

            - Example PM::

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("pm", duration = [(0, 's')])
                  >>> <lhcsmapi.post_mortem.PmDbSignal.PmDbSignal at 0x7f4ec387e780>

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("PM", duration = [(0, 's')])
                  >>> <lhcsmapi.post_mortem.PmDbSignal.PmDbSignal at 0x7f4ec387e780>

            - Example CALS::

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  spark = pytimber.LoggingDB()
                  Signal.executeDbSignalFactory("test_cals", spark=spark, duration = [(0, 's')])
                  >>> <lhcsmapi.test_cals.CalsDbSignal.CalsDbSignal at 0x7f4ea980c1d0>

                  from lhcsmapi.dbsignal.Signal import Signal
                  import pytimber
                  spark = pytimber.LoggingDB()
                  Signal.executeDbSignalFactory("CALS", spark=spark, duration = [(0, 's')])
                  >>> <lhcsmapi.test_cals.CalsDbSignal.CalsDbSignal at 0x7f4ea980c4e0>

            - Example INFLUX::

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("influx", duration = [(0, 's')])
                  >>> <lhcsmapi.influx.InfluxDbSignal.InfluxDbSignal at 0x7f4ea980c9b0>

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("INFLUX", duration = [(0, 's')])
                  >>> <lhcsmapi.influx.InfluxDbSignal.InfluxDbSignal at 0x7f4ea979a128>

            - Expected Fail Response::

                  from lhcsmapi.dbsignal.Signal import Signal
                  Signal.executeDbSignalFactory("nxcals", duration = [(0, 's')])
                  >>> TypeError: NXCALS key does not match any know APIs
                """

        signal_type_upper = signal_type.upper()
        if signal_type_upper == "PM":
            return PmDbSignal.build_signal(**kwargs)
        elif signal_type_upper == "CALS":
            return CalsDbSignal.build_signal(**kwargs)
        elif signal_type_upper == "INFLUX":
            return InfluxDbSignal.build_signal(**kwargs)
        elif signal_type_upper == "NXCALS":
            return NxcalsDbSignal.build_signal(**kwargs)
        else:
            raise TypeError('{} key does not match any known APIs: pm, test_cals, influx'.format(signal_type_upper))

    def get_synchronized_time_df(self, t0=None):

        output = SignalUtilities.synchronize_df(deepcopy(self.df), t0)
        return output
