import scipy.signal as spsig
import pandas as pd
import numpy as np


class SignalProcessing:
    """SignalProcessing class contains static methods for signal processing
    """

    # ToDo - remove and integrate the code with FeatureBuilder
    def first(x):
        return x.iloc[0]

    def last(x):
        return x.iloc[-1]

    __name_to_function = {"first": first, "last": last, "max": max, "min": min, "median": np.median, "std": np.std,
                          "mean": np.mean}

    @staticmethod
    def get_feature_dataframe(df, function_names_to_apply):
        """Method returns a dataFrame according to the input list of features based on the passed DataFrame object
        """

        functions_to_apply = [SignalProcessing.__name_to_function[x] for x in function_names_to_apply]

        if isinstance(df, pd.DataFrame) and not df.empty:
            return df.aggregate(functions_to_apply)
        else:
            return df

    @staticmethod
    def convert_feature_dataframe_into_row(df, timestamp):
        """Method converts a feature dataframe as a set of rows and columns into a single row
        """
        df.replace(np.nan, '', inplace=True)
        df1 = df.stack().swaplevel()
        df1.index = df1.index.map('{0[0]}_{0[1]}'.format)

        df2 = df1.to_frame().T
        values = df2.values[0]
        columns = df2.columns

        df = pd.DataFrame(columns=columns)
        df.loc[0] = values
        df.index = [timestamp]
        df.index.name = 'timestamp'
        df.replace('', np.nan, inplace=True)
        return df

    @staticmethod
    def find_value_with_time_closest_to_index_threshold(series, index_threshold):
        if not isinstance(series, (pd.Series, pd.DataFrame)):
            raise ValueError("First argument has to be an object of Series or DataFrame!")
        array = series.index
        idx = (np.abs(array - index_threshold)).argmin()
        return series.iloc[idx]

    # ToDo -  remove, it already exists in PyeDSL
    @staticmethod
    def filter_signal_names(df, filter_names, kernel_size=3):
        """Method filtering all columns in the passed DataFrame correlating with the SIGNAL_NAMES described in filterNames.
        It does this by applying scipy.signal's medfilt method to these columns
        """
        if isinstance(df, pd.DataFrame) and not df.empty:
            for filter_name in filter_names:
                pd.options.mode.chained_assignment = None
                df[filter_name] = spsig.medfilt(df[filter_name].to_numpy(), kernel_size)
                pd.options.mode.chained_assignment = "warn"
        return df

# ToDo: SignalDerivative class
    @staticmethod
    def calculate_time_derivative_of_signals(df, timeframe, source_signal_names, derivative_signal_names, kernel_size=51):
        """Method calculating time derivatives for resistance values contained in the passed DataFrame object
        """
        if isinstance(df, pd.DataFrame) and not df.empty:
            for source_name, derivative_name in zip(source_signal_names, derivative_signal_names):
                SignalProcessing.__calculate_time_derivative_of_signal(df, timeframe, source_name, derivative_name,
                                                                       kernel_size)

    @staticmethod
    def calculate_time_derivative_of_signal(df, kernel_size=51):
        """Method calculating time derivatives for signal contained in the passed DataFrame object
        """
        if isinstance(df, pd.DataFrame) and (len(df) == 1) and not df.empty:
            signal_name = df.columns[0]
            der_name = 'd{}'.format(signal_name)
            SignalProcessing.__calculate_time_derivative_of_signal(df, df.index, signal_name, der_name, kernel_size)
        else:
            raise ValueError('Wrong input type, please check the assumptions')

    @staticmethod
    def __calculate_time_derivative_of_signal(df, tf, source_signal, derivative_name, kernel_size=51):
        """Method calculating time derivative for a single Signal
        """
        df[derivative_name] = spsig.medfilt(np.gradient(df[source_signal]) / np.gradient(tf), kernel_size)
