import influxdb as influxdb
import pandas as pd
from datetime import datetime
import tzlocal
import warnings

from lhcsmapi.dbsignal.DbSignal import DbSignal
from lhcsmapi.Time import Time as tc


class InfluxDbSignal(DbSignal):
    """Class for encapsulating reading and writing methods of the Influx API
    """

    def __init__(self, client, df=None, query=None, tags={}, measurement=None, t_start=None, t_end=None, tz=None):

        super().__init__('Influx')
        self.client = client
        self.df = df

        self.query = query
        self.tags = tags
        self.measurement = measurement
        self.t_start = t_start
        self.t_end = t_end
        self.tz = tz

        self.protocol = 'json'

    @staticmethod
    def create_data_frame_client(username='', password='', host='dbod-lhc-sm.cern.ch', port=8081, database='test',
                                 ssl=True):
        """"Method creates and returns a DataFrameClient
        """
        return influxdb.DataFrameClient(host, port, username, password, database, ssl)

    @staticmethod
    def build_signal(**kwargs):
        """Method validates kwargs to create and return a InfluxDbSignal object.
        The method looks though kwargs for arguments to create a new DataFrameClient. If kwargs contains the necessary
        arguments, a client is created. If kwargs contain a DataFrameClient object, this is used instead.
        If neither a complete client or parameters to build one is provided, a default client is created.
        The resulting client is used to create a return a InfluxDbSignal object
        """

        # client keys:
        username = InfluxDbSignal.validate_key_in_kwargs('username', str, **kwargs)
        password = InfluxDbSignal.validate_key_in_kwargs('password', str, **kwargs)
        host = InfluxDbSignal.validate_key_in_kwargs('host', str, **kwargs)
        port = InfluxDbSignal.validate_key_in_kwargs('port', int, **kwargs)
        database = InfluxDbSignal.validate_key_in_kwargs('database', str, **kwargs)

        if username and password and host and port and database:
            client = InfluxDbSignal.create_data_frame_client(username, password, host, port, database)
        else:
            client = InfluxDbSignal.create_data_frame_client('root', 'root')

        client = InfluxDbSignal.validate_key_in_kwargs("client", influxdb.DataFrameClient, **kwargs) or client

        # df
        df = InfluxDbSignal.validate_key_in_kwargs('df', pd.DataFrame, **kwargs)

        # influx query keys:
        query = InfluxDbSignal.validate_key_in_kwargs('query', str, **kwargs)
        measurement = InfluxDbSignal.validate_key_in_kwargs('measurement', str, **kwargs)

        t_start = InfluxDbSignal.validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple,
                                                           **kwargs)
        t_end = InfluxDbSignal.validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple,
                                                         **kwargs)
        duration = InfluxDbSignal.validate_key_in_kwargs('duration', tuple, list, **kwargs)
        tz = InfluxDbSignal.validate_key_in_kwargs('tz', str, kwargs) or tzlocal.get_localzone().__str__()

        tags = InfluxDbSignal.validate_key_in_kwargs('tags', dict, **kwargs) or {}

        # calculate start and end time based of passed arguments, if both None, default query with will get all timestamps
        t_start, t_end = tc.get_query_period_in_unix_time(t_start, t_end, duration, tz)

        # get the default tags
        default_tags = {'system': InfluxDbSignal.validate_key_in_kwargs('system', str, **kwargs),
                        'circuit': InfluxDbSignal.validate_key_in_kwargs('circuit', str, **kwargs),
                        'magnet': InfluxDbSignal.validate_key_in_kwargs('magnet', str, **kwargs)}

        # enter default tags into tags if not already present
        for k, v in default_tags.items():
            if v is not None and k not in tags.keys():
                tags[k] = v

        return InfluxDbSignal(client, df, query, tags, measurement, t_start, t_end, tz)

    def write(self, signal=None, **kwargs):
        """Method writes the DataFrame contained in signal to the database specified in the signal instance's client field.
        Returns True if success
        """

        self.df = self.df.dropna(axis='columns')
        self.df = self.df.drop_duplicates()

        try:
            return self.client.write_points(dataframe=self.df,
                                            measurement=self.measurement,
                                            tags=self.tags,
                                            protocol=self.protocol,
                                            time_precision='n')
        except:
            raise Exception('writing the dataframe to the database failed')

    def read(self, signal=None, **kwargs):
        """Method queries the database specified in the client and returns the result as pandas DataFrame
        """

        query_string = self.query

        if not query_string:
            query_string = self.create_influx_db_query(self.measurement)

        try:
            query_result = self.client.query(query_string)
        except:
            raise Exception("Querying InfluxDB with query: {} failed".format(query_string))

        df = query_result.get(self.measurement)
        if df is None or df.empty:
            warnings.warn(
                "The query: {}, returned no data. Therefore an empty DataFrame is returned instead".format(
                    query_string))
            df = pd.DataFrame()

        if not df.empty:
            new_index = df.index.tz_convert(self.tz)
            df = df.set_index([new_index])

        return df

    # region create Query

    def create_influx_db_query(self, measurement, *field_keys):
        """Method returns a query string based on the internal fields
        """

        base_query = self.__create_base_query_string(measurement, *field_keys)
        time_constraint = self.__create_time_constraint_substring()
        tag_constraint = self.__create_tag_constraint_substring()

        if tag_constraint or time_constraint:
            base_query += ' where'

        if tag_constraint and time_constraint:
            time_constraint = ' and' + time_constraint

        return base_query + tag_constraint + time_constraint

    def __create_base_query_string(self, measurement, *field_keys):
        """Method returns a base query describing what fields to get from a certain measurement
        """

        selected_fields = ''

        # set the fields to select
        num_of_field_keys = len(field_keys)
        if num_of_field_keys > 0:
            for i in range(num_of_field_keys):
                selected_fields += field_keys[i]
                if i != num_of_field_keys - 1:
                    selected_fields += ', '
        else:
            selected_fields += '*'

        # create base query - default: 'select * from {measuremenent}'
        return 'select ' + selected_fields + ' from ' + measurement

    def __create_time_constraint_substring(self):
        """Method creates a substring containing the time constraints for a influxDB query
        """

        time_constraint_string = ''

        if self.t_start:
            start_time = tc(self.t_start, 'ns', 'UTC').string_timestamp[:-7]
            time_constraint_string += " time >= \'{}\'".format(start_time)

        if self.t_start and self.t_end:
            end_time = tc(self.t_end, 'ns', 'UTC').string_timestamp[:-7]
            time_constraint_string += " and time <= \'{}\'".format(end_time)

        return time_constraint_string

    def __create_tag_constraint_substring(self):
        """Method creates a substring containing the tag constraints for a influxDB query
        """

        tags_constraint_string = ''
        tag_set = [(k, v) for k, v in self.tags.items()]

        number_of_tags = len(tag_set)
        for i in range(number_of_tags):
            k, v = tag_set[i]
            tags_constraint_string += ' {} = \'{}\''.format(k, v)
            if i < number_of_tags - 1:
                tags_constraint_string += ' and'

        return tags_constraint_string

    # endregion
