import pandas as pd
import numpy as np

from lhcsmapi.Time import Time
from lhcsmapi.dbsignal.post_mortem.PmDbRequest import PmDbRequest
from lhcsmapi.dbsignal.DbSignal import DbSignal


class PmDbSignal(DbSignal):
    """PmDbSignal class encapsulates parsing of a signal obtained from the Post Mortem.
    """

    def __init__(self, system={}, className={}, source={}, signal={}, eventTime={}, time={}, value={},
                 pm_rest_api_path={}, show_warning={}):
        super().__init__('PM')
        self.system = system
        self.className = className
        self.source = source
        self.signal = signal
        self.eventTime = eventTime
        self.value = value
        self.time = time
        self.pm_rest_api_path = pm_rest_api_path
        self.show_warning = show_warning

    def __str__(self):
        return "Db: PM, signal: {}, eventTime = {} ({}), system={}, source={}, className={}".\
            format(self.signal, self.eventTime, Time.to_string(self.eventTime, 'ns'),
                   self.system, self.source, self.className)

    @staticmethod
    def build_signal(**kwargs):
        """**Builds a PmDbSignal object**

            Method builds a PmDbSignal from an input (keyword: argument) dictionary.
            It validates input kwargs and implements static builder design pattern to build a new PmDbSignal object.
            If arguments are not provided, default arguments are used

            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query
            :type kwargs: dictionary
            :return: initialized PmDbSignal object

            - Example::

                from lhcsmapi.post_mortem.PmDbSignal import PmDbSignal
                system = "QPS"
                className = "DQAMCNMB_PMHSU"
                source = "B20L5"
                signal = "I_HDS_4"
                eventTime = 1426220469491000000
                PmDbSignal().buildSignal(system=system, className=className, source=source,
                              signal=signal, eventTime=eventTime)
               >>> <lhcsmapi.post_mortem.PmDbSignal.PmDbSignal at 0x7f7d89ede550>
        """

        system = PmDbSignal.validate_key_in_kwargs('system', str, **kwargs) or {}
        className = PmDbSignal.validate_key_in_kwargs('className', str, **kwargs) or {}
        source = PmDbSignal.validate_key_in_kwargs('source', str, **kwargs) or {}
        signal = PmDbSignal.validate_key_in_kwargs('signal', str, **kwargs) or {}
        eventTime = PmDbSignal.validate_key_in_kwargs('eventTime', int, pd.Timestamp, np.int64, **kwargs) or {}
        time = PmDbSignal.validate_key_in_kwargs('time', list, **kwargs) or {}
        value = PmDbSignal.validate_key_in_kwargs('value', list, **kwargs) or {}
        pm_rest_api_path = PmDbSignal.validate_key_in_kwargs('pm_rest_api_path', str, **kwargs) or "http://pm-api-pro/v2/"
        show_warning = PmDbSignal.validate_key_in_kwargs('showWarning', str, **kwargs) or True

        return PmDbSignal(system, className, source, signal, eventTime, time, value, pm_rest_api_path, show_warning)

    def read(self, signal=None, **kwargs):
        """**Reads a signal from PM database**

            Method returns a pandas Dataframe containing a signal with an index formatted to ns.
            It requires the PmDbSignal object to be built with PmDbSignal.buildSignal() method.

            :param kwargs: dictionary of (keyword: argument) pairs which parametrize an db API query (optional)
            :type kwargs: dictionary
            :return: initialized PmDbSignal object

            - Example::

                from lhcsmapi.post_mortem.PmDbSignal import PmDbSignal
                system = "QPS"
                className = "DQAMCNMB_PMHSU"
                source = "B20L5"
                signal = "I_HDS_4"
                eventTime = 1426220469491000000
                pms = PmDbSignal.buildSignal(system=system, className=className, source=source,
                                              signal=signal, eventTime=eventTime)
                I_HDS_4 = pms.read()
                print(I_HDS_4.head())
                >>>                       I_HDS_4
                >>> 1426220469491020832  0.089648
                >>> 1426220469491026040  0.091555
                >>> 1426220469491031248  0.091555
                >>> 1426220469491036456  0.093463
                >>> 1426220469491041664  0.091555
        """
        try:
            self._parse_signal()
            df = self._get_signal_df(self.signal)
            # remove 0-valued indices
            df = df.loc[df.index != 0]
            # remove duplicated indices
            if 'allow_duplicate_index' in kwargs and (kwargs['allow_duplicate_index'] == True):
                pass
            else:
                df = df.loc[~df.index.duplicated(keep=False)]
                # remove non-monotonic indices
                df['index'] = df.index
                df['diff_index'] = df['index'].diff()
                df.fillna(1, inplace=True)
                df = df[df['diff_index'] > 0]
                df.drop(columns=['index', 'diff_index'], inplace=True)

            return df
        except:
            return pd.DataFrame()

    def _parse_signal(self):
        """Method looks for structure parameter and time in order to return a Dataframe with the signal.
        """
        response = PmDbRequest.get_response("pmdata/signal", False, self.show_warning, self.pm_rest_api_path,
                                            system=self.system, className=self.className, source=self.source,
                                            timestampInNanos=self.eventTime, signal=self.signal)

        if "v2" in self.pm_rest_api_path:
            content = response['content'][0]
            # find SIGNAL_NAMES
            name_of_value = PmDbSignal._get_names_of_value(self, content)
            # find time and value
            PmDbSignal._get_time_value(self, name_of_value, content)
        elif "v3" in self.pm_rest_api_path:
            self.eventTime = response['content'][0]['header']['timestamp']
            self.time = response['content'][0]['values']['circ.AQNTIME']['value']
            self.value = response['content'][0]['values'][self.signal]['value']

    def _get_signal_df(self, *columns):
        """Method returns a pandas DataFrame based on internal time and value properties.
        """
        return pd.DataFrame(self.value, self.time, columns=[col for col in columns])

    def _get_names_of_value(self, content):
        """Method returns a list of SIGNAL_NAMES contained in content.
        """

        size = len(content['namesAndValues'])
        name_of_value = ["" for x in range(0, size)]
        name_of_value[:] = [content['namesAndValues'][n]['name'] for n in range(0, size)]

        return name_of_value

    def _get_time_value(self, name_of_value, content):
        """Method sets time (in ns) and value from the content received with a PM signal query.
        """
        start_time, interv_time = None, None
        self.eventTime = content['rawDataEntry']['pmEventStamp']
        for s in range(0, len(name_of_value)):
            if "START_TIME_NSEC" in name_of_value[s]:  # search for StartTime
                start_time = content['namesAndValues'][s]['value']
            elif "INTERVAL_NSEC" in name_of_value[s]:  # search for Inteval
                interv_time = content['namesAndValues'][s]['value']
            elif any(keyword in name_of_value[s] for keyword in ["AQNTIME", "TIMESTAMP"]):  # search for Interval
                self.time = content['namesAndValues'][s]['value']
            elif self.signal in name_of_value[s]:  # search for SignalName
                self.value = content['namesAndValues'][s]['value']
            else:
                raise ValueError("Keyword {} not supported for this type of query.".format(name_of_value[s]))

        if self.time == {}:  # Time empty, has to be generated from t_start and intervTime
            self.time = [t for t in range(start_time, start_time + interv_time * len(self.value), interv_time)]
