from ipywidgets import Dropdown, Text, HBox, VBox, Select, Label, ToggleButtons, Layout
import getpass

from lhcsmapi.gui.ModuleMediator import ModuleMediator
from lhcsmapi.gui.pc.FgcBrowserAction import FgcBrowserAction
from lhcsmapi.gui.pc.FgcPmSearchBaseModule import FgcPmSearchBaseModule
from lhcsmapi.gui.qh.DateTimeBaseModule import DateTimeBaseModule
from lhcsmapi.gui.qh.QhBrowserAction import QhBrowserAction
from lhcsmapi.metadata.SignalMetadata import SignalMetadata


def get_circuit_names_or_types(circuit_type):
    """ Method returns a list of circuit names or types if the list of names is very long
    Depending on the circuit type the list is modified accordingly:
    - independent circuits: RB - full list of names
    - coupled circuits: RQ, 600A_RCBXHV, 600A_RCDO - list of reduced names to avoid duplications,
    e.g., RQ.A12 instead of RQD.A12 and RQF.A12
    - multiple circuits: 600A - family types in order to avoid very long lists,
    e.g. RU instead of RU.L4

    :param circuit_type: circuit type such as RB, RQ, 600A, IT
    :type circuit_type: str
    :return: list of circuit names or types
    """
    if circuit_type == 'RB':
        return SignalMetadata.get_circuit_names(circuit_type)
    elif circuit_type == 'RQ':
        circuit_names = SignalMetadata.get_circuit_names('RQ')
        circuit_names_rqd = list(filter(lambda name: 'RQD' in name, circuit_names))
        circuit_names_rqd_set = {circuit_name.replace('RQD', 'RQ') for circuit_name in circuit_names_rqd}
        return sorted(list(circuit_names_rqd_set))
    elif circuit_type == '600A':
        circuit_names = SignalMetadata.get_circuit_names('600A')
        circuit_names_without_rcbx = list(filter(lambda name: 'RCBX' not in name, circuit_names))
        circuit_names_without_rcbx_rcd = list(filter(lambda name: 'RCD' not in name, circuit_names_without_rcbx))
        circuit_names_without_rcbx_rcd_rco = list(
            filter(lambda name: 'RCO' not in name, circuit_names_without_rcbx_rcd))
        return sorted(list({circuit_name.split('.')[0] for circuit_name in circuit_names_without_rcbx_rcd_rco}))
    elif circuit_type == '600A_RCBXHV':
        circuit_names = SignalMetadata.get_circuit_names('600A')
        circuit_names_rcbxh = list(filter(lambda name: 'RCBXH' in name, circuit_names))
        circuit_names_rcbxh_set = {circuit_name.replace('RCBXH', 'RCBX') for circuit_name in circuit_names_rcbxh}
        return sorted(list(circuit_names_rcbxh_set))
    elif circuit_type == '600A_RCDO':
        circuit_names = SignalMetadata.get_circuit_names('600A')
        circuit_names_rcd = list(filter(lambda name: 'RCD' in name, circuit_names))
        circuit_names_rcd_set = {circuit_name.replace('RCD', 'RC') for circuit_name in circuit_names_rcd}
        return sorted(list(circuit_names_rcd_set))
    else:
        raise KeyError('Circuit type {} is not supported'.format(circuit_type))


class FgcPmSearchModuleMediator(ModuleMediator):

    def __init__(self, date_time_module: DateTimeBaseModule, fgc_pm_search_module: FgcPmSearchBaseModule,
                 circuit_type='RB'):
        self._circuit_type = circuit_type
        # First row
        circuit_names = get_circuit_names_or_types(circuit_type)
        description = 'Circuit type:' if circuit_type == '600A' else 'Circuit name'
        self.select_circuit_name_dropdown = Dropdown(
            options=circuit_names,
            description=description,
            disabled=False,
        )

        self._date_time_module = date_time_module
        self._date_time_module.mediator = self

        self.done_by_txt = Text(placeholder='Your name', description='Done by:', value=getpass.getuser(),
                                disabled=True)

        self.analysis_mode_tgl_btn = ToggleButtons(
            options=['Automatic', 'Manual'],
            description='Analysis:',
            disabled=False,
            button_style='',
            tooltips=['Select automatic for automatic execution, comment, and acceptance.',
                      'Select manual to execute, comment, and accept each analysis manually.']
        )

        first_row = HBox(children=[self.select_circuit_name_dropdown, self._date_time_module.widget])

        # Second row
        second_row = HBox(children=[self.analysis_mode_tgl_btn, self.done_by_txt])

        # Third row
        self._fgc_pm_search_module = fgc_pm_search_module
        self._fgc_pm_search_module.mediator = self
        third_row = fgc_pm_search_module.widget

        # Fourth row
        self.fgc_pm_sel = Select(
            options=[],
            rows=10,
            disabled=False,
            layout=Layout(width='50%')
        )
        fourth_row = HBox(children=[Label(value='FGC PM Entries:'), self.fgc_pm_sel])

        # Final GUI
        self.widget = VBox(children=[first_row, second_row, third_row, fourth_row])

    def notify(self, sender: object, event: str) -> None:
        start_date_time = self._date_time_module.get_start_date_time()
        end_date_time = self._date_time_module.get_end_date_time()
        circuit_type = self.get_circuit_type()
        circuit_name = self.get_circuit_name()

        if event == QhBrowserAction.DATE_CHANGED:
            pass
        elif event == FgcBrowserAction.FIND_BUTTON_CLICKED:
            self.fgc_pm_entries = self._fgc_pm_search_module.find_fgc_pm_entries_with_qps(start_date_time,
                                                                                          end_date_time, circuit_name)
            options = self._fgc_pm_search_module.convert_fgc_pm_entries_to_options(self.fgc_pm_entries)
            self.fgc_pm_sel.options = options
        else:
            raise KeyError(f'Action {event} not supported!')

    def get_circuit_type(self):
        return self._circuit_type

    def get_circuit_name(self):
        return self.select_circuit_name_dropdown.value

    def is_automatic_mode(self):
        return self.analysis_mode_tgl_btn.value == 'Automatic'

    def get_author(self):
        return self.done_by_txt.value

    def get_fgc_timestamp(self):
        if len(self.fgc_pm_sel.options) == 0:
            raise ValueError('The list of FGC PM Entries should contain at least one element. '
                             'Please adjust circuit signal as well as start and end date.')

        return self._fgc_pm_search_module.get_circuit_timestamps(self.fgc_pm_sel.index, self.fgc_pm_entries)

    def get_fgc_circuit(self):
        if len(self.fgc_pm_sel.options) == 0:
            raise ValueError('The list of FGC PM Entries should contain at least one element. '
                             'Please adjust circuit signal as well as start and end date.')

        return self._fgc_pm_search_module.get_circuit_names(self.fgc_pm_sel.index, self.fgc_pm_entries)
