import json
import os
import re
import pandas as pd
from typing import List, Union, Tuple

from lhcsmapi.metadata.MappingMetadata import MappingMetadata

selected_path = os.path.dirname(__file__)


class SignalMetadata(object):
    """ SignalMetadata class provides static methods for signal names and corresponding metadata. The class is capable
    of replacing wildcard in a signal name to form a complete signal, extracting sector name from a signal name, etc.
    The circuits are divided into main families:

    - RB: main dipole circuits
    - RQ: main quadrupole circuits
    - IPQ: individually powered quadrupoles (IPQ2, IPQ4, IPQ8 - the division depends on the QPS)
    - IPD: individually powered dipoles (IPD2_B1B2, IPD2 - the division depends on the QPS)
    - 600A: circuits with 600A power converters
    - 80-120A: circuits with rated current from 80-120 A
    - 60A: circuits with rated current of 60A

        """

    BMODE = "HX:BMODE"

    __paths = {'RB': 'circuit/RB_METADATA.json',
               'RQ': 'circuit/RQ_METADATA.json',
               'IT': 'circuit/IT_METADATA.json',
               'IPD2_B1B2': 'circuit/IPD2_B1B2_METADATA.json',
               'IPD2': 'circuit/IPD2_METADATA.json',
               'IPQ2': 'circuit/IPQ2_METADATA.json',
               'IPQ4': 'circuit/IPQ4_METADATA.json',
               'IPQ8': 'circuit/IPQ4_METADATA.json',
               '600A': 'circuit/600A_METADATA.json',
               '60A': 'circuit/60A_METADATA.json',
               '80-120A': 'circuit/80-120A_METADATA.json'
               }

    @staticmethod
    def _parse_metadata(circuit_name: str) -> dict:
        """Method parsing circuit metadata from a json file to a python dict.
        The dictionary contains mapping from circuit names to systems (and corresponding wildcards).
        In addition, there is a mapping from system to PM, (NX)CALS database metadata.
        The database metadata contains available signal names (with replacement of wildcards corresponding to
        circuit name) and corresponding parameters needed to contruct a valid query.

        :param circuit_name: circuit name, e.g., RB, RQ, IT
        :type circuit_name: str

        :return: dict with
        """

        try:
            full_path = os.path.join(selected_path, SignalMetadata.__paths[circuit_name.upper()])
        except KeyError:
            raise KeyError("{} is not mapped to a json file".format(circuit_name))

        try:
            with open(full_path) as json_file:
                return json.load(json_file)
        except FileNotFoundError:
            raise FileNotFoundError("Could not load {}".format(full_path))

    @staticmethod
    def get_beam_mode_details() -> pd.DataFrame:
        """ Method reading a csv file with a description of beam mode details and returning a pandas DataFrame.
        https://lhc-commissioning.web.cern.ch/lhc-commissioning/systems/data-exchange/published-topics/beam-modes.htm

        :return: pd.DataFrame with beam mode description (mapping from beam mode index to its meaning)
        """

        full_path = os.path.join(selected_path, 'beam_mode/beam_mode.csv')
        return pd.read_csv(full_path, sep=';')

    @staticmethod
    def get_circuit_types() -> list:
        """ Method returning a list of circuit types to which there exists a metadata dictionary

        :return: list of circuit types with metadata
        """

        return list(SignalMetadata.__paths.keys())

    @staticmethod
    def get_circuit_names(circuit_type: Union[str, List[str]]) -> List[str]:
        """ Method returning a list of circuit names for either an input circuit type or a list of circuit types.

        :param circuit_type: either a str or a list of str with circuit types for which circuit names are returned
        :type circuit_type: str, list
        :return: list of circuit names for input circuit type(s)
        """

        circuit_types = circuit_type if type(circuit_type) is list else [circuit_type]
        circuit_names = []
        for circuit_type in circuit_types:
            circuit_names.extend(SignalMetadata._parse_metadata(circuit_type)['CIRCUITS'].keys())
        return circuit_names

    @staticmethod
    def get_system_types_per_circuit_type(circuit_type: str) -> list:
        """ Method returning a list of systems for an input circuit_type

        :param circuit_type: a circuit type for which a list of systems is requested
        :type circuit_type: str

        :return: list of systems corresponding to an input circuit type
        """
        circuit_metadata = SignalMetadata._parse_metadata(circuit_type)
        types = set()
        for circuit_name in circuit_metadata["CIRCUITS"].keys():
            for key, value in circuit_metadata["CIRCUITS"][circuit_name].items():
                if type(value) is dict:
                    types.add(value['TYPE'])
                elif type(value) is list:
                    for element in value:
                        types.add(element['TYPE'])
        return list(types)

    @staticmethod
    def get_system_types_per_circuit_name(circuit_type: str, circuit_name: str) -> List[str]:
        """ Method returning systems corresponidng to an input circuit type and name

        :param circuit_type: circuit type
        :type circuit_type: str
        :param circuit_name: circuit name corresponding to circuit type
        :type circuit_name: str

        :return: list of systems corresponding to an input circuit type and name (
        """
        circuit_metadata = SignalMetadata._parse_metadata(circuit_type)
        types = set()
        if circuit_name in circuit_metadata["CIRCUITS"].keys():
            for key, value in circuit_metadata["CIRCUITS"][circuit_name].items():
                if type(value) is dict:
                    types.add(value['TYPE'])
                elif type(value) is list:
                    for element in value:
                        types.add(element['TYPE'])
            return list(types)
        else:
            raise ValueError('Circuit name %s not supported for circuit type %s' % (circuit_type, circuit_name))

    @staticmethod
    def get_sector_names_from_components(circuit_type: str, system: str, *keys: Union[str, List[str]]) \
            -> Union[str, List[str]]:
        """ Method reading sector name from a component. Useful in case of PM queries where a search for events within
        a given time interval may return events with systems not in the analysed sector.

        :param circuit_type: circuit type
        :type circuit_type: str
        :param system: name of system
        :type system: name
        :param keys: components for which sector names are retrieved
        :type keys: str, list

        :return: sector(s) as either a str or list
        """
        if len(keys) == 1 and isinstance(keys[0], list):
            keys = keys[0]

        try:
            metadata = SignalMetadata._parse_metadata(circuit_type)['REL_IP_POS_TO_CIRCUIT_NAME']
            if any([system == sys for sys in ['MAGNET', 'BUSBAR', 'BUSBAR_RQD', 'BUSBAR_RQF']]):
                reg_exp = SignalMetadata._parse_metadata(circuit_type)[system]['REL_IP_POS_REG_EXP']
            elif system == "PC":
                reg_exp = SignalMetadata._parse_metadata(circuit_type)[system]['REL_IP_POS_REG_EXP']
                if len(keys) == 1:
                    return re.findall(reg_exp, keys[0])[0]
                else:
                    return [re.findall(reg_exp, key)[0] for key in keys]
            else:
                raise KeyError('System {} does not support this feature'.format(system))
            return SignalMetadata.__get_arc_names_from_magnet_names(metadata, reg_exp, *keys)
        except KeyError as e:
            raise KeyError(str(e) + "in object parsed from {}".format(SignalMetadata.__paths[circuit_type]))

    @staticmethod
    def __get_arc_names_from_magnet_names(metadata, reg_exp, *keys):
        if len(keys) == 1:
            return SignalMetadata.__get_sector_name_from_magnet_name(metadata, reg_exp, keys[0])
        else:
            return [SignalMetadata.__get_sector_name_from_magnet_name(metadata, reg_exp, key) for key in keys]

    @staticmethod
    def __get_sector_name_from_magnet_name(metadata, reg_exp, key):
        found_reg_exp = re.findall(reg_exp, key)
        if len(found_reg_exp) != 1:
            raise ValueError("Key {} does not contain proper MB magnet signal".format(key))
        else:
            key = found_reg_exp[0]
            if key not in metadata.keys():
                raise KeyError("Key {} does not exist ".format(key))
            else:
                return metadata[key]

    @staticmethod
    def get_circuit_name_metadata(circuit_type: str, circuit_name: str) -> dict:
        """ Method returning full metadata dictionary corresponding to a circuit name of a given circuit type.

        :param circuit_type: circuit type
        :type circuit_type: str
        :param circuit_name: circuit name corresponding to circuit type
        :type circuit_name: str

        :return: dict with metadata for a circuit name
        """
        circuit_metadata = SignalMetadata._parse_metadata(circuit_type)
        circuit_names = circuit_metadata["CIRCUITS"].keys()
        if circuit_name in circuit_names:
            return circuit_metadata["CIRCUITS"][circuit_name]
        else:
            raise KeyError("Circuit signal {} not present in {}.".format(circuit_name, circuit_names))

    @staticmethod
    def get_circuit_signal_database_metadata(circuit_type: str, circuit_name: str, system: str, database: str) -> dict:
        """ Method returning metadata corresponding to circuit type, circuit name, system, and database.

        :param circuit_type: circuit type
        :type circuit_type: str
        :param circuit_name: circuit name corresponding to circuit type
        :type circuit_name: str
        :param system: name of a system
        :type system: str
        :param database: name of a database (PM, CALS, NXCALS)
        :type database: str

        :return: dict with metadata
        """
        circuit_metadata = SignalMetadata._parse_metadata(circuit_type)

        # Check if signal type is available
        system_types = SignalMetadata.get_system_types_per_circuit_type(circuit_type)
        if system not in system_types:
            raise KeyError("System type {} not present in {}.".format(system, system_types))

        circuit_signal_type_metadata = circuit_metadata[system]

        # Check if database is available
        database_names = circuit_signal_type_metadata.keys()
        if database not in database_names:
            raise KeyError("Database signal {} not present in {}.".format(database, database_names))

        # update the source wildcard
        circuit_name_metadata = SignalMetadata.get_circuit_name_metadata(circuit_type, circuit_name)
        if database == "PM":
            source = circuit_signal_type_metadata[database]['source']
            if source in circuit_name_metadata.keys():
                circuit_signal_type_metadata[database]['source'] = circuit_name_metadata[source]
            elif source in circuit_name_metadata[system].keys():
                circuit_signal_type_metadata[database]['source'] = circuit_name_metadata[system][source]
        elif database == "NXCALS" and ('device' in circuit_signal_type_metadata[database]):
            device = circuit_signal_type_metadata[database]['device']
            if device in circuit_name_metadata.keys():
                circuit_signal_type_metadata[database]['device'] = circuit_name_metadata[device]
            elif device in circuit_name_metadata[system].keys():
                circuit_signal_type_metadata[database]['device'] = circuit_name_metadata[system][device]
        return circuit_signal_type_metadata[database]

    @staticmethod
    def get_variable_names(circuit_type: str, circuit_name: str, system: str, database: str) -> List[str]:
        """ Method returning an alphabetically sorted list of variable names for a given circuit type, circuit name,
        system, and database.

        :param circuit_type: circuit type
        :type circuit_type: str
        :param circuit_name: circuit name corresponding to circuit type
        :type circuit_name: str
        :param system: name of a system
        :type system: str
        :param database: name of a database (PM, CALS, NXCALS)
        :type database: str

        :return: list of variable names
        """

        circuit_signal_type_db_metadata = \
            SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)

        if database == "PM":
            return sorted(list(circuit_signal_type_db_metadata.keys() - {"system", "className", "source"}))
        elif database == "NXCALS":
            return sorted(list(circuit_signal_type_db_metadata.keys() - {"system", "device", "property"}))
        else:
            return sorted(list(circuit_signal_type_db_metadata.keys()))

    @staticmethod
    def get_signal_name(circuit_type: str, circuit_name: Union[str, List[str]], system: str, database: str,
                        signal: Union[str, List[str]]) -> Union[str, List[str]]:
        """ Method returning a  list of signal names for a given circuit type, circuit name, system, and database.
        The method updates the wildcards with metadata corresponding to a circuit name.

        :param circuit_type: circuit type
        :type circuit_type: str
        :param circuit_name: circuit name corresponding to circuit type
        :type circuit_name: str
        :param system: name of a system
        :type system: str
        :param database: name of a database (PM, CALS, NXCALS)
        :type database: str
        :param signal: name of a signal
        :type signal: str

        :return: list of signal names (with updated wildcards)
        """

        circuit_name = circuit_name if isinstance(circuit_name, list) else [circuit_name]
        signal = signal if isinstance(signal, list) else [signal]

        signal_names: List[str] = []
        for cn in circuit_name:
            for vn in signal:
                var_names = SignalMetadata. \
                    _update_wildcards_in_signal_name_per_circuit_name(circuit_type, cn, database, system, vn)

                if type(var_names) is list:
                    signal_names.extend(var_names)
                else:
                    signal_names.append(var_names)

        return signal_names[0] if len(signal_names) == 1 else signal_names

    @staticmethod
    def _update_wildcards_in_signal_name_per_circuit_name(circuit_type: str, circuit_name: str, database: str,
                                                          system: str, signal: str) -> Union[str, List[str]]:
        circuit_signal_type_db_metadata = SignalMetadata. \
            get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)
        signal = circuit_signal_type_db_metadata[signal]
        # if signal is a list
        if isinstance(signal, list):
            return [SignalMetadata._update_wildcards_in_signal_name(circuit_type, circuit_name, system, vn) for vn in
                    signal]
        # if signal is a scalar
        else:
            return SignalMetadata._update_wildcards_in_signal_name(circuit_type, circuit_name, system, signal)

    @staticmethod
    def _update_wildcards_in_signal_name(circuit_type: str, circuit_name: str, system: str, signal: str) \
            -> Union[str, List[str]]:
        """ Method updating wildcards in a signal name based on circuit name metadata.

        :param circuit_type: circuit type
        :type circuit_type: str
        :param circuit_name: circuit name corresponding to circuit type
        :type circuit_name: str
        :param system: name of a system
        :type system: str
        :param signal: name of a signal
        :type signal: str

        :return: updated signal(s) either as a string or a list
        """

        # get all wildcards to be replaced
        circuit_name_metadata = SignalMetadata.get_circuit_name_metadata(circuit_type, circuit_name)
        # # for circuit_name
        for key, value in circuit_name_metadata.items():
            if "%" in key:
                signal = signal.replace(key, value)
        # # for system
        system_metadata = circuit_name_metadata[system]
        for key, value in system_metadata.items():
            if key != "TYPE" and key in signal:
                if type(value) is list:
                    return [signal.replace(key, val) for val in value]
                else:
                    return signal.replace(key, value)
        return signal

    @staticmethod
    def get_database_names(circuit_type: str, system: str) -> List[str]:
        """ Method returning a list of databases for a given circuit type and system

        :param circuit_type: circuit type
        :type circuit_type: str
        :param system: name of a system
        :type system: str

        :return: list of database names (e.g., PM, CALS, NXCALS)
        """
        circuit_metadata = SignalMetadata._parse_metadata(circuit_type)
        system_types = circuit_metadata.keys()
        if system not in system_types:
            raise KeyError("System type {} not present in {}.".format(system, system_types))

        system_type_metadata = circuit_metadata[system]

        database_names = []
        for key, value in system_type_metadata.items():
            if any(db in key for db in ["PM", "CALS", "NXCALS"]) and (len(value) > 0):
                database_names.append(key)

        return database_names

    @staticmethod
    def get_circuit_name_from_magnet_or_crate(circuit_type: str, magnet_or_crate: str) -> str:
        """ Method returning a circuit name corresponding to a magnet/crate of a given circuit type

        :param circuit_type: circuit type
        :type circuit_type: str
        :param magnet_or_crate: either a magnet or crate name (e.g., MB.A31L2)
        :type magnet_or_crate: str

        :return: str with a circuit name corresponding to a magnet/crate of a given circuit type
        """
        if circuit_type == 'RB':
            reg_exp = SignalMetadata._parse_metadata(circuit_type)['MAGNET']['REL_IP_POS_REG_EXP']
            suffix_to_sector_name = SignalMetadata._parse_metadata(circuit_type)['REL_IP_POS_TO_CIRCUIT_NAME']
            found_reg_exp = re.findall(reg_exp, magnet_or_crate)
            return suffix_to_sector_name[found_reg_exp[0]]
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))

    @staticmethod
    def get_wildcard_for_circuit_system_db_signal(circuit_type: str, circuit_name: str, system: str, database: str,
                                                  signal: str) -> Tuple[str, str]:
        """ Method returning wildcards for a given circuit type, circuit name, system, database, and signal

        :param circuit_type: circuit type
        :type circuit_type: str
        :param circuit_name: circuit name corresponding to circuit type
        :type circuit_name: str
        :param system: name of a system
        :type system: str
        :param database: name of a database (PM, CALS, NXCALS)
        :type database: str
        :param signal: name of a signal
        :type signal: str

        :return: a tuple of wildcard type and wildcard
        """

        if any([system == sys for sys in ['QH', 'QDS']]) and circuit_type == 'RB':
            if database == "PM":
                wildcard = MappingMetadata.get_cells_for_circuit_names(circuit_type, circuit_name)
                return 'Cell', wildcard
            elif database == 'CALS':
                wildcard = MappingMetadata.get_magnets_for_circuit_names(circuit_type, circuit_name)
                return 'Magnet', wildcard
            else:
                raise KeyError("Database type {} not present in {}.".format(database, ['PM', 'CALS']))
        elif 'VF' in system and circuit_type in ['RB', 'RQ']:
            wildcard = MappingMetadata.get_vf_for_circuit_names(circuit_type, circuit_name)
            return 'Magnet', wildcard
        elif "BUSBAR" in system:
            wildcard = MappingMetadata.get_busbars_for_circuit_name(circuit_type, circuit_name)
            return 'Busbar', wildcard
        elif system == 'DIODE_RB' and circuit_type == 'RB':
            if signal == 'U_REF_N1':
                wildcard = MappingMetadata.get_crates_for_circuit_names(circuit_type, circuit_name)
                return 'Crate', wildcard
            else:
                wildcard = MappingMetadata.get_magnets_for_circuit_names(circuit_type, circuit_name)
                return 'Magnet', wildcard
        elif system in ['DIODE_RQD', 'DIODE_RQF'] and circuit_type == 'RQ':
            if signal == 'U_REF_N1':
                wildcard = MappingMetadata.get_crates_for_circuit_names(circuit_type, circuit_name)
                return 'Crate', wildcard
            else:
                wildcard = MappingMetadata.get_magnets_for_circuit_names(circuit_type, circuit_name)
                # remove wildcards for which U_DIODE_RQx is NaN
                magnets_with_nan = SignalMetadata._get_rq_magnet_names_without_u_diode(circuit_name)
                wildcard = [wc for wc in wildcard if wc not in magnets_with_nan]
                return 'Magnet', wildcard
        return ()

    @staticmethod
    def _get_rq_magnet_names_without_u_diode(circuit_name: str) -> List[str]:
        """Method returning RQ magnet names for which a U_DIODE signal is not present

        :param circuit_name: RQ circuit name
        :type circuit_name: str

        :return: list of RQ magnet names without a diode
        """

        magnet_to_qps_crate_df = MappingMetadata.read_magnet_to_cell_qps_crate('RQ')
        mask_magnets_with_nan = (magnet_to_qps_crate_df['Circuit'] == circuit_name) & \
                                (magnet_to_qps_crate_df['Crate U_DIODE_RQx'].isna())
        return magnet_to_qps_crate_df[mask_magnets_with_nan]['Magnet'].values

    @staticmethod
    def update_pm_source_with_wildcard(source_in: str, circuit_type: str, system: str, signal: str,
                                       wildcard: Union[str, List[str]]) -> Union[str, List[str]]:
        """ Method updating wildcard in PM source

        :param source_in: input source with a wildcard
        :type source_in: str
        :param circuit_type: circuit type
        :type circuit_type: str
        :param system: system name
        :type system: str
        :param signal: signal name
        :type signal: str
        :param wildcard: wildcard to be updated
        :type wildcard: str

        :return: str with source (after updating wildcard)
        """

        wildcard = wildcard if isinstance(wildcard, list) else [wildcard]

        if any([system == sys for sys in ['QH', 'QDS']]):
            return wildcard
        elif system == 'BUSBAR':
            return MappingMetadata.get_crate_name_from_busbar_names(circuit_type, wildcard)
        elif 'DIODE' in system:
            if signal == 'U_REF_N1':
                return wildcard
            else:
                return [MappingMetadata.get_crate_name_from_magnet_name(circuit_type, wc) for wc in wildcard]
        elif system == 'VF':
            return [MappingMetadata.get_crate_name_from_magnet_name(circuit_type, wc) for wc in wildcard]
        # for VF, PC, EE, LEADS_EVEN, LEADS_ODD
        else:
            return source_in

    @staticmethod
    def update_pm_signal_names_with_wildcard(signals_in: Union[str, List[str]], system: str, signal: str,
                                             wildcard: Union[str, List[str]]) -> List[str]:
        """ Method updating PM signal names with a wildcard

        :param signals_in: list of signal names with wildcards to be updated
        :type signals_in: list of str
        :param system: system name
        :type system: str
        :param signal: name of signal (used only for U_REF_N1 which requires an update of a qps crate)
        :type signal: str
        :param wildcard: wildcard to be replaced
        :type wildcard: str

        :return: list of str
        """
        wildcard = wildcard if isinstance(wildcard, list) else [wildcard]
        signals_in = signals_in if isinstance(signals_in, list) else [signals_in]

        if system == 'QDS':
            return [signal_name.replace("%CELL%", wc) for wc in wildcard for signal_name in signals_in]
        elif system == 'BUSBAR':
            return [signal_name.replace("%BUSBAR%", wc) for wc in wildcard for signal_name in signals_in]
        elif 'DIODE' in system:
            if signal == 'U_REF_N1':
                return [signal_name.replace("%QPS_CRATE%", wc) for wc in wildcard for signal_name in signals_in]
            else:
                return [signal_name.replace("%MAGNET%", wc) for wc in wildcard for signal_name in signals_in]
        elif system == 'VF':
            return [signal_name.replace("%MAGNET%", wc) for wc in wildcard for signal_name in signals_in]
        # for PC, EE, LEADS_EVEN, LEADS_ODD
        else:
            return signals_in

    @staticmethod
    def update_cals_signal_names_with_wildcard(signals_in: Union[str, List[str]], system: str,
                                               wildcard: Union[str, List[str]]) -> List[str]:
        """ Method updating CALS signal names with a wildcard

        :param signals_in: list of signal names with wildcards to be updated
        :type signals_in: list of str
        :param system: system name
        :type system: str
        :param wildcard: wildcard to be replaced
        :type wildcard: str, list of str

        :return: list of str
        """

        wildcard = wildcard if isinstance(wildcard, list) else [wildcard]
        signals_in = signals_in if isinstance(signals_in, list) else [signals_in]

        if any([system == sys for sys in ['QH', 'QDS', 'DIODE', 'VF']]):
            return [signal_name.replace("%MAGNET%", wc) for wc in wildcard for signal_name in signals_in]
        elif system == 'BUSBAR':
            return [signal_name.replace("%BUSBAR%", wc) for wc in wildcard for signal_name in signals_in]
        # for PC, EE, LEADS_EVEN, LEADS_ODD
        else:
            return signals_in
