import math
import matplotlib.pyplot as plt
from sklearn.linear_model import LinearRegression
import numpy as np
import warnings
from copy import deepcopy
from IPython.display import display, HTML


class Assertion(object):
    def set_dfs(self, dfs):
        self.dfs = dfs

    def set_time_range(self, t_start, t_end):
        self.t_start = t_start
        self.t_end = t_end


class AssertionBuilder(object):

    def __init__(self):
        self.assertion = Assertion()

    def with_signal(self, dfs):
        self.assertion.set_dfs(dfs)

        return AssertionBuilderSignal(self.assertion)

    def with_feature(self, dfs):
        self.assertion.set_dfs(dfs)

        return AssertionBuilderFeature(self.assertion)


def plot_with_value_range(dfs, value_min, value_max):
    fig, ax = plt.subplots(figsize=(15, 7))

    for df in dfs:
        if not df.empty:
            df.plot(ax=ax, grid=True)

    ax.legend(loc='upper right')
    ax.tick_params(labelsize=15)
    ax.set_xlabel("time, [s]", fontsize=15)

    ax.axhspan(ymin=value_min, ymax=value_max, facecolor='xkcd:yellowgreen', alpha=0.5)

    y_lim_min = 0.9 * value_min if value_min > 0 else 1.1 * value_min
    y_lim_max = 1.1 * value_max if value_max > 0 else 0.9 * value_max

    ax.set_ylim((y_lim_min, y_lim_max))

    return fig, ax


class AssertionBuilderFeature(object):

    def __init__(self, assertion):
        self.assertion = assertion

    def has_min_max_value(self, feature, min_value, max_value):
        # Assertion
        features_outside_range = self.assertion.dfs[(self.assertion.dfs[feature] < min_value)
                                                    | (self.assertion.dfs[feature] > max_value)]
        if len(features_outside_range) > 0:
            warnings.warn('The following features are outside of range.')
            display(HTML(features_outside_range.to_html()))

        fig, ax = plt.subplots(figsize=(30, 10))
        self.assertion.dfs[feature].plot(figsize=(30, 10), kind='bar', color='red', grid=True, ax=ax)
        ax.tick_params(labelsize=15)
        ax.axhspan(ymin=min_value, ymax=max_value, facecolor='xkcd:yellowgreen', alpha=0.5)

        return AssertionBuilderFeaturePlot(fig, ax, features_outside_range)


def plot_between_two_references(dfs, envelope_df):
    fig, ax = plt.subplots(figsize=(30, 10))

    for df in dfs:
        df.plot(ax=ax, grid=True)

    ax.legend(loc='lower left')
    ax.set_xlabel('time, [s]', fontsize=15)
    ax.tick_params(labelsize=15)

    ax.fill_between(envelope_df.index, envelope_df['envelope_pos'], envelope_df['envelope_neg'],
                    color='xkcd:yellowgreen', alpha=0.5)

    return fig, ax


class AssertionBuilderSignal(object):

    def __init__(self, assertion):
        self.assertion = assertion

    def with_time_range(self, t_start, t_end):
        self.assertion.set_time_range(t_start, t_end)

        return AssertionBuilderSignalTimeRange(self.assertion)

    def for_time_greater_than(self, t_greater):
        self.assertion.set_time_range(None, t_greater)

        return AssertionBuilderSignalTimeRange(self.assertion)

    def has_min_max_value(self, value_min, value_max):
        # Assertion
        is_signal_correct = True
        for df in self.assertion.dfs:
            col = df.columns[0]
            df_above_thres = df[(df[col] < value_min) | (df[col] > value_max)]
            if len(df_above_thres) > 0:
                warnings.warn('{} outside of the [{}, {}] mV threshold'.format(col, value_min, value_max))
                is_signal_correct = False

        # Plot
        fig, ax = plot_with_value_range(self.assertion.dfs, value_min, value_max)
        return AssertionBuilderSignalPlot(fig, ax, is_signal_correct)

    def compare_to_reference(self, signal_ref_dfs, abs_margin, scaling=1):
        fig, ax = plt.subplots(figsize=(15, 7))
        is_signal_correct = True
        for signal_df, signal_ref_df in zip(self.assertion.dfs, signal_ref_dfs):
            col = signal_ref_df.columns[0]
            col_ref = '{}_REF'.format(col)
            col_min = '{}_MIN'.format(col)
            col_max = '{}_MAX'.format(col)
            signal_ref_df = signal_ref_df * scaling
            signal_ref_df = signal_ref_df.rename(columns={col: col_ref})
            idx = signal_ref_df.index.union(signal_df.index)
            signal_comparison_df = signal_ref_df.reindex(idx).interpolate('index')
            signal_comparison_df = signal_comparison_df.join(signal_df).dropna()

            signal_comparison_df[col_min] = signal_comparison_df[col_ref] - abs_margin
            signal_comparison_df[col_max] = signal_comparison_df[col_ref] + abs_margin

            # Compare two curves
            signal_comparison_df['Comparison'] = signal_comparison_df.apply(
                lambda row: (row[col] >= row[col_min]) and (row[col] <= row[col_max]), axis=1)

            if any(signal_comparison_df['Comparison'] == False):
                warnings.warn('Signal {0} outside of range [-{1}, {1}]!'.format(col, abs_margin))
                is_signal_correct = False

            # Plot
            signal_df.plot(ax=ax)
            signal_ref_df.plot(ax=ax, grid=True, style='--')

            ax.fill_between(signal_comparison_df.index, signal_comparison_df[col_min],
                            signal_comparison_df[col_max], color='xkcd:yellowgreen', alpha=0.5)
            ax.tick_params(labelsize=15)
            ax.set_xlabel("time, [s]", fontsize=15)

        return AssertionBuilderSignalPlot(fig, ax, is_signal_correct)

    def is_between_two_references(self, first_ref, first_scaling, second_ref, second_scaling, corr_thr=5e-3):
        # assumes that first_ref and second_ref have the same index
        envelope_df = deepcopy(first_ref)
        envelope_df['envelope_pos'] = first_ref * first_scaling
        envelope_df['envelope_neg'] = second_ref * second_scaling
        envelope_df = envelope_df.drop(columns=first_ref.columns)

        is_signal_correct = True
        for u_res_df in self.assertion.dfs:
            col = u_res_df.columns[0]
            print('Analysing {}'.format(col))
            idx = envelope_df.index.union(u_res_df.index)
            envelope_df_analysis = envelope_df.reindex(idx).interpolate('index')
            envelope_df_analysis = envelope_df_analysis.join(u_res_df).dropna()
            envelope_df_analysis['comparison'] = envelope_df_analysis.apply(
                lambda row: (row[col] >= row['envelope_neg']) and (row[col] <= row['envelope_pos']), axis=1)
            # Correction for low values
            envelope_df_analysis.loc[abs(envelope_df_analysis[col]) < corr_thr, 'comparison'] = True
            if len(envelope_df_analysis[envelope_df_analysis['comparison'] == False]) > 0:
                warnings.warn('{} outside of the lower: {}*{}, upper: {}*{} reference'
                              .format(col, first_ref.columns[0], first_scaling, second_ref.columns[0], second_scaling))
                is_signal_correct = False

        fig, ax = plot_between_two_references(self.assertion.dfs, envelope_df)
        return AssertionBuilderSignalPlot(fig, ax, is_signal_correct)


def plot_with_time_range(dfs, t_start, t_end):
    fig, ax = plt.subplots(figsize=(15, 7))

    for df in dfs:
        df.plot(ax=ax, grid=True)

    ax.legend(loc='upper right')
    ax.tick_params(labelsize=15)
    ax.set_xlabel("time, [s]", fontsize=15)

    for t_s, t_e in zip(t_start, t_end):
        ax.axvspan(xmin=t_s, xmax=t_e, facecolor='xkcd:goldenrod')

    return fig, ax


class AssertionBuilderSignalTimeRange(object):

    def __init__(self, assertion):
        self.assertion = assertion

    def has_min_max_variation(self, variation_min_max):
        t_start = self.assertion.t_start
        t_end = self.assertion.t_end

        # Assertion
        is_signal_correct = True
        for df in self.assertion.dfs:
            col = df.columns[0]
            for t_s, t_e in zip(t_start, t_end):
                slice_df = df[(df.index > t_s) & (df.index < t_e)]
                variation = slice_df.max() - slice_df.min()
                if variation.values[0] > variation_min_max:
                    warnings.warn('The variation of {} ({:.1f} %) exceeds {} % for constant current from {} to {} s'
                                  .format(col, variation_min_max, variation.values[0], t_s, t_e))
                    is_signal_correct = False

        # Plot
        fig, ax = plot_with_time_range(self.assertion.dfs, t_start, t_end)
        return AssertionBuilderSignalPlot(fig, ax, is_signal_correct)

    def has_min_max_slope(self, slope_min, slope_max, scaling=3600 * 1000, unit='mV/h'):
        t_start = self.assertion.t_start
        t_end = self.assertion.t_end

        # Assertion
        is_signal_correct = True
        for df in self.assertion.dfs:
            col = df.columns[0]
            for t_s, t_e in zip(t_start, t_end):
                slice_df = df[(df.index > t_s) & (df.index < t_e)]
                slope = LinearRegression().fit(np.array(slice_df.index).reshape((-1, 1)), slice_df[col]).coef_[
                            0] * scaling
                if (slope < slope_min) or (slope > slope_max):
                    warnings.warn('The drift of {} is {:.3f} {} for constant current from {} to {} s'
                                  .format(col, slope, unit, t_s, t_e))
                    is_signal_correct = False

        # Plot
        fig, ax = plot_with_time_range(self.assertion.dfs, t_start, t_end)
        return AssertionBuilderSignalPlot(fig, ax, is_signal_correct)

    def has_min_max_value(self, value_min=0, value_max=math.inf):
        t_start = self.assertion.t_start if isinstance(self.assertion.t_start, list) else [self.assertion.t_start]
        t_end = self.assertion.t_end if isinstance(self.assertion.t_end, list) else [self.assertion.t_end]

        # Assertion
        is_signal_correct = True
        for df in self.assertion.dfs:
            col = df.columns[0]
            for t_s, t_e in zip(t_start, t_end):
                if t_s is not None:
                    slice_df = df[(df.index > t_s) & (df.index < t_e)]
                else:
                    slice_df = df[(df.index > t_e)]
                df_above_thres = slice_df[(slice_df[col] < value_min) | (slice_df[col] > value_max)]
                if len(df_above_thres) > 0:
                    warnings.warn('{} outside of the [{}, {}] threshold'.format(col, value_min, value_max))
                    is_signal_correct = False

        fig, ax = plot_with_time_range_min_max_value(self.assertion.dfs, t_start, t_end, value_min, value_max)
        return AssertionBuilderSignalPlot(fig, ax, is_signal_correct)


def plot_with_time_range_min_max_value(dfs, t_start, t_end, value_min, value_max):
    fig, ax = plt.subplots(figsize=(15, 7))

    for df in dfs:
        df.plot(ax=ax, grid=True)

    ax.legend(loc='lower left')
    ax.tick_params(labelsize=15)
    ax.set_xlabel("time, [s]", fontsize=15)
    ax.set_ylim(0, ax.get_ylim()[1])
    for t_s, t_e in zip(t_start, t_end):
        ax.axvspan(xmin=t_e, xmax=ax.get_xlim()[1], ymin=value_min, ymax=value_max / ax.get_ylim()[1],
                   facecolor='xkcd:goldenrod')

    return fig, ax


class AssertionBuilderSignalPlot(object):

    def __init__(self, fig, ax, is_signal_correct):
        self.is_signal_correct = is_signal_correct
        self.ax = ax
        self.fig = fig

    def show_plot(self, **kwargs):
        if 'ylabel' in kwargs:
            self.ax.set_ylabel(kwargs['ylabel'], fontsize=15)

        if 'title' in kwargs:
            self.fig.suptitle(kwargs['title'], fontsize=20)

        if 'data_sec' in kwargs:
            ax2 = self.ax.twinx()
            data_sec = kwargs['data_sec']
            data_sec = data_sec if isinstance(data_sec, list) else [data_sec]

            for data_sec_el in data_sec:
                if 'drawstyle_sec' in kwargs:
                    data_sec_el.plot(ax=ax2, drawstyle=kwargs['drawstyle_sec'])
                else:
                    data_sec_el.plot(ax=ax2)

            if 'ylabel_sec' in kwargs:
                ax2.set_ylabel(kwargs['ylabel_sec'], fontsize=15)

        plt.show()

        return AssertionBuilderSignalPlotResult(self.is_signal_correct)

    def get_assertion_result(self):
        return self.is_signal_correct


class AssertionBuilderSignalPlotResult(object):

    def __init__(self, is_signal_correct):
        self.is_signal_correct = is_signal_correct

    def get_assertion_result(self):
        return self.is_signal_correct


class AssertionBuilderFeaturePlot(object):

    def __init__(self, fig, ax, features_outside_range):
        self.fig = fig
        self.ax = ax
        self.features_outside_range = features_outside_range

    def show_plot(self, **kwargs):
        if 'ylabel' in kwargs:
            self.ax.set_ylabel(kwargs['ylabel'], fontsize=15)

        if 'title' in kwargs:
            self.fig.suptitle(kwargs['title'], fontsize=20)

        if 'xlabel' in kwargs:
            self.ax.set_xlabel(kwargs['xlabel'], fontsize=15)

        plt.show()

        return AssertionBuilderFeaturePlotResult(self.features_outside_range)

    def get_features_outside_range(self):
        return self.features_outside_range


class AssertionBuilderFeaturePlotResult(object):

    def __init__(self, features_outside_range):
        self.features_outside_range = features_outside_range

    def get_features_outside_range(self):
        return self.features_outside_range
