import pandas as pd
import numpy as np

from lhcsmapi.dbsignal.ExpSignalAnalysis import ExpSignalAnalysis, calculate_time_constant_with_optimization, \
    calculate_time_constant_from_log
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities


def flatten_nested_lists(x):
    result = []
    for el in x:
        if isinstance(el, (list, tuple)):
            result.extend(flatten_nested_lists(el))
        else:
            result.append(el)
    return result


def convert_table_df_into_row(df, index):
    df_out = pd.DataFrame(df.stack())
    df_out.index = df_out.index.map('_'.join)
    df_out = df_out.T
    df_out.index = [index]

    return df_out


def first(x):
    return x.iloc[0]


def first20mean(x):
    return np.mean(x.iloc[0:20])


def last(x):
    return x.iloc[-1]


def last20mean(x):
    return np.mean(x.iloc[-20:-1])


def tau_charge(df):
    exponent = 1
    slice_squared_df = df.pow(exponent)
    return ExpSignalAnalysis.calculate_characteristic_time_in_seconds_with_integral(slice_squared_df, exponent)


def tau_energy(df):
    exponent = 2
    slice_squared_df = df.pow(exponent)
    return ExpSignalAnalysis.calculate_characteristic_time_in_seconds_with_integral(slice_squared_df, exponent)


def tau_lin_reg(df):
    return calculate_time_constant_from_log(df)


def tau_exp_fit(df):
    return calculate_time_constant_with_optimization(df)


class Feature(object):
    __name_to_function = {"first": first, "first20mean": first20mean, "last": last, "last20mean": last20mean,
                          "max": max, "min": min, "median": np.median, "std": np.std, "mean": np.mean, 'sum': sum,
                          "tau_charge": tau_charge, "tau_energy": tau_energy, "tau_lin_reg": tau_lin_reg,
                          "tau_exp_fit": tau_exp_fit}

    def set_dfs(self, dfs):
        dfs = dfs if isinstance(dfs, (list, tuple)) else [dfs]
        self.dfs = flatten_nested_lists(dfs)

    def take_signal_window(self, t_start, t_end):
        def get_window(df, t_start, t_end):
            index_start_decay = df.index.get_loc(t_start, method='nearest')
            index_end_decay = df.index.get_loc(t_end, method='nearest')
            return df.iloc[index_start_decay:index_end_decay]

        self.dfs = [get_window(df, t_start, t_end) for df in self.dfs]

    def synchronize_time(self, t0):
        self.dfs = [SignalUtilities.synchronize_df(df, t0) if not df.empty else df for df in self.dfs]

    def subtract_min_value(self):
        self.dfs = [df - df.min() if not df.empty else df for df in self.dfs]

    def calculate_features(self, features, index=0):
        features = features if isinstance(features, list) else [features]
        functions_to_apply = [self.__name_to_function[x] for x in features]

        features_dfs = []
        for df in self.dfs:
            features_df = df.aggregate(functions_to_apply).T
            feature_col_names = {col: '{}:{}'.format(features_df.index[0], col) for col in features_df.columns}
            features_df = features_df.rename(columns=feature_col_names)
            features_df.index = [index]
            features_dfs.append(features_df)

        return pd.concat(features_dfs, axis=1)

    def calculate_multicolumn_features(self, features, prefix):
        features = features if isinstance(features, list) else [features]
        functions_to_apply = [self.__name_to_function[x] for x in features]
        df_out = self.dfs[0].agg(functions_to_apply).T
        df_out.columns = ['%s_%s' % (prefix, col) for col in df_out.columns]

        return df_out


class FeatureBuilder(object):

    def __init__(self):
        self.feature = Feature()

    def with_signal(self, *dfs):
        self.feature.set_dfs(dfs)

        return FeatureBuilderSignal(self.feature)

    def with_multicolumn_signal(self, *dfs):
        self.feature.set_dfs(dfs)

        return FeatureBuilderMultiColumnSignal(self.feature)


class FeatureBuilderSignal(object):

    def __init__(self, feature):
        self.feature = feature

    def with_duration(self, t_start, t_end):
        self.feature.take_signal_window(t_start, t_end)

        return self

    def synchronize_time(self, t_sync):
        self.feature.synchronize_time(t_sync)

        return self

    def subtract_min_value(self):
        self.feature.subtract_min_value()

        return self

    def calculate_features(self, features, index=0):
        return self.feature.calculate_features(features, index)


class FeatureBuilderMultiColumnSignal(object):

    def __init__(self, feature):
        self.feature = feature

    def calculate_features(self, features, prefix):
        self.dfs = self.feature.calculate_multicolumn_features(features, prefix)
        return self

    def convert_into_row(self, index=0):
        self.dfs = convert_table_df_into_row(self.dfs, index)

        return self