import numpy as np
import pandas as pd
import datetime
import warnings

from lhcsmapi.Time import Time
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.FeatureBuilder import convert_table_df_into_row
from lhcsmapi.pyedsl.dbquery.DbQueryBuilder import DbQueryBuilderFactory


def validate_time_definition(kwargs):
    # to correct for an error with checking type of a np.int64 variable by isinstance()
    if 't_start' in kwargs and isinstance(kwargs['t_start'], np.int64):
        kwargs['t_start'] = int(kwargs['t_start'])

    t_start = validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                     np.int64, **kwargs) or 0
    t_end = validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                   np.int64, **kwargs)
    duration = validate_key_in_kwargs('duration', tuple, list, **kwargs)
    return Time.get_query_period_in_unix_time(t_start, t_end, duration, '')


def validate_key_in_kwargs(key, *key_types, **kwargs):
    result = None
    if key in kwargs:
        result = kwargs[key]
        match_found = False
        for key_type in key_types:
            match_found = match_found or isinstance(result, key_type)
        if not match_found:
            raise TypeError("You entered a {} keyword, but the keyword needs to be a {}.".format(key, key_types))

    return result


class Query(object):

    def set_db(self, db):
        self.db = db

    def set_circuit_type(self, circuit_type):
        self.circuit_type = circuit_type

    def set_t_start(self, t_start):
        self.t_start = t_start

    def set_t_end(self, t_end):
        self.t_end = t_end

    def set_timestamp(self, timestamp):
        self.timestamp = timestamp

    def get_time_params(self):
        if hasattr(self, 'timestamp') and self.db == 'PM':
            return {'eventTime': self.timestamp}
        elif hasattr(self, 't_start') and self.db in ['NXCALS', 'CALS', 'PM', 'AFT']:
            return {'t_start': self.t_start, 't_end': self.t_end}
        else:
            raise ValueError('Wrong time definition!')


class QueryBuilder(object):
    """ Builder class to query databases (PM, CALS, NXCALS, AFT).
        The general syntax for PM (NX)CALS is
        QueryBuilder().{with_pm()/with_cals()/with_nxcals()}.with_circuit_type().{with_duration()/with_timestamp()}.with_metadata().{signal_query()/feature_query()}

        For AFT the metadata is not provided which simplifies the syntax to
        QueryBuilder().with_aft().context_query()
        QueryBuilder().with_aft().with_duration().fault_query()

        - EXAMPLE PM (event query)
        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        source_timestamp_df = QueryBuilder().with_pm().with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end)\
                        .with_metadata(circuit_name=circuit_name, system='PC').event_query().df

        print(source_timestamp_df)
        >>>              source            timestamp
        >>> 0  RPTE.UA23.RB.A12  1421164827600000000

        - EXAMPLE PM (signal query)
        timestamp_sync = 1421164827600000000
        source_fgc = 'RPTE.UA23.RB.A12'

        i_meas_df = QueryBuilder().with_pm().with_circuit_type('RB').with_timestamp(timestamp_sync)\
                .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query()\
                .synchronize_time(timestamp_sync).convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>          STATUS.I_MEAS
        >>> -359.98      -0.150693
        >>> -359.96      -0.152328
        >>> -359.94      -0.151348
        >>> -359.92      -0.150366
        >>> -359.90      -0.149385

        - Example NXCALS (assuming connection to NXCALS cluster)::
        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        i_meas_df = QueryBuilder().with_nxcals().with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end).\
                with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query(spark).synchronize_time().convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>       I_MEAS
        >>> 0.0     0.07
        >>> 25.5    1.13
        >>> 26.0    1.72
        >>> 26.5    2.82
        >>> 27.0    4.32

        - EXAMPLE CALS

        import pytimber
        ldb = pytimber.LoggingDB()

        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        i_meas_df = QueryBuilder().with_cals().with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end).\
                        with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query(ldb).synchronize_time().convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>      RPTE.UA23.RB.A12:I_MEAS
        >>> 0.0                     0.07
        >>> 0.0                     1.13
        >>> 0.0                     1.72
        >>> 0.0                     2.82
        >>> 0.0                     4.32

        """

    _msg = 'Set database name using with_pm(), with_cals(ldb), with_nxcals(spark) method.'

    def __init__(self):
        self.query = Query()

    def __getattr__(self, name):
        if name not in ['query', 'with_pm', 'with_cals', 'with_nxcals']:
            raise AttributeError('{}() is not supported. {}'.format(name, self._msg))

    def __repr__(self):
        return self._msg

    def with_pm(self):
        self.query.set_db('PM')
        self.query.dbconnector = None
        return QueryBuilderWithDb(self.query)

    def with_cals(self, dbconnector):
        self.query.set_db('CALS')
        self.query.dbconnector = dbconnector
        return QueryBuilderWithDb(self.query)

    def with_nxcals(self, dbconnector):
        self.query.set_db('NXCALS')
        self.query.dbconnector = dbconnector
        return QueryBuilderWithDb(self.query)

    def with_aft(self, dbconnector):
        self.query.set_db('AFT')
        self.query.dbconnector = dbconnector
        return QueryBuilderWithDb(self.query)


class QueryBuilderWithDb(object):
    _msg = 'Set time definition: for PM signal query, with_timestamp(),' \
           '\n for PM event query or CALS, NXCALS signal query, with_duration()'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_timestamp', 'with_duration']:
            raise AttributeError('{}() is not supported. {}'.format(name, self._msg))

    def __repr__(self):
        return 'Database name properly set to {}. {}'.format(self.query.db, self._msg)

    def with_timestamp(self, timestamp):
        if self.query.db == 'PM':
            self.query.set_timestamp(Time.to_unix_timestamp(timestamp))
            return QueryBuilderWithDbTimeDef(self.query)
        else:
            raise ValueError('Timestamp can be only set for PM database, got {}'.format(self.query.db))

    def with_duration(self, **duration):
        t_start, t_end = validate_time_definition(duration)
        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')

        self.query.set_t_start(t_start)
        self.query.set_t_end(t_end)

        return QueryBuilderWithDbTimeDef(self.query)

    def context_query(self, context, context_id=None, **kwargs):
        if self.query.db is not 'AFT':
            NotImplementedError('This method is not implemented for this database %s' % self.query.db)

        is_timestamp = False
        is_t_start = True
        self.query.db_query_builder = DbQueryBuilderFactory.create('AFT', is_timestamp, is_t_start, None)
        time_params = {}
        kwargs['dbconnector'] = self.query.dbconnector
        kwargs['context_id'] = context_id
        return self.query.db_query_builder.execute_context_query(time_params, context, **kwargs)

class QueryBuilderWithDbTimeDef(object):
    _msg = 'Set generic query parameter using with_query_parameters() method, ' \
           'or a circuit signal using with_circuit_type() method.'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_circuit_type', 'with_query_parameters']:
            raise AttributeError('{}() is not supported. \n{}'.format(name, self._msg))

    def __repr__(self):
        if hasattr(self.query, 'timestamp'):
            return 'Timestamp set to timestamp={}. {}'.format(self.query.timestamp, self._msg)
        else:
            return 'Query duration set to t_start={}, t_end={}. {}' \
                .format(self.query.t_start, self.query.t_end, self._msg)

    def with_circuit_type(self, circuit_type):
        #ToDo Check if circuit_type is supported
        self.query.set_circuit_type(circuit_type.upper())
        return QueryBuilderWithDbTimeDefCircuitType(self.query)

    def with_query_parameters(self, **kwargs):
        db = self.query.db
        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')

        self.query.db_query_builder = DbQueryBuilderFactory.create(db, is_timestamp, is_t_start)
        self.query.db_query_builder.create_query_parameters(kwargs)

        return QueryBuilderWithDbTimeDefMetadata(self.query)

    def fault_query(self, **kwargs):
        if self.query.db is not 'AFT':
            NotImplementedError('This method is not implemented for this database %s' % self.query.db)

        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')

        self.query.db_query_builder = DbQueryBuilderFactory.create('AFT', is_timestamp, is_t_start, None)
        time_params = self.query.get_time_params()
        kwargs['dbconnector'] = self.query.dbconnector
        return self.query.db_query_builder.execute_fault_query(time_params_dict=time_params, **kwargs)

class QueryBuilderWithDbTimeDefCircuitType(object):
    _msg = 'Set metadata using with_metadata() method, e.g., ' \
           'with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\').'

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_metadata']:
            raise AttributeError('{}() not supported. \n{}'.format(name, self._msg))

    def __repr__(self):
        return 'Circuit_type set to {}. {}'.format(self.query.circuit_type, self.msg)

    def with_metadata(self, **kwargs):
        db = self.query.db
        is_timestamp = hasattr(self.query, 'timestamp')
        is_t_start = hasattr(self.query, 't_start')
        circuit_type = self.query.circuit_type

        self.query.db_query_builder = DbQueryBuilderFactory.create(db, is_timestamp, is_t_start, circuit_type)
        self.query.db_query_builder.create_query_metadata(kwargs)
        self.query.db_query_builder.update_query_metadata(kwargs)

        return QueryBuilderWithDbTimeDefMetadata(self.query)


class QueryBuilderWithDbTimeDefMetadata(object):

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if hasattr(self.query, 'timestamp') and name not in ['query', 'signal_query']:
            raise AttributeError('{}() not supported, for PM with timestamp you can call signal_query()'.format(name))
        elif hasattr(self.query, 't_start') and self.query.db == 'PM' and name not in ['query', 'event_query']:
            raise AttributeError('{}() not supported, for PM with duration you can call event_query()'.format(name))
        elif name not in ['query', 'signal_query']:
            raise AttributeError('{}() is not supported, you can call signal_query()'.format(name))

    def __repr__(self):
        if hasattr(self.query, 'timestamp') and self.query.db == 'PM':
            return 'Query signal using signal_query() method, i.e., signal_query().'
        elif hasattr(self.query, 't_start') and self.query.db == 'PM':
            return 'Query signal using event_query() method, i.e., event_query().'
        else:
            return 'Query events using event_query() method, i.e., signal_query().'

    def signal_query(self, allow_duplicate_index=False, **kwargs):
        time_params = self.query.get_time_params()
        circuit_dfs = self.query.db_query_builder.execute_signal_query(time_params, dbconnector=self.query.dbconnector,
                                                                       allow_duplicate_index=allow_duplicate_index,
                                                                       **kwargs)

        return SignalProcessing(circuit_dfs, self.query.db)

    def event_query(self, **kwargs):
        time_params = self.query.get_time_params()
        event_source_timestamp_df = self.query.db_query_builder.execute_event_query(time_params, **kwargs)

        return EventProcessing(event_source_timestamp_df)

    def feature_query(self, features, **kwargs):
        time_params = self.query.get_time_params()

        return FeatureProcessing(self.query.db_query_builder.execute_feature_query(time_params, features, **kwargs,
                                                                                   dbconnector=self.query.dbconnector))

    def context_query(self, contexts, **kwargs):
        time_params = self.query.get_time_params()
        context_df = self.query.db_query_builder.execute_context_query(time_params, contexts, **kwargs)

        return ContextProcessing(context_df)


class ContextProcessing(object):
    def __init__(self, df):
        self.df = df


class FeatureProcessing(object):
    def __init__(self, df):
        self.df = df

    def sort_busbar_location(self, circuit_type, circuit_name, split=':'):
        if 'nxcals_variable_name' in self.df.columns:
            circuit_name = SignalMetadata.get_circuit_names(circuit_type) if circuit_type == 'RB' else circuit_name
            self.df['name_prefix'] = self.df.apply(lambda row: row['nxcals_variable_name'].split(split)[0], axis=1)
            busbars_ordered_df = MappingMetadata.read_busbar_to_magnet_mapping(circuit_type)
            circuit_name = circuit_name if isinstance(circuit_name, list) else [circuit_name]
            is_busbar_in_circuit_name = busbars_ordered_df['Circuit'].isin(circuit_name)
            busbars_ordered_per_circuit_df = busbars_ordered_df[is_busbar_in_circuit_name]
            self.df['name_prefix'] = pd.Categorical(self.df['name_prefix'],
                                                    categories=busbars_ordered_per_circuit_df['Bus Bar Segment Name'],
                                                    ordered=True)
            self.df.sort_values(by='name_prefix', inplace=True)
            self.df.drop(columns='name_prefix', inplace=True)
            self.df.reset_index(inplace=True, drop=True)
        else:
            warnings.warn('This operation is only supported to feature DataFrame with nxcals_variable_name column,'
                          ' got {}'.format(self.df.columns))

        return self

    def correct_voltage_sign(self, variable_name='nxcals_variable_name', class_compare=0, correction_col='mean', polarity=-1):
        """ correct_voltage_sign provides alteration of polarity to groups of data, by comparing a selected columns
        value, from a selected class to a chosen expected polarity. E.g. Comparing the value of the 'mean' column for
        a Class representing inductive voltage at ramp, to a polarity of -1, will result in swapped polarity for all
        related classes (including selected class) if statement yields true

        Example of Usages:
            - Create default check and correction: Checks class == 0, column that is being checked and
            corrected is 'mean', and we are checking for negative polarity.
            feature_query takes in function 'translate_udf' that creates the 'class'-column needed to copmlete
            correction of voltage sign

            df = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start_injs, t_end=t_end_sbs) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name=['RB.A12', 'RB.A45'], system='BUSBAR', signal='U_MAG',
                    wildcard={'BUSBAR': ['DCBB.8L2.R', 'DCBA.15R4.R']}) \
                .feature_query(['mean'], function=translate_udf) \
                .correct_voltage_sign() \
                .df

            print(df)
            >>> nxcals_variable_name 	class 	mean
            >>> DCBA.15R4.R:U_MAG 	2 	-0.001142
            >>> DCBA.15R4.R:U_MAG 	0 	0.632161
            >>> DCBA.15R4.R:U_MAG 	1 	-0.001239
            >>> DCBB.8L2.R:U_MAG 	1 	0.001861
            >>> DCBB.8L2.R:U_MAG 	0 	0.636315
            >>> DCBB.8L2.R:U_MAG 	2 	0.001983


        Assumes:
            - featuer_query includes function to classify, and add, classes to dataframe
            - For each variable_name there is only one of each class type
            - Data from selected correction_col contains numbers value
            - There are groups of data with same variable name


        :param variable_name: Name of column used to group data points together for value comparison. Default: nxcals_variable_name
        :type variable_name: str

        :param class_compare: Class checked for polarity, action performed if polarity chosen equals true. Default: Class == 0
        :type class_compare: str or num

        :param correction_col: Chosen value for polarity change. Default: mean
        :type correction_col: str

        :param polarity: Check for selected polarity: Default: -1
        :type polarity: whole number, either -1 or 1

        :return: self.df with altered polarity to selected column
        """
        if (correction_col not in self.df) or ('class' not in self.df):
            warnings.warn('This operation is only supported to feature DataFrame with class column, and the selected '
                          '%s column. Got %s' % (correction_col, self.df.columns))

        def correction(df, class_cmpr, cor_col, pol, num_of_rows=3):
            if len(df) < num_of_rows:
                return df

            class_change = df['class'].unique()
            idx = df.index.unique()

            if (class_cmpr in class_change) & (df[cor_col].isnull().values.any() == False):
                if np.sign(df[df['class'] == class_cmpr][cor_col].iloc[0]) == pol:
                    for i in range(len(class_change)):
                        df = df.assign(mean=df[cor_col].replace({df[cor_col][idx[int(class_change[i])]]:df[cor_col][idx[int(class_change[i])]] * -1}))

            return df

        self.df = self.df.groupby(variable_name).apply(lambda row: correction(row, class_compare, correction_col, polarity))
        return self

    def calculate_max_abs(self):
        if ('min' in self.df.columns) and ('max' in self.df.columns):
            self.df['max_abs'] = self.df.apply(lambda row: max(row['max'], abs(row['min'])), axis=1)
        else:
            warnings.warn('This operation is only supported to feature DataFrame with min, max columns,'
                          ' got {}'.format(self.df.columns))
        return self

    def sort_values(self, by):
        self.df.sort_values(by=by, inplace=True)
        self.df.reset_index(inplace=True, drop=True)
        return self

    def convert_into_row(self, index=0):
        # table format
        self.df = self.df.set_index(self.df['nxcals_variable_name'], drop=True) \
            .drop(columns='nxcals_variable_name')
        # row format
        self.df = convert_table_df_into_row(self.df, index)
        return self

class EventProcessing(object):

    def __init__(self, df):
        self.df = df

    def filter_source(self, circuit_name, system):
        if self.df.empty:
            return self

        if system in ['QH', 'QDS']:
            if 'RQ' in circuit_name:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (SignalMetadata.get_sector_names_from_components(
                    'RQ', 'MAGNET', row['source'])[0] == circuit_name) |
                                                                     (SignalMetadata.get_sector_names_from_components(
                                                                         'RQ', 'MAGNET', row['source'])[
                                                                          1] == circuit_name), axis=1)
            elif 'RB' in circuit_name:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (
                        SignalMetadata.get_sector_names_from_components('RB', 'MAGNET', row['source']) == circuit_name),
                                                         axis=1)
            else:
                return self

            self.df = self.df[self.df['is_circuit_ok']]
            self.df.drop(columns=['is_circuit_ok'], inplace=True)
            self.df.reset_index(drop=True, inplace=True)
        return self

    def drop_duplicate_source(self):
        if self.df.empty:
            return self

        self.df = self.df.drop_duplicates(subset=['source']).sort_values(by=['timestamp']).reset_index(drop=True)
        return self

    def sort_values(self, by):
        if self.df.empty:
            return self

        self.df.sort_values(by=by, inplace=True)
        self.df.reset_index(inplace=True, drop=True)
        return self


class SignalProcessing(object):

    def __init__(self, dfs, db):
        self.dfs = dfs
        self.db = db

    def __getattr__(self, name):
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        if name not in method_list:
            raise AttributeError(name + '{}() is not supported, you can call {}'.format(name, method_list))

    def __repr__(self):
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        return 'Available methods are {}'.format(method_list)

    def synchronize_time(self, t0=None):
        self.dfs = [SignalUtilities.synchronize_df(df, t0) if not df.empty else df for df in self.dfs]
        return self

    def convert_index_to_sec(self):
        self.dfs = [SignalUtilities.convert_indices_to_sec(df) if not df.empty else df for df in self.dfs]
        return self

    def create_col_from_index(self, col):
        for df in self.dfs:
            df[col] = df.index
        return self

    def filter_median(self, window=3):
        self.dfs = [df.rolling(window=window, min_periods=1).median() if not df.empty else df for df in self.dfs]
        return self

    def remove_values_for_time_less_than(self, time_threshold):
        self.dfs = [df[df.index > time_threshold] if not df.empty else df for df in self.dfs]
        return self

    def remove_initial_offset(self, time_threshold):
        self.dfs = [df.subtract(df[df.index < time_threshold].mean()) if not df.empty else df for df in self.dfs]
        return self

    def filter_values(self, values):
        self.dfs = [df[df[df.columns[0]].isin(values)] for df in self.dfs]
        return self

    def map_values(self, mapping):
        self.dfs = [pd.DataFrame(df.apply(lambda col: mapping[col[df.columns[0]]], axis=1), columns=df.columns)
                    for df in self.dfs]
        return self

    def overwrite_sampling_time(self, t_sampling, t_end):
        def overwrite_sampling_time(df, t_sampling_, t_end_):
            new_index = np.arange(0, len(df), t_end_) * t_sampling_
            df.set_index(new_index, inplace=True)
            return df

        self.dfs = [overwrite_sampling_time(df, t_sampling, t_end) for df in self.dfs]
        return self

