import warnings

import pandas as pd
import numpy as np
from sklearn.linear_model import LinearRegression

from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.pyedsl.QueryBuilder import FeatureProcessing

def convert_current_into_mean_vector(circuit_type, i_busbar_feature_df):
    gb = i_busbar_feature_df.groupby('device')
    x = {SignalMetadata.get_sector_names_from_components(circuit_type, 'PC', x): gb.get_group(x)['mean'].values for x
         in gb.groups}
    return pd.DataFrame(x)


def convert_voltage_into_mean_vector(u_res_stat_df):
    gb = u_res_stat_df.groupby('class')
    dfs = [gb.get_group(x) for x in gb.groups]

    out = dfs[0]
    columns = set(out.columns) - {'nxcals_variable_name'}
    out = out.rename(columns={col: '%s_0' % col for col in columns})

    for index in range(1, len(dfs)):
        out = out.merge(dfs[index], on='nxcals_variable_name')
        # add index suffix to columns
        out = out.rename(columns={col: '%s_%d' % (col, index) for col in columns})
    # Add sector name here
    out['mean_vector'] = out.filter(regex='mean_').values.tolist()
    return out


def calculate_resistance_from_lin_fit(circuit_type, out, sector_to_i_meas_mean_df):
    first_busbar = out.loc[0, 'nxcals_variable_name']
    RES_NAME = first_busbar.split(':')[1].replace('U_', 'R_')

    # ToDo Add sector name
    if circuit_type == 'RB':
        out['sector'] = out.apply(
            lambda row: SignalMetadata.get_sector_names_from_components(circuit_type, 'BUSBAR',
                                                                        row['nxcals_variable_name'.split(':')[0]]),
            axis=1)
    elif circuit_type == 'RQ':
        def get_sector_name(nxcals_variable_name):
            busbar_name = 'BUSBAR_RQD' if nxcals_variable_name[3] == 'D' else 'BUSBAR_RQF'
            sector_index = 0 if nxcals_variable_name[3] == 'D' else 1
            return SignalMetadata.get_sector_names_from_components(circuit_type, busbar_name,
                                                                   nxcals_variable_name.split(':')[0])[sector_index]

        out['sector'] = out.apply(lambda row: get_sector_name(row['nxcals_variable_name']), axis=1)
    else:
        raise ValueError('Circuit type {} not supported!'.format(circuit_type))

    # Take slope coefficient of the linear fit of the V-I characteristics
    if len(sector_to_i_meas_mean_df) == 1:
        out[RES_NAME] = out.apply(
            lambda col: abs(col['mean_vector'][0]) / sector_to_i_meas_mean_df[col['sector']].values[0], axis=1)
    else:
        out[RES_NAME] = out.apply(lambda col: abs(
            LinearRegression().fit(sector_to_i_meas_mean_df[col['sector']].values.reshape((-1, 1)),
                                   col['mean_vector']).coef_[0]), axis=1)
    return out


class ResistanceBusbar(object):

    def __init__(self, u_busbar_feature_df):
        self.u_busbar_feature_df = u_busbar_feature_df
        self.i_busbar_feature_df = None

    def set_current(self, i_busbar_feature_df):
        self.i_busbar_feature_df = i_busbar_feature_df


class ResistanceBuilder(object):

    def with_busbar_voltage(self, u_busbar_feature_df):
        u_busbar_feature_df = u_busbar_feature_df[u_busbar_feature_df['class'] != 0] if 'class' in u_busbar_feature_df else u_busbar_feature_df
        resistance_busbar = ResistanceBusbar(u_busbar_feature_df)

        return ResistanceBuilderBusbarVoltage(resistance_busbar)


class ResistanceBuilderBusbarVoltage(object):

    def __init__(self, resistance_busbar):
        self.resistance_busbar = resistance_busbar

    def with_busbar_current(self, i_busbar_feature_df):
        i_busbar_feature_df = i_busbar_feature_df[i_busbar_feature_df['class'] != 0] if 'class' in i_busbar_feature_df else i_busbar_feature_df
        self.resistance_busbar.set_current(i_busbar_feature_df)

        return ResistanceBuilderBusbarVoltageCurrent(self.resistance_busbar)


class ResistanceBuilderBusbarVoltageCurrent(object):

    def __init__(self, resistance_busbar):
        self.resistance_busbar = resistance_busbar
        self.df = None

    def calculate_mean_resistance(self, circuit_type):
        i_busbar_mean_vector_df = convert_current_into_mean_vector(circuit_type,
                                                                   self.resistance_busbar.i_busbar_feature_df)
        u_busbar_mean_vector_df = convert_voltage_into_mean_vector(self.resistance_busbar.u_busbar_feature_df)

        self.df = calculate_resistance_from_lin_fit(circuit_type, u_busbar_mean_vector_df, i_busbar_mean_vector_df)
        return self

    def convert_to_row(self, index):
        if 'R_RES' in self.df.columns:
            out_df = self.df[['nxcals_variable_name', 'mean_0', 'std_0', 'mean_1', 'std_1', 'R_RES']]
        elif 'R_MAG' in self.df.columns:
            out_df = self.df[['nxcals_variable_name', 'mean_0', 'std_0', 'mean_1', 'std_1', 'R_MAG']]
        else:
            raise KeyError('R_RES or R_MAG not present in DF: {}'.format(self.df.columns))

        out_df['nxcals_variable_name_prefix'] = out_df['nxcals_variable_name'].apply(lambda row: row.split(':')[0])
        out_df.index = out_df['nxcals_variable_name_prefix']
        out_df.drop(columns=['nxcals_variable_name', 'nxcals_variable_name_prefix'], inplace=True)
        out_df = out_df.rename(
            columns={'mean_0': 'mean_inj', 'std_0': 'std_inj', 'mean_1': 'mean_sb', 'std_1': 'std_sb'})

        out_df = pd.DataFrame(out_df.stack())
        out_df.index = out_df.index.map('_'.join)
        out_df = out_df.T
        out_df.reset_index(inplace=True, drop=True)
        out_df.index = pd.Index([index])

        self.df = out_df
        return self

    # ====== FILL MISSING COLUMNS ======
    def fill_missing_columns(self, check_missing='R_MAG', circuit_type='RB', circuit_name='*', split='_'):
        circuit_name = SignalMetadata.get_circuit_names(
            circuit_type) if circuit_type == 'RB' else circuit_name  # Circuit Names
        busbars_ordered_df = MappingMetadata.read_busbar_to_magnet_mapping(circuit_type)  # Get all signal names that was asked for

        # Gett all column names in df
        list_df_columns = list(self.df.filter(regex='{}$'.format(check_missing), axis=1).columns)
        list_df_columns = [elem.split(split)[0] for elem in list_df_columns]

        columns_missing = list(set(busbars_ordered_df['Bus Bar Segment Name']) - set(list_df_columns))

        if len(columns_missing) > 0:
            for col in columns_missing:
                self.df.insert(0, ('{}{}{}'.format(col, split, check_missing)), np.nan)

        return self

    # ====== SORT COLUMNS ======
    def sort_columns_busbar_location(self, check_missing='R_MAG', circuit_type='RB', circuit_name='*', split='_'):
        self.df.drop([col for col in self.df.columns if col[-1] not in '_R_MAG'], axis=1,inplace=True)

        self.df = self.df.T
        self.df['nxcals_variable_name'] = self.df.index
        self.df.reset_index(inplace=True, drop=True)

        FeatureProcessing.sort_busbar_location(self, circuit_type=circuit_type, circuit_name=circuit_name, split='_')  # Can use API sort_busbar_location if updated with selective split

        self.df = self.df.set_index(self.df['nxcals_variable_name'])
        self.df = self.df.rename(columns={'nxcals_variable_name': 'Index'})
        self.df = self.df.T
        self.df = self.df.drop(['Index'])

        return self
