from abc import ABC, abstractmethod

import pyspark.sql.functions as func
from pyspark.sql.window import Window
from pyspark.sql.functions import row_number, monotonically_increasing_id, col

# NXCALS libraries
try:
    from cern.nxcals.pyquery.builders import *
except:
    pass


def check_pyspark_functions(functions, argument):
    functions_dictionary = {'max': func.max, 'min': func.min, 'mean': func.mean, 'std': func.stddev,
                            'count': func.count}
    if set(functions) - functions_dictionary.keys() == set():
        return [functions_dictionary[f](argument) for f in functions if
                          f in functions_dictionary.keys()]
    else:
        raise ValueError('Input {} not valid'.format(functions - functions_dictionary.keys()))


class NxcalsFeatureQuery(ABC):

    def __init__(self, time_column):
        self.time_column = time_column
        self.query = None

    @abstractmethod
    def initialize_query(self, spark, system, t_start, t_end):
        pass

    @abstractmethod
    def populate_query(self, kwargs):
        pass

    @abstractmethod
    def build_dataset(self, signal):
        pass

    def decimate_columns(self, kwargs):
        if 'decimation' in kwargs:
            decimation = kwargs['decimation']
            shift = kwargs['shift'] if 'shift' in kwargs else 0

            self.query = self.query.sort(self.time_column) \
                .withColumn('id', monotonically_increasing_id()) \
                .withColumn('idx', row_number().over(Window.orderBy('id'))) \
                .filter((col('idx') + shift) % decimation == 0)

    def translate_column(self, kwargs):
        if 'function' in kwargs:
            translate_udf = kwargs['function']
            self.query = self.query.withColumn("class", translate_udf(col(self.time_column))) \
                .filter(col("class") != -1)

    @abstractmethod
    def get_col_names(self, kwargs):
        pass

    def group_by(self, cols):
        self.query = self.query.groupBy(*cols)

    @abstractmethod
    def get_features(self, features, query_params=None):
        pass

    def aggregate(self, features):
        return self.query.agg(*features).collect()


class DevicePropertyNxcalsFeatureQuery(NxcalsFeatureQuery):

    def __init__(self):
        super().__init__('acqStamp')

    def initialize_query(self, spark, system, t_start, t_end):
        self.query = DevicePropertyQuery.builder(spark).system(system) \
            .startTime(t_start).endTime(t_end)

    def populate_query(self, kwargs):
        self.query = self.query.entity().device(kwargs['metadata_device']).property(kwargs['metadata_property'])

    def build_dataset(self, query_params):
        signal = query_params['signal'].values[0]
        self.query = self.query.buildDataset().select("acqStamp", signal, "device")

    def get_features(self, features, query_params=None):
        signal = query_params['signal'].values[0]
        return check_pyspark_functions(features, signal)

    def get_col_names(self, kwargs):
        return ['device', 'class'] if 'function' in kwargs else ['device']


class VariableNxcalsFeatureQuery(NxcalsFeatureQuery):

    def __init__(self):
        super().__init__('nxcals_timestamp')

    def initialize_query(self, spark, system, t_start, t_end):
        self.query = VariableQuery.builder(spark).system(system).startTime(t_start).endTime(t_end)

    def populate_query(self, kwargs):
        self.query = self.query.variable(kwargs['signal'])

    def build_dataset(self, query_params):
        self.query = self.query.buildDataset().select("nxcals_timestamp", 'nxcals_value', 'nxcals_variable_name')

    def get_col_names(self, kwargs):
        return ['nxcals_variable_name', 'class'] if 'function' in kwargs else ['nxcals_variable_name']

    def get_features(self, features, query_params=None):
        return check_pyspark_functions(features, 'nxcals_value')

