import unittest
from unittest import mock
import pandas as pd

from lhcsmapi.Time import Time as tc
from lhcsmapi.dbsignal.cals.CalsDbSignal import CalsDbSignal


class TestCalsSignal(unittest.TestCase):

    def setUp(self):

        self.NAME = 'test'
        self.START = 0.0
        self.END = 3e-9
        self.DATA = [1.0, 1.2, 1.4]
        self.INDEX_SECS = [tc.to_unix_timestamp_in_sec(i, 's') for i in [0, 1, 2]]
        self.INDEX_TIMESTAMPS = [tc.to_pandas_timestamp(i, 's') for i in self.INDEX_SECS]

        self.mockLoggingDB = mock.Mock()
        self.defaultCALSSignal = CalsDbSignal(ldb=self.mockLoggingDB)
        self.df = pd.DataFrame(self.DATA, self.INDEX_TIMESTAMPS, columns=[self.NAME])
        self.calsDF = pd.DataFrame(self.DATA, self.INDEX_SECS, columns=[self.NAME])

    def tearDown(self):
        pass

    def assertEqualSignals(self, first_cals_signal, second_cals_signal):
        """Method assert each property of firstCALSSignal against the the same property of secondCALSSignal"""
        self.assertEqual(first_cals_signal.signal_name, second_cals_signal.signal_name)
        self.assertEqual(first_cals_signal.t_start, second_cals_signal.t_start)
        self.assertEqual(first_cals_signal.t_end, second_cals_signal.t_end)

    #region createSignal tests

    def test_toString(self):
        # arrange
        t_start = '2018-12-12 10:00:00+01:00'
        t_end = '2018-12-13 12:00:00+01:00'
        signal_name = 'RPTE.UA23.RB.A12:I_MEAS'

        cds = CalsDbSignal.build_signal(signal=signal_name, t_start=t_start, t_end=t_end, ldb=1)
        # act
        cdsStringAct = str(cds)

        # assert
        cdsStringExp = "Db: CALS, Signal: RPTE.UA23.RB.A12:I_MEAS, Start time = 2018-12-12 10:00:00+01:00, End time = 2018-12-13 12:00:00+01:00"
        self.assertEqual(cdsStringExp, cdsStringAct)

    def test_createCALSSignal_noParameters_defaultCALSBehaviorCreated(self):
        """Tests that the default CalsSignal is created if no arguments are passed to the createSignalCALS method"""
        calsSignal = CalsDbSignal.build_signal(ldb=self.mockLoggingDB, duration = [(0, 's')])
        self.assertEqualSignals(calsSignal, self.defaultCALSSignal)

    def test_createCALSSignal_startTimeAndEndTimeProvided_CorrectCALSBehaviorCreated(self):
        """Tests that the correct CalsSignal is created if start and end times are passed to the creaeSignalCALS method"""
        
        # arrange:
        arbitraryCALSSignal = CalsDbSignal(["test"], 0.0, 1.0, ldb=self.mockLoggingDB) # common situation is signal t_start and t_end

        # act:
        calsSignal = CalsDbSignal.build_signal(signal='test', t_start=0.0, t_end=(1.0, 's'), ldb=self.mockLoggingDB)

        # assert:
        self.assertEqualSignals(calsSignal, arbitraryCALSSignal)

    def test_createCALSSignal_startTimeAndSingleDurationComponentProvided_CorrectCALSBehavior(self):
        """Tests that the correct CalsSignal is created if start and end times are passed to the creaeSignalCALS method"""

        # arrange:
        arbitraryCALSSignal = CalsDbSignal(["test"], 0.0, 1.0, ldb=self.mockLoggingDB)  # common situation is signal t_start and t_end

        # act:
        calsSignal = CalsDbSignal.build_signal(signal='test', t_start=0, duration=(1.0, 's'), ldb=self.mockLoggingDB)

        # assert:
        self.assertEqualSignals(calsSignal, arbitraryCALSSignal)

    def test_createCALSSignal_startTimeAndTwoDurationComponentsProvided_CorrectCALSBehavior(self):
        """Tests that the correct CalsSignal is created if start and a single duration component are passed to the creaeSignalCALS method"""

        # arrange:
        arbitraryCALSSignal = CalsDbSignal(["test"], 0.0, 2.0, ldb=self.mockLoggingDB)  # common situation is signal t_start and t_end

        # act:
        calsSignal = CalsDbSignal.build_signal(signal='test', t_start=(1, 's'), duration=[(1.0, 's'), (1.0, 's')], ldb=self.mockLoggingDB)

        # assert:
        self.assertEqualSignals(calsSignal, arbitraryCALSSignal)

    def test_createCALSSignal_parameterIsInvalidType_TypeErrorIsThrown(self):
        """T[1.0, 1.2, 1.4]ests when a keyword that is of an invalid type, a TypeError is thrown"""

        with self.assertRaises(TypeError):
            CalsDbSignal.build_signal(signal=1, t_start=0, t_end=1, ldb=1)

    #endregion


    def test_getSignal_returnsCorrectDataFrame(self):
        """Tests that getSignal returns correct pandas DataFrame"""

        # arrange:
        queryResult = {self.NAME: [self.INDEX_SECS, self.DATA]}

        self.mockLoggingDB.get.return_value = queryResult

        uut = CalsDbSignal(self.NAME, self.START, self.END, ldb=self.mockLoggingDB)

        # act
        result = uut.get_signal(self.NAME, self.START, self.END, ldb=self.mockLoggingDB)

        # assert that pytimber LoggingDB's get method was called with the right arguments
        self.mockLoggingDB.get.assert_called_with(self.NAME, self.START, self.END)

        # assert that a correctly formatted DataFrame is returned
        pd.testing.assert_frame_equal(result, self.calsDF)


    def test_getSignalDF_correctDataFrame(self):
        """Tests that getSignalDF returns correct pandas DataFrame consisting of more than one signal"""

        # arrange:createSignalCALS
        columns = ['signal1', 'signal2', 'signal3']

        self.mockLoggingDB.get.side_effect = [{columns[0]: [self.INDEX_SECS, self.DATA]},
                                              {columns[1]: [self.INDEX_SECS, self.DATA]},
                                              {columns[2]: [self.INDEX_SECS, self.DATA]}]

        df = pd.DataFrame({columns[0]: self.DATA, columns[1]: self.DATA, columns[2]: self.DATA}, self.INDEX_SECS)

        uut = CalsDbSignal(columns, self.START, self.END, ldb=self.mockLoggingDB)

        # act
        result = uut.get_signal_df(columns, self.START, self.END, ldb=self.mockLoggingDB)

        # assert that pytimber LoggingDB's get method was called with the right arguments
        self.mockLoggingDB.get.assert_any_call(columns[0], self.START, self.END)
        self.mockLoggingDB.get.assert_any_call(columns[1], self.START, self.END)
        self.mockLoggingDB.get.assert_any_call(columns[2], self.START, self.END)

        # assert that a correctly formatted DataFrame is returned
        pd.testing.assert_frame_equal(result, df)

    def test_read_multipleNamesProvided_returnsCorrectDataFrameWithPandasTimestampsIndex(self):
        """Tests that read method correctly get a the correct DataFrame. That it then converts the index
        to pandas Timestamps, when the object is created with multiple signal SIGNAL_NAMES"""
        # arrange:
        columns = ['signal1', 'signal2', 'signal3']

        self.mockLoggingDB.get.side_effect = [{columns[0]: [self.INDEX_SECS, self.DATA]},
                                              {columns[1]: [self.INDEX_SECS, self.DATA]},
                                              {columns[2]: [self.INDEX_SECS, self.DATA]}]

        df = pd.DataFrame({columns[0]: self.DATA, columns[1]: self.DATA, columns[2]: self.DATA}, self.INDEX_SECS)

        uut = CalsDbSignal(columns, self.START, self.END, ldb=self.mockLoggingDB)

        # act
        result = uut.read()

        # assert that pytimber LoggingDB's get method was called with the right arguments
        self.mockLoggingDB.get.assert_any_call(columns[0], self.START, self.END)
        self.mockLoggingDB.get.assert_any_call(columns[1], self.START, self.END)
        self.mockLoggingDB.get.assert_any_call(columns[2], self.START, self.END)

        # assert that a correctly formatted DataFrame is returned
        pd.testing.assert_frame_equal(result, df)

    def test_read_singleNameProvided_returnsCorrectDataFrame(self):
        """Tests that read method correctly get a the correct DataFrame. That it then converts the index
        to pandas Timestamps, when the object is created with a single signal signal"""

        self.mockLoggingDB.get.return_value = {self.NAME: [self.INDEX_SECS, self.DATA]}

        uut = CalsDbSignal([self.NAME], self.START, self.END, ldb=self.mockLoggingDB)

        # act
        result = uut.read()

        # assert that pytimber LoggingDB's get method was called with the right arguments
        self.mockLoggingDB.get.assert_called_with(self.NAME, self.START, self.END)

        # assert that a correctly formatted DataFrame is returned
        pd.testing.assert_frame_equal(result, self.calsDF)

    def test_getSignal_queryFailed_RaisesError(self):
        """Tests  that when a query fails a exception is thrown"""

        self.mockLoggingDB.get.side_effect = Exception
        uut = uut = CalsDbSignal([self.NAME], self.START, self.END, ldb=self.mockLoggingDB)

        with self.assertRaises(Exception):
            uut.get_signal(self.NAME)

