import unittest
from pathlib import Path
import os
import pandas as pd

from lhcsmapi.metadata.SignalMetadata import SignalMetadata


class TestSignalMetadata(unittest.TestCase):

    def test_getCircuitTypes(self):
        # arrange

        # act
        circuit_types_act = SignalMetadata.get_circuit_types()

        # assert
        circuit_types_exp = ['IPD2', 'IPD2_B1B2', 'IPQ2', 'IPQ4', 'IPQ8', 'IT', 'RB', 'RQ', '60A', '80-120A', '600A']
        circuit_types_act.sort()
        circuit_types_exp.sort()
        self.assertEqual(circuit_types_exp, circuit_types_act)

    def test_RB_getCircuitNames(self):
        # arrange
        circuit_type = "RB"

        # act
        circuit_names_act = SignalMetadata.get_circuit_names(circuit_type)

        # assert
        circuit_names_exp = ["RB.A12", "RB.A23", "RB.A34", "RB.A45", "RB.A56", "RB.A67", "RB.A78", "RB.A81"]
        circuit_names_act.sort()
        circuit_names_exp.sort()
        self.assertEqual(circuit_names_exp, circuit_names_act)

    def test_RQ_getCircuitNames(self):
        # arrange
        circuit_type = 'RQ'

        # act
        circuit_names_act = SignalMetadata.get_circuit_names(circuit_type)

        # assert
        circuit_names_exp = ['RQD.A12', 'RQD.A23', 'RQD.A34', 'RQD.A45', 'RQD.A56', 'RQD.A67',  'RQD.A78', 'RQD.A81',
                             'RQF.A12', 'RQF.A23', 'RQF.A34', 'RQF.A45', 'RQF.A56', 'RQF.A67', 'RQF.A78', 'RQF.A81']
        circuit_names_act.sort()
        circuit_names_exp.sort()
        self.assertEqual(circuit_names_exp, circuit_names_act)

    def test_RB_getVariableTypesPerCircuit(self):
        # arrange
        circuit_type = "RB"

        # act
        system_types_per_circuit_act = SignalMetadata.get_system_types_per_circuit_type(circuit_type)

        # assert
        system_types_per_circuit_exp = ['CIP', 'CRYO', 'PIC', 'PC', 'QDS', 'QH', 'BUSBAR', 'DIODE_RB', 'VF',
                                        'LEADS_EVEN', 'LEADS_ODD', 'EE']
        system_types_per_circuit_act.sort()
        system_types_per_circuit_exp.sort()
        self.assertEqual(system_types_per_circuit_exp, system_types_per_circuit_act)

    def test_RB_getVariableNames_CALS(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A56"
        system = "PC"
        database = "CALS"

        # act
        variables_act = SignalMetadata.get_variable_names(circuit_type, circuit_name, system, database)

        # assert
        variables_exp = ['I_REF', 'I_MEAS', 'V_MEAS', 'V_REF', 'I_EARTH', 'I_EARTH_PCNT']
        variables_act.sort()
        variables_exp.sort()
        self.assertEqual(variables_exp, variables_act)

    def test_RQ_getVariableNames_CALS(self):
        # arrange
        circuit_type = "RQ"
        circuit_name = "RQF.A81"
        system = "PC"
        database = "CALS"

        # act
        signals_act = SignalMetadata.get_variable_names(circuit_type, circuit_name, system, database)

        # assert
        signals_exp = ['I_REF', 'I_MEAS', 'V_MEAS', 'V_REF', 'I_EARTH', 'I_EARTH_PCNT']
        signals_act.sort()
        signals_exp.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_RB_getVariableNames_CALS_VariableTypeKeyError(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A34"
        system = "PCX"
        database = "CALS"

        # act

        # assert
        with self.assertRaises(KeyError):
            SignalMetadata.get_variable_names(circuit_type, circuit_name, system, database)

    def test_RQ_getVariableNames_CALS_VariableTypeKeyError(self):
        # arrange
        circuit_type = "RQ"
        circuit_name = "RQD.A78"
        system = "PCX"
        database = "CALS"

        # act

        # assert
        with self.assertRaises(KeyError):
            SignalMetadata.get_variable_names(circuit_type, circuit_name, system, database)

    def test_RB_getVariableNames_CALS_DatabaseNameKeyError(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A45"
        system = "EE"
        database = "NXCALS"

        # act

        # assert
        with self.assertRaises(KeyError):
            SignalMetadata.get_variable_names(circuit_type, circuit_name, system, database)

    def test_RQ_getVariableNames_CALS_DatabaseNameKeyError(self):
        # arrange
        circuit_type = "RQ"
        circuit_name = "RQF.A34"
        system = "CIP"
        database = "NXCALS"

        # act

        # assert
        with self.assertRaises(KeyError):
            SignalMetadata.get_variable_names(circuit_type, circuit_name, system, database)

    def test_RB_getVariableNames_PM(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "PC"
        database = "PM"

        # act
        signals_act = SignalMetadata.get_variable_names(circuit_type, circuit_name, system, database)

        # assert
        signals_exp = ['V_REF', 'I_MEAS', 'I_EARTH', 'V_MEAS', 'I_EARTH_PCNT', 'I_A', 'I_REF', 'I_B', 'PROPERTY',
                       'SYMBOL', 'ACTION']
        signals_act.sort()
        signals_exp.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_RQ_getVariableNames_PM(self):
        # arrange
        circuit_type = "RQ"
        circuit_name = "RQF.A12"
        system = "PC"
        database = "PM"

        # act
        signals_act = SignalMetadata.get_variable_names(circuit_type, circuit_name, system, database)

        # assert
        signals_exp = ['V_REF', 'I_MEAS', 'IEARTH', 'V_MEAS', 'I_EARTH_PCNT', 'I_A', 'I_REF', 'I_B', 'PROPERTY',
                       'SYMBOL', 'ACTION']
        signals_act.sort()
        signals_exp.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_RB_getVariableName_PC_CALS_I_MEAS(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system_type = "PC"
        database_name = "CALS"
        variable_name = "I_MEAS"

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = "RPTE.UA23.RB.A12:I_MEAS"
        self.assertEqual(variable_name_exp, variable_name_act)

    def test_RQ_getVariableName_PC_CALS_I_MEAS(self):
        # arrange
        circuit_type = "RQ"
        circuit_name = "RQD.A12"
        system_type = "PC"
        database_name = "CALS"
        variable_name = "I_MEAS"

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = "RPHE.UA23.RQD.A12:I_MEAS"
        self.assertEqual(variable_name_exp, variable_name_act)

    @staticmethod
    def import_dfb_info(circuit_type):
        path = Path(os.path.dirname(__file__))
        file_path = os.path.join(str(path.parent),
                                 '../test/resources/metadata/dfb/{}_DfbInfo.csv'.format(circuit_type))
        return pd.read_csv(file_path)

    def test_RB_getVariableName_LEADS_EVEN_CALS_TT893(self):
        # arrange
        circuit_type = "RB"
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        system_type = "LEADS_EVEN"
        database_name = "CALS"
        variable_name = "TT893"
        dfb_info = TestSignalMetadata.import_dfb_info('RB')
        dfb_info['EVEN_ODD'] = dfb_info.apply(lambda col: int(col['DFB'][-1]), axis=1)

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = dfb_info[dfb_info['EVEN_ODD'] % 2 == 0]['TT893'].values.tolist()
        variable_name_act.sort()
        variable_name_exp.sort()
        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RB_getVariableName_LEADS_ODD_CALS_TT893(self):
        # arrange
        circuit_type = "RB"
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        system_type = "LEADS_ODD"
        database_name = "CALS"
        variable_name = "TT893"
        dfb_info = TestSignalMetadata.import_dfb_info('RB')
        dfb_info['EVEN_ODD'] = dfb_info.apply(lambda col: int(col['DFB'][-1]), axis=1)

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = dfb_info[dfb_info['EVEN_ODD'] % 2 == 1]['TT893'].values.tolist()
        variable_name_act.sort()
        variable_name_exp.sort()
        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RQ_getVariableName_LEADS_CALS_TT893(self):
        # arrange
        circuit_type = "RQ"
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        system_type = "LEADS"
        database_name = "CALS"
        variable_name = "TT893"
        dfb_info = TestSignalMetadata.import_dfb_info('RQ')

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = dfb_info['TT893'].values.tolist()
        variable_name_act.sort()
        variable_name_exp.sort()
        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RB_getVariableName_LEADS_EVEN_CALS_TT891A(self):
        # arrange
        circuit_type = "RB"
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        system_type = "LEADS_EVEN"
        database_name = "CALS"
        variable_name = "TT891A"
        dfb_info = TestSignalMetadata.import_dfb_info('RB')
        dfb_info['EVEN_ODD'] = dfb_info.apply(lambda col: int(col['DFB'][-1]), axis=1)

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = dfb_info[dfb_info['EVEN_ODD'] % 2 == 0]['TT891A'].values.tolist()
        variable_name_act.sort()
        variable_name_exp.sort()
        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RB_getVariableName_LEADS_ODD_CALS_TT891A(self):
        # arrange
        circuit_type = "RB"
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        system_type = "LEADS_ODD"
        database_name = "CALS"
        variable_name = "TT891A"
        dfb_info = TestSignalMetadata.import_dfb_info('RB')
        dfb_info['EVEN_ODD'] = dfb_info.apply(lambda col: int(col['DFB'][-1]), axis=1)

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = dfb_info[dfb_info['EVEN_ODD'] % 2 == 1]['TT891A'].values.tolist()
        variable_name_act.sort()
        variable_name_exp.sort()
        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RQ_getVariableName_LEADS_CALS_TT891A(self):
        # arrange
        circuit_type = "RQ"
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        system_type = "LEADS"
        database_name = "CALS"
        variable_name = "TT891A"
        dfb_info = TestSignalMetadata.import_dfb_info('RQ')

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = dfb_info['TT891A'].values.tolist()
        variable_name_act.sort()
        variable_name_exp.sort()
        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RB_getVariableName_LEADS_EVEN_CALS_CV891_rb(self):
        # arrange
        circuit_type = "RB"
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        system_type = "LEADS_EVEN"
        database_name = "CALS"
        variable_name = "CV891"
        dfb_info = TestSignalMetadata.import_dfb_info('RB')
        dfb_info['EVEN_ODD'] = dfb_info.apply(lambda col: int(col['DFB'][-1]), axis=1)

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = dfb_info[dfb_info['EVEN_ODD'] % 2 == 0]['CV891'].values.tolist()
        variable_name_act.sort()
        variable_name_exp.sort()
        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RB_getVariableName_LEADS_ODD_CALS_CV891_rb(self):
        # arrange
        circuit_type = "RB"
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        system_type = "LEADS_ODD"
        database_name = "CALS"
        variable_name = "CV891"
        dfb_info = TestSignalMetadata.import_dfb_info('RB')
        dfb_info['EVEN_ODD'] = dfb_info.apply(lambda col: int(col['DFB'][-1]), axis=1)

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = dfb_info[dfb_info['EVEN_ODD'] % 2 == 1]['CV891'].values.tolist()
        variable_name_act.sort()
        variable_name_exp.sort()
        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RQ_getVariableName_LEADS_CALS_CV891(self):
        # arrange
        circuit_type = "RQ"
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        system_type = "LEADS"
        database_name = "CALS"
        variable_name = "CV891"
        dfb_info = TestSignalMetadata.import_dfb_info('RQ')

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = dfb_info['CV891'].values.tolist()
        variable_name_act.sort()
        variable_name_exp.sort()
        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RB_getVariableName_EE_CALS_T_RES_BODY(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system_type = "EE"
        database_name = "CALS"
        variable_name = ["T_RES_BODY_1", "T_RES_BODY_2", "T_RES_BODY_3"]

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = ['DQRB.RR17.RB.A12:T_RES_BODY_1', 'DQRB.UA23.RB.A12:T_RES_BODY_1',
                             'DQRB.RR17.RB.A12:T_RES_BODY_2', 'DQRB.UA23.RB.A12:T_RES_BODY_2',
                             'DQRB.RR17.RB.A12:T_RES_BODY_3', 'DQRB.UA23.RB.A12:T_RES_BODY_3']

        variable_name_act.sort()
        variable_name_exp.sort()

        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RQ_getVariableName_EE_CALS_T_RES_BODY(self):
        # arrange
        circuit_type = "RQ"
        circuit_name = ["RQD.A12", "RQF.A12"]
        system_type = "EE"
        database_name = "CALS"
        variable_name = "T_RES"

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = ['DQRQ.UA23.RQD.A12:T_RES', 'DQRQ.UA23.RQF.A12:T_RES']

        variable_name_act.sort()
        variable_name_exp.sort()

        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RB_getVariableName_EE_CALS_ST_RES_OVERTEMP(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system_type = "EE"
        database_name = "CALS"
        variable_name = "ST_RES_OVERTEMP"

        # act
        variable_name_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system_type, database_name,
                                                           variable_name)

        # assert
        variable_name_exp = ['DQRB.RR17.RB.A12:ST_RES_OVERTEMP', 'DQRB.UA23.RB.A12:ST_RES_OVERTEMP']

        variable_name_act.sort()
        variable_name_exp.sort()

        self.assertListEqual(variable_name_exp, variable_name_act)

    def test_RQ_getVariableName_EE_CALS_ST_RES_OVERTEMP(self):
        # arrange
        circuit_type = "RQ"
        circuit_name = ["RQD.A12", "RQF.A12"]
        system = "EE"
        database = "CALS"
        variable = "ST_RES_OVERTEMP"

        # act
        signals_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system, database, variable)

        # assert
        signals_exp = ['DQRQ.UA23.RQD.A12:ST_RES_OVERTEMP', 'DQRQ.UA23.RQF.A12:ST_RES_OVERTEMP']

        signals_act.sort()
        signals_exp.sort()

        self.assertListEqual(signals_exp, signals_act)

    def test_RB_getVariableName_PC_NXCALS_I_MEAS(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "PC"
        database = "NXCALS"
        variable = "I_MEAS"

        # act
        signal_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system, database, variable)

        # assert
        signal_exp = "I_MEAS"
        self.assertEqual(signal_exp, signal_act)

    def test_RQ_getVariableName_PC_NXCALS_I_MEAS(self):
        # arrange
        circuit_type = "RQ"
        circuit_name = "RQD.A12"
        system = "PC"
        database = "NXCALS"
        variable = "I_MEAS"

        # act
        signal_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system, database, variable)

        # assert
        signal_exp = "I_MEAS"
        self.assertEqual(signal_exp, signal_act)

    def test_RB_getVariableName_PC_NXCALS_system(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "PC"
        database = "NXCALS"

        # act
        metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)
        system_act, device_act, property_act = metadata['system'], metadata['device'], metadata['property']

        # assert
        system_exp = "CMW"
        device_exp = "RPTE.UA23.RB.A12"
        property_exp = "SUB"
        self.assertEqual(system_exp, system_act)
        self.assertEqual(device_exp, device_act)
        self.assertEqual(property_exp, property_act)

    def test_RBs_getVariableName_PC_NXCALS_system(self):
        # arrange
        circuit_type = "RB"
        system = "PC"
        database = "NXCALS"

        # act
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        pc_device_property_act = []
        for circuit_name in circuit_names:
            metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)
            pc_device_property_act.append((metadata['device'], metadata['property']))

        # assert
        pc_device_property_exp = [("RPTE.UA23.RB.A12", 'SUB'),
                                  ("RPTE.UA27.RB.A23", "SUB"),
                                  ("RPTE.UA43.RB.A34", "SUB"),
                                  ("RPTE.UA47.RB.A45", "SUB"),
                                  ("RPTE.UA63.RB.A56", "SUB"),
                                  ("RPTE.UA67.RB.A67", "SUB"),
                                  ("RPTE.UA83.RB.A78", "SUB"),
                                  ("RPTE.UA87.RB.A81", "SUB")]
        pc_device_property_exp.sort()
        pc_device_property_act.sort()
        self.assertListEqual(pc_device_property_exp, pc_device_property_act)

    def test_RQs_getVariableName_PC_NXCALS_system(self):
        # arrange
        circuit_type = "RQ"
        system = "PC"
        database = "NXCALS"

        # act
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        pc_device_property_act = []
        for circuit_name in circuit_names:
            metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)
            pc_device_property_act.append((metadata['device'], metadata['property']))

        # assert
        pc_device_property_exp = [('RPHE.UA23.RQD.A12', 'SUB'),
                                  ('RPHE.UA27.RQD.A23', 'SUB'),
                                  ('RPHE.UA43.RQD.A34', 'SUB'),
                                  ('RPHE.UA47.RQD.A45', 'SUB'),
                                  ('RPHE.UA63.RQD.A56', 'SUB'),
                                  ('RPHE.UA67.RQD.A67', 'SUB'),
                                  ('RPHE.UA83.RQD.A78', 'SUB'),
                                  ('RPHE.UA87.RQD.A81', 'SUB'),
                                  ('RPHE.UA23.RQF.A12', 'SUB'),
                                  ('RPHE.UA27.RQF.A23', 'SUB'),
                                  ('RPHE.UA43.RQF.A34', 'SUB'),
                                  ('RPHE.UA47.RQF.A45', 'SUB'),
                                  ('RPHE.UA63.RQF.A56', 'SUB'),
                                  ('RPHE.UA67.RQF.A67', 'SUB'),
                                  ('RPHE.UA83.RQF.A78', 'SUB'),
                                  ('RPHE.UA87.RQF.A81', 'SUB')
                                  ]
        pc_device_property_exp.sort()
        pc_device_property_act.sort()
        self.assertListEqual(pc_device_property_exp, pc_device_property_act)

    def test_RQD_getVariableName_BUSBAR_NXCALS_system(self):
        # arrange
        circuit_type = "RQ"
        circuit_name = "RQD.A12"
        system = "BUSBAR"
        database = "NXCALS"

        # act
        metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)
        system_act = metadata['system']

        # assert
        system_exp = "CMW"
        self.assertEqual(system_exp, system_act)

    def test_RB_getVariableName_QH_I_HDS(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "QH"
        database = "CALS"
        variable = "I_HDS"

        # act
        signals_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system,
                                                           database, variable)

        # assert
        signals_exp = ["%MAGNET%:I_HDS_1", "%MAGNET%:I_HDS_2", "%MAGNET%:I_HDS_3", "%MAGNET%:I_HDS_4"]
        self.assertEqual(signals_exp, signals_act)

    def test_RB_getVariableName_PM_QH_I_HDS_metadata(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "QH"
        database = "PM"

        # act
        metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)
        system_act, className_act, source_act = metadata['system'], metadata['className'], metadata['source']

        # assert
        source_exp = "%CELL%"
        className_exp = "DQAMCNMB_PMHSU"
        system_exp = "QPS"
        self.assertEqual(source_exp, source_act)
        self.assertEqual(className_exp, className_act)
        self.assertEqual(system_exp, system_act)

    def test_RB_getVariableName_PM_PC_metadata(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "PC"
        database = "PM"

        # act
        metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)
        system_act, className_act, source_act = metadata['system'], metadata['className'], metadata['source']

        # assert
        source_exp = "RPTE.UA23.RB.A12"
        className_exp = "51_self_pmd"
        system_exp = "FGC"
        self.assertEqual(source_exp, source_act)
        self.assertEqual(className_exp, className_act)
        self.assertEqual(system_exp, system_act)

    def test_RB_getVariableName_PM_LEADS_EVEN_metadata(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A34"
        system = "LEADS_EVEN"
        database = "PM"

        # act
        metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)
        system_act, className_act, source_act = metadata['system'], metadata['className'], metadata['source']

        # assert
        source_exp = "RB.A34"
        className_exp = "DQAMGNDRBEVEN"
        system_exp = "QPS"
        self.assertEqual(source_exp, source_act)
        self.assertEqual(className_exp, className_act)
        self.assertEqual(system_exp, system_act)

    def test_RB_getVariableName_PM_LEADS_ODD_metadata(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A45"
        system = "LEADS_ODD"
        database = "PM"

        # act
        metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system,
                                                                       database)
        system_act, className_act, source_act = metadata['system'], metadata['className'], metadata['source']

        # assert
        source_exp = "RB.A45"
        className_exp = "DQAMGNDRBODD"
        system_exp = "QPS"
        self.assertEqual(source_exp, source_act)
        self.assertEqual(className_exp, className_act)
        self.assertEqual(system_exp, system_act)

    def test_RB_getVariableName_PM_EE_metadata(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A23"
        system = "EE"
        database = "PM"

        # act
        metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, database)
        system_act, className_act, source_act = metadata['system'], metadata['className'], metadata['source']

        # assert
        source_exp = ["UJ33.RB.A23", "UA27.RB.A23"]
        className_exp = "DQAMSNRB"
        system_exp = "QPS"

        source_act.sort()
        source_exp.sort()
        self.assertEqual(source_exp, source_act)
        self.assertEqual(className_exp, className_act)
        self.assertEqual(system_exp, system_act)

    def test_RB_getVariableName_CALS_PIC_ST_ABORT_PIC(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "PIC"
        database = "CALS"
        variable = "ST_ABORT_PIC"

        # act
        variable_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system, database, variable)

        # assert
        variable_exp = ["RB.A12.ODD:ST_ABORT_PIC", "RB.A12.EVEN:ST_ABORT_PIC"]
        variable_act.sort()
        variable_exp.sort()
        self.assertEqual(variable_exp, variable_act)

    def test_RB_getVariableNames_PM_PC_I_MEAS_I_A_I_REF(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "PC"
        database = "PM"
        variables = ["I_REF", "I_MEAS", "V_REF", "V_MEAS", "I_EARTH", "I_EARTH_PCNT", "I_A", "I_B"]

        # act
        signals_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system, database, variables)

        # assert
        signals_exp = ["STATUS.I_REF", "STATUS.I_MEAS", "STATUS.V_REF", "STATUS.V_MEAS", "IEARTH.IEARTH",
                           "STATUS.I_EARTH_PCNT", "IAB.I_A", "IAB.I_B"]
        signals_act.sort()
        signals_exp.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_RB_getVariableNames_PM_QDS_U_QS0_U_1_U_2(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "QDS"
        database = "PM"
        variables = ['U_QS0', 'U_1', 'U_2']

        # act
        signals_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system, database, variables)

        # assert
        signals_exp = ['%CELL%:U_1', '%CELL%:U_2', '%CELL%:U_QS0']
        signals_act.sort()
        signals_exp.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_RB_getVariableName_CALS_LEADS_ODD_U_BB_2(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system_type = "LEADS_ODD"
        database = "CALS"
        variable = "U_BB_2"

        # act
        signal_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system_type, database, variable)

        # assert
        signal_exp = "RB.A12:U_BB_2"
        self.assertEqual(signal_exp, signal_act)

    def test_RBs_getVariableName_CALS_LEADS_ODD_U_BB_2(self):
        # arrange
        circuit_type = "RB"
        circuit_names = ["RB.A12", "RB.A23"]
        system = "LEADS_ODD"
        database = "CALS"
        variable = "U_BB_2"

        # act
        signals_act = SignalMetadata.get_signal_name(circuit_type, circuit_names, system, database, variable)

        # assert
        signals_exp = ["RB.A12:U_BB_2", "RB.A23:U_BB_2"]
        signals_act.sort()
        signals_exp.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_RB_getVariableName_CALS_BUSBAR_U_RES(self):
        # arrange
        circuit_type = "RB"
        circuit_name = "RB.A12"
        system = "BUSBAR"
        database = "CALS"
        variable = "U_RES"

        # act
        signal_act = SignalMetadata.get_signal_name(circuit_type, circuit_name, system, database, variable)

        # assert
        signal_exp = "%BUSBAR%:U_RES"
        self.assertEqual(signal_exp, signal_act)

    def test_RB_getDatabaseNames_LEADS_ODD(self):
        # arrange
        circuit_type = "RB"
        system = "LEADS_ODD"

        # act
        databases_act = SignalMetadata.get_database_names(circuit_type, system)

        # assert
        databases_exp = ['PM', 'CALS', 'NXCALS']
        databases_act.sort()
        databases_exp.sort()
        self.assertEqual(databases_exp, databases_act)

    def test_RB_getDatabaseNames_PIC(self):
        # arrange
        circuit_type = "RB"
        system_type = "PIC"

        # act
        databases_act = SignalMetadata.get_database_names(circuit_type, system_type)

        # assert
        databases_exp = ['CALS', 'NXCALS']
        databases_act.sort()
        databases_exp.sort()
        self.assertEqual(databases_exp, databases_act)

    def test_RB_getCircuitNameFromMagnetOrCrate(self):
        # arrange
        circuit_type = 'RB'
        magnet_name = "A30L2"

        # act
        circuit_name_act = SignalMetadata.get_circuit_name_from_magnet_or_crate(circuit_type, magnet_name)

        # assert
        circuit_name_exp = 'RB.A12'
        self.assertEqual(circuit_name_exp, circuit_name_act)

    def test_RB_getWildcardForCircuitSystemDbSignal_RB_RBA12_QH_PM_U_HDS(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'QH'
        db = 'PM'
        signal = 'U_HDS'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name,
                                                                                system, db, signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Cell'
        wildcard_values_exp = ['A8R1', 'B8R1', 'A9R1', 'B9R1', 'A10R1', 'B10R1', 'A11R1', 'B11R1', 'A12R1', 'B12R1',
                               'C12R1', 'A13R1', 'B13R1', 'C13R1', 'A14R1', 'B14R1', 'C14R1', 'A15R1', 'B15R1', 'C15R1',
                               'A16R1', 'B16R1', 'C16R1', 'A17R1', 'B17R1', 'C17R1', 'A18R1', 'B18R1', 'C18R1', 'A19R1',
                               'B19R1', 'C19R1', 'A20R1', 'B20R1', 'C20R1', 'A21R1', 'B21R1', 'C21R1', 'A22R1', 'B22R1',
                               'C22R1', 'A23R1', 'B23R1', 'C23R1', 'A24R1', 'B24R1', 'C24R1', 'A25R1', 'B25R1', 'C25R1',
                               'A26R1', 'B26R1', 'C26R1', 'A27R1', 'B27R1', 'C27R1', 'A28R1', 'B28R1', 'C28R1', 'A29R1',
                               'B29R1', 'C29R1', 'A30R1', 'B30R1', 'C30R1', 'A31R1', 'B31R1', 'C31R1', 'A32R1', 'B32R1',
                               'C32R1', 'A33R1', 'B33R1', 'C33R1', 'A34R1', 'B34R1', 'C34R1', 'C34L2', 'B34L2', 'A34L2',
                               'C33L2', 'B33L2', 'A33L2', 'C32L2', 'B32L2', 'A32L2', 'C31L2', 'B31L2', 'A31L2', 'C30L2',
                               'B30L2', 'A30L2', 'C29L2', 'B29L2', 'A29L2', 'C28L2', 'B28L2', 'A28L2', 'C27L2', 'B27L2',
                               'A27L2', 'C26L2', 'B26L2', 'A26L2', 'C25L2', 'B25L2', 'A25L2', 'C24L2', 'B24L2', 'A24L2',
                               'C23L2', 'B23L2', 'A23L2', 'C22L2', 'B22L2', 'A22L2', 'C21L2', 'B21L2', 'A21L2', 'C20L2',
                               'B20L2', 'A20L2', 'C19L2', 'B19L2', 'A19L2', 'C18L2', 'B18L2', 'A18L2', 'C17L2', 'B17L2',
                               'A17L2', 'C16L2', 'B16L2', 'A16L2', 'C15L2', 'B15L2', 'A15L2', 'C14L2', 'B14L2', 'A14L2',
                               'C13L2', 'B13L2', 'A13L2', 'C12L2', 'B12L2', 'A12L2', 'B11L2', 'A11L2', 'B10L2', 'A10L2',
                               'B9L2', 'A9L2', 'B8L2', 'A8L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_QH_CALS_U_HDS(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'QH'
        db = 'CALS'
        signal = 'U_HDS'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db, signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Magnet'
        wildcard_values_exp = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1', 'MB.B9R1', 'MB.A10R1', 'MB.B10R1', 'MB.A11R1',
                               'MB.B11R1', 'MB.A12R1', 'MB.B12R1', 'MB.C12R1', 'MB.A13R1', 'MB.B13R1', 'MB.C13R1',
                               'MB.A14R1', 'MB.B14R1', 'MB.C14R1', 'MB.A15R1', 'MB.B15R1', 'MB.C15R1', 'MB.A16R1',
                               'MB.B16R1', 'MB.C16R1', 'MB.A17R1', 'MB.B17R1', 'MB.C17R1', 'MB.A18R1', 'MB.B18R1',
                               'MB.C18R1', 'MB.A19R1', 'MB.B19R1', 'MB.C19R1', 'MB.A20R1', 'MB.B20R1', 'MB.C20R1',
                               'MB.A21R1', 'MB.B21R1', 'MB.C21R1', 'MB.A22R1', 'MB.B22R1', 'MB.C22R1', 'MB.A23R1',
                               'MB.B23R1', 'MB.C23R1', 'MB.A24R1', 'MB.B24R1', 'MB.C24R1', 'MB.A25R1', 'MB.B25R1',
                               'MB.C25R1', 'MB.A26R1', 'MB.B26R1', 'MB.C26R1', 'MB.A27R1', 'MB.B27R1', 'MB.C27R1',
                               'MB.A28R1', 'MB.B28R1', 'MB.C28R1', 'MB.A29R1', 'MB.B29R1', 'MB.C29R1', 'MB.A30R1',
                               'MB.B30R1', 'MB.C30R1', 'MB.A31R1', 'MB.B31R1', 'MB.C31R1', 'MB.A32R1', 'MB.B32R1',
                               'MB.C32R1', 'MB.A33R1', 'MB.B33R1', 'MB.C33R1', 'MB.A34R1', 'MB.B34R1', 'MB.C34R1',
                               'MB.C34L2', 'MB.B34L2', 'MB.A34L2', 'MB.C33L2', 'MB.B33L2', 'MB.A33L2', 'MB.C32L2',
                               'MB.B32L2', 'MB.A32L2', 'MB.C31L2', 'MB.B31L2', 'MB.A31L2', 'MB.C30L2', 'MB.B30L2',
                               'MB.A30L2', 'MB.C29L2', 'MB.B29L2', 'MB.A29L2', 'MB.C28L2', 'MB.B28L2', 'MB.A28L2',
                               'MB.C27L2', 'MB.B27L2','MB.A27L2', 'MB.C26L2', 'MB.B26L2', 'MB.A26L2', 'MB.C25L2',
                               'MB.B25L2', 'MB.A25L2', 'MB.C24L2', 'MB.B24L2', 'MB.A24L2', 'MB.C23L2', 'MB.B23L2',
                               'MB.A23L2', 'MB.C22L2', 'MB.B22L2', 'MB.A22L2', 'MB.C21L2', 'MB.B21L2', 'MB.A21L2',
                               'MB.C20L2', 'MB.B20L2', 'MB.A20L2', 'MB.C19L2', 'MB.B19L2', 'MB.A19L2', 'MB.C18L2',
                               'MB.B18L2', 'MB.A18L2', 'MB.C17L2', 'MB.B17L2', 'MB.A17L2', 'MB.C16L2', 'MB.B16L2',
                               'MB.A16L2', 'MB.C15L2', 'MB.B15L2', 'MB.A15L2', 'MB.C14L2', 'MB.B14L2', 'MB.A14L2',
                               'MB.C13L2', 'MB.B13L2', 'MB.A13L2', 'MB.C12L2', 'MB.B12L2', 'MB.A12L2', 'MB.B11L2',
                               'MB.A11L2', 'MB.B10L2', 'MB.A10L2', 'MB.B9L2', 'MB.A9L2', 'MB.B8L2', 'MB.A8L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_QDS_PM_U_HDS(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'QDS'
        db = 'PM'
        signal = 'U_HDS'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db, signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Cell'
        wildcard_values_exp = ['A8R1', 'B8R1', 'A9R1', 'B9R1', 'A10R1', 'B10R1', 'A11R1', 'B11R1', 'A12R1', 'B12R1',
                               'C12R1', 'A13R1', 'B13R1', 'C13R1', 'A14R1', 'B14R1', 'C14R1', 'A15R1', 'B15R1', 'C15R1',
                               'A16R1', 'B16R1', 'C16R1', 'A17R1', 'B17R1', 'C17R1', 'A18R1', 'B18R1', 'C18R1', 'A19R1',
                               'B19R1', 'C19R1', 'A20R1', 'B20R1', 'C20R1', 'A21R1', 'B21R1', 'C21R1', 'A22R1', 'B22R1',
                               'C22R1', 'A23R1', 'B23R1', 'C23R1', 'A24R1', 'B24R1', 'C24R1', 'A25R1', 'B25R1', 'C25R1',
                               'A26R1', 'B26R1', 'C26R1', 'A27R1', 'B27R1', 'C27R1', 'A28R1', 'B28R1', 'C28R1', 'A29R1',
                               'B29R1', 'C29R1', 'A30R1', 'B30R1', 'C30R1', 'A31R1', 'B31R1', 'C31R1', 'A32R1', 'B32R1',
                               'C32R1', 'A33R1', 'B33R1', 'C33R1', 'A34R1', 'B34R1', 'C34R1', 'C34L2', 'B34L2', 'A34L2',
                               'C33L2', 'B33L2', 'A33L2', 'C32L2', 'B32L2', 'A32L2', 'C31L2', 'B31L2', 'A31L2', 'C30L2',
                               'B30L2', 'A30L2', 'C29L2', 'B29L2', 'A29L2', 'C28L2', 'B28L2', 'A28L2', 'C27L2', 'B27L2',
                               'A27L2', 'C26L2', 'B26L2', 'A26L2', 'C25L2', 'B25L2', 'A25L2', 'C24L2', 'B24L2', 'A24L2',
                               'C23L2', 'B23L2', 'A23L2', 'C22L2', 'B22L2', 'A22L2', 'C21L2', 'B21L2', 'A21L2', 'C20L2',
                               'B20L2', 'A20L2', 'C19L2', 'B19L2', 'A19L2', 'C18L2', 'B18L2', 'A18L2', 'C17L2', 'B17L2',
                               'A17L2', 'C16L2', 'B16L2', 'A16L2', 'C15L2', 'B15L2', 'A15L2', 'C14L2', 'B14L2', 'A14L2',
                               'C13L2', 'B13L2', 'A13L2', 'C12L2', 'B12L2', 'A12L2', 'B11L2', 'A11L2', 'B10L2', 'A10L2',
                               'B9L2', 'A9L2', 'B8L2', 'A8L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_QDS_CALS_U_HDS(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'QDS'
        db = 'CALS'
        signal = 'U_HDS'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db, signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Magnet'
        wildcard_values_exp = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1', 'MB.B9R1', 'MB.A10R1', 'MB.B10R1', 'MB.A11R1',
                               'MB.B11R1', 'MB.A12R1', 'MB.B12R1','MB.C12R1', 'MB.A13R1', 'MB.B13R1', 'MB.C13R1',
                               'MB.A14R1', 'MB.B14R1', 'MB.C14R1', 'MB.A15R1', 'MB.B15R1', 'MB.C15R1', 'MB.A16R1',
                               'MB.B16R1', 'MB.C16R1', 'MB.A17R1', 'MB.B17R1', 'MB.C17R1', 'MB.A18R1', 'MB.B18R1',
                               'MB.C18R1', 'MB.A19R1', 'MB.B19R1', 'MB.C19R1', 'MB.A20R1', 'MB.B20R1', 'MB.C20R1',
                               'MB.A21R1', 'MB.B21R1', 'MB.C21R1', 'MB.A22R1', 'MB.B22R1', 'MB.C22R1', 'MB.A23R1',
                               'MB.B23R1', 'MB.C23R1', 'MB.A24R1', 'MB.B24R1', 'MB.C24R1', 'MB.A25R1', 'MB.B25R1',
                               'MB.C25R1', 'MB.A26R1', 'MB.B26R1', 'MB.C26R1', 'MB.A27R1', 'MB.B27R1', 'MB.C27R1',
                               'MB.A28R1', 'MB.B28R1', 'MB.C28R1', 'MB.A29R1', 'MB.B29R1', 'MB.C29R1', 'MB.A30R1',
                               'MB.B30R1', 'MB.C30R1', 'MB.A31R1', 'MB.B31R1', 'MB.C31R1', 'MB.A32R1', 'MB.B32R1',
                               'MB.C32R1', 'MB.A33R1', 'MB.B33R1', 'MB.C33R1', 'MB.A34R1', 'MB.B34R1', 'MB.C34R1',
                               'MB.C34L2', 'MB.B34L2', 'MB.A34L2', 'MB.C33L2', 'MB.B33L2', 'MB.A33L2', 'MB.C32L2',
                               'MB.B32L2', 'MB.A32L2', 'MB.C31L2', 'MB.B31L2', 'MB.A31L2', 'MB.C30L2', 'MB.B30L2',
                               'MB.A30L2', 'MB.C29L2', 'MB.B29L2', 'MB.A29L2', 'MB.C28L2', 'MB.B28L2', 'MB.A28L2',
                               'MB.C27L2', 'MB.B27L2', 'MB.A27L2', 'MB.C26L2', 'MB.B26L2', 'MB.A26L2', 'MB.C25L2',
                               'MB.B25L2', 'MB.A25L2', 'MB.C24L2', 'MB.B24L2', 'MB.A24L2', 'MB.C23L2', 'MB.B23L2',
                               'MB.A23L2', 'MB.C22L2', 'MB.B22L2', 'MB.A22L2', 'MB.C21L2', 'MB.B21L2', 'MB.A21L2',
                               'MB.C20L2', 'MB.B20L2', 'MB.A20L2', 'MB.C19L2', 'MB.B19L2', 'MB.A19L2', 'MB.C18L2',
                               'MB.B18L2', 'MB.A18L2', 'MB.C17L2', 'MB.B17L2', 'MB.A17L2', 'MB.C16L2', 'MB.B16L2',
                               'MB.A16L2', 'MB.C15L2', 'MB.B15L2', 'MB.A15L2', 'MB.C14L2', 'MB.B14L2', 'MB.A14L2',
                               'MB.C13L2', 'MB.B13L2', 'MB.A13L2', 'MB.C12L2', 'MB.B12L2', 'MB.A12L2', 'MB.B11L2',
                               'MB.A11L2', 'MB.B10L2', 'MB.A10L2', 'MB.B9L2', 'MB.A9L2', 'MB.B8L2', 'MB.A8L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RQ_RQDA12_QH_CALS_U_HDS(self):
        # arrange
        circuit = 'RQ'
        circuit_name = 'RQD.A12'
        system = 'QH'
        db = 'CALS'
        signal = 'U_HDS'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db,
                                                                                signal)

        # assert
        wildcard_exp = ()
        self.assertEqual(wildcard_exp, wildcard_act)

    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_QH_NXCALS_U_HDS(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'QH'
        db = 'NXCALS'
        signal = 'U_HDS'

        # act

        # assert
        with self.assertRaises(KeyError):
            SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db, signal)

    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_VF_PM_U_EARTH_RB(self):
        # arrange
        circuit_type = 'RB'
        circuit_name = 'RB.A12'
        system = 'VF'
        db = 'PM'
        signal = 'U_EARTH_RB'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit_type, circuit_name, system, db,
                                                                                signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Magnet'
        wildcard_values_exp = ['MB.A8L2', 'MB.A10L2', 'MB.C12L2', 'MB.C14L2', 'MB.C16L2', 'MB.C18L2', 'MB.C20L2',
                               'MB.C22L2',  'MB.C24L2',  'MB.C26L2', 'MB.C28L2', 'MB.C30L2', 'MB.C32L2', 'MB.C34L2',
                               'MB.A33R1', 'MB.C31R1', 'MB.C29R1', 'MB.C27R1', 'MB.C25R1', 'MB.C23R1', 'MB.C21R1',
                               'MB.C19R1', 'MB.C17R1', 'MB.C15R1', 'MB.C13R1', 'MB.B11R1', 'MB.B8R1', 'MB.A8R1',
                               'MB.A10R1', 'MB.C12R1', 'MB.C14R1', 'MB.C16R1', 'MB.C18R1', 'MB.C20R1', 'MB.C22R1',
                               'MB.C24R1', 'MB.C26R1', 'MB.C28R1', 'MB.C30R1', 'MB.C32R1', 'MB.C34R1', 'MB.C33L2',
                               'MB.C31L2', 'MB.C29L2', 'MB.C27L2', 'MB.C25L2', 'MB.C23L2', 'MB.C21L2', 'MB.C19L2',
                               'MB.C17L2', 'MB.C15L2', 'MB.C13L2', 'MB.B11L2', 'MB.B8L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RQ_RQDA12_VF_PM_U_EARTH_RQD(self):
        # arrange
        circuit_type = 'RQ'
        circuit_name = 'RQD.A12'
        system = 'VF_RQD'
        db = 'PM'
        signal = 'U_EARTH_RQD'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit_type, circuit_name, system, db,
                                                                                signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Magnet'
        wildcard_values_exp = ['DFLAS.7L2.2', 'MQ.11L2', 'MQ.11R1', 'MQ.12L2', 'MQ.12R1', 'MQ.13L2', 'MQ.13R1',
                               'MQ.14L2', 'MQ.14R1', 'MQ.15L2', 'MQ.15R1', 'MQ.16L2', 'MQ.16R1', 'MQ.17L2', 'MQ.17R1',
                               'MQ.18L2', 'MQ.18R1', 'MQ.19L2', 'MQ.19R1', 'MQ.20L2', 'MQ.20R1', 'MQ.21L2', 'MQ.21R1',
                               'MQ.22L2', 'MQ.22R1', 'MQ.23L2', 'MQ.23R1', 'MQ.24L2', 'MQ.24R1', 'MQ.25L2', 'MQ.25R1',
                               'MQ.26L2', 'MQ.26R1', 'MQ.27L2', 'MQ.27R1', 'MQ.28L2', 'MQ.28R1', 'MQ.29L2', 'MQ.29R1',
                               'MQ.30L2', 'MQ.30R1', 'MQ.31L2', 'MQ.31R1', 'MQ.32L2', 'MQ.32R1', 'MQ.33L2', 'MQ.33R1',
                               'MQ.34R1']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RQ_RQFA12_VF_PM_U_EARTH_RQF(self):
        # arrange
        circuit_type = 'RQ'
        circuit_name = 'RQF.A12'
        system = 'VF_RQF'
        db = 'PM'
        signal = 'U_EARTH_RQF'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit_type, circuit_name, system, db, signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Magnet'
        wildcard_values_exp = ['DFLAS.7L2.4', 'MQ.11L2', 'MQ.11R1', 'MQ.12L2', 'MQ.12R1', 'MQ.13L2', 'MQ.13R1',
                               'MQ.14L2', 'MQ.14R1', 'MQ.15L2', 'MQ.15R1', 'MQ.16L2', 'MQ.16R1', 'MQ.17L2', 'MQ.17R1',
                               'MQ.18L2', 'MQ.18R1', 'MQ.19L2', 'MQ.19R1', 'MQ.20L2', 'MQ.20R1', 'MQ.21L2', 'MQ.21R1',
                               'MQ.22L2', 'MQ.22R1', 'MQ.23L2', 'MQ.23R1', 'MQ.24L2', 'MQ.24R1', 'MQ.25L2', 'MQ.25R1',
                               'MQ.26L2', 'MQ.26R1', 'MQ.27L2', 'MQ.27R1', 'MQ.28L2', 'MQ.28R1', 'MQ.29L2', 'MQ.29R1',
                               'MQ.30L2', 'MQ.30R1', 'MQ.31L2', 'MQ.31R1', 'MQ.32L2', 'MQ.32R1', 'MQ.33L2', 'MQ.33R1',
                               'MQ.34R1']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_VF_CALS_U_EARTH(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'VF'
        db = 'CALS'
        signal = 'U_EARTH'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db, signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Magnet'
        wildcard_values_exp = ['MB.A8L2', 'MB.A10L2', 'MB.C12L2', 'MB.C14L2', 'MB.C16L2', 'MB.C18L2', 'MB.C20L2',
                               'MB.C22L2', 'MB.C24L2', 'MB.C26L2', 'MB.C28L2', 'MB.C30L2', 'MB.C32L2', 'MB.C34L2',
                               'MB.A33R1', 'MB.C31R1', 'MB.C29R1', 'MB.C27R1', 'MB.C25R1', 'MB.C23R1', 'MB.C21R1',
                               'MB.C19R1', 'MB.C17R1', 'MB.C15R1', 'MB.C13R1', 'MB.B11R1', 'MB.B8R1', 'MB.A8R1',
                               'MB.A10R1', 'MB.C12R1', 'MB.C14R1', 'MB.C16R1', 'MB.C18R1', 'MB.C20R1', 'MB.C22R1',
                               'MB.C24R1', 'MB.C26R1', 'MB.C28R1', 'MB.C30R1', 'MB.C32R1', 'MB.C34R1', 'MB.C33L2',
                               'MB.C31L2', 'MB.C29L2', 'MB.C27L2', 'MB.C25L2', 'MB.C23L2', 'MB.C21L2', 'MB.C19L2',
                               'MB.C17L2', 'MB.C15L2', 'MB.C13L2', 'MB.B11L2', 'MB.B8L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)


    def test_getWildcardForCircuitSystemDbSignal_RQ_RQDA12_CALS_U_DIODE_RQD(self):
        # arrange
        circuit = 'RQ'
        circuit_name = 'RQD.A12'
        system = 'DIODE_RQD'
        db = 'CALS'
        signal = 'U_DIODE_RQD'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db,
                                                                                signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Magnet'
        wildcard_values_exp = ['MQ.11L2', 'MQ.11R1', 'MQ.12L2', 'MQ.12R1', 'MQ.13L2', 'MQ.13R1', 'MQ.14L2', 'MQ.14R1',
                               'MQ.15L2', 'MQ.15R1', 'MQ.16L2', 'MQ.16R1', 'MQ.17L2', 'MQ.17R1', 'MQ.18L2', 'MQ.18R1',
                               'MQ.19L2', 'MQ.19R1', 'MQ.20L2', 'MQ.20R1', 'MQ.21L2', 'MQ.21R1', 'MQ.22L2', 'MQ.22R1',
                               'MQ.23L2', 'MQ.23R1', 'MQ.24L2', 'MQ.24R1', 'MQ.25L2', 'MQ.25R1', 'MQ.26L2', 'MQ.26R1',
                               'MQ.27L2', 'MQ.27R1', 'MQ.28L2', 'MQ.28R1', 'MQ.29L2', 'MQ.29R1', 'MQ.30L2', 'MQ.30R1',
                               'MQ.31L2', 'MQ.31R1', 'MQ.32L2', 'MQ.32R1', 'MQ.33L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RQ_RQFA12_CALS_U_DIODE_RQF(self):
        # arrange
        circuit = 'RQ'
        circuit_name = 'RQF.A12'
        system = 'DIODE_RQF'
        db = 'CALS'
        signal = 'U_DIODE_RQF'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db,
                                                                                signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Magnet'
        wildcard_values_exp = ['MQ.11L2', 'MQ.11R1', 'MQ.12L2', 'MQ.12R1', 'MQ.13L2', 'MQ.13R1', 'MQ.14L2', 'MQ.14R1',
                               'MQ.15L2', 'MQ.15R1', 'MQ.16L2', 'MQ.16R1', 'MQ.17L2', 'MQ.17R1', 'MQ.18L2', 'MQ.18R1',
                               'MQ.19L2', 'MQ.19R1', 'MQ.20L2', 'MQ.20R1', 'MQ.21L2', 'MQ.21R1', 'MQ.22L2', 'MQ.22R1',
                               'MQ.23L2', 'MQ.23R1', 'MQ.24L2', 'MQ.24R1', 'MQ.25L2', 'MQ.25R1', 'MQ.26L2', 'MQ.26R1',
                               'MQ.27L2', 'MQ.27R1', 'MQ.28L2', 'MQ.28R1', 'MQ.29L2', 'MQ.29R1', 'MQ.30L2', 'MQ.30R1',
                               'MQ.31L2', 'MQ.31R1', 'MQ.32L2', 'MQ.32R1', 'MQ.33L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)


    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_BUSBAR_CALS_U_RES(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'BUSBAR'
        db = 'CALS'
        signal = 'U_RES'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db,
                                                                                signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Busbar'
        wildcard_values_exp = ['DCBB.8L2.R', 'DCBB.9L2.R', 'DCBB.10L2.R', 'DCBB.11L2.R', 'DCBB.A12L2.R', 'DCBB.B12L2.R',
                               'DCBB.13L2.R', 'DCBB.A14L2.R', 'DCBB.B14L2.R', 'DCBB.15L2.R', 'DCBB.A16L2.R',
                               'DCBB.B16L2.R', 'DCBB.17L2.R', 'DCBB.A18L2.R', 'DCBB.B18L2.R', 'DCBB.19L2.R',
                               'DCBB.A20L2.R', 'DCBB.B20L2.R', 'DCBB.21L2.R', 'DCBB.A22L2.R', 'DCBB.B22L2.R',
                               'DCBB.23L2.R', 'DCBB.A24L2.R', 'DCBB.B24L2.R', 'DCBB.25L2.R', 'DCBB.A26L2.R',
                               'DCBB.B26L2.R', 'DCBB.27L2.R', 'DCBB.A28L2.R', 'DCBB.B28L2.R', 'DCBB.29L2.R',
                               'DCBB.A30L2.R', 'DCBB.B30L2.R', 'DCBB.31L2.R', 'DCBB.A32L2.R', 'DCBB.B32L2.R',
                               'DCBB.33L2.R', 'DCBB.A34L2.R', 'DCBB.B34L2.R', 'DCBB.34R1.R', 'DCBB.B33R1.R',
                               'DCBB.A33R1.R', 'DCBB.32R1.R', 'DCBB.B31R1.R', 'DCBB.A31R1.R', 'DCBB.30R1.R',
                               'DCBB.B29R1.R', 'DCBB.A29R1.R', 'DCBB.28R1.R', 'DCBB.B27R1.R', 'DCBB.A27R1.R',
                               'DCBB.26R1.R', 'DCBB.B25R1.R', 'DCBB.A25R1.R', 'DCBB.24R1.R', 'DCBB.B23R1.R',
                               'DCBB.A23R1.R', 'DCBB.22R1.R', 'DCBB.B21R1.R', 'DCBB.A21R1.R', 'DCBB.20R1.R',
                               'DCBB.B19R1.R', 'DCBB.A19R1.R', 'DCBB.18R1.R', 'DCBB.B17R1.R', 'DCBB.A17R1.R',
                               'DCBB.16R1.R', 'DCBB.B15R1.R', 'DCBB.A15R1.R', 'DCBB.14R1.R', 'DCBB.B13R1.R',
                               'DCBB.A13R1.R', 'DCBB.12R1.R', 'DCBB.11R1.R', 'DCBB.10R1.R', 'DCBB.9R1.R', 'DCBB.8R1.R',
                               'DCBD.7R1.R', 'DCBQ.7R1.L', 'DCBQ.8R1.L', 'DCBQ.9R1.L', 'DCBQ.10R1.L', 'DCBQ.11R1.L',
                               'DCBB.12R1.L', 'DCBB.A13R1.L', 'DCBQ.13R1.L', 'DCBB.14R1.L', 'DCBB.A15R1.L',
                               'DCBQ.15R1.L', 'DCBB.16R1.L', 'DCBB.A17R1.L', 'DCBQ.17R1.L', 'DCBB.18R1.L',
                               'DCBB.A19R1.L', 'DCBQ.19R1.L', 'DCBB.20R1.L', 'DCBB.A21R1.L', 'DCBQ.21R1.L',
                               'DCBB.22R1.L', 'DCBB.A23R1.L', 'DCBQ.23R1.L', 'DCBB.24R1.L', 'DCBB.A25R1.L',
                               'DCBQ.25R1.L', 'DCBB.26R1.L', 'DCBB.A27R1.L', 'DCBQ.27R1.L', 'DCBB.28R1.L',
                               'DCBB.A29R1.L', 'DCBQ.29R1.L', 'DCBB.30R1.L', 'DCBB.A31R1.L', 'DCBQ.31R1.L',
                               'DCBB.32R1.L', 'DCBB.A33R1.L', 'DCBQ.33R1.L', 'DCBB.34R1.L', 'DCBB.B34L2.L',
                               'DCBQ.33L2.L', 'DCBB.33L2.L', 'DCBB.B32L2.L', 'DCBQ.31L2.L', 'DCBB.31L2.L',
                               'DCBB.B30L2.L', 'DCBQ.29L2.L', 'DCBB.29L2.L', 'DCBB.B28L2.L', 'DCBQ.27L2.L',
                               'DCBB.27L2.L', 'DCBB.A26L2.L', 'DCBQ.25L2.L', 'DCBB.25L2.L', 'DCBB.B24L2.L',
                               'DCBQ.23L2.L', 'DCBB.23L2.L', 'DCBB.B22L2.L', 'DCBQ.21L2.L', 'DCBB.21L2.L',
                               'DCBB.B20L2.L', 'DCBQ.19L2.L', 'DCBB.19L2.L', 'DCBB.B18L2.L', 'DCBQ.17L2.L',
                               'DCBB.17L2.L', 'DCBB.B16L2.L', 'DCBQ.15L2.L', 'DCBB.15L2.L', 'DCBB.B14L2.L',
                               'DCBQ.13L2.L', 'DCBB.13L2.L', 'DCBB.A12L2.L', 'DCBQ.11L2.L', 'DCBQ.10L2.L',
                               'DCBQ.9L2.L', 'DCBQ.8L2.L', 'DCBD.7L2.L']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_BUSBAR_PM_U_RES(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'BUSBAR'
        db = 'PM'
        signal = 'U_RES'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db,
                                                                                signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Busbar'
        wildcard_values_exp = ['DCBB.8L2.R', 'DCBB.9L2.R', 'DCBB.10L2.R', 'DCBB.11L2.R', 'DCBB.A12L2.R', 'DCBB.B12L2.R',
                               'DCBB.13L2.R', 'DCBB.A14L2.R', 'DCBB.B14L2.R', 'DCBB.15L2.R', 'DCBB.A16L2.R',
                               'DCBB.B16L2.R', 'DCBB.17L2.R', 'DCBB.A18L2.R', 'DCBB.B18L2.R', 'DCBB.19L2.R',
                               'DCBB.A20L2.R', 'DCBB.B20L2.R', 'DCBB.21L2.R', 'DCBB.A22L2.R', 'DCBB.B22L2.R',
                               'DCBB.23L2.R', 'DCBB.A24L2.R', 'DCBB.B24L2.R', 'DCBB.25L2.R', 'DCBB.A26L2.R',
                               'DCBB.B26L2.R', 'DCBB.27L2.R', 'DCBB.A28L2.R', 'DCBB.B28L2.R', 'DCBB.29L2.R',
                               'DCBB.A30L2.R', 'DCBB.B30L2.R', 'DCBB.31L2.R', 'DCBB.A32L2.R', 'DCBB.B32L2.R',
                               'DCBB.33L2.R', 'DCBB.A34L2.R', 'DCBB.B34L2.R', 'DCBB.34R1.R', 'DCBB.B33R1.R',
                               'DCBB.A33R1.R', 'DCBB.32R1.R', 'DCBB.B31R1.R', 'DCBB.A31R1.R', 'DCBB.30R1.R',
                               'DCBB.B29R1.R', 'DCBB.A29R1.R', 'DCBB.28R1.R', 'DCBB.B27R1.R', 'DCBB.A27R1.R',
                               'DCBB.26R1.R', 'DCBB.B25R1.R', 'DCBB.A25R1.R', 'DCBB.24R1.R', 'DCBB.B23R1.R',
                               'DCBB.A23R1.R', 'DCBB.22R1.R', 'DCBB.B21R1.R', 'DCBB.A21R1.R', 'DCBB.20R1.R',
                               'DCBB.B19R1.R', 'DCBB.A19R1.R', 'DCBB.18R1.R', 'DCBB.B17R1.R', 'DCBB.A17R1.R',
                               'DCBB.16R1.R', 'DCBB.B15R1.R', 'DCBB.A15R1.R', 'DCBB.14R1.R', 'DCBB.B13R1.R',
                               'DCBB.A13R1.R', 'DCBB.12R1.R', 'DCBB.11R1.R', 'DCBB.10R1.R', 'DCBB.9R1.R', 'DCBB.8R1.R',
                               'DCBD.7R1.R', 'DCBQ.7R1.L', 'DCBQ.8R1.L', 'DCBQ.9R1.L', 'DCBQ.10R1.L', 'DCBQ.11R1.L',
                               'DCBB.12R1.L', 'DCBB.A13R1.L', 'DCBQ.13R1.L', 'DCBB.14R1.L', 'DCBB.A15R1.L',
                               'DCBQ.15R1.L', 'DCBB.16R1.L', 'DCBB.A17R1.L', 'DCBQ.17R1.L', 'DCBB.18R1.L',
                               'DCBB.A19R1.L', 'DCBQ.19R1.L', 'DCBB.20R1.L', 'DCBB.A21R1.L', 'DCBQ.21R1.L',
                               'DCBB.22R1.L', 'DCBB.A23R1.L', 'DCBQ.23R1.L', 'DCBB.24R1.L', 'DCBB.A25R1.L',
                               'DCBQ.25R1.L', 'DCBB.26R1.L', 'DCBB.A27R1.L', 'DCBQ.27R1.L', 'DCBB.28R1.L',
                               'DCBB.A29R1.L', 'DCBQ.29R1.L', 'DCBB.30R1.L', 'DCBB.A31R1.L', 'DCBQ.31R1.L',
                               'DCBB.32R1.L', 'DCBB.A33R1.L', 'DCBQ.33R1.L', 'DCBB.34R1.L', 'DCBB.B34L2.L',
                               'DCBQ.33L2.L', 'DCBB.33L2.L', 'DCBB.B32L2.L', 'DCBQ.31L2.L', 'DCBB.31L2.L',
                               'DCBB.B30L2.L', 'DCBQ.29L2.L', 'DCBB.29L2.L', 'DCBB.B28L2.L', 'DCBQ.27L2.L',
                               'DCBB.27L2.L', 'DCBB.A26L2.L', 'DCBQ.25L2.L', 'DCBB.25L2.L', 'DCBB.B24L2.L',
                               'DCBQ.23L2.L', 'DCBB.23L2.L', 'DCBB.B22L2.L', 'DCBQ.21L2.L', 'DCBB.21L2.L',
                               'DCBB.B20L2.L', 'DCBQ.19L2.L', 'DCBB.19L2.L', 'DCBB.B18L2.L', 'DCBQ.17L2.L',
                               'DCBB.17L2.L', 'DCBB.B16L2.L', 'DCBQ.15L2.L', 'DCBB.15L2.L', 'DCBB.B14L2.L',
                               'DCBQ.13L2.L', 'DCBB.13L2.L', 'DCBB.A12L2.L', 'DCBQ.11L2.L', 'DCBQ.10L2.L',
                               'DCBQ.9L2.L', 'DCBQ.8L2.L', 'DCBD.7L2.L']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_DIODE_PM_U_REF_N1(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'DIODE_RB'
        db = 'PM'
        signal = 'U_REF_N1'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db,
                                                                               signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Crate'
        wildcard_values_exp = ['B8L2', 'B10L2', 'B12L2', 'B14L2', 'B16L2', 'B18L2', 'B20L2', 'B22L2', 'B24L2', 'B26L2',
                               'B28L2', 'B30L2', 'B32L2', 'B34L2', 'B33R1', 'B31R1', 'B29R1', 'B27R1', 'B25R1', 'B23R1',
                               'B21R1', 'B19R1', 'B17R1', 'B15R1', 'B13R1', 'B11R1', 'B9R1', 'B8R1', 'B10R1', 'B12R1',
                               'B14R1', 'B16R1', 'B18R1', 'B20R1', 'B22R1', 'B24R1', 'B26R1', 'B28R1', 'B30R1', 'B32R1',
                               'B34R1', 'B33L2', 'B31L2', 'B29L2', 'B27L2', 'B25L2', 'B23L2', 'B21L2', 'B19L2', 'B17L2',
                               'B15L2', 'B13L2', 'B11L2', 'B9L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_getWildcardForCircuitSystemDbSignal_RB_RBA12_DIODE_PM_U_DIODE_RB(self):
        # arrange
        circuit = 'RB'
        circuit_name = 'RB.A12'
        system = 'DIODE_RB'
        db = 'PM'
        signal = 'U_DIODE_RB'

        # act
        wildcard_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit, circuit_name, system, db,
                                                                                signal)
        wildcard_name_act = wildcard_act[0]
        wildcard_values_act = wildcard_act[1]

        # assert
        wildcard_name_exp = 'Magnet'
        wildcard_values_exp = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1', 'MB.B9R1', 'MB.A10R1', 'MB.B10R1', 'MB.A11R1',
                               'MB.B11R1', 'MB.A12R1', 'MB.B12R1', 'MB.C12R1', 'MB.A13R1', 'MB.B13R1', 'MB.C13R1',
                               'MB.A14R1', 'MB.B14R1', 'MB.C14R1', 'MB.A15R1', 'MB.B15R1', 'MB.C15R1', 'MB.A16R1',
                               'MB.B16R1', 'MB.C16R1', 'MB.A17R1', 'MB.B17R1', 'MB.C17R1', 'MB.A18R1', 'MB.B18R1',
                               'MB.C18R1', 'MB.A19R1', 'MB.B19R1', 'MB.C19R1', 'MB.A20R1', 'MB.B20R1', 'MB.C20R1',
                               'MB.A21R1', 'MB.B21R1', 'MB.C21R1', 'MB.A22R1', 'MB.B22R1', 'MB.C22R1', 'MB.A23R1',
                               'MB.B23R1', 'MB.C23R1', 'MB.A24R1', 'MB.B24R1', 'MB.C24R1', 'MB.A25R1', 'MB.B25R1',
                               'MB.C25R1', 'MB.A26R1', 'MB.B26R1', 'MB.C26R1', 'MB.A27R1', 'MB.B27R1', 'MB.C27R1',
                               'MB.A28R1', 'MB.B28R1', 'MB.C28R1', 'MB.A29R1', 'MB.B29R1', 'MB.C29R1', 'MB.A30R1',
                               'MB.B30R1', 'MB.C30R1', 'MB.A31R1', 'MB.B31R1', 'MB.C31R1', 'MB.A32R1', 'MB.B32R1',
                               'MB.C32R1', 'MB.A33R1', 'MB.B33R1', 'MB.C33R1', 'MB.A34R1', 'MB.B34R1', 'MB.C34R1',
                               'MB.C34L2', 'MB.B34L2', 'MB.A34L2', 'MB.C33L2', 'MB.B33L2', 'MB.A33L2', 'MB.C32L2',
                               'MB.B32L2', 'MB.A32L2', 'MB.C31L2', 'MB.B31L2', 'MB.A31L2', 'MB.C30L2', 'MB.B30L2',
                               'MB.A30L2', 'MB.C29L2', 'MB.B29L2', 'MB.A29L2', 'MB.C28L2', 'MB.B28L2', 'MB.A28L2',
                               'MB.C27L2', 'MB.B27L2', 'MB.A27L2', 'MB.C26L2', 'MB.B26L2', 'MB.A26L2', 'MB.C25L2',
                               'MB.B25L2', 'MB.A25L2', 'MB.C24L2', 'MB.B24L2', 'MB.A24L2', 'MB.C23L2', 'MB.B23L2',
                               'MB.A23L2', 'MB.C22L2', 'MB.B22L2', 'MB.A22L2', 'MB.C21L2', 'MB.B21L2', 'MB.A21L2',
                               'MB.C20L2', 'MB.B20L2', 'MB.A20L2', 'MB.C19L2', 'MB.B19L2', 'MB.A19L2', 'MB.C18L2',
                               'MB.B18L2', 'MB.A18L2', 'MB.C17L2', 'MB.B17L2', 'MB.A17L2', 'MB.C16L2', 'MB.B16L2',
                               'MB.A16L2', 'MB.C15L2', 'MB.B15L2', 'MB.A15L2', 'MB.C14L2', 'MB.B14L2', 'MB.A14L2',
                               'MB.C13L2', 'MB.B13L2', 'MB.A13L2', 'MB.C12L2', 'MB.B12L2', 'MB.A12L2', 'MB.B11L2',
                               'MB.A11L2', 'MB.B10L2', 'MB.A10L2', 'MB.B9L2', 'MB.A9L2', 'MB.B8L2', 'MB.A8L2']

        self.assertEqual(wildcard_name_exp, wildcard_name_act)
        wildcard_values_exp.sort()
        wildcard_values_act.sort()
        self.assertEqual(wildcard_values_exp, wildcard_values_act)

    def test_updatePMSourceWithWildcard_RB_QH_U_HDS(self):
        # arrange
        source_in = '*'
        circuit = 'RB'
        system = 'QH'
        signal = 'U_HDS'
        wildcard = ['A8R1', 'B8R1', 'A9R1']

        # act
        source_act = SignalMetadata.update_pm_source_with_wildcard(source_in, circuit, system, signal, wildcard)

        # assert
        source_exp = wildcard
        source_exp.sort()
        source_act.sort()
        self.assertEqual(source_exp, source_act)

    def test_updatePMSourceWithWildcard_RB_QDS_U_HDS(self):
        # arrange
        source_in = '*'
        circuit = 'RB'
        system = 'QDS'
        signal = 'U_HDS'
        wildcard = 'A8R1'

        # act
        source_act = SignalMetadata.update_pm_source_with_wildcard(source_in, circuit, system, signal, wildcard)

        # assert
        source_exp = [wildcard]
        self.assertEqual(source_exp, source_act)

    def test_updatePMSourceWithWildcard_RB_BUSBAR_U_RES(self):
        # arrange
        source_in = '*'
        circuit = 'RB'
        system = 'BUSBAR'
        signal = 'U_RES'
        wildcard = ['DCBB.8L2.R', 'DCBB.9L2.R', 'DCBB.10L2.R']

        # act
        source_act = SignalMetadata.update_pm_source_with_wildcard(source_in, circuit, system, signal, wildcard)

        # assert
        source_exp = ['B10L2', 'B8L2', 'B8L2']
        source_exp.sort()
        source_act.sort()
        self.assertEqual(source_exp, source_act)

    def test_updatePMSourceWithWildcard_RB_DIODE_U_REF_N1(self):
        # arrange
        source_in = '*'
        circuit = 'RB'
        system = 'DIODE_RB'
        signal = 'U_REF_N1'
        wildcard = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1']

        # act
        source_act = SignalMetadata.update_pm_source_with_wildcard(source_in, circuit, system, signal, wildcard)

        # assert
        source_exp = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1']
        source_exp.sort()
        source_act.sort()
        self.assertEqual(source_exp, source_act)

    def test_updatePMSourceWithWildcard_RB_DIODE_U_DIODE_RB(self):
        # arrange
        source_in = '*'
        circuit = 'RB'
        system = 'DIODE_RB'
        signal = 'U_DIODE_RB'
        wildcard = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1']

        # act
        source_act = SignalMetadata.update_pm_source_with_wildcard(source_in, circuit, system, signal, wildcard)

        # assert
        source_exp = ['B10R1', 'B8R1', 'B9R1']
        source_exp.sort()
        source_act.sort()
        self.assertEqual(source_exp, source_act)

    def test_updatePMSourceWithWildcard_RB_VF_U_EARTH(self):
        # arrange
        source_in = "%QPS_CRATE%"
        circuit = 'RB'
        system = 'VF'
        signal = 'U_EARTH'
        wildcard = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1']

        # act
        source_act = SignalMetadata.update_pm_source_with_wildcard(source_in, circuit, system, signal, wildcard)

        # assert
        source_exp = ['B10R1', 'B8R1', 'B9R1']
        source_exp.sort()
        source_act.sort()
        self.assertEqual(source_exp, source_act)

    # UPDATE PM SIGNAL NAMES WITH WILDCARD
    def test_updatePMSignalNamesWithWildcard_RB_QH_U_HDS(self):
        # arrange
        signals_in = ['U_HDS_1', 'U_HDS_2', 'U_HDS_3', 'U_HDS_4']
        system = 'QH'
        signal = 'U_HDS'
        wildcard = ['A8R1', 'B8R1', 'A9R1']

        # act
        signals_act = SignalMetadata.update_pm_signal_names_with_wildcard(signals_in, system, signal, wildcard)

        # assert
        signals_exp = ['U_HDS_1', 'U_HDS_2', 'U_HDS_3', 'U_HDS_4']
        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_updatePMSignalNamesWithWildcard_RB_QDS_U_HDS(self):
        # arrange
        signals_in = ["%CELL%:U_QS0", "%CELL%:U_1", "%CELL%:U_2"]
        system = 'QDS'
        signal = 'U_HDS'
        wildcard = ['A8R1', 'B8R1', 'A9R1']

        # act
        signals_act = SignalMetadata.update_pm_signal_names_with_wildcard(signals_in, system, signal, wildcard)

        # assert
        signals_exp = ['A8R1:U_QS0', 'A8R1:U_1', 'A8R1:U_2', 'B8R1:U_QS0', 'B8R1:U_1', 'B8R1:U_2', 'A9R1:U_QS0',
                       'A9R1:U_1', 'A9R1:U_2']
        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_updatePMSignalNamesWithWildcard_RB_BUSBAR_U_RES(self):
        # arrange
        signals_in = ["%BUSBAR%:U_MAG", "%BUSBAR%:U_RES", "%BUSBAR%:ST_BUSBAR_OK", "%BUSBAR%:ST_PWR_PERM"]
        system = 'BUSBAR'
        signal = 'U_RES'
        wildcard = ['DCBB.8L2.R', 'DCBB.9L2.R', 'DCBB.10L2.R']

        # act
        signals_act = SignalMetadata.update_pm_signal_names_with_wildcard(signals_in, system, signal, wildcard)

        # assert
        signals_exp = ['DCBB.8L2.R:U_MAG', 'DCBB.8L2.R:U_RES', 'DCBB.8L2.R:ST_BUSBAR_OK', 'DCBB.8L2.R:ST_PWR_PERM',
                       'DCBB.9L2.R:U_MAG', 'DCBB.9L2.R:U_RES', 'DCBB.9L2.R:ST_BUSBAR_OK', 'DCBB.9L2.R:ST_PWR_PERM',
                       'DCBB.10L2.R:U_MAG', 'DCBB.10L2.R:U_RES', 'DCBB.10L2.R:ST_BUSBAR_OK', 'DCBB.10L2.R:ST_PWR_PERM']
        signals_act.sort()
        signals_exp.sort()
        self.assertEqual(signals_act, signals_exp)

    def test_updatePMSignalNamesWithWildcard_RB_DIODE_U_REF_N1(self):
        # arrange
        signals_in = ["DQQDS.%QPS_CRATE%.RB.A12:U_REF_N1"]
        system = 'DIODE'
        signal = 'U_REF_N1'
        wildcard = ['B8L2', 'B10L2', 'B12L2']

        # act
        signals_act = SignalMetadata.update_pm_signal_names_with_wildcard(signals_in, system, signal, wildcard)

        # assert
        signals_exp = ['DQQDS.B8L2.RB.A12:U_REF_N1', 'DQQDS.B10L2.RB.A12:U_REF_N1', 'DQQDS.B12L2.RB.A12:U_REF_N1']
        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_updatePMSignalNamesWithWildcard_RB_DIODE_U_DIODE_RB(self):
        # arrange
        signals_in = "%MAGNET%:ST_MAGNET_OK_SQ_RB"
        system = 'DIODE'
        signal = 'U_DIODE_RB'
        wildcard = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1']

        # act
        signals_act = SignalMetadata.update_pm_signal_names_with_wildcard(signals_in, system, signal, wildcard)

        # assert
        signals_exp = ['MB.A8R1:ST_MAGNET_OK_SQ_RB', 'MB.B8R1:ST_MAGNET_OK_SQ_RB', 'MB.A9R1:ST_MAGNET_OK_SQ_RB']
        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_updatePMSignalNamesWithWildcard_RB_VF_U_EARTH(self):
        # arrange
        signals_in = "%MAGNET%:U_EARTH_RB"
        system = 'VF'
        signal = 'U_EARTH_RB'
        wildcard = ['MB.A8L2', 'MB.A10L2', 'MB.C12L2', 'MB.C14L2']

        # act
        signals_act = SignalMetadata.update_pm_signal_names_with_wildcard(signals_in, system, signal, wildcard)

        # assert
        signals_exp = ['MB.A8L2:U_EARTH_RB', 'MB.A10L2:U_EARTH_RB', 'MB.C12L2:U_EARTH_RB', 'MB.C14L2:U_EARTH_RB']
        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_updatePMSignalNamesWithWildcard_RB_PC_I_MEAS(self):
        # arrange
        signals_in = "STATUS.I_REF"
        system = 'PC'
        signal = 'I_REF'
        wildcard = ''

        # act
        signals_act = SignalMetadata.update_pm_signal_names_with_wildcard(signals_in, system, signal, wildcard)

        # assert
        signals_exp = ["STATUS.I_REF"]
        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    # UPDATE CALS SIGNAL NAMES WITH WILDCARD
    def test_updateCALSSignalNamesWithWildcard_RB_QH_U_HDS(self):
        # arrange
        signals_in = ["%MAGNET%:U_HDS_1", "%MAGNET%:U_HDS_2", "%MAGNET%:U_HDS_3", "%MAGNET%:U_HDS_4"]
        system = 'QH'
        wildcard = ['A8R1', 'B8R1', 'A9R1']

        # act
        signals_act = SignalMetadata.update_cals_signal_names_with_wildcard(signals_in, system, wildcard)

        # assert
        signals_exp = ['A8R1:U_HDS_1', 'A8R1:U_HDS_2', 'A8R1:U_HDS_3', 'A8R1:U_HDS_4', 'B8R1:U_HDS_1',
                          'B8R1:U_HDS_2', 'B8R1:U_HDS_3', 'B8R1:U_HDS_4', 'A9R1:U_HDS_1', 'A9R1:U_HDS_2',
                          'A9R1:U_HDS_3', 'A9R1:U_HDS_4']

        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_updateVSLDSignalNamesWithWildcard_RB_QDS_U_HDS(self):
        # arrange
        signals_in = ["%MAGNET%:U_QS0", "%MAGNET%:U_1", "%MAGNET%:U_2"]
        system = 'QDS'
        wildcard = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1']

        # act
        signals_act = SignalMetadata.update_cals_signal_names_with_wildcard(signals_in, system, wildcard)

        # assert
        signals_exp = ['MB.A8R1:U_QS0', 'MB.A8R1:U_1', 'MB.A8R1:U_2', 'MB.B8R1:U_QS0', 'MB.B8R1:U_1', 'MB.B8R1:U_2',
                          'MB.A9R1:U_QS0', 'MB.A9R1:U_1', 'MB.A9R1:U_2']
        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_updateCALSSignalNamesWithWildcard_RB_BUSBAR_U_RES(self):
        # arrange
        signals_in = ["%BUSBAR%:U_MAG", "%BUSBAR%:U_RES", "%BUSBAR%:ST_BUSBAR_OK", "%BUSBAR%:ST_PWR_PERM"]
        system = 'BUSBAR'
        wildcard = ['DCBB.8L2.R', 'DCBB.9L2.R', 'DCBB.10L2.R']

        # act
        signals_act = SignalMetadata.update_cals_signal_names_with_wildcard(signals_in, system, wildcard)

        # assert
        signals_exp = ['DCBB.8L2.R:U_MAG', 'DCBB.8L2.R:U_RES', 'DCBB.8L2.R:ST_BUSBAR_OK', 'DCBB.8L2.R:ST_PWR_PERM',
                       'DCBB.9L2.R:U_MAG', 'DCBB.9L2.R:U_RES', 'DCBB.9L2.R:ST_BUSBAR_OK', 'DCBB.9L2.R:ST_PWR_PERM',
                       'DCBB.10L2.R:U_MAG', 'DCBB.10L2.R:U_RES', 'DCBB.10L2.R:ST_BUSBAR_OK', 'DCBB.10L2.R:ST_PWR_PERM']
        signals_act.sort()
        signals_exp.sort()
        self.assertEqual(signals_act, signals_exp)

    def test_updateCALSSignalNamesWithWildcard_RB_DIODE_U_DIODE_RB(self):
        # arrange
        signals_in = "%MAGNET%:ST_MAGNET_OK_SQ_RB"
        system = 'DIODE'
        wildcard = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1']

        # act
        signals_act = SignalMetadata.update_cals_signal_names_with_wildcard(signals_in, system, wildcard)

        # assert
        signals_exp = ['MB.A8R1:ST_MAGNET_OK_SQ_RB', 'MB.B8R1:ST_MAGNET_OK_SQ_RB', 'MB.A9R1:ST_MAGNET_OK_SQ_RB']
        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_updateCALSSignalNamesWithWildcard_RB_VF_U_EARTH(self):
        # arrange
        signals_in = "%MAGNET%:U_EARTH_RB"
        system = 'VF'
        wildcard = ['MB.A8L2', 'MB.A10L2', 'MB.C12L2', 'MB.C14L2']

        # act
        signals_act = SignalMetadata.update_cals_signal_names_with_wildcard(signals_in, system, wildcard)

        # assert
        signals_exp = ['MB.A8L2:U_EARTH_RB', 'MB.A10L2:U_EARTH_RB', 'MB.C12L2:U_EARTH_RB', 'MB.C14L2:U_EARTH_RB']
        signals_exp.sort()
        signals_act.sort()
        self.assertEqual(signals_exp, signals_act)

    def test_updateCALSSignalNamesWithWildcard_RB_PC_I_MEAS(self):
        # arrange
        signal_in = "STATUS.I_REF"
        system = 'PC'
        wildcard = ''

        # act
        signal_act = SignalMetadata.update_cals_signal_names_with_wildcard(signal_in, system, wildcard)

        # assert
        signal_exp = ["STATUS.I_REF"]
        signal_exp.sort()
        signal_act.sort()
        self.assertEqual(signal_exp, signal_act)

    def test_RB_Metadata_getSectorNamesFromMagnetNames(self):
        """Tests that Metadata returns correct  sector signal from magnet SIGNAL_NAMES"""

        # arrange
        circuit_type = "RB"
        magnet = "B20L5"
        component = "MAGNET"

        # act
        sector_act = SignalMetadata.get_sector_names_from_components(circuit_type, component, magnet)

        # assert
        sector_exp = "RB.A45"
        self.assertEqual(sector_exp, sector_act)

    def test_RB_Metadata_getSectorNamesFromBusbarNames(self):
        """Tests that Metadata returns correct  sector signal from busbar SIGNAL_NAMES"""

        # arrange
        circuit_type = "RB"
        busbar = "DCBA.10L1.L:U_MAG"
        component = "BUSBAR"

        # act
        sector_act = SignalMetadata.get_sector_names_from_components(circuit_type, component, busbar)

        # assert
        sector_exp = "RB.A81"
        self.assertEqual(sector_exp, sector_act)

    def test_RB_Metadata_getSectorNamesFromCurrentName(self):
        """Tests that Metadata returns correct sector signal from current signal"""

        # arrange
        circuit_type = "RB"
        current = "RPTE.UA23.RB.A12:I_MEAS"
        component = "PC"

        # act
        sector_act = SignalMetadata.get_sector_names_from_components(circuit_type, component, current)

        # assert
        sector_exp = "RB.A12"
        self.assertEqual(sector_exp, sector_act)

    def test_RB_Metadata_getSectorNamesFromCurrentNames(self):
        """Tests that Metadata returns correct sector signal from current SIGNAL_NAMES"""

        # arrange
        circuit_type = "RB"
        current = ["RPTE.UA23.RB.A12:I_MEAS", "RPTE.UA27.RB.A23:I_MEAS"]
        component = "PC"

        # act
        sector_act = SignalMetadata.get_sector_names_from_components(circuit_type, component, current)

        # assert
        sector_exp = ["RB.A12", "RB.A23"]
        self.assertListEqual(sector_exp, sector_act)

    def test_RB_get_wildcard_for_circuit_system_db_signal(self):
        # arrange
        circuit_type = 'RB'
        circuit_name = 'RB.A12'
        system = 'DIODE_RB'
        signal = 'U_DIODE_RB'
        db = 'CALS'

        # act
        name_act, wildcards_act = SignalMetadata.get_wildcard_for_circuit_system_db_signal(circuit_type, circuit_name,
                                                                                           system, db, signal)

        # assert
        name_exp = 'Magnet'
        wildcards_exp = ['MB.A8R1', 'MB.B8R1', 'MB.A9R1', 'MB.B9R1', 'MB.A10R1', 'MB.B10R1', 'MB.A11R1', 'MB.B11R1',
                         'MB.A12R1', 'MB.B12R1', 'MB.C12R1', 'MB.A13R1', 'MB.B13R1', 'MB.C13R1', 'MB.A14R1', 'MB.B14R1',
                         'MB.C14R1', 'MB.A15R1', 'MB.B15R1', 'MB.C15R1', 'MB.A16R1', 'MB.B16R1', 'MB.C16R1', 'MB.A17R1',
                         'MB.B17R1', 'MB.C17R1', 'MB.A18R1', 'MB.B18R1', 'MB.C18R1', 'MB.A19R1', 'MB.B19R1', 'MB.C19R1',
                         'MB.A20R1', 'MB.B20R1', 'MB.C20R1', 'MB.A21R1', 'MB.B21R1', 'MB.C21R1', 'MB.A22R1', 'MB.B22R1',
                         'MB.C22R1', 'MB.A23R1', 'MB.B23R1', 'MB.C23R1', 'MB.A24R1', 'MB.B24R1', 'MB.C24R1', 'MB.A25R1',
                         'MB.B25R1', 'MB.C25R1', 'MB.A26R1', 'MB.B26R1', 'MB.C26R1', 'MB.A27R1', 'MB.B27R1', 'MB.C27R1',
                         'MB.A28R1', 'MB.B28R1', 'MB.C28R1', 'MB.A29R1', 'MB.B29R1', 'MB.C29R1', 'MB.A30R1', 'MB.B30R1',
                         'MB.C30R1', 'MB.A31R1', 'MB.B31R1', 'MB.C31R1', 'MB.A32R1', 'MB.B32R1', 'MB.C32R1', 'MB.A33R1',
                         'MB.B33R1', 'MB.C33R1', 'MB.A34R1', 'MB.B34R1', 'MB.C34R1', 'MB.C34L2', 'MB.B34L2', 'MB.A34L2',
                         'MB.C33L2', 'MB.B33L2', 'MB.A33L2', 'MB.C32L2', 'MB.B32L2', 'MB.A32L2', 'MB.C31L2', 'MB.B31L2',
                         'MB.A31L2', 'MB.C30L2', 'MB.B30L2', 'MB.A30L2', 'MB.C29L2', 'MB.B29L2', 'MB.A29L2', 'MB.C28L2',
                         'MB.B28L2', 'MB.A28L2', 'MB.C27L2', 'MB.B27L2', 'MB.A27L2', 'MB.C26L2', 'MB.B26L2', 'MB.A26L2',
                         'MB.C25L2', 'MB.B25L2', 'MB.A25L2', 'MB.C24L2', 'MB.B24L2', 'MB.A24L2', 'MB.C23L2', 'MB.B23L2',
                         'MB.A23L2', 'MB.C22L2', 'MB.B22L2', 'MB.A22L2', 'MB.C21L2', 'MB.B21L2', 'MB.A21L2', 'MB.C20L2',
                         'MB.B20L2', 'MB.A20L2', 'MB.C19L2', 'MB.B19L2', 'MB.A19L2', 'MB.C18L2', 'MB.B18L2', 'MB.A18L2',
                         'MB.C17L2', 'MB.B17L2', 'MB.A17L2', 'MB.C16L2', 'MB.B16L2', 'MB.A16L2', 'MB.C15L2', 'MB.B15L2',
                         'MB.A15L2', 'MB.C14L2', 'MB.B14L2', 'MB.A14L2', 'MB.C13L2', 'MB.B13L2', 'MB.A13L2', 'MB.C12L2',
                         'MB.B12L2', 'MB.A12L2', 'MB.B11L2', 'MB.A11L2', 'MB.B10L2', 'MB.A10L2', 'MB.B9L2', 'MB.A9L2',
                         'MB.B8L2', 'MB.A8L2']

        wildcards_exp.sort()
        wildcards_act.sort()

        self.assertEqual(name_exp, name_act)
        self.assertListEqual(wildcards_exp, wildcards_act)


    def test_get_signal_name_ipq4_rq4l1b1_pc_i_meas_cals(self):
        # arrange

        # act
        signal_act = SignalMetadata.get_signal_name('IPQ4', 'RQ4.L1B1', 'PC', 'CALS', 'I_MEAS')

        # assert
        signal_ref = 'RPHH.RR13.RQ4.L1B1:I_MEAS'

        self.assertEqual(signal_ref, signal_act)

