import unittest
from unittest.mock import patch
import requests
from lhcsmapi.dbsignal.post_mortem.PmDbSignal import PmDbSignal as SignalPM
import pandas as pd
from datetime import datetime

class TestPmDbSignal(unittest.TestCase):

    def setUp(self):
        self.SYSTEM = "QPS"
        self.CLASS_NAME = "DQAMCNMB_PMHSU"
        self.SOURCE = "B20L5"
        self.SIGNAL_NAME = "I_HDS_4"
        self.TIMESTAMP = 1426220469491000000
        self.defaultSignalPM = SignalPM()

    def tearDown(self):
        pass

    def assertEqualSignals(self, firstSignalPM, secondSignalPM):
        """Assert each property of firstSignalPM against of the same property of secondSignalPM"""
        self.assertEqual(firstSignalPM.system, secondSignalPM.system)
        self.assertEqual(firstSignalPM.className, secondSignalPM.className)
        self.assertEqual(firstSignalPM.source, secondSignalPM.source)
        self.assertEqual(firstSignalPM.signal, secondSignalPM.signal)
        self.assertEqual(firstSignalPM.eventTime, secondSignalPM.eventTime)
        self.assertEqual(firstSignalPM.value, secondSignalPM.value)
        self.assertEqual(firstSignalPM.time, secondSignalPM.time)

    def test_toString(self):
        # arrange
        pms = SignalPM().build_signal(system=self.SYSTEM, className=self.CLASS_NAME, source=self.SOURCE,
                                      signal=self.SIGNAL_NAME, eventTime=self.TIMESTAMP)
        # act
        pmsStringAct = str(pms)

        # assert
        pmsStringExp = 'Db: PM, signal: I_HDS_4, eventTime = 1426220469491000000 (2015-03-13 05:21:09.491000+01:00), ' \
                       'system=QPS, source=B20L5, className=DQAMCNMB_PMHSU'
        self.assertEqual(pmsStringExp, pmsStringAct)

    def test_createSignalPM_noParameters_defaultSignalCreated(self):
        """Tests that the default CalsSignal is created if no arguments are passed to the createSignalPM method"""

        # act:
        pmSignal = SignalPM.build_signal()

        # assert:
        self.assertEqualSignals(pmSignal, self.defaultSignalPM)

    def test_createSignalPM_initParametersProvided_CorrectSignalCreated(self):
        """Tests that the correct CalsSignal is created if arbitrary arguments are passed to the createSignalPM method"""

        # arrange:
        now = datetime.now()
        arbitraryPMSignal = SignalPM("System", "className", "source", "test", pd.Timestamp(now))

        # act:
        pmSignal = SignalPM.build_signal(system="System",
                                         className="className",
                                         source="source",
                                         signal="test",
                                         eventTime=pd.Timestamp(now))

        # assert:
        self.assertEqualSignals(pmSignal, arbitraryPMSignal)

    def test_createSignalPM_parameterIsInvalidType_TypeErrorIsThrown(self):
        """Tests that when kwargs contain a username keyword that isn't a string, a TypeError
        is thrown"""

        now = datetime.now()

        with self.assertRaises(TypeError):
            SignalPM.build_signal(system=1000,
                                  className="className",
                                  source="source",
                                  signal="test",
                                  eventTime=pd.Timestamp(now))


    # needs read and getTimeValue tests


    def test_getNamesOfValues_correctDBAccessWriteCalled(self):
        """Test that write correctly call the write method of the correct DbSignal type dependent on the api parameter"""

        # arrange:
        uut = SignalPM()

        testCases = [("zero", {'namesAndValues': []}, []),
                     ('one', {'namesAndValues': [{'name': 'name1'}]}, ['name1']),
                     ('many', {'namesAndValues': [{'name': 'name1'}, {'name': 'name2'}]}, ['name1', 'name2'])]

        for testName, content, valueNames in testCases:
            with self.subTest(name=testName):

                # act:
                result = uut._get_names_of_value(content)

                # assert:
                self.assertEqual(result, valueNames)

    def test_parseSignal_queryingFailed_RaisesError(self):

        pm_signal = SignalPM("System", "className", "source", "test", pd.Timestamp(datetime.now()))

        with patch.object(requests, 'get') as mock_get:
            mock_get.side_effect = Exception()

            with self.assertRaises(Exception):
                pm_signal.parseSignal()
