import unittest
import warnings

import pandas as pd
import numpy as np
from pathlib import Path
import os

from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.metadata.SignalMetadata import SignalMetadata

# === USED FOR MOCK
from lhcsmapi.pyedsl.dbquery.DbQueryBuilder import NxcalsDbQueryBuilder
from unittest.mock import patch


def read_csv(file_name):
    path = Path(os.path.dirname(__file__))
    full_path = os.path.join(str(path.parent), 'resources/pyedsl/{}'.format(file_name))
    return pd.read_csv(full_path, index_col=0)


class TestQueryBuilder(unittest.TestCase):

    # with_db().with_query_parameters()
    # PM: event query
    def test_eq_query_parameters_qps_16l2_source(self):
        """Tests PM metadata for 1 signal, 1 system, 1 source, 1 className"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm()\
            .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_query_parameters(system='QPS', source='16L2', className='DQAMCNMQ_PMHSU') \
            .query.db_query_builder._query_params

        # assert metadata
        metadata_ref = read_csv('query_builder/pm_eq/QPS_system_source_className.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # PM: signal query
    def test_sq_query_parameters_rqda12_fgc_i_meas(self):
        """Tests PM metadata for 1 signal, 1 system, 1 source, 1 className"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1426220469520000000) \
            .with_query_parameters(signal='STATUS.I_MEAS', system='FGC', source='RPTE.UA47.RB.A45',
                                   className='51_self_pmd') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/PC_system_source_className_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # PM: signal query
    def test_sq_query_parameters_rqda12_fgc_i_meas_i_ref(self):
        """Tests PM metadata for 2 signals, 2 systems, 2 sources, 2 classNames"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1426220469520000000) \
            .with_query_parameters(signal=['STATUS.I_MEAS', 'STATUS.I_REF'], system=['FGC', 'FGC'],
                                   source=['RPTE.UA47.RB.A45', 'RPTE.UA47.RB.A45'],
                                   className=['51_self_pmd', '51_self_pmd']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/PC_system_source_className_I_MEAS_I_REF.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # PM: signal query
    def test_sq_query_parameters_rqda12_fgc_i_meas_i_ref_polymorphic(self):
        """Tests PM metadata for 2 signals, 1 systems, 1 sources, 1 classNames"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1426220469520000000) \
            .with_query_parameters(signal=['STATUS.I_MEAS', 'STATUS.I_REF'], system='FGC',
                                   source='RPTE.UA47.RB.A45',
                                   className='51_self_pmd') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/PC_system_source_className_I_MEAS_I_REF.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # CALS: signal_query
    def test_sq_query_parameters_u_diode_1_signal(self):
        """Tests CALS metadata for 1 signal"""
        # arrange
        # act
        ldb = None

        query_params = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=int(1544622149613000000), duration=[(50, 's'), (150, 's')]) \
            .with_query_parameters(signal='MQ.12L2:U_DIODE_RQD') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/cals_sq/MQ.12L2_U_DIODE_RQD.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # CALS: signal_query
    def test_sq_query_parameters_u_diode_2_signals(self):
        """Tests CALS metadata for 2 signal"""
        # arrange
        # act
        ldb = None

        query_params = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=int(1544622149613000000), duration=[(50, 's'), (150, 's')]) \
            .with_query_parameters(signal=['MQ.12L2:U_DIODE_RQD', 'MQ.14L2:U_DIODE_RQD']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/cals_sq/MQ.12L2_MQ.14L2_U_DIODE_RQD.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_query_parameters_i_meas_2_circuits_ipq(self):
        """Tests CALS metadata for 2 signal"""
        # arrange
        # act
        ldb = None

        query_params = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start='2017-04-03 16:43:50.298', t_end='2017-04-03 16:43:51.254') \
            .with_circuit_type('IPQ4') \
            .with_metadata(circuit_name=['RQ4.L1B1', 'RQ4.L1B2'], system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/cals_sq/IPQ_RQ4.L1B1_RQ4.L1B2_PC_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # NXCALS: signal_query
    def test_sq_query_parameters_i_meas(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""
        # arrange
        # act
        spark = None
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_query_parameters(nxcals_system='CMW', nxcals_device='RPHE.UA23.RQD.A12',
                                   nxcals_property='SUB', signal='I_MEAS')\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/FGC_system_device_property_I_MEAS_1_circuit.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

    # NXCALS: signal_query
    def test_sq_query_parameters_i_meas_2_circuits(self):
        """Tests NXCALS metadata for 2 circuit_names, 2 systems, 2 signal, 0 sources, 0 wildcards"""
        # arrange
        # act
        spark = None
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_query_parameters(nxcals_system=['CMW', 'CMW'],
                                   nxcals_device=['RPHE.UA23.RQD.A12','RPHE.UA23.RQF.A12'],
                                   nxcals_property=['SUB', 'SUB'], signal=['I_MEAS', 'I_MEAS']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/FGC_system_device_property_I_MEAS_2_circuits.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

    # NXCALS: signal_query
    def test_sq_query_parameters_i_meas_2_circuits_single_system(self):
        """Tests NXCALS metadata for 2 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""
        # arrange
        # act
        spark = None

        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_query_parameters(nxcals_system='CMW',
                                   nxcals_device=['RPHE.UA23.RQD.A12', 'RPHE.UA23.RQF.A12'],
                                   nxcals_property='SUB', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/FGC_system_device_property_I_MEAS_2_circuits.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

    # with_db().with_circuit_type().with_duration()/with_timestamp().with_metadata()

    # PM: event_query
    # 1 circuit_name
    # 1 system
    # 1 source (optional)
    def test_eq_metadata_rqda12_pc_fgc_1_source(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='PC') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_eq/RQD.A12_PC_1_SOURCE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_eq_metadata_rba12_ee(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 2 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system='EE') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_eq/RB.A12_EE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_eq_metadata_rqda12_qps_qh_wildcard_source(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        # assert
        with self.assertRaises(KeyError) as context:
            QueryBuilder().with_pm() \
                .with_duration(timestamp=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
                .with_circuit_type('RQ') \
                .with_metadata(circuit_name='RQD.A12', system='QH')

        self.assertTrue('source' in context.exception.args[0])

        # assert time definition

    def test_eq_metadata_rqda12_qps_qh_16l2_source(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 1 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='QH', source='16L2') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_eq/RQD.A12_QH_16L2_SOURCE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition
    def test_eq_metadata_rqda12_qps_qh_no_source(self):
        """Tests event query PM metadata for 1 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        with self.assertRaises(KeyError) as context:
            QueryBuilder().with_pm() \
                .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
                .with_circuit_type('RQ') \
                .with_metadata(circuit_name='RQD.A12', system='QH')

        # assert
        self.assertTrue('source' in context.exception.args[0])

    def test_eq_metadata_rqda12_rqfa12_pc_fgc_1_source(self):
        """Tests event query PM metadata for 2 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_duration(timestamp=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name=['RQD.A12', 'RQF.A12'], system='PC') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_eq/RQD.A12_RQF.A12_PC_1_SOURCE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_eq_metadata_rqda12_rqfa12_qps_qh_wildcard_source(self):
        """Tests event query PM metadata for 2 circuit_name, 1 systems, 0 sources, 0 wildcards"""
        # arrange
        # act
        # assert
        with self.assertRaises(KeyError) as context:
            QueryBuilder().with_pm() \
                .with_duration(t_start=1544622149620000000, duration=[(10, 's'), (400, 's')]) \
                .with_circuit_type('RQ') \
                .with_metadata(circuit_name=['RQD.A12', 'RQF.A12'], system='QH')

        self.assertTrue('source' in context.exception.args[0])

    def test_s1_metadata_rb_pm_t_res_body(self):
        """Tests PM metadata for 1 circuit_names, 1 systems, 3 signals, 1 sources, 0 wildcards"""
        # arrange
        metadata_ee = SignalMetadata.get_circuit_signal_database_metadata('RB', 'RB.A12', 'EE', 'PM')

        # act
        QueryBuilder().with_pm() \
            .with_timestamp(1544631694896000000) \
            .with_query_parameters(signal=['T_RES_BODY_1', 'T_RES_BODY_2', 'T_RES_BODY_3'],
                                   system=metadata_ee['system'],
                                   className=metadata_ee['className'], source='RR17.RB.A12') \
            .query.db_query_builder._query_params

    # signal_query
    # 1..N circuit_names
    # 1..N systems
    # 1..N signals
    # 1..N sources
    # 1..N wildcards

    # PM: signal_query
    # 1..N circuit names
    # 1 system
    # 1..N signals
    # 1..N sources
    # 1..N wildcards
    def test_sq_metadata_rqda12_qps_qh_16l2_source(self):
        """Tests PM metadata for 1 circuit_name, 1 systems, 1 signals, 1 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(timestamp=1544622149620000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='QH', source='16L2',
                           signal='U_HDS', wildcard={'CELL': '16L2'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_QH_16L2_SOURCE.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rqda12_pm_fgc_i_meas(self):
        """Tests PM metadata for 1 circuit_name, 1 systems, 1 signals, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1544622149620000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_PC_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rqda12_rqfa12_pm_fgc_i_meas(self):
        """Tests PM metadata for 2 circuit_name, 1 systems, 1 signals, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1544622149620000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name=['RQD.A12', 'RQF.A12'], system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_RQF.A12_PC_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rqda12_pm_qds_u_1_ext_u_2_ext(self):
        """Tests PM metadata for 1 circuit_names, 1 systems, 2 signals, 1 sources, 1 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm()\
            .with_timestamp(1544622149598000000) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='QDS', signal=['U_1_EXT', 'U_2_EXT'],
                           source='16L2', wildcard={'CELL': '16L2'})\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RQD.A12_QDS_U_1_EXT_U_2_EXT.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rba12_pm_leads_even_leads_odd_u_hts(self):
        """Tests PM metadata for 1 circuit_names, 2 systems, 1 signals, 1 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm()\
            .with_timestamp(1544622149598000000) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system=['LEADS_EVEN', 'LEADS_ODD'], signal='U_HTS')\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RB.A12_LEADS_EVEN_LEADS_ODD_U_HTS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rba12_pm_leads_even_leads_odd_u_hts_u_res(self):
        """Tests PM metadata for 1 circuit_names, 2 systems, 2 signals, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1544622149598000000) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system=['LEADS_EVEN', 'LEADS_ODD'], signal=['U_HTS', 'U_RES']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RB.A12_LEADS_EVEN_LEADS_ODD_U_HTS_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rba12_rba23_pm_leads_even_leads_odd_u_hts_u_res(self):
        """Tests PM metadata for 2 circuit_names, 2 systems, 2 signals, 0 sources, 0 wildcards"""
        # arrange
        # act
        query_params = QueryBuilder().with_pm() \
            .with_timestamp(1544622149598000000) \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name=['RB.A12', 'RB.A23'], system=['LEADS_EVEN', 'LEADS_ODD'],
                           signal=['U_HTS', 'U_RES'])\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/pm_sq/RB.A12_RB.A23_LEADS_EVEN_LEADS_ODD_U_HTS_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # CALS: signal_query
    # 1..N circuit names
    # 1..N systems
    # 1..N signals
    # 1..N sources
    # 1..N wildcards
    def test_sq_metadata_rqda12_cals_u_diode_single_magnet(self):
        """Tests CALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 1 wildcards"""
        # arrange
        # act
        ldb = None

        query_params = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=int(1544622149613000000), duration=[(50, 's'), (150, 's')]) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='DIODE_RQD', signal='U_DIODE_RQD',
                           wildcard={'MAGNET': 'MQ.12L2'})\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/cals_sq/RQD.A12_DIODE_RQD_SINGLE_MAGNET.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

        # CALS: signal_query

    def test_sq_query_parameters_i_meas_1_signal(self):
        """Tests CALS metadata for 1 signal"""
        # arrange
        # act
        ldb = None

        query_params = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start='2017-04-13 18:05:08.730', t_end='2017-04-13 18:05:09.690') \
            .with_circuit_type('600A') \
            .with_metadata(circuit_name='ROD.A12B1', system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/cals_sq/ROD.A12B1_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_sq_metadata_rqda12_cals_u_diode_many_magnets(self):
        """Tests CALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, N wildcards"""
        # arrange
        # act
        ldb = None

        query_params = QueryBuilder().with_cals(ldb) \
            .with_duration(t_start=int(1544622149613000000), duration=[(50, 's'), (150, 's')]) \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='DIODE_RQD', signal='U_DIODE_RQD', wildcard={'MAGNET': '*'})\
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/cals_sq/RQD.A12_DIODE_RQD_MANY_MAGNETS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # NXCALS: signal_query
    # 1..N circuit names
    # 1 system
    # 1..N signals
    # 1..N sources
    # 1..N wildcards

    def test_metadata_nxcals_signal_query_i_meas_single_circuit(self):
        """Tests CALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""
        # arrange
        t_start = '2018-12-12 00:00:00+01:00'
        t_end = '2018-12-13 00:00:00+01:00'
        circuit_type = 'RQ'
        circuit_name = 'RQD.A12'
        signal = 'I_MEAS'
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_circuit_type(circuit_type) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal=signal) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/RQD.A12_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_signal_query_rqd_i_meas(self):
        """Tests CALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""
        # arrange
        circuit_name = [cn for cn in SignalMetadata.get_circuit_names('RQ') if 'D' in cn]
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_sq/RQD_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    # feature_query
    # 1..N circuit_names
    # 1..N systems
    # 1..N signals
    # 1..N sources
    # 1..N wildcards

    def test_metadata_nxcals_feature_query_rb_i_meas(self):
        """Tests NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='*', system='PC', signal='I_MEAS') \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RB_I_MEAS.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rba12_u_res(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='RB.A12', system='BUSBAR', signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RB.A12_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rqda12_u_res(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQD.A12', system='BUSBAR', signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RQD.A12_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rqfa12_u_res(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name='RQF.A12', system='BUSBAR', signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RQF.A12_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rqda12_rqfa12_u_res(self):
        """Tests NXCALS metadata for 1 circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RQ') \
            .with_metadata(circuit_name=['RQD.A12', 'RQF.A12'], system='BUSBAR',
                           signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RQD.A12_RQF.A12_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_metadata_nxcals_feature_query_rb_u_res(self):
        """Tests NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2018-12-12 00:00:00+01:00', t_end='2018-12-13 00:00:00+01:00') \
            .with_circuit_type('RB') \
            .with_metadata(circuit_name='*', system='BUSBAR', signal='U_RES', wildcard={'BUSBAR': '*'}) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RB_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def test_query_parameters_nxcals_feature_query_rqd_u_res(self):
        """Query NXCALS query parameters for 1 circuit_names, 1 systems, 3 signal, 0 sources, 0 wildcards"""

        # arrange
        spark = None

        # act
        query_params = QueryBuilder().with_nxcals(spark) \
            .with_duration(t_start='2017-04-24 19:29:45.679', t_end='2017-04-24 19:51:14.809') \
            .with_query_parameters(nxcals_system='WINCCOA',
                                   signal=['DCQFD.7L2.R:U_RES', 'DCQFB.C21R1.R:U_RES', 'DCQFQ.12L2.L:U_RES']) \
            .query.db_query_builder._query_params

        # assert
        metadata_ref = read_csv('query_builder/nxcals_fq/RQ_U_RES.csv')
        pd.testing.assert_frame_equal(metadata_ref, query_params)

        # assert time definition

    def explode(query_params: pd.DataFrame, column: str):
        import numpy as np
        if any([row_type == list for row_type in query_params[column].apply(type).values]):
            return query_params.reindex(query_params.index.repeat(query_params[column].str.len())) \
                .assign(**{column: np.concatenate(query_params[column].values)}) \
                .reset_index(drop=True)
        else:
            return query_params

    # ======================= MOCK
    # ======================================================
    def test_metadata_nxcals_feature_query_rb_i_meas_fill_missing_rows(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """
        # arrange
        spark = None
        t_start='2018-12-12 00:00:00+01:00'
        t_end='2018-12-13 00:00:00+01:00'
        circuit_type = 'RB'
        api_mock = read_csv('query_builder/nxcals_fq/RB_I_MEAS_Missing_Rows.csv')

        with patch.object(NxcalsDbQueryBuilder, 'nxcals_feature_query') as mock_read:
            mock_read.return_value = api_mock
            # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start, t_end=t_end) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name='*', system='PC', signal='I_MEAS') \
                .feature_query(['mean', 'std', 'max', 'min', 'count']) \
                .df

            # assert
            metadata_ref = read_csv('query_builder/nxcals_fq/RB_I_MEAS_Missing_Rows_Filled.csv')
            pd.testing.assert_frame_equal(metadata_ref, query_params)


    def test_metadata_nxcals_feature_query_rb_u_mag_fill_missing_rows_class(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """
        # arrange
        spark = None
        t_start_injs = 1526898157236000000
        t_end_sbs = 1526903352338000000
        circuit_type = 'RB'
        translate_udf = None
        api_mock = read_csv('query_builder/nxcals_fq/RB_U_MAG_Missing_Rows.csv')

        with patch.object(NxcalsDbQueryBuilder, 'nxcals_feature_query') as mock_read:
            mock_read.return_value = api_mock
            # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start_injs, t_end=t_end_sbs) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name=['RB.A12','RB.A23', 'RB.A45'], system='BUSBAR', signal='U_MAG',
                               wildcard={'BUSBAR': ['DCBB.8L2.R', 'DCBA.8R2.R', 'DCBA.15R4.R']}) \
                .feature_query(['mean', 'max', 'min'], function=translate_udf) \
                .correct_voltage_sign() \
                .df

            # assert
            metadata_ref = read_csv('query_builder/nxcals_fq/RB_U_MAG_Missing_Rows_Filled.csv')
            pd.testing.assert_frame_equal(metadata_ref, query_params)



    def test_metadata_nxcals_feature_query_rb_i_meas_features(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """

        # arrange
        spark = None
        t_start='2018-12-12 00:00:00+01:00'
        t_end='2018-12-13 00:00:00+01:00'
        circuit_type = 'RB'
        api_mock = read_csv('query_builder/nxcals_fq/RB_I_MEAS_Feature_RBA45.csv')

        with patch.object(NxcalsDbQueryBuilder, 'execute_feature_query') as mock_read:
            mock_read.return_value = api_mock

            # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start, t_end=t_end) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name='*', system='PC', signal='I_MEAS') \
                .feature_query(['mean', 'std', 'max', 'min', 'count']) \
                .df

            mean_qp_0 = query_params['mean'][0]
            mean_qp_1 = query_params['mean'][1]
            mean_qp_2 = query_params['mean'][2]

            # assert
            metadata_ref = read_csv('query_builder/nxcals_fq/RB_I_MEAS_Raw_No_Feature_RBA45.csv')
            i_meas_df = metadata_ref

            #i_meas_df.index[0]
            list_0 = []
            list_1 = []
            list_2 = []

            t_start_injs = 1526898157236000000
            t_end_injs = 1526899957236000000
            t_start_sbs = 1526901552338000000
            t_end_sbs = 1526903352338000000

            for i in range(len(i_meas_df)):
                timestamp = i_meas_df.index[i]

                if (timestamp >= t_start_injs and timestamp < t_end_injs):
                    list_1.append(i_meas_df['I_MEAS'][timestamp])

                elif (timestamp >= t_end_injs and timestamp < t_start_sbs):
                    list_0.append(i_meas_df['I_MEAS'][timestamp])

                elif (timestamp >= t_start_sbs and timestamp <= t_end_sbs):
                    list_2.append(i_meas_df['I_MEAS'][timestamp])
                else:
                    print(i)

            mean_ref_0 = sum(list_0) / len(list_0)
            mean_ref_1 = sum(list_1) / len(list_1)
            mean_ref_2 = sum(list_2) / len(list_2)

            self.assertAlmostEqual(mean_qp_0, mean_ref_0)
            self.assertAlmostEqual(mean_qp_1, mean_ref_1)
            self.assertAlmostEqual(mean_qp_2, mean_ref_2)


    def test_metadata_nxcals_feature_query_rb_u_mag_correct_voltage_sign(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """

        # arrange
        spark = None
        t_start_injs = 1526898157236000000
        t_end_sbs = 1526903352338000000
        circuit_type = 'RB'
        translate_udf = None

        api_mock = read_csv('query_builder/nxcals_fq/RB_U_MAG_RAW_Voltage_Sign_RB12_RB45.csv')

        with patch.object(NxcalsDbQueryBuilder, 'execute_feature_query') as mock_read:
            mock_read.return_value = api_mock
                # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start_injs, t_end=t_end_sbs) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name=['RB.A12', 'RB.A45'], system='BUSBAR', signal='U_MAG', wildcard={'BUSBAR': ['DCBB.8L2.R', 'DCBA.15R4.R']}) \
                .feature_query(['mean', 'max', 'min'], function=translate_udf) \
                .correct_voltage_sign() \
                .df

            # assert
            metadata_ref = read_csv('query_builder/nxcals_fq/RB_U_MAG_Corrected_Voltage_Sign_RB12_RB45.csv')
            pd.testing.assert_frame_equal(metadata_ref, query_params)


    def test_metadata_nxcals_feature_query_rb_u_mag_correct_voltage_sign_All(self):
        """Tests pyeDSL feature query with fill_missing_rows function,
            NXCALS metadata for N circuit_names, 1 systems, 1 signal, 0 sources, 0 wildcards
        """

        # arrange
        spark = None
        t_start_injs = 1526898157236000000
        t_end_sbs = 1526903352338000000
        circuit_type = 'RB'
        translate_udf = None

        api_mock = read_csv('query_builder/nxcals_fq/RB_U_MAG_RAW_Voltage_Sign_All.csv')

        with patch.object(NxcalsDbQueryBuilder, 'execute_feature_query') as mock_read:
            mock_read.return_value = api_mock
            # act
            query_params = QueryBuilder().with_nxcals(spark) \
                .with_duration(t_start=t_start_injs, t_end=t_end_sbs) \
                .with_circuit_type(circuit_type) \
                .with_metadata(circuit_name='*', system='BUSBAR', signal='U_MAG', wildcard={'BUSBAR': '*'}) \
                .feature_query(['mean', 'max', 'min'], function=translate_udf) \
                .correct_voltage_sign() \
                .df

            # assert
            metadata_ref = read_csv('query_builder/nxcals_fq/RB_U_MAG_Corrected_Voltage_Sign_All.csv')
            pd.testing.assert_frame_equal(metadata_ref, query_params)
